@@ -5,7 +5,7 @@
  * functions that adjust the top line in the window and invalidate the
  * framing, are hard.
  *
- * $Header: /usr/build/vile/vile/RCS/basic.c,v 1.157 2007/12/23 21:50:05 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/basic.c,v 1.170 2014/04/20 23:51:32 tom Exp $
  *
  */
 
@@ -18,20 +18,23 @@
  * Replace "DOT.o++".
  */
 #if OPT_MULTIBYTE
+#define lp_bytes_at0(lp, off) bytes_at0(lvalue(lp), llength(lp), off)
+#define tb_bytes_at0(tb, off) bytes_at0(tb_values(tb), (int) tb_length(tb), off)
+
 static int
-bytes_at0(LINE *lp, int off)
+bytes_at0(const char *value, int length, int off)
 {
-    return ((llength(lp) > off)
+    return ((length > off && off >= 0)
 	    ? vl_conv_to_utf32((UINT *) 0,
-			       lvalue(lp) + off,
-			       llength(lp) - off)
+			       value + off,
+			       (B_COUNT) (length - off))
 	    : 0);
 }
 
 int
 bytes_at(LINE *lp, int off)
 {
-    int rc = bytes_at0(lp, off);
+    int rc = lp_bytes_at0(lp, off);
     return rc ? rc : 1;
 }
 
@@ -43,13 +46,37 @@ int
 bytes_before(LINE *lp, int off)
 {
     int rc = 0;
+    int first = off;
+    int legal = FALSE;
 
     while (off-- > 0) {
 	++rc;
-	if (bytes_at0(lp, off) == rc) {
+	if (lp_bytes_at0(lp, off) == rc) {
+	    legal = TRUE;
 	    break;
 	}
     }
+    if (!legal && first)
+	rc = 1;
+    return rc;
+}
+
+int
+tb_bytes_before(TBUFF *tb, int off)
+{
+    int rc = 0;
+    int first = off;
+    int legal = FALSE;
+
+    while (off-- > 0) {
+	++rc;
+	if (tb_bytes_at0(tb, off) == rc) {
+	    legal = TRUE;
+	    break;
+	}
+    }
+    if (!legal && first)
+	rc = 1;
     return rc;
 }
 
@@ -136,8 +163,8 @@ mb_cellwidth(WINDOW *wp, const char *text, int limit)
     UINT value;
     int rc = COLS_UTF8;		/* "\uXXXX" */
 
-    vl_conv_to_utf32(&value, text, limit);
-    if (FoldTo8bits(value)) {
+    vl_conv_to_utf32(&value, text, (B_COUNT) limit);
+    if (isPrint(value) && FoldTo8bits((int) value)) {
 	rc = 1;
 	if (w_val(wp, WMDUNICODE_AS_HEX)) {
 	    rc = COLS_UTF8;
@@ -145,7 +172,7 @@ mb_cellwidth(WINDOW *wp, const char *text, int limit)
 	    rc = COLS_8BIT;
 	}
     } else if (term_is_utfXX()) {
-	rc = vl_wcwidth(value);
+	rc = vl_wcwidth((int) value);
 	if (rc <= 0)
 	    rc = COLS_UTF8;
     }
@@ -417,7 +444,7 @@ gotoline(int f, int n)
     } else {
 	status = vl_gotoline(n);
 	if (status != TRUE)
-	    mlwarn("[Not that many lines in buffer: %d]", absol(n));
+	    mlwarn("[Not that many lines in buffer: %ld]", absol((long) n));
     }
     return status;
 }
@@ -780,7 +807,7 @@ gotobop(int f, int n)
 	if (!sameline(DOT, odot) &&
 	    (pre_op_dot.o > lastchar(pre_op_dot.l)) &&
 	    ((fc >= 0 && DOT.o <= fc) || fc < 0)) {
-	    regionshape = FULLLINE;
+	    regionshape = rgn_FULLLINE;
 	}
     }
     return TRUE;
@@ -833,7 +860,7 @@ gotoeop(int f, int n)
 	}
 	/* if we started at the start of line, eat the whole line */
 	if (!sameline(DOT, odot) && was_at_bol)
-	    regionshape = FULLLINE;
+	    regionshape = rgn_FULLLINE;
     }
     return TRUE;
 }
@@ -906,7 +933,7 @@ gotobosent(int f, int n)
     savepos = DOT;
     exp = b_val_rexp(curbp, VAL_SENTENCES)->reg;
 
-    while (s && (is_at_end_of_line(DOT) || isSpace(char_at(DOT)))) {
+    while (s && (is_at_end_of_line(DOT) || isSpace(CharAtDot()))) {
 	s = backchar(TRUE, 1);
 	if (is_empty_line(DOT) && !empty)
 	    return TRUE;
@@ -917,7 +944,7 @@ gotobosent(int f, int n)
 	return gotobob(f, n);
     }
     s = forwchar(TRUE, RegexpLen(exp));
-    while (s && (is_at_end_of_line(DOT) || isSpace(char_at(DOT)))) {
+    while (s && (is_at_end_of_line(DOT) || isSpace(CharAtDot()))) {
 	s = forwchar(TRUE, 1);
 	extra++;
     }
@@ -957,7 +984,7 @@ gotoeosent(int f, int n)
 	    DOT = curbp->b_line;
 	} else if (empty || !is_at_end_of_line(DOT)) {
 	    s = forwchar(TRUE, RegexpLen(exp));
-	    while (s && (is_at_end_of_line(DOT) || isSpace(char_at(DOT)))) {
+	    while (s && (is_at_end_of_line(DOT) || isSpace(CharAtDot()))) {
 		LINE *lp = DOT.l;
 		s = forwchar(TRUE, 1);
 		if (lp != DOT.l)
@@ -966,7 +993,7 @@ gotoeosent(int f, int n)
 	}
     } else {
 	s = forwchar(TRUE, RegexpLen(exp));
-	while (s && (is_at_end_of_line(DOT) || isSpace(char_at(DOT))))
+	while (s && (is_at_end_of_line(DOT) || isSpace(CharAtDot())))
 	    s = forwchar(TRUE, 1);
     }
     return TRUE;
@@ -1019,20 +1046,26 @@ int
 next_column(LINE *lp, int off, int col)
 {
     int rc = 1;
-    int c = lgetc(lp, off);
+    int c;
 
-    if (c == '\t') {
-	rc = next_tabcol(col) - col;
-    }
+    if (off < llength(lp)) {
+	c = lgetc(lp, off);
+
+	if (c == '\t') {
+	    rc = next_tabcol(col) - col;
+	}
 #if OPT_MULTIBYTE
-    else if (b_is_utfXX(curbp)) {
-	if (bytes_at(lp, off) > 1) {
-	    rc = mb_cellwidth(curwp, lvalue(lp) + off, llength(lp) - off);
+	else if (b_is_utfXX(curbp)) {
+	    if (bytes_at(lp, off) > 1) {
+		rc = mb_cellwidth(curwp, lvalue(lp) + off, llength(lp) - off);
+	    }
 	}
-    }
 #endif
-    else if (!isPrint(c)) {
-	rc = NonPrintingCols(c);
+	else if (!isPrint(CharOf(c))) {
+	    rc = NonPrintingCols(c);
+	}
+    } else {
+	rc = 0;
     }
     return col + rc;
 }
@@ -1050,11 +1083,11 @@ column_after(int c, int col, int list)
     }
 #if OPT_MULTIBYTE
     else if (b_is_utfXX(curbp)) {
-	if (vl_conv_to_utf8((UCHAR *) 0, c, 10) > 1)
+	if (vl_conv_to_utf8((UCHAR *) 0, (UINT) c, (B_COUNT) 10) > 1)
 	    rc = col + COLS_UTF8;	/* "\uXXXX" */
     }
 #endif
-    else if (!isPrint(c)) {
+    else if (!isPrint(CharOf(c))) {
 	rc = col + NonPrintingCols(c);
     }
     return rc;
@@ -1070,17 +1103,17 @@ column_after(int c, int col, int list)
 int
 column_sizes(WINDOW *wp, const char *text, unsigned limit, int *used)
 {
-    int rc = NonPrintingCols(*text);
+    int rc = NonPrintingCols(CharOf(*text));
 
     *used = 1;
 #if OPT_MULTIBYTE
     if (b_is_utfXX(wp->w_bufp)) {
-	*used = vl_conv_to_utf32((UINT *) 0, text, limit);
+	*used = vl_conv_to_utf32((UINT *) 0, text, (B_COUNT) limit);
 	if (*used > 1) {
 	    rc = COLS_UTF8;	/* "\uXXXX" */
 	} else if (*used < 1) {
 	    *used = 1;		/* probably a broken character... */
-	} else if (isPrint(*text)) {
+	} else if (isPrint(CharOf(*text))) {
 	    rc = 1;
 	}
     } else
@@ -1088,7 +1121,7 @@ column_sizes(WINDOW *wp, const char *text, unsigned limit, int *used)
     (void) wp;
     (void) limit;
 #endif
-    if (isPrint(*text)) {
+    if (isPrint(CharOf(*text))) {
 	rc = 1;
     }
     return rc;
@@ -1607,7 +1640,7 @@ gorectnmmark(int f GCC_UNUSED, int n GCC_UNUSED)
     if (s != TRUE)
 	return s;
 
-    regionshape = RECTANGLE;
+    regionshape = rgn_RECTANGLE;
     return gonmmark(c);
 }
 
@@ -1668,15 +1701,18 @@ godotplus(int f, int n)
     int s;
 
     if (!f || n == 1) {
-	return firstnonwhite(FALSE, 1);
+	s = firstnonwhite(FALSE, 1);
+    } else if (n < 1) {
+	s = FALSE;
+    } else {
+	s = forwline(TRUE, n - 1);
+	if (s && is_header_line(DOT, curbp)) {
+	    s = backline(FALSE, 1);
+	}
+	if (s == TRUE) {
+	    (void) firstnonwhite(FALSE, 1);
+	}
     }
-    if (n < 1)
-	return (FALSE);
-    s = forwline(TRUE, n - 1);
-    if (s && is_header_line(DOT, curbp))
-	s = backline(FALSE, 1);
-    if (s == TRUE)
-	(void) firstnonwhite(FALSE, 1);
     return s;
 }
 
@@ -1692,12 +1728,12 @@ swapmark(void)
 
     if (samepoint(MK, nullmark)) {
 	mlforce("BUG: No mark ");
-	return;
+    } else {
+	odot = DOT;
+	DOT = MK;
+	MK = odot;
+	curwp->w_flag |= WFMOVE;
     }
-    odot = DOT;
-    DOT = MK;
-    MK = odot;
-    curwp->w_flag |= WFMOVE;
     return;
 }
 
