@@ -3,7 +3,7 @@
  *
  *	written 11-feb-86 by Daniel Lawrence
  *
- * $Header: /usr/build/vile/vile/RCS/bind.c,v 1.316 2007/11/25 19:13:29 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/bind.c,v 1.352 2010/09/06 17:45:49 tom Exp $
  *
  */
 
@@ -151,22 +151,26 @@ old_namebst(BI_NODE * a)
     if (!(a->value.n_flags & NBST_READONLY)) {
 	CMDFUNC *cmd = TYPECAST(CMDFUNC, a->value.n_cmd);
 	if (cmd != 0) {
+	    if ((cmd->c_flags & CMD_TYPE) != CMD_FUNC) {
 #if OPT_ONLINEHELP
-	    if (cmd->c_help)
-		free(TYPECAST(char, cmd->c_help));
+		if (cmd->c_help)
+		    free(TYPECAST(char, cmd->c_help));
 #endif
 #if OPT_MACRO_ARGS
-	    if (cmd->c_args) {
-		int n;
-		for (n = 0; cmd->c_args[n].pi_text != 0; ++n) {
-		    free(cmd->c_args[n].pi_text);
+		if (cmd->c_args) {
+		    int n;
+		    for (n = 0; cmd->c_args[n].pi_text != 0; ++n) {
+			free(cmd->c_args[n].pi_text);
+		    }
+		    free(cmd->c_args);
 		}
-		free(cmd->c_args);
-	    }
 #endif
-	    free(cmd);
+		free(cmd);
+		free(TYPECAST(char, BI_KEY(a)));
+	    }
+	} else {
+	    free(TYPECAST(char, BI_KEY(a)));
 	}
-	free(TYPECAST(char, BI_KEY(a)));
     }
     free(a);
     endofDisplay();
@@ -196,17 +200,32 @@ static void
 dpy_namebst(BI_NODE * a GCC_UNUSED, int level GCC_UNUSED)
 {
 #if OPT_TRACE
-    TRACE(("%s%p -> %s (%d)\n",
-	   trace_indent(level, '.'), a, BI_KEY(a), a->balance));
+    char *indent = strmalloc(trace_indent(level, '.'));
+    TRACE(("[%d]%s%p -> %s%s (%d)\n",
+	   level, indent,
+	   (void *) a,
+	   (a->value.n_flags & NBST_READONLY)
+	   ? "*"
+	   : "",
+	   BI_KEY(a), a->balance));
+    free(indent);
 #endif
 }
 
+static void
+xcg_namebst(BI_NODE * a, BI_NODE * b)
+{
+    BI_DATA temp = a->value;
+    a->value = b->value;
+    b->value = temp;
+}
+
 #define BI_DATA0 {{0}, 0, {0,0,0}}
 #define BI_TREE0 0, 0, BI_DATA0
 static BI_TREE namebst =
-{new_namebst, old_namebst, dpy_namebst, BI_TREE0};
+{new_namebst, old_namebst, dpy_namebst, xcg_namebst, BI_TREE0};
 static BI_TREE redefns =
-{new_namebst, old_namebst, dpy_namebst, BI_TREE0};
+{new_namebst, old_namebst, dpy_namebst, xcg_namebst, BI_TREE0};
 #endif /* OPT_NAMEBST */
 
 /*----------------------------------------------------------------------------*/
@@ -235,7 +254,7 @@ vl_help(int f GCC_UNUSED, int n GCC_UNUSED)
 	returnCode(FALSE);
 
     if (bp->b_active == FALSE) {	/* never been used */
-	hname = cfg_locate(helpfile, FL_ANYWHERE | FL_READABLE);
+	hname = cfg_locate(helpfile, LOCATE_HELP);
 	if (!hname) {
 	    mlforce("[Sorry, can't find the help information]");
 	    (void) zotbuf(bp);
@@ -438,7 +457,7 @@ mac_token2kcod(int check)
     char *value;
     int c;
 
-    if ((value = mac_tokval(&tok)) == 0) {
+    if ((value = mac_unquotedarg(&tok)) == 0) {
 	c = -1;
     } else {
 	c = prc2kcod(value);
@@ -756,7 +775,7 @@ update_binding_list(BUFFER *bp GCC_UNUSED)
 }
 
 static int
-describe_any_bindings(char *lookup, int whichcmd)
+describe_any_bindings(char *lookup, CMDFLAGS whichcmd)
 {
     last_lookup_string = lookup;
     last_whichcmds = whichcmd;
@@ -768,7 +787,7 @@ describe_any_bindings(char *lookup, int whichcmd)
 int
 desbind(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return describe_any_bindings((char *) 0, 0);
+    return describe_any_bindings((char *) 0, (CMDFLAGS) 0);
 }
 
 static int
@@ -776,7 +795,7 @@ describe_alternate_bindings(BINDINGS * bs)
 {
     int code;
     bindings_to_describe = bs;
-    code = describe_any_bindings((char *) 0, 0);
+    code = describe_any_bindings((char *) 0, (CMDFLAGS) 0);
     bindings_to_describe = &dft_bindings;
     return code;
 }
@@ -821,11 +840,11 @@ dessubstr(int f GCC_UNUSED, int n GCC_UNUSED)
     int s;
     static char substring[NSTRING];
 
-    s = mlreply("Apropos string: ", substring, sizeof(substring));
+    s = mlreply("Apropos string: ", substring, (UINT) sizeof(substring));
     if (s != TRUE)
 	return (s);
 
-    return describe_any_bindings(substring, 0);
+    return describe_any_bindings(substring, (CMDFLAGS) 0);
 }
 
 static char described_cmd[NLINE + 1];	/* string to match cmd names to */
@@ -847,7 +866,7 @@ desfunc(int f GCC_UNUSED, int n GCC_UNUSED)
 	s = no_such_function(fnp);
     } else {
 	append_to_binding_list = TRUE;
-	s = describe_any_bindings(described_cmd, 0);
+	s = describe_any_bindings(described_cmd, (CMDFLAGS) 0);
 	append_to_binding_list = FALSE;
     }
     return s;
@@ -981,7 +1000,7 @@ prompt_describe_key(BINDINGS * bs)
     described_cmd[0] = '^';
     (void) strcpy(described_cmd + 1, temp.n_name);
     append_to_binding_list = TRUE;
-    s = describe_any_bindings(described_cmd, 0);
+    s = describe_any_bindings(described_cmd, (CMDFLAGS) 0);
     append_to_binding_list = FALSE;
 
     mlwrite("Key sequence '%s' is bound to function \"%s\"",
@@ -1018,7 +1037,14 @@ des_s_key(int f GCC_UNUSED, int n GCC_UNUSED)
 static char *
 quoted(char *dst, const char *src)
 {
-    return strcat(strcat(strcpy(dst, "\""), src), "\"");
+    char *result = 0;
+    if (dst != 0) {
+	result = strcat(strcat(strcpy(dst, "\""), src), "\"");
+    } else {
+	char temp[NLINE];
+	bputsn_xcolor(quoted(temp, src), -1, XCOLOR_STRING);
+    }
+    return result;
 }
 
 /* returns the number of columns used by the given string */
@@ -1026,7 +1052,7 @@ static unsigned
 converted_len(char *buffer)
 {
     unsigned len = 0, c;
-    while ((c = *buffer++) != EOS) {
+    while ((c = CharOf(*buffer++)) != EOS) {
 	if (c == '\t')
 	    len |= 7;
 	len++;
@@ -1038,11 +1064,17 @@ static void
 quote_and_pad(char *dst, const char *src)
 {
     quoted(dst, src);
-    if (converted_len(dst) >= 32)
-	strcat(dst, "\t");
-    else
-	while (converted_len(dst) < 32)
+    if (dst != 0) {
+	if (converted_len(dst) >= 32)
 	    strcat(dst, "\t");
+	else
+	    while (converted_len(dst) < 32)
+		strcat(dst, "\t");
+    } else {
+	do {
+	    bputc('\t');
+	} while (getccol(FALSE) < 32);
+    }
 }
 
 #if OPT_MENUS || OPT_NAMEBST
@@ -1061,7 +1093,17 @@ to_tabstop(char *buffer)
 static void
 convert_kcode(int c, char *buffer)
 {
-    (void) kcod2prc(c, to_tabstop(buffer));
+    if (buffer != 0) {
+	(void) kcod2prc(c, to_tabstop(buffer));
+    } else {
+	char temp[NLINE];
+	if (llength(DOT.l) > 0
+	    && !isBlank(lvalue(DOT.l)[llength(DOT.l) - 1])) {
+	    bputc('\t');
+	}
+	(void) kcod2prc(c, temp);
+	bputsn_xcolor(temp, -1, XCOLOR_STRING);
+    }
 }
 
 static void
@@ -1088,6 +1130,24 @@ convert_cmdfunc(BINDINGS * bs, const CMDFUNC * cmd, char *outseq)
 }
 #endif /* OPT_MENUS || OPT_NAMEBST */
 
+static int
+add_newline(void)
+{
+    int rc = TRUE;
+    if (buf_head(curbp) != DOT.l)
+	rc = bputc('\n');
+    return rc;
+}
+
+static int
+add_line(const char *text)
+{
+    int rc = add_newline();
+    if (rc)
+	rc = bputsn(text, -1);
+    return rc;
+}
+
 #if OPT_ONLINEHELP
 static int
 show_onlinehelp(const CMDFUNC * cmd)
@@ -1106,7 +1166,7 @@ show_onlinehelp(const CMDFUNC * cmd)
     } else {
 	(void) lsprintf(outseq, "  ( no help for this command )");
     }
-    if (!addline(curbp, outseq, -1))
+    if (!add_line(outseq))
 	return FALSE;
 #if OPT_PROCEDURES
     if ((cmd->c_flags & CMD_TYPE) == CMD_PROC)
@@ -1117,8 +1177,8 @@ show_onlinehelp(const CMDFUNC * cmd)
 	flags &= !(UNDO | REDO);
 #endif
     if (flags & (RANGE | UNDO | REDO | GLOBOK)) {
-	char *gaps = "";
-	char *next = outseq;
+	const char *gaps = "";
+	char *next;
 
 	next = lsprintf(outseq, "  ( ");
 	if (cmd->c_flags & (UNDO | REDO)) {
@@ -1135,7 +1195,7 @@ show_onlinehelp(const CMDFUNC * cmd)
 	    next = lsprintf(next, "%smay follow global command", gaps);
 	}
 	(void) lsprintf(next, " )");
-	if (!addline(curbp, outseq, -1))
+	if (!add_line(outseq))
 	    return FALSE;
     }
 #if OPT_MACRO_ARGS
@@ -1147,7 +1207,7 @@ show_onlinehelp(const CMDFUNC * cmd)
 			    ? cmd->c_args[i].pi_text
 			    : choice_to_name(&fsm_paramtypes_blist,
 					     cmd->c_args[i].pi_type));
-	    if (!addline(curbp, outseq, -1))
+	    if (!add_line(outseq))
 		return FALSE;
 	}
     }
@@ -1189,29 +1249,14 @@ clearflag_func(BI_NODE * n, const void *d GCC_UNUSED)
     return 0;
 }
 
-static int
-addsynonym_func(BI_NODE * node, const void *d)
-{
-    const CMDFUNC *func = (const CMDFUNC *) d;
-    char outseq[NLINE];		/* output buffer for text */
-
-    if (node->value.n_cmd == func &&
-	!(node->value.n_flags & NBST_DONE)) {
-	strcpy(outseq, "  or\t");
-	quoted(outseq + 5, BI_KEY(node));
-	if (!addline(curbp, outseq, -1))
-	    return 1;
-    }
-
-    return 0;
-}
+#define isShortCmd(s) ((int) strlen(s) <= SHORT_CMD_LEN)
 
 static int
 makebind_func(BI_NODE * node, const void *d)
 {
     const struct bindlist_data *data = (const struct bindlist_data *) d;
-    char outseq[NLINE];		/* output buffer for keystroke sequence */
     const CMDFUNC *cmd = node->value.n_cmd;
+    int n;
 
     /* has this been listed? */
     if (node->value.n_flags & NBST_DONE)
@@ -1222,7 +1267,7 @@ makebind_func(BI_NODE * node, const void *d)
 	return 0;
 
     /* try to avoid alphabetizing by the real short names */
-    if (data->min && (int) strlen(BI_KEY(node)) <= data->min)
+    if (data->min && isShortCmd(BI_KEY(node)))
 	return 0;
 
     /* failed lookup by substring? */
@@ -1230,24 +1275,40 @@ makebind_func(BI_NODE * node, const void *d)
 	return 0;
 
     /* add in the command name */
-    quote_and_pad(outseq, BI_KEY(node));
-    convert_cmdfunc(data->bs, cmd, outseq);
-
-    /* dump the line */
-    if (!addline(curbp, outseq, -1))
-	return 1;
+    quote_and_pad(0, BI_KEY(node));
+    convert_cmdfunc(data->bs, cmd, 0);
 
     node->value.n_flags |= NBST_DONE;
 
-    /* add synonyms */
-    btree_walk(&namebst.head, addsynonym_func, cmd);
-
+    if (cmd->c_alias != 0) {
+	const char *top = BI_KEY(node);
+	const char *name;
+	for (n = 0; (name = cmd->c_alias[n]) != 0; ++n) {
+	    if (data->min) {
+		if ((isShortCmd(name)
+		     || ((strcmp(name, top) > 0)
+			 && !isShortCmd(top)))
+		    && add_newline()
+		    && bputsn("  or\t", -1)) {
+		    quoted(0, name);
+		}
+	    } else {
+		if ((isShortCmd(name)
+		     && isShortCmd(top)
+		     && strcmp(name, top) > 0)
+		    && add_newline()
+		    && bputsn("  or\t", -1)) {
+		    quoted(0, name);
+		}
+	    }
+	}
+    }
 #if OPT_ONLINEHELP
     if (!show_onlinehelp(cmd))
 	return FALSE;
 #endif
     /* blank separator */
-    if (!addline(curbp, "", -1))
+    if (!add_line(""))
 	return 1;
 
     return 0;
@@ -1260,7 +1321,7 @@ makebindlist(int whichmask, void *mstring)
 {
     struct bindlist_data data;
 
-    data.mask = whichmask;
+    data.mask = (UINT) whichmask;
     data.min = SHORT_CMD_LEN;
     data.apropos = (char *) mstring;
     data.bs = bindings_to_describe;
@@ -1296,7 +1357,7 @@ makefuncdesc(int j, char *listed)
     quote_and_pad(outseq, nametbl[j].n_name);
 
     /* dump the line */
-    if (!addline(curbp, outseq, -1))
+    if (!add_line(outseq))
 	return FALSE;
 
     /* then look for synonyms */
@@ -1312,7 +1373,7 @@ makefuncdesc(int j, char *listed)
 	if (nametbl[i].n_cmd != cmd)
 	    continue;
 	(void) quoted(outseq + 5, nametbl[i].n_name);
-	if (!addline(curbp, outseq, -1))
+	if (!add_line(outseq))
 	    return FALSE;
     }
 
@@ -1321,7 +1382,7 @@ makefuncdesc(int j, char *listed)
 	return FALSE;
 #endif
     /* blank separator */
-    if (!addline(curbp, "", -1))
+    if (!add_line(""))
 	return FALSE;
 
     return TRUE;
@@ -1414,13 +1475,110 @@ ourstrstr(const char *haystack, const char *needle, int anchor)
 }
 #endif /* OPT_REBIND || OPT_EVAL */
 
+#if SYS_UNIX
+/*
+ * Only source a startup file on Unix if it (and the directory in which it
+ * resides) happens to be in a directory owned by the current user (or root),
+ * and neither is writable by other users.
+ */
+static int
+is_our_file(char *fname, int check_owner)
+{
+    int status = FALSE;
+    struct stat sb;
+
+    if (stat(fname, &sb) == 0) {
+	if ((sb.st_mode & 002) != 0) {
+	    TPRINTF(("Ignoring world-writable \"%s\"\n", fname));
+	} else if ((sb.st_mode & 020) != 0) {
+	    TPRINTF(("Ignoring group-writable \"%s\"\n", fname));
+	} else if (check_owner
+		   && (sb.st_uid != getuid() && sb.st_uid != 0)) {
+	    TPRINTF(("\"%s\" is not our %s\n",
+		     fname, S_ISDIR(sb.st_mode) ? "directory" : "file"));
+	} else {
+	    status = TRUE;
+	}
+    } else {
+	TPRINTF(("\"%s\": %s\n", fname, strerror(errno)));
+    }
+    return status;
+}
+#endif
+
+static int
+check_file_access(char *fname, UINT mode)
+{
+    int result = ffaccess(fname, mode);
+
+    TRACE((T_CALLED "check_file_access(%s, %#x) = %d\n", fname, mode, result));
+#if SYS_UNIX
+#define CHK_OWNER (FL_HOME | FL_CDIR)
+    if (result) {
+	int doit = FALSE;
+	int owner;
+#ifdef GVAL_CHK_ACCESS
+	int check;
+#endif
+
+	/*
+	 * Suppress ownership check for absolute pathnames.
+	 */
+	if (is_abs_pathname(fname))
+	    mode &= ~CHK_OWNER;
+
+	/*
+	 * Check ownership for current- and home-directories.
+	 */
+	owner = (mode < (CHK_OWNER * 2));
+
+	if (mode & FL_EXECABLE) {
+	    doit = TRUE;
+	}
+#ifdef GVAL_CHK_ACCESS
+	else if ((check = global_g_val(GVAL_CHK_ACCESS)) != 0) {
+	    /*
+	     * The values for check-access mode are single bits.  But we treat
+	     * those as an ordered list.  Hence, setting "home" implies we
+	     * also check "current", since that is the first item in the list.
+	     */
+	    if ((mode & FL_ALWAYS) && !((UINT) check & FL_ALWAYS)) {
+		doit = FALSE;
+	    } else if ((mode
+			& ((UINT) check * 2 - 1)
+			& ~(FL_EXECABLE | FL_WRITEABLE | FL_READABLE)) != 0) {
+		doit = TRUE;
+	    }
+	}
+#endif
+	if (doit) {
+	    char *dname = (char *) malloc(NFILEN + strlen(fname) + 10);
+	    char *leaf;
+
+	    if (dname != 0) {
+		leaf = pathleaf(lengthen_path(strcpy(dname, fname)));
+		if ((leaf - 1) != dname)
+		    *--leaf = EOS;
+		if (!is_our_file(dname, owner) || !is_our_file(fname, owner)) {
+		    mlforce("[Skipping '%s' (insecure permissions)]", fname);
+		    result = ABORT;
+		}
+		free(dname);
+	    }
+	}
+    }
+#endif
+
+    returnCode(result);
+}
+
 static char *
 locate_fname(char *dir_name, char *fname, UINT mode)
 {
     static char fullpath[NFILEN];	/* expanded path */
 
     if (!isEmpty(dir_name)
-	&& ffaccess(pathcat(fullpath, dir_name, fname), mode))
+	&& check_file_access(pathcat(fullpath, dir_name, fname), mode) == TRUE)
 	return (fullpath);
 
     return 0;
@@ -1478,28 +1636,6 @@ PATH_value(void)
 #endif
 #endif /* OPT_PATHLOOKUP */
 
-#if SYS_UNIX
-/*
- * Only source a startup file on Unix if it (and the directory in which it
- * resides) happens to be in a directory owned by the current user, and neither
- * is writable by other users.
- */
-static int
-is_our_file(char *fname)
-{
-    int status = FALSE;
-    struct stat sb;
-
-    if (stat(fname, &sb) == 0) {
-	if ((sb.st_mode & 0022) == 0
-	    && sb.st_uid == getuid()) {
-	    status = TRUE;
-	}
-    }
-    return status;
-}
-#endif
-
 /* config files that vile is interested in may live in various places,
  * and which may be constrained to various r/w/x modes.  the files may
  * be
@@ -1515,10 +1651,10 @@ char *
 cfg_locate(char *fname, UINT which)
 {
     char *sp;
-    UINT mode = (which & (FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
+    UINT mode = (which & (FL_ALWAYS | FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
 
 #define FL_BIT(name) ((which & FL_##name) ? " " #name : "")
-    TRACE((T_CALLED "cfg_locate('%s',%s%s%s%s%s%s%s%s%s)\n", NonNull(fname),
+    TRACE((T_CALLED "cfg_locate('%s',%s%s%s%s%s%s%s%s%s%s)\n", NonNull(fname),
 	   FL_BIT(EXECABLE),
 	   FL_BIT(WRITEABLE),
 	   FL_BIT(READABLE),
@@ -1527,7 +1663,8 @@ cfg_locate(char *fname, UINT which)
 	   FL_BIT(EXECDIR),
 	   FL_BIT(STARTPATH),
 	   FL_BIT(PATH),
-	   FL_BIT(LIBDIR)));
+	   FL_BIT(LIBDIR),
+	   FL_BIT(ALWAYS)));
 
     /* take care of special cases */
     if (!fname || !fname[0] || isSpace(fname[0]))
@@ -1537,56 +1674,39 @@ cfg_locate(char *fname, UINT which)
 
     /* look in the current directory */
     if (which & FL_CDIR) {
-	if (ffaccess(fname, mode)) {
-#if SYS_UNIX
-	    int success = FALSE;
-
-	    if ((mode & FL_EXECABLE) || (which == LOCATE_SOURCE)) {
-		char *dname = malloc(NFILEN + strlen(fname) + 10);
-		char *leaf;
-
-		if (dname != 0) {
-		    leaf = pathleaf(lengthen_path(strcpy(dname, fname)));
-		    if ((leaf - 1) != dname)
-			*--leaf = EOS;
-		    if (is_our_file(dname) && is_our_file(fname)) {
-			success = TRUE;
-		    } else {
-			mlforce("[Skipping '%s' (insecure permissions)]", fname);
-		    }
-		    free(dname);
-		}
-	    } else {
-		success = TRUE;
-	    }
-	    if (success)
-#endif
-		returnString(fname);
+	if (check_file_access(fname, FL_CDIR | mode) == TRUE) {
+	    returnString(fname);
 	}
     }
 
     if ((which & FL_HOME)	/* look in the home directory */
-	&&((sp = locate_fname(home_dir(), fname, mode)) != 0))
+	&&((sp = locate_fname(home_dir(), fname, FL_HOME | mode)) != 0))
 	returnString(sp);
 
     if ((which & FL_EXECDIR)	/* look in vile's bin directory */
-	&&((sp = locate_fname(exec_pathname, fname, mode)) != 0))
+	&&((sp = locate_fname(exec_pathname, fname, FL_EXECDIR | mode)) != 0))
 	returnString(sp);
 
     if ((which & FL_STARTPATH)	/* look along "VILE_STARTUP_PATH" */
-	&&(sp = locate_file_in_list(startup_path, fname, mode)) != 0)
+	&&(sp = locate_file_in_list(startup_path,
+				    fname,
+				    FL_STARTPATH | mode)) != 0)
 	returnString(sp);
 
     if (which & FL_PATH) {	/* look along "PATH" */
 #if OPT_PATHLOOKUP
-	if ((sp = locate_file_in_list(PATH_value(), fname, mode)) != 0)
+	if ((sp = locate_file_in_list(PATH_value(),
+				      fname,
+				      FL_PATH | mode)) != 0)
 	    returnString(sp);
 #endif /* OPT_PATHLOOKUP */
 
     }
 
     if ((which & FL_LIBDIR)	/* look along "VILE_LIBDIR_PATH" */
-	&&(sp = locate_file_in_list(libdir_path, fname, mode)) != 0)
+	&&(sp = locate_file_in_list(libdir_path,
+				    fname,
+				    FL_LIBDIR | mode)) != 0)
 	returnString(sp);
 
     returnString(NULL);
@@ -1605,7 +1725,42 @@ ask_which_file(char *fname)
 static void
 list_one_fname(char *fname, UINT mode)
 {
-    bprintf("\n%s\t%s", ffaccess(fname, mode) ? "*" : "", fname);
+    int tag = check_file_access(fname, mode);
+
+    bprintf("\n%s\t", ((tag == ABORT)
+		       ? "?"
+		       : ((tag == TRUE)
+			  ? "*"
+			  : "")));
+
+#if OPT_HYPERTEXT
+    if (tag == TRUE) {
+	REGION myRegion;
+	MARK myDot;
+	TBUFF *myBuff = 0;
+	TBUFF *myFile = 0;
+
+	memset(&myRegion, 0, sizeof(myRegion));
+	myRegion.r_orig = DOT;
+	bprintf("%s", fname);
+	myRegion.r_end = DOT;
+
+	tb_scopy(&myBuff, "view ");
+	path_quote(&myFile, fname);
+	bsl_to_sl_inplace(tb_values(myFile));
+	tb_sappend0(&myBuff, tb_values(myFile));
+
+	myDot = DOT;
+	attributeregion_in_region(&myRegion,
+				  rgn_EXACT,
+				  VAREV,
+				  tb_values(myBuff));
+	tb_free(&myBuff);
+	tb_free(&myFile);
+	DOT = myDot;
+    } else
+#endif
+	bprintf("%s", fname);
 }
 
 static void
@@ -1634,8 +1789,9 @@ list_which_file_in_list(char *list, char *fname, UINT mode)
 static void
 list_which(LIST_ARGS)
 {
+    UINT uflag = (UINT) flag;
     char *fname = (char *) ptr;
-    UINT mode = (flag & (FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
+    UINT mode = (uflag & (FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
 
     /* take care of special cases */
     if (!fname || !fname[0] || isSpace(fname[0]))
@@ -1648,37 +1804,37 @@ list_which(LIST_ARGS)
 	    fname);
 
     /* look in the current directory */
-    if (flag & FL_CDIR) {
+    if (uflag & FL_CDIR) {
 	bprintf("\n$cwd");
-	list_one_fname(fname, mode);
+	list_one_fname(fname, FL_CDIR | mode);
     }
 
-    if (flag & FL_HOME) {	/* look in the home directory */
+    if (uflag & FL_HOME) {	/* look in the home directory */
 	bprintf("\n$HOME");
-	list_which_fname(home_dir(), fname, mode);
+	list_which_fname(home_dir(), fname, FL_HOME | mode);
     }
 
-    if (flag & FL_EXECDIR) {	/* look in vile's bin directory */
+    if (uflag & FL_EXECDIR) {	/* look in vile's bin directory */
 	bprintf("\n$exec-path");
-	list_which_fname(exec_pathname, fname, mode);
+	list_which_fname(exec_pathname, fname, FL_EXECDIR | mode);
     }
 
-    if (flag & FL_STARTPATH) {	/* look along "VILE_STARTUP_PATH" */
+    if (uflag & FL_STARTPATH) {	/* look along "VILE_STARTUP_PATH" */
 	bprintf("\n$startup-path");
-	list_which_file_in_list(startup_path, fname, mode);
+	list_which_file_in_list(startup_path, fname, FL_STARTPATH | mode);
     }
 
-    if (flag & FL_PATH) {	/* look along "PATH" */
+    if (uflag & FL_PATH) {	/* look along "PATH" */
 #if OPT_PATHLOOKUP
 	bprintf("\n$PATH");
-	list_which_file_in_list(PATH_value(), fname, mode);
+	list_which_file_in_list(PATH_value(), fname, FL_PATH | mode);
 #endif /* OPT_PATHLOOKUP */
 
     }
 
-    if (flag & FL_LIBDIR) {	/* look along "VILE_LIBDIR_PATH" */
+    if (uflag & FL_LIBDIR) {	/* look along "VILE_LIBDIR_PATH" */
 	bprintf("\n$libdir-path");
-	list_which_file_in_list(libdir_path, fname, mode);
+	list_which_file_in_list(libdir_path, fname, FL_LIBDIR | mode);
     }
 }
 
@@ -1690,7 +1846,7 @@ show_which_file(char *fname, UINT mode, int f, int n)
 	mlwrite("%s", result);
     }
     if (f)
-	liststuff(WHICH_BufName, n > 2, list_which, mode, fname);
+	liststuff(WHICH_BufName, n > 2, list_which, (int) mode, fname);
     return (result != 0);
 }
 
@@ -1726,7 +1882,7 @@ which_exec(int f, int n)
 char *
 kcod2pstr(int c, char *seq, int limit)
 {
-    seq[0] = (char) kcod2escape_seq(c, &seq[1], limit - 1);
+    seq[0] = (char) kcod2escape_seq(c, &seq[1], (size_t) (limit - 1));
     return seq;
 }
 
@@ -1749,7 +1905,7 @@ static const struct {
 #endif
 
 #define ADD_KCODE(src) \
-	if ((ptr - base) + (len = strlen(src)) + 3 < limit) { \
+	if (((size_t) (ptr - base) + (len = strlen(src)) + 3) < limit) { \
 	    strcpy(ptr, src); \
 	    ptr += len; \
 	}
@@ -1763,13 +1919,13 @@ kcod2escape_seq(int c, char *ptr, size_t limit)
     char *base = ptr;
 
     if (base != 0 && limit > 5) {
-	if (c & CTLA)
+	if ((UINT) c & CTLA)
 	    *ptr++ = (char) cntl_a;
-	else if (c & CTLX)
+	else if ((UINT) c & CTLX)
 	    *ptr++ = (char) cntl_x;
 
 #if OPT_KEY_MODIFY
-	if (c & mod_KEY) {
+	if ((UINT) c & mod_KEY) {
 	    size_t len;
 	    unsigned n;
 	    for (n = 0; n < TABLESIZE(key_modifiers); ++n) {
@@ -1791,7 +1947,7 @@ kcod2escape_seq(int c, char *ptr, size_t limit)
 #endif
 	}
 #endif
-	if (c & SPEC)
+	if ((UINT) c & SPEC)
 	    *ptr++ = (char) poundc;
 	*ptr++ = (char) c;
 	*ptr = EOS;
@@ -1850,9 +2006,9 @@ kcod2prc(int c, char *seq)
     char temp[NSTRING];
     int length;
 
-    length = kcod2pstr(c, temp, sizeof(temp))[0];
+    length = kcod2pstr(c, temp, (int) sizeof(temp))[0];
 #if OPT_KEY_MODIFY
-    if ((c & mod_KEY) != 0 && (length != 0)) {
+    if (((UINT) c & mod_KEY) != 0 && (length != 0)) {
 	(void) strcpy(seq, temp + 1);
 	if (length < (int) (1 + strlen(temp + 1) + (CharOf(c) == 0))) {
 	    (void) bytes2prc(seq + length - 1, temp + length, 1);
@@ -1860,7 +2016,7 @@ kcod2prc(int c, char *seq)
     } else
 #endif
 	(void) bytes2prc(seq, temp + 1, length);
-    TRACE(("kcod2prc(%#x) ->%s\n", c, seq));
+    TRACE2(("kcod2prc(%#x) ->%s\n", c, seq));
     return seq;
 }
 #endif
@@ -1894,11 +2050,17 @@ cmdfunc2keycode(BINDINGS * bs, const CMDFUNC * f)
 const CMDFUNC *
 kcod2fnc(const BINDINGS * bs, int c)
 {
-    if (isSpecial(c)) {
-	KBIND *kp = kcode2kbind(bs, c);
-	return (kp != 0) ? kp->k_cmd : 0;
+    const CMDFUNC *result = 0;
+
+    if (bs != 0) {
+	if (isSpecial(c)) {
+	    KBIND *kp = kcode2kbind(bs, c);
+	    result = (kp != 0) ? kp->k_cmd : 0;
+	} else {
+	    result = bs->kb_normal[CharOf(c)];
+	}
     }
-    return bs->kb_normal[CharOf(c)];
+    return result;
 }
 
 /* fnc2kcod: translate a function pointer to a keycode */
@@ -2129,7 +2291,7 @@ decode_prefix(const char *kk, UINT * prefix)
 		    bits = CTLX;
 		if (isCntrl(poundc) && ch == (UCHAR) toalpha(poundc))
 		    bits = SPEC;
-	    } else if (!strncmp(kk, "FN", 2)) {
+	    } else if (!strncmp(kk, "FN", (size_t) 2)) {
 		bits = SPEC;
 	    }
 	    if (bits != 0)
@@ -2168,13 +2330,13 @@ prc2kcod(const char *kk)
 		break;
 	    if (s == kk || s[1] == '\n' || s[1] == '\0')
 		break;
-	    mklower(vl_strncpy(temp, kk, len + 1));
+	    mklower(vl_strncpy(temp, kk, (size_t) (len + 1)));
 	    if (isLower(temp[0]))
 		temp[0] = (char) toUpper(temp[0]);
 	    for (n = 0; n < TABLESIZE(key_modifiers); ++n) {
-		if (!strncmp(key_modifiers[n].name, temp, len)) {
+		if (!strncmp(key_modifiers[n].name, temp, (size_t) len)) {
 		    found = TRUE;
-		    pref |= (key_modifiers[n].code | mod_KEY);
+		    pref |= ((UINT) key_modifiers[n].code | mod_KEY);
 		    break;
 		}
 	    }
@@ -2200,7 +2362,7 @@ prc2kcod(const char *kk)
     }
 #endif
 
-    if (strlen(kk) > 2 && !strncmp(kk, "M-", 2)) {
+    if (strlen(kk) > 2 && !strncmp(kk, "M-", (size_t) 2)) {
 	pref |= HIGHBIT;
 	kk += 2;
     }
@@ -2208,7 +2370,7 @@ prc2kcod(const char *kk)
     if (*kk == '^' && kk[1] != EOS) {	/* control character */
 	kcod = (UCHAR) kk[1];
 	if (isLower(kcod))
-	    kcod = toUpper(kcod);
+	    kcod = (UINT) toUpper(kcod);
 	kcod = tocntrl(kcod);
 	kk += 2;
     } else {			/* any single char, control or not */
@@ -2257,7 +2419,7 @@ kbd_engl(const char *prompt, char *buffer)
 /* sound the alarm! */
 #if OPT_TRACE
 void
-trace_alarm(char *file, int line)
+trace_alarm(const char *file, int line)
 #else
 void
 kbd_alarm(void)
@@ -2276,22 +2438,47 @@ kbd_alarm(void)
     warnings++;
 }
 
+void
+kbd_start(KBD_DATA * data)
+{
+    beginDisplay();
+
+#if OPT_MULTIBYTE
+    make_local_b_val(bminip, VAL_FILE_ENCODING);
+    if (cmd_encoding == enc_AUTO) {
+	set_b_val(bminip, VAL_FILE_ENCODING, b_val(curbp, VAL_FILE_ENCODING));
+    } else {
+	set_b_val(bminip, VAL_FILE_ENCODING, cmd_encoding);
+    }
+#endif
+
+    data->save_bp = curbp;
+    data->save_wp = curwp;
+    data->save_mk = MK;
+
+    curbp = bminip;
+    curwp = wminip;
+    MK = DOT;
+}
+
+void
+kbd_finish(KBD_DATA * data)
+{
+    curbp = data->save_bp;
+    curwp = data->save_wp;
+    MK = data->save_mk;
+
+    endofDisplay();
+}
+
 /* put a character to the keyboard-prompt */
 int
 kbd_putc(int c)
 {
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    KBD_DATA save;
     int status;
 
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    curbp = bminip;
-    curwp = wminip;
-    savemk = MK;
-    MK = DOT;
+    kbd_start(&save);
     if ((kbd_expand <= 0) && isreturn(c)) {
 	status = kbd_erase_to_end(0);
     } else {
@@ -2304,10 +2491,7 @@ kbd_putc(int c)
 	TRACE(("mini:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
 #endif
     }
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
-    endofDisplay();
+    kbd_finish(&save);
 
     return status;
 }
@@ -2324,58 +2508,35 @@ kbd_puts(const char *s)
 void
 kbd_erase(void)
 {
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    if (vl_echo && !clhide) {
+	KBD_DATA save;
 
-    if (!vl_echo || clhide)
-	return;
-
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    curbp = bminip;
-    curwp = wminip;
-    savemk = MK;
-    MK = DOT;
-    if (DOT.o > 0) {
-	backchar_to_bol(TRUE, 1);
-	ldel_chars(1, FALSE);
-    }
+	kbd_start(&save);
+	if (DOT.o > 0) {
+	    backchar_to_bol(TRUE, 1);
+	    ldel_chars((B_COUNT) 1, FALSE);
+	}
 #ifdef VILE_DEBUG
-    TRACE(("MINI:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
+	TRACE(("MINI:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
 #endif
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
-    endofDisplay();
+	kbd_finish(&save);
+    }
 }
 
 int
 kbd_erase_to_end(int column)
 {
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
-
     if (vl_echo && !clhide) {
-	beginDisplay();
-	savebp = curbp;
-	savewp = curwp;
-	curbp = bminip;
-	curwp = wminip;
-	savemk = MK;
-	MK = DOT;
+	KBD_DATA save;
+
+	kbd_start(&save);
 	if (llength(DOT.l) > 0) {
 	    DOT.o = column;
 	    if (llength(DOT.l) > DOT.o)
-		ldel_bytes(llength(DOT.l) - DOT.o, FALSE);
+		ldel_bytes((B_COUNT) (llength(DOT.l) - DOT.o), FALSE);
 	    TRACE(("NULL:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
 	}
-	curbp = savebp;
-	curwp = savewp;
-	MK = savemk;
-	endofDisplay();
+	kbd_finish(&save);
     }
 
     return TRUE;
@@ -2525,7 +2686,7 @@ makecmpllist(int case_insensitive, void *cinfop)
 	    maxlen = l;
     }
 
-    slashcol = (int) (pathleaf(buf) - buf);
+    slashcol = (size_t) (pathleaf(buf) - buf);
     if (slashcol != 0) {
 	char b[NLINE];
 	(void) strncpy(b, buf, slashcol);
@@ -2540,7 +2701,7 @@ makecmpllist(int case_insensitive, void *cinfop)
     if (cmplcols == 0)
 	cmplcols = 1;
 
-    nentries = (int) ((last - first) / size_entry);
+    nentries = (int) ((size_t) (last - first) / size_entry);
     cmplrows = nentries / cmplcols;
     cmpllen = term.cols / cmplcols;
     if (cmplrows * cmplcols < nentries)
@@ -2550,7 +2711,9 @@ makecmpllist(int case_insensitive, void *cinfop)
 	for (j = 0; j < cmplcols; j++) {
 	    int idx = cmplrows * j + i;
 	    if (idx < nentries) {
-		const char *s = THIS_NAME(first + (idx * size_entry)) + slashcol;
+		const char *s = (THIS_NAME(first
+					   + ((size_t) idx * size_entry))
+				 + slashcol);
 		if (j == cmplcols - 1)
 		    bprintf("%s\n", s);
 		else
@@ -2663,11 +2826,12 @@ fill_partial(
     const char *this_name = THIS_NAME(first);
 
     TRACE(("fill_partial(%d:%.*s) first=%s, last=%s\n",
-	   pos, (int) pos,
+	   (int) pos, (int) pos,
 	   TRACE_NULL(buf),
 	   TRACE_NULL(THIS_NAME(first)),
 	   TRACE_NULL(THIS_NAME(last))));
 
+    assert(buf != 0);
 #if 0				/* case insensitive reply correction doesn't work reliably yet */
     if (!clexec && case_insensitive) {
 	int spos = pos;
@@ -2692,11 +2856,11 @@ fill_partial(
     for_ever {
 	buf[n] = this_name[n];	/* add the next char in */
 	buf[n + 1] = EOS;
-	TRACE(("...added(%d:%c)\n", n, buf[n]));
+	TRACE(("...added(%d:%c)\n", (int) n, buf[n]));
 
 	/* scan through the candidates */
 	for (p = first; p != last; p = NEXT_DATA(p)) {
-	    if (StrNcmp(&THIS_NAME(p)[n], &buf[n], 1) != 0) {
+	    if (StrNcmp(&THIS_NAME(p)[n], &buf[n], (size_t) 1) != 0) {
 		buf[n] = EOS;
 		if (n == pos
 #if OPT_POPUPCHOICE
@@ -2709,7 +2873,7 @@ fill_partial(
 		    )
 		    kbd_alarm();
 		kbd_flush();	/* force out alarm or partial completion */
-		TRACE(("...fill_partial %d\n", n));
+		TRACE(("...fill_partial %d\n", (int) n));
 		return n;
 	    }
 	}
@@ -3029,7 +3193,7 @@ kbd_engl_stat(const char *prompt, char *buffer, int stated)
 #if COMPLETE_FILES
     init_filec(FILECOMPLETION_BufName);
 #endif
-    kbd_flags |= stated;
+    kbd_flags |= (KBD_OPTIONS) stated;
     code = kbd_reply(
 			prompt,	/* no-prompt => splice */
 			&temp,	/* in/out buffer */
@@ -3049,18 +3213,24 @@ insert_namebst(const char *name, const CMDFUNC * cmd, int ro)
 {
     int result;
 
+    TRACE((T_CALLED "insert_namebst(%s,%s)\n", name, ro ? "ro" : "rw"));
     if (name != 0) {
 	BI_DATA temp, *p;
 
 	if ((p = btree_search(&namebst, name)) != 0) {
 	    if ((p->n_flags & NBST_READONLY)) {
-		if (btree_insert(&redefns, p) == 0
-		    || !btree_delete(&namebst, name)) {
-		    return FALSE;
+		if (btree_insert(&redefns, p) == 0) {
+		    returnCode(FALSE);
+		} else {
+		    if (!btree_delete(&namebst, name)) {
+			returnCode(FALSE);
+		    }
 		}
 		mlwrite("[Redefining builtin '%s']", name);
-	    } else if (!delete_namebst(name, TRUE)) {
-		return FALSE;
+	    } else {
+		if (!delete_namebst(name, TRUE, TRUE)) {
+		    returnCode(FALSE);
+		}
 	    }
 	}
 
@@ -3072,7 +3242,7 @@ insert_namebst(const char *name, const CMDFUNC * cmd, int ro)
     } else {
 	result = FALSE;
     }
-    return result;
+    returnCode(result);
 }
 
 static void
@@ -3115,44 +3285,47 @@ remove_cmdfunc_ref(BINDINGS * bs, const CMDFUNC * cmd)
  * Lookup a name in the binary-search tree, remove it if found
  */
 int
-delete_namebst(const char *name, int release)
+delete_namebst(const char *name, int release, int redefining)
 {
     BI_DATA *p = btree_search(&namebst, name);
-    int code;
+    int code = TRUE;
 
+    TRACE((T_CALLED "delete_namebst(%s,%d) %p\n", name, release, p));
     /* not a named procedure */
-    if (!p)
-	return TRUE;
+    if ((p = btree_search(&namebst, name)) != 0) {
 
-    /* we may have to free some stuff */
-    if (p && release) {
-	remove_cmdfunc_ref(&dft_bindings, p->n_cmd);
-	remove_cmdfunc_ref(&ins_bindings, p->n_cmd);
-	remove_cmdfunc_ref(&cmd_bindings, p->n_cmd);
+	/* we may have to free some stuff */
+	if (p && release) {
+	    remove_cmdfunc_ref(&dft_bindings, p->n_cmd);
+	    remove_cmdfunc_ref(&ins_bindings, p->n_cmd);
+	    remove_cmdfunc_ref(&cmd_bindings, p->n_cmd);
 
-	/* free stuff */
+	    /* free stuff */
 #if OPT_PERL
-	if (p->n_cmd->c_flags & CMD_PERL)
-	    perl_free_sub(CMD_U_PERL(p->n_cmd));
+	    if (p->n_cmd->c_flags & CMD_PERL)
+		perl_free_sub(CMD_U_PERL(p->n_cmd));
 #endif
 
-	if (!(p->n_flags & NBST_READONLY)) {
-	    beginDisplay();
-	    free(TYPECAST(char, p->n_cmd->c_help));
-	    free(TYPECAST(char, p->n_cmd));
-	    endofDisplay();
+	    if (!(p->n_flags & NBST_READONLY)) {
+		beginDisplay();
+		free(TYPECAST(char, p->n_cmd->c_help));
+		free(TYPECAST(char, p->n_cmd));
+		endofDisplay();
+	    }
+	    p->n_cmd = 0;	/* ...so old_namebst won't free this too */
 	}
-	p->n_cmd = 0;		/* ...so old_namebst won't free this too */
-    }
 
-    if ((code = btree_delete(&namebst, name)) == TRUE) {
-	if ((p = btree_search(&redefns, name)) != 0) {
-	    mlwrite("[Restoring builtin '%s']", name);
-	    code = (btree_insert(&namebst, p) != 0);
-	    (void) btree_delete(&redefns, p->bi_key);
+	if ((code = btree_delete(&namebst, name)) == TRUE) {
+	    if (!redefining) {
+		if ((p = btree_search(&redefns, name)) != 0) {
+		    mlwrite("[Restoring builtin '%s']", name);
+		    code = (btree_insert(&namebst, p) != 0);
+		    (void) btree_delete(&redefns, p->bi_key);
+		}
+	    }
 	}
     }
-    return code;
+    returnCode(code);
 }
 
 /*
@@ -3172,7 +3345,7 @@ rename_namebst(const char *oldname, const char *newname)
 
     /* remove the entry if the new name is not a procedure (bracketed) */
     if (!is_scratchname(newname))
-	return delete_namebst(oldname, TRUE);
+	return delete_namebst(oldname, TRUE, FALSE);
 
     /* add the new name */
     strip_brackets(name, newname);
@@ -3181,7 +3354,7 @@ rename_namebst(const char *oldname, const char *newname)
 	return FALSE;
 
     /* delete the old (but don't free the data) */
-    return delete_namebst(oldname, FALSE);
+    return delete_namebst(oldname, FALSE, FALSE);
 }
 
 int
@@ -3197,9 +3370,10 @@ search_namebst(const char *name)
 void
 build_namebst(const NTAB * nptr, int lo, int hi)
 {
-    for (; lo < hi; lo++)
+    for (; lo < hi; lo++) {
 	if (!insert_namebst(nptr[lo].n_name, nptr[lo].n_cmd, TRUE))
 	    tidy_exit(BADEXIT);
+    }
 }
 
 /*
@@ -3213,7 +3387,7 @@ kbd_complete_bst(
 		    char *buf,
 		    size_t *pos)
 {
-    unsigned cpos = (unsigned) *pos;
+    size_t cpos = *pos;
     int status = FALSE;
     const char **nptr;
 
@@ -3312,6 +3486,7 @@ free_ext_bindings(BINDINGS * bs)
 void
 bind_leaks(void)
 {
+    btree_printf(&namebst);
 #if OPT_REBIND
     free_all_bindings(&dft_bindings);
     free_ext_bindings(&ins_bindings);
