@@ -1,12 +1,14 @@
 /*
- * $Id: btree.c,v 1.23 2006/04/25 20:13:46 tom Exp $
- * Copyright 1997-2005 by Thomas E. Dickey
+ * $Id: btree.c,v 1.51 2010/09/06 15:42:19 tom Exp $
+ * Copyright 1997-2008,2010 by Thomas E. Dickey
  *
  * Maintains a balanced binary tree (aka AVL tree) of unspecified nodes.  The
  * algorithm is taken from "The Art of Computer Programming -- Volume 3 --
  * Sorting and Searching", by Donald Knuth.  Knuth presents the insertion and
  * search algorithms in assembly language for MIX, and gives an overview of the
  * deletion algorithm for the "interested reader".
+ *
+ * See 6.2.2 for deletion, as well as notes in 6.2.3 for rebalancing.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -16,6 +18,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <ctype.h>
 
 #if !defined(DEBUG_BTREE) && !defined(NDEBUG)
 #define NDEBUG
@@ -39,6 +42,8 @@
 #define beginDisplay()		/* nothing */
 #define endofDisplay()		/* nothing */
 
+#define	NonNull(s)	((s == 0) ? "" : s)
+
 #include <time.h>
 
 #else
@@ -62,12 +67,16 @@
 
 #ifdef DEBUG_BTREE
 # ifndef TRACE
-#  if DEBUG_BTREE > 0
+#  if DEBUG_BTREE > 1
 #   define TRACE(p)	printf p ; fflush(stdout)
 #  endif
 # endif
-#else
+#elif 1
 # undef TRACE
+# undef OPT_TRACE
+# define OPT_TRACE 0
+# define DEBUG_BTREE -1
+#else
 # define DEBUG_BTREE -1
 #endif
 
@@ -97,7 +106,7 @@ static void dump_nodes(BI_TREE * funcs, BI_NODE * p, int level);
 #endif
 
 /*
- * FIXME
+ * FIXME - this should compare unsigned chars to be completely consistent.
  */
 static int
 btree_strcmp(const char *a, const char *b)
@@ -124,7 +133,7 @@ btree_insert(BI_TREE * funcs, BI_DATA * data)
     int a;
     BI_DATA *value = 0;
 
-    TRACE(("inserting '%s'\n", data->bi_key));
+    TRACE(("btree_insert(%p,%s)\n", funcs, data->bi_key));
     if (p == 0) {
 	if ((p = (*funcs->allocat) (data)) == 0)
 	    return 0;
@@ -196,6 +205,11 @@ btree_insert(BI_TREE * funcs, BI_DATA * data)
 		    assert(B(r) == -a);
 		    TRACE(("(A9: Double rotation)\n"));
 		    p = LINK(-a, r);
+		    assert(p != 0);
+		    if (p == 0) {
+			TRACE(("(BUG: null pointer)\n"));
+			break;
+		    }
 		    LINK(-a, r) = LINK(a, p);
 		    LINK(a, p) = r;
 		    LINK(a, s) = LINK(-a, p);
@@ -228,277 +242,201 @@ btree_insert(BI_TREE * funcs, BI_DATA * data)
     return (value);
 }
 
-static BI_NODE *
-parent_of(BI_TREE * funcs, BI_NODE * s)
-{
-    BI_NODE *r = &(funcs->head);
-    BI_NODE *p = RLINK(r);	/* 'p' => down the tree        */
-    int a;
+#define MAXSTK 100
 
-    while ((a = COMPARE(KEY(s), KEY(p))) != 0) {
-	assert(LINK(a, p) != 0);
-	r = p;
-	p = LINK(a, p);
-    }
-    TRACE(("parent of '%s' is '%s'\n", KEY(s), KEY(r)));
-    return r;
+typedef struct {
+    BI_NODE *ptr;
+    int dir;
+} STACKDATA;
+
+static BI_NODE *
+single_rotation(BI_NODE * item, int dir)
+{
+    BI_NODE *temp = item->links[!dir];
+
+    item->links[!dir] = temp->links[dir];
+    temp->links[dir] = item;
+
+    return temp;
 }
 
-#define MAXSTK 20
-#define PUSH(A,P) { \
-	TRACE(("@%d:push #%d a=%2d, B=%2d, p='%s' -> '%s'\n", __LINE__, \
-		k, A, B(P), P?KEY(P):"", LINK(A,P)?KEY(LINK(A,P)):"")); \
-	stack[k].a = A;\
-	stack[k].p = P;\
-	k++; }
+static BI_NODE *
+double_rotation(BI_NODE * item, int dir)
+{
+    BI_NODE *temp = item->links[!dir]->links[dir];
+
+    item->links[!dir]->links[dir] = temp->links[!dir];
+    temp->links[!dir] = item->links[!dir];
+    item->links[!dir] = temp;
+
+    temp = item->links[!dir];
+    item->links[!dir] = temp->links[dir];
+    temp->links[dir] = item;
+
+    return temp;
+}
+
+static void
+readjust_balance(BI_NODE * item, int dir, int balance)
+{
+    BI_NODE *next = item->links[dir];
+    BI_NODE *last = next->links[!dir];
+
+    if (last->balance == 0) {
+	item->balance = next->balance = 0;
+    } else if (last->balance == balance) {
+	item->balance = -balance;
+	next->balance = 0;
+    } else {
+	/* last->balance == -balance */
+	item->balance = 0;
+	next->balance = balance;
+    }
+
+    last->balance = 0;
+}
+
+static BI_NODE *
+remove_balance(BI_NODE * item, int dir, int *done)
+{
+    BI_NODE *next = item->links[!dir];
+    int balance = (dir == 0) ? -1 : 1;
+
+    if (next->balance == -balance) {
+	item->balance = next->balance = 0;
+	item = single_rotation(item, dir);
+    } else if (next->balance == balance) {
+	readjust_balance(item, !dir, -balance);
+	item = double_rotation(item, dir);
+    } else {
+	/* next->balance == 0 */
+	item->balance = -balance;
+	next->balance = balance;
+	item = single_rotation(item, dir);
+	*done = 1;
+    }
+
+    return item;
+}
 
 int
 btree_delete(BI_TREE * funcs, const char *data)
 {
-    struct {
-	BI_NODE *p;
-	int a;
-    } stack[MAXSTK];
-    int k = 0;
-    /* (A1:Initialize) */
+    TRACE(("btree_delete(%p,%s)\n", funcs, NonNull(data)));
 
-    BI_NODE *t = &(funcs->head), *p, *q, *r, *s;
+    if (funcs->count != 0) {
+	STACKDATA stack[MAXSTK];
+	BI_NODE *item;
+	int top = 0;
+	int done = 0;
 
-    int a, b;
-    char *value;
-
-    if ((p = t) == 0
-	|| (p = LINK(a = 1, p)) == 0) {
-	return 0;
-    }
-    /* (A2:Compare) */
-    while ((a = COMPARE(data, value = KEY(p))) != 0) {
-	if ((q = LINK(a, p)) == 0) {
-	    value = 0;
-	    break;
-	}
-	/* (A3,A4: move left/right accordingly) */
-	t = p;
-	p = LINK(a, p);
-	/* ...continue comparing */
-    }
-
-    if (value != 0) {		/* we found the node to delete, in p */
-	TRACE(("deleting node '%s'\n", value));
-	q = p;
-	p = t;
-	a = (q == RLINK(p)) ? 1 : -1;
-
-	TRACE(("@%d, p='%s'\n", __LINE__, p ? KEY(p) : ""));
-	TRACE(("@%d, q='%s'\n", __LINE__, q ? KEY(q) : ""));
-	TRACE(("@%d, a=%d\n", __LINE__, a));
-
-	/* D1: Is RLINK null? */
-	if (RLINK(q) == 0) {
-	    TRACE(("D1\n"));
-
-	    LINK(a, p) = LLINK(q);
-	    s = p;
-#if 0				/* LATER? */
-	} else if (LLINK(q) == 0) {	/* D1.5 */
-	    TRACE(("D1.5\n"));
-	    LINK(a, p) = RLINK(q);
-	    s = RLINK(q);
-#endif
-	} else {		/* D2: Find successor */
-	    TRACE_SUBTREE("SUBTREE(p)-before\n", funcs, p);
-	    r = RLINK(q);
-	    if (LLINK(r) == 0) {
-		TRACE(("D2, r='%s', q='%s'\n", KEY(r), KEY(q)));
-		TRACE(("DIFF: %d\n", B(r) - B(q)));
-
-		LLINK(r) = LLINK(q);
-		LINK(a, p) = r;
-		s = r;
-		TRACE(("(D2) replace balance %2d with %2d, a=%2d\n",
-		       B(s), B(q), a));
-		B(s) = B(q);
-		a = 1;		/* for RLINK(q) */
-
-		TRACE_SUBTREE("SUBTREE(p)-after\n", funcs, p);
-		TRACE(("@%d, p='%s'\n", __LINE__, KEY(p)));
-		TRACE(("@%d, r='%s'\n", __LINE__, KEY(r)));
-	    } else {		/* D3: Find null LLINK */
-		TRACE(("D3: find null LLINK\n"));
-
-		TRACE(("@%d, r='%s'\n", __LINE__, r ? KEY(r) : ""));
-		s = r;
-		do {
-		    r = s;
-		    s = LLINK(r);
-		} while (LLINK(s) != 0);
-
-		LLINK(s) = LLINK(q);
-		LLINK(r) = RLINK(s);
-		RLINK(s) = RLINK(q);
-		TRACE(("@%d, p='%s', a=%d\n", __LINE__, KEY(p), a));
-		TRACE(("@%d, r='%s'\n", __LINE__, KEY(r)));
-		TRACE(("@%d, s='%s'\n", __LINE__, KEY(s)));
-		LINK(a, p) = s;
-		TRACE(("(D3) replace balance %2d with %2d, a=%2d\n",
-		       B(s), B(q), a));
-		B(s) = B(q);
-		s = r;
-		a = -1;		/* ...since we followed left */
-	    }
-	}
-	b = a;
-	/* D4: Free the node */
-	assert(q != 0);
-	(*funcs->dealloc) (q);
-	funcs->count -= 1;
-	TRACE_TREE("after delinking:", funcs);
-
-	/* Construct the auxiliary stack */
-	a = 1;
-	q = &(funcs->head);
-	if (s != 0
-	    && s != q
-	    && q != 0) {
-	    TRACE(("Construct stack from '%s' down to '%s', final a=%d\n",
-		   q ? KEY(q) : "",
-		   s ? KEY(s) : "",
-		   b));
-	    while (s != q) {
-		PUSH(a, q);
-		q = LINK(a, q);
-		a = COMPARE(KEY(s), KEY(q));
-	    }
-	    PUSH(b, q);
-	    TRACE(("...done building stack\n"));
+	/* (A1:Initialize) */
+	item = &(funcs->head);
+	if (item == 0
+	    || (item = LINK(1, item)) == 0) {
+	    TRACE(("...not found @%d\n", __LINE__));
+	    return 0;
 	}
 
-	/* Rebalance the tree */
-	for_ever {
-	    if (--k <= 0) {
-		TRACE(("shorten the whole tree\n"));
-		funcs->depth -= 1;
-		break;
-	    }
-	    p = stack[k].p;
-	    a = stack[k].a;
-	    TRACE(("processing #%d '%s' B = %d, a = %d (%p)\n",
-		   k, KEY(p), B(p), a, LINK(a, p)));
-	    if (B(p) == a) {
-		TRACE(("Case (i)\n"));
-		B(p) = 0;
-	    } else if (B(p) == 0) {
-		TRACE(("Case (ii)\n"));
-		B(p) = -a;
-		break;
+	for (;;) {
+	    /* (A2:Compare) */
+	    if (item == NULL || KEY(item) == 0) {
+		TRACE(("...not found @%d\n", __LINE__));
+		return 0;
 	    } else {
-		TRACE(("Case (iii): Rebalancing needed!\n"));
-
-		q = LINK(-a, p);
-		assert(q != 0);
-		assert(k > 0);
-		assert(B(p) == -a);
-
-		t = stack[k - 1].p;
-		TRACE(("t:%s, balance:%d\n", KEY(t), B(t)));
-		TRACE(("A:p:%s, balance:%d\n", KEY(p), B(p)));
-		TRACE(("B:q:%s, balance:%d\n", KEY(q), B(q)));
-		TRACE_TREE("before rebalancing:", funcs);
-
-		if (B(q) == -a) {
-		    TRACE(("CASE 1: single rotation\n"));
-
-		    r = q;
-
-		    TRACE(("link LINK(-a,p) -> LINK( a,q) = '%s'\n",
-			   LINK(a, q) ? KEY(LINK(a, q)) : ""));
-		    TRACE(("link LINK( a,q) -> p          = '%s'\n",
-			   p ? KEY(p) : ""));
-
-		    LINK(-a, p) = LINK(a, q);
-		    LINK(a, q) = p;
-
-		    B(p) = B(q) = 0;
-
-		    t = parent_of(funcs, p);
-
-		    TRACE(("Finish by linking '%s' to %sLINK(%s)\n",
-			   KEY(r),
-			   (p == RLINK(t))
-			   ? "R"
-			   : (p == LLINK(t))
-			   ? "L" : "?",
-			   KEY(t)));
-
-		    t->links[(p == RLINK(t))] = r;
-
-		} else if (B(q) == a) {
-		    TRACE(("CASE 2: double rotation\n"));
-
-		    r = LINK(a, q);
-#if DEBUG_BTREE > 1
-		    TRACE(("a = '%d'\n", a));
-		    TRACE(("p = '%s'\n", p ? KEY(p) : ""));
-		    TRACE(("q = '%s'\n", q ? KEY(q) : ""));
-		    TRACE(("r = '%s'\n", r ? KEY(r) : ""));
-		    TRACE(("link LINK( a,q) -> LINK(-a,r) = '%s'\n",
-			   LINK(-a, r) ? KEY(LINK(-a, r)) : ""));
-		    TRACE(("link LINK(-a,r) -> q          = '%s'\n",
-			   KEY(q)));
-		    TRACE(("link LINK(-a,p) -> LINK(a,r)  = '%s'\n",
-			   LINK(a, r) ? KEY(LINK(a, r)) : ""));
-		    TRACE(("link LINK( a,r) -> p          = '%s'\n",
-			   KEY(p)));
-#endif
-		    LINK(a, q) = LINK(-a, r);
-		    LINK(-a, r) = q;
-		    LINK(-a, p) = LINK(a, r);
-		    LINK(a, r) = p;
-
-		    TRACE(("compute balance for '%s', %d vs a=%d\n",
-			   KEY(r), B(r), a));
-
-		    if (B(r) == -a) {
-			B(p) = a;
-			B(q) = 0;
-		    } else if (B(r) == 0) {
-			B(p) = 0;
-			B(q) = 0;
-		    } else {	/* (B(r) == a) */
-			B(p) = 0;
-			B(q) = -a;
-		    }
-
-		    B(r) = 0;
-
-		    a = stack[k - 1].a;
-		    t = stack[k - 1].p;
-
-		    TRACE(("Finish by linking '%s' to %sLINK(%s)\n",
-			   KEY(r),
-			   (a > 0) ? "R" : "L",
-			   KEY(t)));
-
-		    LINK(a, t) = r;
-
-		} else {
-		    TRACE(("CASE 3: single rotation, end\n"));
-
-		    LINK(-a, p) = LINK(a, q);
-		    LINK(a, q) = p;
-
-		    B(q) = -B(p);
-
-		    LINK(stack[k - 1].a, stack[k - 1].p) = q;
+		if (COMPARE(KEY(item), data) == 0) {
 		    break;
 		}
+	    }
 
-		TRACE_TREE("after rebalancing:", funcs);
+	    /* Push direction and node onto stack */
+	    assert(top < MAXSTK);
+	    if (top >= MAXSTK) {
+		TRACE(("(BUG: stack overflow)\n"));
+		return 0;
+	    }
+	    stack[top].dir = (COMPARE(KEY(item), data) < 0);
+	    stack[top].ptr = item;
+	    ++top;
 
+	    /* (A3,A4: move left/right accordingly) */
+	    item = item->links[stack[top - 1].dir];
+	}
+
+	/* D1: Is RLINK null? */
+	/* D1.5: Is LLINK null? */
+	if (item->links[0] == NULL || item->links[1] == NULL) {
+	    /* Which child is not null? */
+	    int dir = item->links[0] == NULL;
+
+	    /* Fix parent */
+	    if (top != 0)
+		stack[top - 1].ptr->links[stack[top - 1].dir] = item->links[dir];
+	    else
+		LINK(1, &(funcs->head)) = item->links[dir];
+
+	    TRACE(("...freed %p @%d\n", BI_KEY(item), __LINE__));
+	    (*funcs->dealloc) (item);
+	    funcs->count -= 1;
+	} else {
+	    /* D2: Find the inorder successor */
+	    BI_NODE *child = item->links[1];
+
+	    /* Save the path */
+	    assert(top < MAXSTK);
+	    if (top >= MAXSTK) {
+		TRACE(("(BUG: stack overflow)\n"));
+		return 0;
+	    }
+	    stack[top].dir = 1;
+	    stack[top].ptr = item;
+	    ++top;
+
+	    /* D3: Find null LLINK */
+	    while (child->links[0] != NULL) {
+		stack[top].dir = 0;
+		stack[top].ptr = child;
+		++top;
+		child = child->links[0];
+	    }
+
+	    /* Swap data */
+	    (*funcs->exchang) (item, child);
+	    /* Unlink successor and adjust parent */
+	    stack[top - 1].ptr->links[stack[top - 1].ptr == item] =
+		child->links[1];
+
+	    /* D4: Free the node */
+	    TRACE(("...freed %p @%d\n", BI_KEY(child), __LINE__));
+	    (*funcs->dealloc) (child);
+	    funcs->count -= 1;
+	}
+
+	/* Walk back up the stack */
+	while (--top >= 0 && !done) {
+	    /* Update balance factors */
+	    stack[top].ptr->balance += (stack[top].dir != 0) ? -1 : +1;
+
+	    /* Terminate, or rebalance as necessary */
+	    if (abs(stack[top].ptr->balance) == 1) {
+		break;
+	    } else if (abs(stack[top].ptr->balance) > 1) {
+		stack[top].ptr = remove_balance(stack[top].ptr,
+						stack[top].dir, &done);
+
+		/* Fix parent */
+		if (top != 0) {
+		    stack[top - 1].ptr->links[stack[top - 1].dir] =
+			stack[top].ptr;
+		} else {
+		    LINK(1, &(funcs->head)) = stack[0].ptr;
+		}
 	    }
 	}
     }
-    return (value != 0);
+
+    TRACE(("...found @%d\n", __LINE__));
+    return 1;
 }
 
 BI_DATA *
@@ -579,7 +517,7 @@ dump_nodes(BI_TREE * funcs, BI_NODE * p, int level)
     if (p) {
 	dump_nodes(funcs, LLINK(p), level + 1);
 	(*funcs->display) (p, level);
-#if DEBUG_BTREE > 0
+#if DEBUG_BTREE > 1
 	if (LLINK(p) != 0
 	    && COMPARE(KEY(LLINK(p)), KEY(p)) > 0)
 	    TRACE((" OOPS:L"));
@@ -621,7 +559,7 @@ btree_printf(BI_TREE * funcs)
 BI_NODE *
 btree_pmatch(BI_NODE * n, const int mode, const char *name)
 {
-    BI_NODE *m;
+    BI_NODE *m = 0;
     size_t len = strlen(name);
 
     while (n != 0) {
@@ -699,7 +637,7 @@ btree_parray(BI_TREE * tree, char *name, size_t len)
 
     if (top != 0) {
 	int i = 0;
-	size_t cnt = btree_pcount(top, name, len);
+	size_t cnt = (size_t) btree_pcount(top, name, len);
 	beginDisplay();
 	nptr = castalloc(const char *, sizeof(const char *) * (cnt + 1));
 	endofDisplay();
@@ -763,13 +701,13 @@ btree_verify(BI_TREE * funcs, BI_NODE * p)
 	    if (balance > 0) {
 		if (balance > 1) {
 		    ok = 0;
-		    TRACE(("OOPS: '%s' is unbalanced\n", KEY(p)));
+		    TRACE(("OOPS: '%s' is unbalanced: %d\n", KEY(p), balance));
 		}
 		balance = 1;
 	    } else {
 		if (balance < -1) {
 		    ok = 0;
-		    TRACE(("OOPS: '%s' is unbalanced\n", KEY(p)));
+		    TRACE(("OOPS: '%s' is unbalanced: %d\n", KEY(p), balance));
 		}
 		balance = -1;
 	    }
@@ -804,7 +742,7 @@ btree_verify(BI_TREE * funcs, BI_NODE * p)
 
 /******************************************************************************/
 #undef typecalloc
-#define typecalloc(cast) (cast *)calloc(sizeof(cast),1)
+#define typecalloc(cast) (cast *)calloc(sizeof(cast), (size_t) 1)
 
 static BI_NODE *
 new_node(BI_DATA * data)
@@ -819,6 +757,7 @@ static void
 old_node(BI_NODE * node)
 {
     beginDisplay();
+    free(BI_KEY(node));
     free(node);
     endofDisplay();
 }
@@ -832,11 +771,20 @@ dpy_node(BI_NODE * a, int level)
     fflush(stdout);
 }
 
+static void
+xcg_node(BI_NODE * a, BI_NODE * b)
+{
+    BI_DATA value = a->value;
+    a->value = b->value;
+    b->value = value;
+}
+
 static BI_TREE text_tree =
 {
     new_node,
     old_node,
     dpy_node,
+    xcg_node,
     0,
     0,
     {
@@ -849,8 +797,8 @@ static BI_TREE text_tree =
 
 #define MAX_VEC 10000
 
-int
-main(int argc, char *argv[])
+static void
+test_btree(FILE *fp)
 {
     BI_DATA temp;
     BI_NODE *top;
@@ -861,28 +809,30 @@ main(int argc, char *argv[])
     const char **list;
 
     memset(&temp, 0, sizeof(temp));
-    for (n = 1; n < argc; n++) {
-	temp.bi_key = argv[n];
-	btree_insert(&text_tree, &temp);
-    }
 
-    btree_printf(&text_tree);
-    btree_verify(&text_tree, RLINK(&(text_tree.head)));
-
-    while (!done && fgets(buffer, sizeof(buffer), stdin) != 0) {
+    while (!done && fgets(buffer, sizeof(buffer), fp) != 0) {
 	char *t = buffer;
 	char *s = t + strlen(t);
 
 	if (s != buffer && *--s == '\n')
 	    *s = 0;
 
+	if (isalpha(*buffer))
+	    printf("%s\n", buffer);
+
 	switch (*t++) {
+	case '[':
+	case ']':
+	case '!':
+	case '#':
+	case '*':
+	    break;
 	default:
 	    printf("Commands are f(ind) i(nsert), d(elete), l(ist), p(rint), r(andom)\n");
 	    break;
 	case 'f':
 	    n = (btree_search(&text_tree, t) != 0);
-	    printf("** found(%s) %d\n", t, n);
+	    printf("** find(%s) %d\n", t, n);
 	    break;
 	case 'i':
 	    temp.bi_key = t;
@@ -936,7 +886,7 @@ main(int argc, char *argv[])
 		    if (!btree_verify(&text_tree, RLINK(&(text_tree.head)))) {
 			printf("OOPS: Random %s-verify '%s' failed\n", name, buffer);
 			btree_printf(&text_tree);
-			return (1);
+			return;
 		    }
 		    printf("** result: %d\n", ok);
 		}
@@ -944,14 +894,34 @@ main(int argc, char *argv[])
 	case 'q':
 	    done = 1;
 	    break;
-	case '#':
-	    break;
 	}
 	btree_verify(&text_tree, RLINK(&(text_tree.head)));
     }
+}
+
+int
+main(int argc, char *argv[])
+{
+    if (argc > 1) {
+	int n;
+
+	for (n = 1; n < argc; ++n) {
+	    FILE *fp = fopen(argv[n], "r");
+	    if (fp != 0) {
+		printf("** read commands from \"%s\"\n", argv[n]);
+		test_btree(fp);
+		fclose(fp);
+	    } else {
+		perror(argv[n]);
+	    }
+	}
+    } else {
+	printf("** read commands from <stdin>\n");
+	test_btree(stdin);
+    }
+
     btree_freeup(&text_tree);
 
-    printf("done!\n\n");
     return 0;
 }
 #endif /* DEBUG_BTREE >= BTREE_DRIVER */
