@@ -1,5 +1,5 @@
 /*
- * $Id: charsets.c,v 1.52 2007/11/25 17:30:05 tom Exp $
+ * $Id: charsets.c,v 1.71 2010/09/08 21:15:08 tom Exp $
  *
  * see
  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/intl/unicode_42jv.asp
@@ -15,7 +15,6 @@
 #if OPT_ICONV_FUNCS
 #include <iconv.h>
 #include <locale.h>
-#include <langinfo.h>
 #endif
 /* *INDENT-OFF* */
 static const UCHAR mark_NONE[]    = { 0x00 };
@@ -30,7 +29,7 @@ static const UCHAR mark_UTF32BE[] = { 0x00, 0x00, 0xfe, 0xff };
 typedef struct {
     BOM_CODES code;
     const UCHAR *mark;
-    unsigned size;
+    size_t size;
 } BOM_TABLE;
 
 #define DATA(name) { bom_##name, mark_##name, sizeof(mark_##name) }
@@ -48,7 +47,7 @@ static const BOM_TABLE bom_table[] = {
 /******************************************************************************/
 
 static void
-allow_decoder(BUFFER *bp, unsigned need)
+allow_decoder(BUFFER *bp, size_t need)
 {
     if (need > bp->decode_utf_len) {
 	bp->decode_utf_len = (need + 1) * 2;
@@ -59,7 +58,7 @@ allow_decoder(BUFFER *bp, unsigned need)
 }
 
 static void
-allow_encoder(BUFFER *bp, unsigned need)
+allow_encoder(BUFFER *bp, size_t need)
 {
     if (need > bp->encode_utf_len) {
 	bp->encode_utf_len = (need + 1) * 2;
@@ -108,22 +107,24 @@ vl_conv_to_utf8(UCHAR * target, UINT source, B_COUNT limit)
 	case 3:
 	    target[2] = (UCHAR) (0x80 | (CH(0) & 0x3f));
 	    target[1] = (UCHAR) (0x80 | (CH(0) >> 6) | ((CH(1) & 0x0f) << 2));
-	    target[0] = (UCHAR) (0xe0 | ((CH(1) & 0xf0) >> 4));
+	    target[0] = (UCHAR) (0xe0 | ((int) (CH(1) & 0xf0) >> 4));
 	    break;
 
 	case 4:
 	    target[3] = (UCHAR) (0x80 | (CH(0) & 0x3f));
 	    target[2] = (UCHAR) (0x80 | (CH(0) >> 6) | ((CH(1) & 0x0f) << 2));
-	    target[1] = (UCHAR) (0x80 | ((CH(1) & 0xf0) >> 4) | ((CH(2) &
-								  0x03) << 4));
-	    target[0] = (UCHAR) (0xf0 | ((CH(2) & 0x1f) >> 2));
+	    target[1] = (UCHAR) (0x80 |
+				 ((int) (CH(1) & 0xf0) >> 4) |
+				 ((int) (CH(2) & 0x03) << 4));
+	    target[0] = (UCHAR) (0xf0 | ((int) (CH(2) & 0x1f) >> 2));
 	    break;
 
 	case 5:
 	    target[4] = (UCHAR) (0x80 | (CH(0) & 0x3f));
 	    target[3] = (UCHAR) (0x80 | (CH(0) >> 6) | ((CH(1) & 0x0f) << 2));
-	    target[2] = (UCHAR) (0x80 | ((CH(1) & 0xf0) >> 4) | ((CH(2) &
-								  0x03) << 4));
+	    target[2] = (UCHAR) (0x80 |
+				 ((int) (CH(1) & 0xf0) >> 4) |
+				 ((int) (CH(2) & 0x03) << 4));
 	    target[1] = (UCHAR) (0x80 | (CH(2) >> 2));
 	    target[0] = (UCHAR) (0xf8 | (CH(3) & 0x03));
 	    break;
@@ -134,7 +135,7 @@ vl_conv_to_utf8(UCHAR * target, UINT source, B_COUNT limit)
 	    target[3] = (UCHAR) (0x80 | (CH(1) >> 4) | ((CH(2) & 0x03) << 4));
 	    target[2] = (UCHAR) (0x80 | (CH(2) >> 2));
 	    target[1] = (UCHAR) (0x80 | (CH(3) & 0x3f));
-	    target[0] = (UCHAR) (0xfc | ((CH(3) & 0x40) >> 6));
+	    target[0] = (UCHAR) (0xfc | ((int) (CH(3) & 0x40) >> 6));
 	    break;
 	}
 	TRACE2(("decode %#08x %02X.%02X.%02X.%02X %d:%.*s\n", source,
@@ -190,7 +191,8 @@ vl_conv_to_utf32(UINT * target, const char *source, B_COUNT limit)
 		break;
 	}
 	if (j != rc) {
-	    rc = 1;
+	    TRACE2(("check failed %d/%d in vl_conv_to_utf32\n", j, rc));
+	    rc = 0;
 	}
     }
 
@@ -213,14 +215,14 @@ vl_conv_to_utf32(UINT * target, const char *source, B_COUNT limit)
 }
 
 static const BOM_TABLE *
-find_mark_info(int code)
+find_mark_info(BOM_CODES code)
 {
     const BOM_TABLE *result = 0;
     unsigned n;
 
     for (n = 0; n < TABLESIZE(bom_table); ++n) {
 	const BOM_TABLE *mp = bom_table + n;
-	if (mp->code == (BOM_CODES) code) {
+	if (mp->code == code) {
 	    result = mp;
 	    break;
 	}
@@ -237,6 +239,98 @@ get_bom(BUFFER *bp)
     return rc;
 }
 
+/*
+ * If we read a file without a byteorder-mark, it is using one of the
+ * le-assumed or be-assumed values for the corresponding mode.  See what the
+ * file-encoding is, and choose a specific byteorder-mark, needed when we
+ * write the data to a file.
+ */
+static BOM_CODES
+inferred_bom2(BUFFER *bp, BOM_CODES code)
+{
+    BOM_CODES result = code;
+
+    switch (result) {
+    case bom_LE_ASSUMED:
+	switch (b_val(bp, VAL_FILE_ENCODING)) {
+	case enc_UTF16:
+	    result = bom_UTF16LE;
+	    break;
+	case enc_UTF32:
+	    result = bom_UTF32LE;
+	    break;
+	default:
+	    result = bom_NONE;
+	    break;
+	}
+	break;
+    case bom_BE_ASSUMED:
+	switch (b_val(bp, VAL_FILE_ENCODING)) {
+	case enc_UTF16:
+	    result = bom_UTF16BE;
+	    break;
+	case enc_UTF32:
+	    result = bom_UTF32BE;
+	    break;
+	default:
+	    result = bom_NONE;
+	    break;
+	}
+	break;
+    case bom_NONE:
+	switch (b_val(bp, VAL_FILE_ENCODING)) {
+	case enc_UTF16:
+	    result = bom_UTF16LE;
+	    break;
+	case enc_UTF32:
+	    result = bom_UTF32LE;
+	    break;
+	}
+	break;
+    default:
+	break;
+    }
+#if OPT_TRACE > 1
+    if (result != code)
+	TRACE2(("inferred_bom(%s) %s\n",
+		byteorder2s(code),
+		byteorder2s(result)));
+#endif
+    return result;
+}
+
+static BOM_CODES
+inferred_bom(BUFFER *bp, const BOM_TABLE * mp)
+{
+    return inferred_bom2(bp, mp->code);
+}
+
+/*
+ * If the buffer has no explicit byteorder-mark, but the encoding is UTF-16
+ * or UTF-32, we still need to know the assumed or implicit byteorder-mark.
+ *
+ * When reading the buffer, and no byteorder-mark is found for UTF-16/-32,
+ * we store an implied BOM in the buffer attributes, and set an assumed
+ * byteorder-mark that the user can see/modify.  If the assumed BOM is reset,
+ * e.g., to auto or none, we can still use the implied BOM, e.g., for writing
+ * the file.
+ */
+static const BOM_TABLE *
+find_mark_info2(BUFFER *bp)
+{
+    const BOM_TABLE *mp = find_mark_info(get_bom(bp));
+
+    if ((mp != 0) &&
+	(mp->size == 0) &&
+	(b_val(bp, VAL_FILE_ENCODING) > enc_UTF8)) {
+	mp = find_mark_info(inferred_bom(bp, mp));
+    } else if (mp == 0 &&
+	       (b_val(bp, VAL_FILE_ENCODING) > enc_UTF8)) {
+	mp = find_mark_info(inferred_bom2(bp, (BOM_CODES) b_val(bp, VAL_BYTEORDER_MARK)));
+    }
+    return mp;
+}
+
 static int
 line_has_mark(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 {
@@ -255,26 +349,30 @@ dump_as_utfXX(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 {
 #define BYTE_OF(k,n) (char) (bp->decode_utf_buf[k] >> ((n) * 8))
     int rc = 0;
-    const BOM_TABLE *mp = find_mark_info(get_bom(bp));
+    const BOM_TABLE *mp = find_mark_info2(bp);
 
     if (mp != 0 && mp->size > 1) {
-	unsigned j = 0;
-	unsigned k = 0;
-	unsigned need = nbuf + strlen(ending);
-	unsigned lend = strlen(ending);
+	size_t j = 0;
+	size_t k = 0;
+	size_t need = (size_t) nbuf + strlen(ending);
+	size_t lend = strlen(ending);
 
 	allow_encoder(bp, need * mp->size);
 	allow_decoder(bp, need);
 	while (j < (unsigned) nbuf) {
 	    int skip = vl_conv_to_utf32(bp->decode_utf_buf + k++,
 					buf + j,
-					nbuf - j);
-	    j += skip;
+					(B_COUNT) ((UINT) nbuf - j));
+	    if (skip == 0)
+		goto finish;
+	    j += (UINT) skip;
 	}
 	while (*ending != 0) {
-	    (void) vl_conv_to_utf32(bp->decode_utf_buf + k++,
-				    ending++,
-				    lend--);
+	    int skip = vl_conv_to_utf32(bp->decode_utf_buf + k++,
+					ending++,
+					(B_COUNT) (lend--));
+	    if (skip == 0)
+		goto finish;
 	}
 	need = k;
 
@@ -283,6 +381,10 @@ dump_as_utfXX(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 	    case bom_NONE:
 		/* FALLTHRU */
 	    case bom_UTF8:
+		/* FALLTHRU */
+	    case bom_LE_ASSUMED:
+		/* FALLTHRU */
+	    case bom_BE_ASSUMED:
 		/* ignored */
 		break;
 	    case bom_UTF16LE:
@@ -307,28 +409,58 @@ dump_as_utfXX(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 		break;
 	    }
 	}
-	rc = j;
+	rc = (int) j;
     }
+  finish:
     return rc;
 #undef BYTE_OF
 }
 
+static void
+set_byteorder_mark(BUFFER *bp, int value)
+{
+    if (value != ENUM_UNKNOWN
+	&& value != global_b_val(VAL_BYTEORDER_MARK)) {
+	make_local_b_val(bp, VAL_BYTEORDER_MARK);
+	set_b_val(bp, VAL_BYTEORDER_MARK, value);
+
+	TRACE(("set_byteorder_mark for '%s' to %s\n",
+	       bp->b_bname,
+	       byteorder2s(b_val(bp, VAL_BYTEORDER_MARK))));
+    }
+}
+
+static void
+set_encoding(BUFFER *bp, int value)
+{
+    if (value != ENUM_UNKNOWN
+	&& value != global_b_val(VAL_FILE_ENCODING)) {
+	make_local_b_val(bp, VAL_FILE_ENCODING);
+	set_b_val(bp, VAL_FILE_ENCODING, value);
+
+	TRACE(("set_encoding for '%s' to %s\n",
+	       bp->b_bname,
+	       encoding2s(b_val(bp, VAL_FILE_ENCODING))));
+    }
+}
+
 static int
 load_as_utf8(BUFFER *bp, LINE *lp)
 {
 #define CH(n) ((UCHAR)(lgetc(lp, n)))
     int rc = FALSE;
-    const BOM_TABLE *mp = find_mark_info(get_bom(bp));
+    const BOM_TABLE *mp = find_mark_info2(bp);
 
     if (mp != 0 && mp->size > 1) {
 	int pass;
 	unsigned j, k;
-	unsigned need = llength(lp);
+	unsigned need = (UINT) llength(lp);
 	unsigned used;
 
 	TRACE2(("load_as_utf8:%d:%s\n", need, lp_visible(lp)));
 	allow_decoder(bp, need);
 	if (bp->decode_utf_buf != 0) {
+	    rc = TRUE;
 	    if (need) {
 		for (j = k = 0; j < need; ++k) {
 		    UCHAR ch = CH(j);
@@ -337,34 +469,38 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 			++j;	/* see remove_crlf_nulls() */
 			continue;
 		    }
-		    switch (mp->code) {
+		    switch (inferred_bom(bp, mp)) {
 		    case bom_NONE:
 			/* FALLTHRU */
 		    case bom_UTF8:
+			/* FALLTHRU */
+		    case bom_LE_ASSUMED:
+			/* FALLTHRU */
+		    case bom_BE_ASSUMED:
 			/* ignored */
 			break;
 		    case bom_UTF16LE:
 			bp->decode_utf_buf[k] = (CH(j)
-						 + (CH(j + 1) << 8));
+						 + (UINT) (CH(j + 1) << 8));
 			break;
 		    case bom_UTF16BE:
 			bp->decode_utf_buf[k] = (CH(j + 1)
-						 + (CH(j) << 8));
+						 + (UINT) (CH(j) << 8));
 			break;
 		    case bom_UTF32LE:
 			bp->decode_utf_buf[k] = (CH(j + 0)
-						 + (CH(j + 1) << 8)
-						 + (CH(j + 2) << 16)
-						 + (CH(j + 3) << 24));
+						 + (UINT) (CH(j + 1) << 8)
+						 + (UINT) (CH(j + 2) << 16)
+						 + (UINT) (CH(j + 3) << 24));
 			break;
 		    case bom_UTF32BE:
 			bp->decode_utf_buf[k] = (CH(j + 3)
-						 + (CH(j + 2) << 8)
-						 + (CH(j + 1) << 16)
-						 + (CH(j + 0) << 24));
+						 + (UINT) (CH(j + 2) << 8)
+						 + (UINT) (CH(j + 1) << 16)
+						 + (UINT) (CH(j + 0) << 24));
 			break;
 		    }
-		    j += mp->size;
+		    j += (UINT) mp->size;
 		}
 		used = k;
 
@@ -373,10 +509,10 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 		    for (j = k = 0; j < used; ++j) {
 			int nn = vl_conv_to_utf8(buffer,
 						 bp->decode_utf_buf[j],
-						 used + 1 - j);
+						 (B_COUNT) (used + 1 - j));
 			if (buffer != 0)
 			    buffer += nn;
-			k += nn;
+			k += (UINT) nn;
 		    }
 		    if (pass == 1) {
 			TRACE2(("need %d, have %d\n", k, lp->l_size));
@@ -388,21 +524,20 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 			     * of the buffer, do not want to allow undo.  Just
 			     * go ahead and reallocate the line's text buffer.
 			     */
-			    if ((ntext = castalloc(char, k)) == NULL) {
+			    if ((ntext = castalloc(char, (size_t) k)) == NULL) {
 				rc = FALSE;
 				break;
 			    }
 			    ltextfree(lp, bp);
 			    lvalue(lp) = ntext;
 			    lp->l_size = k;
-			    llength(lp) = k;
+			    llength(lp) = (int) k;
 			} else {
-			    llength(lp) = k;
+			    llength(lp) = (int) k;
 			}
 		    }
 		}
 	    }
-	    rc = TRUE;
 	} else {
 	    bp->decode_utf_len = 0;
 	}
@@ -419,60 +554,66 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 static void
 remove_crlf_nulls(BUFFER *bp, UCHAR * buffer, B_COUNT * length)
 {
-    const BOM_TABLE *mp = find_mark_info(get_bom(bp));
+    const BOM_TABLE *mp = find_mark_info2(bp);
     UCHAR mark_cr[4];
     UCHAR mark_lf[4];
-    unsigned marklen = 0;
+    size_t marklen = 0;
 
-    memset(mark_cr, 0, sizeof(mark_cr));
-    memset(mark_lf, 0, sizeof(mark_lf));
+    if (mp != 0) {
+	memset(mark_cr, 0, sizeof(mark_cr));
+	memset(mark_lf, 0, sizeof(mark_lf));
 
-    switch (mp->code) {
-    case bom_NONE:
-	/* FALLTHRU */
-    case bom_UTF8:
-	/* ignored */
-	break;
-    case bom_UTF16LE:
-	marklen = 2;
-	mark_cr[0] = '\r';
-	mark_lf[0] = '\n';
-	break;
-    case bom_UTF16BE:
-	marklen = 2;
-	mark_cr[1] = '\r';
-	mark_lf[1] = '\n';
-	break;
-    case bom_UTF32LE:
-	marklen = 4;
-	mark_cr[0] = '\r';
-	mark_lf[0] = '\n';
-	break;
-    case bom_UTF32BE:
-	marklen = 4;
-	mark_cr[3] = '\r';
-	mark_lf[3] = '\n';
-	break;
-    }
-    if (marklen != 0) {
-	B_COUNT dst = 0;
-	B_COUNT src = 0;
-	char skip = 0;
-	while (src < *length) {
-	    if (!memcmp(mark_cr, buffer + src, marklen))
-		skip = '\r';
-	    else if (!memcmp(mark_lf, buffer + src, marklen))
-		skip = '\n';
-	    if (skip) {
-		buffer[dst++] = skip;
-		skip = 0;
-	    } else {
-		memcpy(buffer + dst, buffer + src, marklen);
-		dst += marklen;
+	switch (mp->code) {
+	case bom_NONE:
+	    /* FALLTHRU */
+	case bom_UTF8:
+	    /* FALLTHRU */
+	case bom_LE_ASSUMED:
+	    /* FALLTHRU */
+	case bom_BE_ASSUMED:
+	    /* ignored */
+	    break;
+	case bom_UTF16LE:
+	    marklen = 2;
+	    mark_cr[0] = '\r';
+	    mark_lf[0] = '\n';
+	    break;
+	case bom_UTF16BE:
+	    marklen = 2;
+	    mark_cr[1] = '\r';
+	    mark_lf[1] = '\n';
+	    break;
+	case bom_UTF32LE:
+	    marklen = 4;
+	    mark_cr[0] = '\r';
+	    mark_lf[0] = '\n';
+	    break;
+	case bom_UTF32BE:
+	    marklen = 4;
+	    mark_cr[3] = '\r';
+	    mark_lf[3] = '\n';
+	    break;
+	}
+	if (marklen != 0) {
+	    B_COUNT dst = 0;
+	    B_COUNT src = 0;
+	    char skip = 0;
+	    while (src < *length) {
+		if (!memcmp(mark_cr, buffer + src, marklen))
+		    skip = '\r';
+		else if (!memcmp(mark_lf, buffer + src, marklen))
+		    skip = '\n';
+		if (skip) {
+		    buffer[dst++] = (UCHAR) skip;
+		    skip = 0;
+		} else {
+		    memcpy(buffer + dst, buffer + src, marklen);
+		    dst += marklen;
+		}
+		src += marklen;
 	    }
-	    src += marklen;
+	    *length = dst;
 	}
-	*length = dst - 1;
     }
 }
 
@@ -487,19 +628,21 @@ riddled_buffer(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 {
     int result = 0;
     B_COUNT total = 0;
-    int offset = 0;
-    unsigned j, k;
+    size_t offset = 0;
+    size_t j, k;
 
     if (mp->size && !(mp->size % 2)) {
-	TRACE(("checking if %s / %d-byte\n",
+	TRACE(("checking if %s / %u-byte\n",
 	       byteorder2s(mp->code),
-	       mp->size));
+	       (UINT) mp->size));
 
 	/* Check the line-length.  If it is not a multiple of the pattern
 	 * size, just give up.
 	 */
 	if ((length + offset) % mp->size) {
-	    TRACE(("length %ld vs pattern %d - give up\n", length, mp->size));
+	    TRACE(("length %ld vs pattern %u - give up\n",
+		   length,
+		   (UINT) mp->size));
 	} else {
 	    /*
 	     * Now walk through the line and measure the pattern against it.
@@ -519,7 +662,9 @@ riddled_buffer(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 		}
 	    }
 	}
-	result = (int) ((100.0 * total) / length);
+	result = (int) (length
+			? (((100.0 * (double) total) / (double) length))
+			: 0);
 
 	TRACE(("...%ld/%ld ->%d%%\n", total, length, result));
     }
@@ -527,31 +672,7 @@ riddled_buffer(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 }
 
 static void
-set_byteorder_mark(BUFFER *bp, int value)
-{
-    if (value != ENUM_UNKNOWN
-	&& value != global_b_val(VAL_BYTEORDER_MARK)) {
-	make_local_b_val(bp, VAL_BYTEORDER_MARK);
-	set_b_val(bp, VAL_BYTEORDER_MARK, value);
-    }
-}
-
-static void
-set_encoding(BUFFER *bp, int value)
-{
-    if (value != ENUM_UNKNOWN
-	&& value != global_b_val(VAL_FILE_ENCODING)) {
-	make_local_b_val(bp, VAL_FILE_ENCODING);
-	set_b_val(bp, VAL_FILE_ENCODING, value);
-
-	TRACE(("set_encoding for '%s' to %s\n",
-	       bp->b_bname,
-	       encoding2s(b_val(bp, VAL_FILE_ENCODING))));
-    }
-}
-
-static void
-set_encoding_from_bom(BUFFER *bp, int bom_value)
+set_encoding_from_bom(BUFFER *bp, BOM_CODES bom_value)
 {
     const BOM_TABLE *mp;
     int result;
@@ -661,14 +782,18 @@ decode_bom(BUFFER *bp, UCHAR * buffer, B_COUNT * length)
     }
 
     if (b_val(bp, VAL_BYTEORDER_MARK) > bom_NONE
-	&& (mp = find_mark_info(b_val(bp, VAL_BYTEORDER_MARK))) != 0
+	&& (mp = find_mark_info((BOM_CODES) b_val(bp,
+						  VAL_BYTEORDER_MARK))) != 0
 	&& line_has_mark(mp, buffer, *length)) {
 	for (n = 0; n < *length - mp->size; ++n) {
 	    buffer[n] = buffer[n + mp->size];
 	}
+	while (n < *length) {
+	    buffer[n++] = 0;
+	}
 	*length -= mp->size;
 
-	set_encoding_from_bom(bp, b_val(bp, VAL_BYTEORDER_MARK));
+	set_encoding_from_bom(bp, (BOM_CODES) b_val(bp, VAL_BYTEORDER_MARK));
 	code = TRUE;
     }
     returnCode(code);
@@ -714,7 +839,7 @@ deduce_charset(BUFFER *bp, UCHAR * buffer, B_COUNT * length, int always)
 	   encoding2s(b_val(bp, VAL_FILE_ENCODING))));
 
     bp->implied_BOM = bom_NONE;
-    if (b_is_enc_DEFAULT(bp)) {
+    if (b_is_enc_AUTO(bp)) {
 	unsigned n;
 	int match = 0;
 	int found = -1;
@@ -723,14 +848,27 @@ deduce_charset(BUFFER *bp, UCHAR * buffer, B_COUNT * length, int always)
 	    int check = riddled_buffer(&bom_table[n], buffer, *length);
 	    if (check > match) {
 		match = check;
-		found = n;
+		found = (int) n;
 	    }
 	}
 	if (found > 0 && match >= b_val(bp, VAL_PERCENT_UTF8)) {
 	    bp->implied_BOM = bom_table[found].code;
 	    set_encoding_from_bom(bp, bp->implied_BOM);
 	    TRACE(("...found_charset %s\n",
-		   byteorder2s(get_bom(bp))));
+		   byteorder2s(bp->implied_BOM)));
+
+	    switch (bp->implied_BOM) {
+	    case bom_UTF16BE:
+	    case bom_UTF32BE:
+		set_byteorder_mark(bp, bom_BE_ASSUMED);
+		break;
+	    case bom_UTF16LE:
+	    case bom_UTF32LE:
+		set_byteorder_mark(bp, bom_LE_ASSUMED);
+		break;
+	    default:
+		break;
+	    }
 	    rc = TRUE;
 	} else if (always) {
 	    TRACE(("...try looking for UTF-8\n"));
@@ -752,13 +890,13 @@ deduce_charset(BUFFER *bp, UCHAR * buffer, B_COUNT * length, int always)
 int
 check_utf8(UCHAR * buffer, B_COUNT length)
 {
-    unsigned n;
+    B_COUNT n;
     int check = TRUE;
     int skip = 0;
     int found;
     UINT target;
 
-    for (n = found = 0; n < length - 1; n += skip) {
+    for (n = 0, found = 0; n < length - 1; n += (B_COUNT) skip) {
 	skip = vl_conv_to_utf32(&target,
 				(char *) (buffer + n),
 				length - n);
@@ -796,8 +934,8 @@ write_bom(BUFFER *bp)
     const BOM_TABLE *mp;
     int status = FIOSUC;
 
-    if ((mp = find_mark_info(b_val(bp, VAL_BYTEORDER_MARK))) != 0) {
-	status = ffputline((const char *) mp->mark, mp->size, NULL);
+    if ((mp = find_mark_info((BOM_CODES) b_val(bp, VAL_BYTEORDER_MARK))) != 0) {
+	status = ffputline((const char *) mp->mark, (int) mp->size, NULL);
     }
     return status;
 }
@@ -828,7 +966,7 @@ chgd_byteorder(BUFFER *bp,
 	       int testing)
 {
     if (!testing && !glob_vals) {
-	set_encoding_from_bom(bp, args->local->vp->i);
+	set_encoding_from_bom(bp, (BOM_CODES) args->local->vp->i);
     }
     return TRUE;
 }
@@ -846,8 +984,16 @@ chgd_fileencode(BUFFER *bp,
     if (testing) {
 	;
     } else {
-	if (!glob_vals) {
-	    set_bom_from_encoding(bp, args->local->vp->i);
+	int new_encoding = args->local->vp->i;
+	if (glob_vals) {
+	    if (new_encoding == enc_POSIX) {
+		rebuild_charclasses(0, 127);
+	    } else {
+		rebuild_charclasses(global_g_val(GVAL_PRINT_LOW),
+				    global_g_val(GVAL_PRINT_HIGH));
+	    }
+	} else {
+	    set_bom_from_encoding(bp, new_encoding);
 	}
 	set_bufflags(glob_vals, WFHARD | WFMODE);
     }
