@@ -12,7 +12,7 @@
 */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/estruct.h,v 1.663 2008/05/30 19:09:19 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/estruct.h,v 1.719 2010/09/08 08:44:59 tom Exp $
  */
 
 #ifndef _estruct_h
@@ -94,6 +94,10 @@
 # define SYS_CYGWIN 1
 #endif
 
+#ifdef __MINGW32__
+#define SYS_MINGW 1
+#endif
+
 #ifdef __EMX__
 # define SYS_OS2_EMX 1		/* makefile.emx, or configure-script */
 #endif
@@ -139,6 +143,10 @@
 #define SYS_CYGWIN		0	/* Unix'ed Win32		*/
 #endif
 
+#ifndef SYS_MINGW
+#define SYS_MINGW		0	/* MINGW32			*/
+#endif
+
 #ifndef SYS_MSDOS
 #define SYS_MSDOS		0	/* MS-DOS			*/
 #endif
@@ -172,7 +180,7 @@
  * probably correct for MSC (Microsoft C) and ZTC (Zortech), but I'm not
  * sure of those.  (It implies a lot of ANSI and POSIX behavior.)
  */
-#if CC_TURBO || CC_WATCOM || CC_MSC || CC_DJGPP || SYS_WINNT || CC_CSETPP || CC_MSVC || CC_LCC_WIN32
+#if CC_TURBO || CC_WATCOM || CC_MSC || CC_DJGPP || SYS_WINNT || CC_CSETPP || CC_MSVC || CC_LCC_WIN32 || SYS_OS2_EMX 
 # define CC_NEWDOSCC 1
 #endif
 
@@ -206,6 +214,7 @@
 #endif
 
 #if SYS_WINNT
+# define HAVE_ENVIRON		1
 # define HAVE_PUTENV		1
 #endif
 
@@ -266,9 +275,26 @@
 # endif
 #endif
 
+#if SYS_MSDOS || SYS_OS2 || SYS_VMS || SYS_WINNT
+#ifdef OPT_PERL
+typedef unsigned short	mode_t;
+/* perl's win32.h typedef's mode_t */
+#else
+#define mode_t int
+#endif
+#endif
+
+#if SYS_OS2 && !SYS_OS2_EMX
+#define inline /* nothing */
+#endif
+
+#if SYS_OS2 || SYS_OS2_EMX || SYS_WINNT || SYS_MSDOS
+#define MISSING_EXTERN_CRYPT 1
+#endif
+
 #endif /* HAVE_CONFIG_H */
 
-#if defined(WIN32) || CC_TURBO
+#if (defined(WIN32) || CC_TURBO || defined(WIN32_LEAN_AND_MEAN))
 #include "w32vile.h"
 #endif
 
@@ -479,10 +505,6 @@
 #define NO_WIDGETS 0
 #endif
 
-#ifndef OL_WIDGETS
-#define OL_WIDGETS 0
-#endif
-
 /*
  * Constants for ifdef'ing out chunks of code
  */
@@ -499,7 +521,7 @@
 #define	OPT_COLOR (DISP_ANSI || IBM_VIDEO || DISP_TERMCAP || DISP_CURSES || DISP_X11)
 #define	OPT_16_COLOR (IBM_VIDEO || DISP_TERMCAP || DISP_CURSES || DISP_X11)
 
-#define OPT_DUMBTERM (DISP_TERMCAP || DISP_CURSES || DISP_VMSVT)
+#define OPT_DUMBTERM (DISP_TERMCAP || DISP_CURSES || DISP_VMSVT || DISP_X11)
 
 /* Feature turnon/turnoff */
 #define OPT_CACHE_VCOL  !SMALLER /* cache mk_to_vcol() starting point          */
@@ -599,14 +621,16 @@
 /* systems with MSDOS-like filename syntax */
 #define OPT_MSDOS_PATH  (SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX)
 #define OPT_CASELESS	(SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX || SYS_CYGWIN || SYS_VMS)
-#define OPT_UNC_PATH	(SYS_WINNT)
+#define OPT_UNC_PATH	(SYS_WINNT || SYS_CYGWIN)
 
 /* individual features that are (normally) controlled by SMALLER */
+#define OPT_AUTOCOLOR	(!SMALLER && OPT_COLOR)	/* autocolor support */
 #define OPT_BNAME_CMPL  !SMALLER		/* name-completion for buffers */
 #define OPT_B_LIMITS    !SMALLER		/* left-margin */
 #define OPT_CURTOKENS   !SMALLER		/* cursor-tokens mode */
 #define OPT_ENUM_MODES  !SMALLER		/* fixed-string modes */
 #define OPT_EVAL        !SMALLER		/* expression-evaluation */
+#define OPT_EXTRA_COLOR (!SMALLER && OPT_COLOR)	/* modeline, etc */
 #define OPT_FILEBACK    (!SMALLER && !SYS_VMS)	/* file backup style */
 #define OPT_FINDERR     !SMALLER		/* finderr support. */
 #define OPT_FLASH       !SMALLER		/* visible-bell */
@@ -621,7 +645,7 @@
 #define OPT_MACRO_ARGS	(!SMALLER && OPT_EVAL)	/* macro argument parsing */
 #define OPT_MLFORMAT    !SMALLER		/* modeline-format */
 #define OPT_MODELINE    !SMALLER		/* vi-style modeline-support */
-#define OPT_MULTIBYTE   (!SMALLER && OPT_LOCALE) /* experimental multibyte support */
+#define OPT_MULTIBYTE   (!SMALLER && OPT_LOCALE) /* multibyte support */
 #define OPT_NAMEBST     !SMALLER		/* name's stored in a bst */
 #define OPT_ONLINEHELP  !SMALLER		/* short per-command help */
 #define OPT_POPUPCHOICE !SMALLER		/* popup-choices mode */
@@ -678,7 +702,9 @@
 
 	/* menus */
 #define	OPT_MENUS	(!SMALLER && DISP_X11 && (MOTIF_WIDGETS||ATHENA_WIDGETS))
-#define OPT_MENUS_COLORED 0	/* (MOTIF_WIDGETS && OPT_MENUS) */
+#ifndef OPT_MENUS_COLORED
+#define OPT_MENUS_COLORED 0
+#endif
 
 	/* icons */
 #define OPT_X11_ICON	DISP_X11 /* use compiled-in X icon */
@@ -694,42 +720,38 @@
 #define OWN_SELECTION()	/*EMPTY*/
 #endif
 
-/*
- * Unix vi-style file encryption is available only on some platforms
- */
-#ifdef HAVE_CRYPT
-#define OPT_ENCRYPT     !SMALLER
-#else
-#define OPT_ENCRYPT     0
-#endif
-
 /*
  * Symbols that turn on tables related to OPT_ENUM_MODES in nefsms.h
  */
 #define OPT_COLOR_SCHEMES          (OPT_ENUM_MODES && !SMALLER && OPT_COLOR)
 
-#define OPT_BACKUP_CHOICES	   (OPT_ENUM_MODES && OPT_FILEBACK)
-#define OPT_BOOL_CHOICES	   !SMALLER
+#define OPT_ACCESS_CHOICES         !SMALLER
+#define OPT_BACKUP_CHOICES         (OPT_ENUM_MODES && OPT_FILEBACK)
+#define OPT_BOOL_CHOICES           !SMALLER
 #define OPT_BYTEORDER_MARK_CHOICES OPT_MULTIBYTE
-#define OPT_CHARCLASS_CHOICES	   OPT_SHOW_CTYPE
-#define OPT_COLOR_CHOICES	   (OPT_ENUM_MODES && OPT_COLOR)
+#define OPT_CHARCLASS_CHOICES      OPT_SHOW_CTYPE
+#define OPT_CMD_ENCODING_CHOICES   OPT_MULTIBYTE
+#define OPT_COLOR_CHOICES          (OPT_ENUM_MODES && OPT_COLOR)
 #define OPT_CURTOKENS_CHOICES      OPT_CURTOKENS
 #define OPT_DIRECTIVE_CHOICES      !SMALLER
+#define OPT_ENCRYPT                !SMALLER
 #define OPT_FILE_ENCODING_CHOICES  OPT_MULTIBYTE
 #define OPT_FORBUFFERS_CHOICES     !SMALLER
-#define OPT_HILITE_CHOICES	   (OPT_ENUM_MODES && OPT_HILITEMATCH)
+#define OPT_HILITE_CHOICES         (OPT_ENUM_MODES && OPT_HILITEMATCH)
 #define OPT_LOOKUP_CHOICES         !SMALLER
 #define OPT_MMQUALIFIERS_CHOICES   OPT_MAJORMODE
 #define OPT_PARAMTYPES_CHOICES     OPT_MACRO_ARGS
 #define OPT_PATH_CHOICES           !SMALLER
-#define OPT_POPUP_CHOICES	   (OPT_ENUM_MODES && OPT_POPUPCHOICE)
+#define OPT_POPUP_CHOICES          (OPT_ENUM_MODES && OPT_POPUPCHOICE)
 #define OPT_READERPOLICY_CHOICES   !SMALLER
 #define OPT_RECORDATTRS_CHOICES    (OPT_ENUM_MODES && SYS_VMS)
 #define OPT_RECORDFORMAT_CHOICES   (OPT_ENUM_MODES && SYS_VMS)
 #define OPT_RECORDSEP_CHOICES      !SMALLER
 #define OPT_SHOWFORMAT_CHOICES     !SMALLER
+#define OPT_TITLE_ENCODING_CHOICES OPT_MULTIBYTE
 #define OPT_VIDEOATTRS_CHOICES     (OPT_ENUM_MODES && OPT_COLOR_SCHEMES)
 #define OPT_VTFLASHSEQ_CHOICES     (OPT_ENUM_MODES && VTFLASH_HOST && OPT_FLASH)
+#define OPT_XCOLORS_CHOICES        (OPT_ENUM_MODES && OPT_EXTRA_COLOR)
 
 /*
  * Special characters used in globbing
@@ -738,7 +760,7 @@
 #define	GLOB_SINGLE	'?'
 #define	GLOB_ELLIPSIS	"..."	/* implemented on VMS-only */
 #define	GLOB_RANGE	"[]"
-#define GLOB_ENVIRON	"$"	/* unimplemented */
+#define	GLOB_ENVIRON	"$"	/* unimplemented */
 #define	GLOB_NEGATE	"^!"
 
 /*
@@ -990,14 +1012,12 @@ extern void endofDisplay(void);
  * we use those definitions to avoid trouble when using OS/2 include
  * files.
  */
-# include <os2def.h>
-#else
-# if !defined(WIN32)
+#  include <os2def.h>
+#elif !(defined(WIN32) || defined(WIN32_LEAN_AND_MEAN))
 #  define UCHAR  unsigned char
 #  define UINT   unsigned int
 #  define USHORT unsigned short
 #  define ULONG  unsigned long
-# endif
 #endif
 
 /*	internal constants	*/
@@ -1099,7 +1119,7 @@ extern void endofDisplay(void);
 #define mod_NOMOD  (~(mod_KEY|mod_SHIFT|mod_CTRL|mod_ALT))
 #endif
 
-#define kcod2key(c)	((c) & (UINT)(iBIT(MaxCBits)-1)) /* strip off the above prefixes */
+#define kcod2key(c)	(int)((UINT)(c) & (UINT)(iBIT(MaxCBits)-1)) /* strip off the above prefixes */
 #define	isSpecial(c)	(((UINT)(c) & (UINT)~(iBIT(MaxCBits)-1)) != 0)
 
 #define	char2int(c)	((int)kcod2key(c))	/* mask off sign-extension, etc. */
@@ -1118,6 +1138,10 @@ extern void endofDisplay(void);
 #define isEscaped(s)	((s)[-1] == BACKSLASH)
 #define isTab(c)	((c) == '\t')
 
+#define isErrorVal(s)	((s) == error_val)
+#define isLegalVal(s)	((s) != 0 && !isErrorVal(s))
+#define isLegalExp(s,x) ((s = (x)) != 0 && !isErrorVal(s))
+
 /* protect against losing namespaces */
 #undef	FALSE
 #undef	TRUE
@@ -1171,6 +1195,8 @@ typedef struct {
 typedef enum {
 	bom_NONE = 0
 	, bom_UTF8
+	, bom_LE_ASSUMED
+	, bom_BE_ASSUMED
 	, bom_UTF16LE
 	, bom_UTF16BE
 	, bom_UTF32LE
@@ -1183,6 +1209,22 @@ typedef enum {
 	, CT_REGEX
 } CURTOKENS;
 
+/*
+ * Enumerated type for extra-colors.
+ */
+typedef enum {
+	XCOLOR_NONE = 0
+	, XCOLOR_ENUM
+	, XCOLOR_HYPERTEXT
+	, XCOLOR_ISEARCH
+	, XCOLOR_MODELINE
+	, XCOLOR_NUMBER
+	, XCOLOR_REGEX
+	, XCOLOR_STRING
+	, XCOLOR_WARNING
+	, XCOLOR_MAX		/* last entry, for array-size */
+} XCOLOR_CODES;
+
 /*	Directive definitions	*/
 
 typedef	enum {
@@ -1244,9 +1286,13 @@ typedef enum {
 /*
  * True if the buffer contents are in UTF-8 (or -16, -32).
  */
+#if OPT_MULTIBYTE
 #define b_is_utfXX(bp)       ((b_val(bp, VAL_FILE_ENCODING) >= enc_UTF8) \
 			   || (b_val(bp, VAL_FILE_ENCODING) == enc_LOCALE \
 			    && vl_encoding >= enc_UTF8))
+#else
+#define b_is_utfXX(bp)       0
+#endif
 
 /*
  * Resolve file-encoding and the special symbols "auto" and "locale" to
@@ -1309,7 +1355,8 @@ typedef struct {
 } PARAM_INFO;
 
 typedef enum {
-	RS_AUTO = 0
+	RS_AUTO = -1
+	, RS_DEFAULT = 0
 	, RS_LF			/* unix */
 	, RS_CR			/* mac */
 	, RS_CRLF		/* dos */
@@ -1322,9 +1369,9 @@ typedef enum {
 } READERPOLICY_CHOICES;
 
 #if CRLF_LINES
-#define RS_DEFAULT RS_CRLF
+#define RS_SYS_DEFAULT RS_CRLF
 #else
-#define RS_DEFAULT RS_LF
+#define RS_SYS_DEFAULT RS_LF
 #endif
 
 typedef enum {
@@ -1345,6 +1392,7 @@ typedef enum {
 #define FL_STARTPATH iBIT(6)	/* look in environment $VILE_STARTUP_PATH */
 #define FL_PATH      iBIT(7)	/* look in environment $PATH */
 #define FL_LIBDIR    iBIT(8)	/* look in environment $VILE_LIBDIR_PATH */
+#define FL_ALWAYS    iBIT(9)	/* file is not a script, but data */
 
 #define FL_ANYWHERE  (FL_CDIR|FL_HOME|FL_EXECDIR|FL_STARTPATH|FL_PATH|FL_LIBDIR)
 
@@ -1357,6 +1405,8 @@ typedef enum {
 #endif
 
 #define LOCATE_EXEC   (FL_PATH | FL_LIBDIR | FL_EXECABLE)
+#define LOCATE_HELP   (FL_READABLE | FL_ALWAYS | FL_ANYWHERE)
+#define LOCATE_TAGS   (FL_READABLE | FL_ALWAYS | FL_CDIR)
 
 /* definitions for name-completion */
 #define	NAMEC		name_cmpl /* char for forcing name-completion */
@@ -1705,11 +1755,11 @@ typedef struct	LINE {
 	/*
 	 * Macros for referencing fields in the LINE struct.
 	 */
-#define lgetc(lp, n)		char2int(lvalue(lp)[(n)])
+#define lgetc(lp, n)		CharOf(lvalue(lp)[(n)])
 #define lputc(lp, n, c)		(lvalue(lp)[(n)]=(char)(c))
 #define lvalue(lp)		((lp)->l_text)
 #define llength(lp)		((lp)->l_used)
-#define line_length(lp)		((B_COUNT)(llength(lp))+len_rs) /* counting recordsep */
+#define line_length(lp)		((B_COUNT)llength(lp)+len_rs) /* counting recordsep */
 
 #define liscopied(lp)		((lp)->l.l_undo_cookie == current_undo_cookie)
 #define lsetcopied(lp)		((lp)->l.l_undo_cookie = current_undo_cookie)
@@ -1717,12 +1767,12 @@ typedef struct	LINE {
 
 #define lismarked(lp)		((lp)->l.l_flag & LGMARK)
 #define lsetmarked(lp)		((lp)->l.l_flag |= LGMARK)
-#define lsetnotmarked(lp)	((lp)->l.l_flag &= ~LGMARK)
+#define lsetnotmarked(lp)	((lp)->l.l_flag &= (USHORT) ~LGMARK)
 #define lflipmark(lp)		((lp)->l.l_flag ^= LGMARK)
 
 #define listrimmed(lp)		((lp)->l.l_flag & LTRIMMED)
 #define lsettrimmed(lp)		((lp)->l.l_flag |= LTRIMMED)
-#define lsetnottrimmed(lp)	((lp)->l.l_flag &= ~LTRIMMED)
+#define lsetnottrimmed(lp)	((lp)->l.l_flag &= (USHORT) ~LTRIMMED)
 #define lsetclear(lp)		((lp)->l.l_flag = (lp)->l.l_undo_cookie = 0)
 
 #define lisreal(lp)		(llength(lp) >= 0)
@@ -1834,9 +1884,9 @@ typedef UCHAR VIDEO_ATTR;
  */
 #define GROW(ptr, type, oldsize, newsize) \
 { \
-	int tmpold = oldsize; \
+	UINT tmpold = (UINT) oldsize; \
 	type *tmpp; \
-	tmpp = typeallocn(type, newsize); \
+	tmpp = typeallocn(type, (UINT) newsize); \
 	if (tmpp == NULL) \
 		return FALSE; \
  \
@@ -1847,7 +1897,7 @@ typedef UCHAR VIDEO_ATTR;
 		tmpold = 0; \
 	} \
 	ptr = tmpp; \
-	(void) memset ((char *)(ptr+tmpold), 0, (newsize - tmpold) * sizeof(type)); \
+	(void) memset ((char *)(ptr+tmpold), 0, ((UINT) newsize - tmpold) * sizeof(type)); \
 }
 
 /*
@@ -1910,6 +1960,7 @@ struct VALNAMES {
 		ChgdFunc side_effect;
 };
 /* the values of VALNAMES->type */
+#define VALTYPE_UNKNOWN  -1
 #define VALTYPE_INT    0
 #define VALTYPE_STRING 1
 #define VALTYPE_BOOL   2
@@ -2058,6 +2109,16 @@ typedef struct {
 typedef int FUID;
 #endif
 
+#if (OPT_AUTOCOLOR || OPT_ELAPSED) && !defined(VL_ELAPSED)
+#ifdef HAVE_GETTIMEOFDAY
+#define VL_ELAPSED struct timeval
+#elif SYS_WINNT
+/* this is in w32vile.h */
+#else
+#define VL_ELAPSED time_t
+#endif
+#endif
+
 /*
  * Text is kept in buffers.  A buffer header, described below, exists
  * for every buffer in the system.  The buffers are kept in a big
@@ -2113,6 +2174,10 @@ typedef struct	BUFFER {
 	char	*b_fname;		/* File name			*/
 	int	b_fnlen;		/* length of filename		*/
 	char	b_bname[NBUFN];		/* Buffer name			*/
+#if OPT_AUTOCOLOR
+	double	last_autocolor_time;	/* millisecond for last autocolor */
+	long	next_autocolor_time;	/* count for skipping autocolor */
+#endif
 #if OPT_CURTOKENS
 	struct VAL buf_fname_expr;	/* $buf-fname-expr		*/
 #endif
@@ -2146,9 +2211,9 @@ typedef struct	BUFFER {
 #if OPT_MULTIBYTE
 	BOM_CODES implied_BOM;		/* fix decode/encode if BOM missing */
 	UINT	*decode_utf_buf;	/* workspace for decode_charset() */
-	UINT	decode_utf_len;
+	size_t	decode_utf_len;
 	char	*encode_utf_buf;	/* workspace for encode_charset() */
-	UINT	encode_utf_len;
+	size_t	encode_utf_len;
 #endif
 #if OPT_PERL || OPT_TCL || OPT_PLUGIN
 	void *	b_api_private;		/* pointer to private perl, tcl, etc.
@@ -2312,9 +2377,14 @@ typedef struct	BUFFER {
 #define HILITE_DIRTY	2
 #endif
 
+/* macros related to external record separator */
 #define len_record_sep(bp)	((bp)->b_recordsep_len)
 #define get_record_sep(bp)	((bp)->b_recordsep_str)
 
+/* macros related to internal record separator */
+#define use_record_sep(bp)	((b_val(bp, VAL_RECORD_SEP) == RS_CR) ? '\r' : '\n')
+#define is_record_sep(bp,c)	((c) == use_record_sep(bp))
+
 /* macro for iterating over the marks associated with the current buffer */
 
 #if OPT_PERL || OPT_TCL || OPT_PLUGIN
@@ -2368,6 +2438,8 @@ extern MARK *api_mark_iterator(BUFFER *bp, int *iter);
     } one_time
 #endif /* OPT_VIDEO_ATTRS */
 
+typedef int WIN_ID;
+
 /*
  * There is a window structure allocated for every active display window. The
  * windows are kept in a big list, in top to bottom screen order, with the
@@ -2396,13 +2468,19 @@ typedef struct	WINDOW {
 	int	w_ruler_col;
 #endif
 #if OPT_PERL || OPT_TCL
-	ULONG	w_id;			/* Unique window id */
+	WIN_ID	w_id;			/* Unique window id */
 #endif
 }	WINDOW;
 
 #define is_visible_window(wp) ((wp)->w_toprow >= 0)
 #define is_fake_window(wp) (!(is_visible_window(wp)))
 
+#if OPT_MULTIBYTE
+#define is_delinked_bp(bp) ((bp) == bminip || (bp) == btempp)
+#else
+#define is_delinked_bp(bp) ((bp) == bminip)
+#endif
+
 #define	for_each_window(wp) for (wp = wheadp; wp; wp = wp->w_wndp)
 #define for_each_visible_window(wp) \
 		for_each_window(wp) if (is_visible_window(wp))
@@ -2430,6 +2508,15 @@ typedef struct	WINDOW {
 #define MK Mark
 #endif
 
+/*
+ * Data to save/restore when switching temporarily to the minibuffer for I/O.
+ */
+typedef struct {
+	BUFFER *save_bp;
+	WINDOW *save_wp;
+	MARK save_mk;
+} KBD_DATA;
+
 	/* we use left-margin for protecting the prefix-area of [Registers]
 	 * from cut/paste selection.
 	 */
@@ -2464,7 +2551,8 @@ typedef struct	{
 	int	rows;			/* current row count		*/
 	int	maxcols;		/* max column count		*/
 	int	cols;			/* current column count		*/
-	ENC_CHOICES encoding;		/* tell what the display can do	*/
+	void	(*set_enc)(ENC_CHOICES); /* tell what the display can do */
+	ENC_CHOICES (*get_enc)(void);	/* tell what the display can do	*/
 	void	(*open) (void);		/* Open terminal at the start.	*/
 	void	(*close) (void);	/* Close terminal at end.	*/
 	void	(*kopen) (void);	/* keyboard open		*/
@@ -2503,7 +2591,7 @@ typedef struct	{
 	void	(*mevent) (void);	/* mouse event			*/
 }	TERM;
 
-#define term_is_utfXX()         ((int) term.encoding >= enc_UTF8)
+#define term_is_utfXX()         (term.get_enc() >= enc_UTF8)
 
 #if DISP_CURSES && defined(HAVE_ADDNWSTR)
 #define WIDE_CURSES 1
@@ -2596,11 +2684,12 @@ typedef	struct {
 #define INIT_UNION(n) {n}
 #endif /* CC_CANNOT_INIT_UNIONS */
 	CMDFLAGS c_flags;	/* what sort of command is it? */
+	const char **c_alias;	/* all names by which this is known */
 #if OPT_MACRO_ARGS
 	PARAM_INFO *c_args;	/* if nonnull, lists types of parameters */
 #endif
 #if OPT_TRACE
-	const char *c_name;
+	const char *c_name;	/* preferred name, for tracing */
 #endif
 #if OPT_ONLINEHELP
 	const char *c_help;	/* short help message for the command */
@@ -2738,11 +2827,12 @@ typedef struct {
 
 /* these flags determine the type of cu.* */
 #define typBIT(n) cmdBIT(n+28)	/* ...to simplify adding bits */
-/* bits 27-28 */
+/* bits 27-29 */
 #define CMD_FUNC 0L		/* this is the default (CmdFunc) */
 #define CMD_PROC typBIT(0)	/* named procedure (BUFFER *) */
-#define CMD_PERL typBIT(1)	/* perl subroutine (AV *) */
-#define CMD_TYPE (CMD_PROC | CMD_PERL) /* type mask */
+#define CMD_OPER typBIT(1)	/* user-defined operator */
+#define CMD_PERL typBIT(2)	/* perl subroutine (AV *) */
+#define CMD_TYPE (CMD_PROC | CMD_OPER | CMD_PERL) /* type mask */
 
 #define SPECIAL_BANG_ARG -42	/* arg passed as 'n' to functions which
 					were invoked by their "xxx!" name */
@@ -2845,19 +2935,27 @@ extern void exit (int code);
 extern void _exit (int code);
 #endif	/* HAVE_STDLIB_H */
 
+#if !(defined(intptr_t))
+/* get intptr_t */
+#if defined(HAVE_INTTYPES_H)
+#include <inttypes.h>
+#elif defined(HAVE_STDINT_H)
+#include <stdint.h>
+#elif !(defined(WIN32) || defined(WIN32_LEAN_AND_MEAN))
+typedef long intptr_t;
+#endif
+#endif
+
 /* array/table size */
 #define	TABLESIZE(v)	(sizeof(v)/sizeof(v[0]))
 
 #define PERCENT(num,den) ((den) ? (int)((100.0 * (num))/(den)) : 100)
 
 /* Quiet compiler warnings on places where we're being blamed incorrectly,
- * e.g., for casting away const, or for alignment problems.  It's generally
- * legal in c89 to cast a pointer to long w/o loss of precision.
- *
- * FIXME: c99 may require a wider type.
+ * e.g., for casting away const, or for alignment problems.
  */
 #ifdef __GNUC__
-#define TYPECAST(type,ptr) (type*)((long)(ptr))
+#define TYPECAST(type,ptr) (type*)((intptr_t)(ptr))
 #else
 #define TYPECAST(type,ptr) (type*)(ptr)
 #endif
@@ -2870,7 +2968,7 @@ extern void ExitProgram(int code);
 #define ExitProgram(code)   oleauto_exit(code)
 #else
 #if !defined(VILE_ERROR_ABORT)
-#define	ExitProgram(code)	exit(code)
+#define	ExitProgram(code)	exit_program(code)
 #endif
 #endif
 
@@ -2926,7 +3024,10 @@ extern void ExitProgram(int code);
 /*
  * If we have va_copy(), use it.
  */
-#if defined(va_copy) || defined(HAVE_VA_COPY)
+#if defined(HAVE___VA_COPY)
+#define begin_va_copy(dst,src)	__va_copy(dst, src)
+#define end_va_copy(dst)	va_end(dst)
+#elif defined(va_copy) || defined(HAVE_VA_COPY)
 #define begin_va_copy(dst,src)	va_copy(dst, src)
 #define end_va_copy(dst)	va_end(dst)
 #else
@@ -3046,24 +3147,27 @@ extern void show_elapsed(void);
 /* Normally defined in "trace.h" */
 #ifndef TRACE
 #define TRACE(p) /* nothing */
-#define returnCode(c)   return(c)
-#define returnPtr(c)    return(c)
-#define returnString(c) return(c)
-#define returnVoid()    return
+#define returnCString(c) return(c)
+#define returnCode(c)    return(c)
+#define returnPtr(c)     return(c)
+#define returnString(c)  return(c)
+#define returnVoid()     return
 #endif
 
 #if OPT_TRACE > 1
-#define TRACE2(params)   TRACE(params)
-#define return2Code(c)   returnCode(c)
-#define return2Ptr(c)    returnPtr(c)
-#define return2String(c) returnString(c)
-#define return2Void()    returnVoid()
+#define TRACE2(params)    TRACE(params)
+#define return2CString(c) returnCString(c)
+#define return2Code(c)    returnCode(c)
+#define return2Ptr(c)     returnPtr(c)
+#define return2String(c)  returnString(c)
+#define return2Void()     returnVoid()
 #else
 #define TRACE2(params) /*nothing*/
-#define return2Code(c)   return(c)
-#define return2Ptr(c)    return(c)
-#define return2String(c) return(c)
-#define return2Void()    return
+#define return2CString(c) return(c)
+#define return2Code(c)    return(c)
+#define return2Ptr(c)     return(c)
+#define return2String(c)  return(c)
+#define return2Void()     return
 #endif
 
 #if OPT_EVAL || OPT_DEBUGMACROS
@@ -3087,6 +3191,12 @@ extern void show_elapsed(void);
 #endif
 #endif /* gcc workarounds */
 
+#if defined(__GNUC__) && defined(_FORTIFY_SOURCE)
+#define IGNORE_RC(func) ignore_unused = (int) func
+#else
+#define IGNORE_RC(func) (void) func
+#endif /* gcc workarounds */
+
 /*
  * Local prototypes (must follow NO_LEAKS definition)
  */
