@@ -2,7 +2,7 @@
  *	eval.c -- function and variable evaluation
  *	original by Daniel Lawrence
  *
- * $Header: /usr/build/vile/vile/RCS/eval.c,v 1.379 2008/06/01 17:41:02 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/eval.c,v 1.429 2010/06/09 20:56:25 tom Exp $
  *
  */
 
@@ -56,49 +56,134 @@ static void free_vars_cmpl(void);
 /*--------------------------------------------------------------------------*/
 #if OPT_SHOW_CTYPE
 
-/* list the current character-classes into the current buffer */
-/* ARGSUSED */
+static int show_ctypes_f;
+static int show_ctypes_n;
+
+#define same_string(a,b) (((a) && (b) && !strcmp(a,b)) || !((a) || (b)))
+
 static void
-makectypelist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
+one_ctype_row(VL_CTYPE2 * enc, int ch)
 {
-    UINT i, j;
-    CHARTYPE k;
+    UINT j;
+    CHARTYPE test;
+    CHARTYPE used;
     const char *s;
 
 #if OPT_MULTIBYTE
-    /*
-     * Keep the buffer encoding 8-bit to make it work with the character
-     * classes for the narrow locale.
-     */
+    int rc;
+    char temp[10];
+    int use_locale = same_string(enc->locale, vl_wide_enc.locale);
+#else
+    (void) enc;
+#endif
+
+    bprintf("\n%d\t", (int) ch);
+    if ((ch == '\n') || (ch == '\t')) {		/* vtlistc() may not do these */
+	bprintf("^%c", (int) ('@' | ch));
+    } else {
+#if OPT_MULTIBYTE
+	if (use_locale
+	    && (rc = vl_conv_to_utf8((UCHAR *) temp,
+				     (UINT) ch,
+				     sizeof(temp))) > 1) {
+	    bputsn(temp, rc);
+	} else
+#endif
+	    bputc(ch);
+    }
+    bputc('\t');
+
+#if OPT_MULTIBYTE
+    used = vl_ctype_bits(ch, use_locale);
+#else
+    used = vl_chartypes_[ch];
+#endif
+
+    for (j = 0; j != vl_UNUSED; j++) {
+	if ((s = choice_to_name(&fsm_charclass_blist, (int) j)) != 0) {
+
+	    test = (CHARTYPE) (1 << j);
+
+	    if (j != 0 && *s != EOS)
+		bputc(' ');
+	    bprintf("%*s",
+		    (int) strlen(s),
+		    ((used & test)
+		     ? s
+		     : "-"));
+	}
+    }
+}
+
+/* list the current character-classes into the current buffer */
+/* ARGSUSED */
+static void
+make_ctype_list(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
+{
+    int first = (show_ctypes_n * 256);
+    int last = first + N_chars;
+    int i;
+    VL_CTYPE2 *enc = &vl_real_enc;
+
+#if OPT_MULTIBYTE
     make_local_b_val(curbp, VAL_FILE_ENCODING);
-    set_b_val(curbp, VAL_FILE_ENCODING, enc_8BIT);
+    if ((vl_encoding > enc_8BIT) && show_ctypes_f) {
+	set_b_val(curbp, VAL_FILE_ENCODING, vl_encoding);
+	enc = &vl_wide_enc;
+    } else {
+	/*
+	 * Keep the buffer encoding 8-bit to make it work with the character
+	 * classes for the narrow locale.
+	 */
+	set_b_val(curbp, VAL_FILE_ENCODING, enc_8BIT);
+    }
+#endif
+
+#if defined(WMDSHOWCHAR) && OPT_MULTIBYTE
+    /*
+     * If "showchar" is set, and the current buffer is not this one, show
+     * the row for the current character as well.  Slow, but as noted, useful
+     * for debugging.
+     */
+    if (w_val(curwp, WMDSHOWCHAR)) {
+	WINDOW *wp = recomputing_win();
+	BUFFER *bp = recomputing_buf();
+
+	if (wp != 0 && bp != 0 && (curbp != bp)) {
+	    VL_CTYPE2 *my_enc = ((b_val(bp, VAL_FILE_ENCODING) >= enc_UTF8)
+				 ? &vl_wide_enc
+				 : &vl_real_enc);
+	    BUFFER *savebp = curbp;
+	    int savech;
+
+	    bprintf("-- Current Character in %s, locale %s (%s) ",
+		    bp->b_bname,
+		    (my_enc->locale
+		     ? my_enc->locale
+		     : NONNULL(enc->locale)),
+		    NONNULL(encoding2s(b_val(bp, VAL_FILE_ENCODING))));
+	    bpadc('-', term.cols - DOT.o);
+	    bputc('\n');
+
+	    curbp = bp;
+	    savech = char_at_mark(wp->w_dot);
+	    curbp = savebp;
+
+	    one_ctype_row(my_enc, savech);
+	    bputc('\n');
+	    bputc('\n');
+	}
+    }
 #endif
 
     bprintf("--- Printable Characters for locale %s (%s) ",
-	    NONNULL(vl_real_enc.locale),
-	    NONNULL(vl_real_enc.encoding));
+	    NONNULL(enc->locale),
+	    NONNULL(enc->encoding));
     bpadc('-', term.cols - DOT.o);
     bputc('\n');
 
-    for (i = 0; i < N_chars; i++) {
-	bprintf("\n%d\t", (int) i);
-	if ((i == '\n') || (i == '\t'))		/* vtlistc() may not do these */
-	    bprintf("^%c", (int) ('@' | i));
-	else
-	    bprintf("%c", (int) i);
-	bputc('\t');
-	for (j = 0; j != vl_UNUSED; j++) {
-	    if ((s = choice_to_name(&fsm_charclass_blist, j)) != 0) {
-		k = (1 << j);
-		if (j != 0)
-		    bputc(' ');
-		bprintf("%*s",
-			(int) strlen(s),
-			(vl_chartypes_[i] & k)
-			? s
-			: "-");
-	    }
-	}
+    for (i = first; i < last; i++) {
+	one_ctype_row(enc, i);
     }
 }
 
@@ -107,27 +192,46 @@ static int
 show_CharClasses(BUFFER *bp GCC_UNUSED)
 {
     return liststuff(PRINTABLECHARS_BufName,
-		     FALSE, makectypelist, 0, (void *) 0);
+		     -TRUE, make_ctype_list, 0, (void *) 0);
 }
 
 #if OPT_UPBUFF
-static void
+void
 update_char_classes(void)
 {
     update_scratch(PRINTABLECHARS_BufName, show_CharClasses);
 }
-
-#else
-#define update_char_classes()	/*nothing */
 #endif
 
 /* ARGSUSED */
 int
-desprint(int f GCC_UNUSED, int n GCC_UNUSED)
+desprint(int f, int n GCC_UNUSED)
 {
+    show_ctypes_f = f;
+    show_ctypes_n = 0;		/* always page 0 with this command */
     return show_CharClasses(curbp);
 }
 
+#if OPT_MULTIBYTE
+/*
+ * Show just wide-characters, if they are available.  The 'f' parameter is only
+ * true if 'n' is greater than zero, so a separate function from desprint() is
+ * needed if we are allowing the user to select pages of printing information.
+ */
+int
+deswprint(int f, int n)
+{
+    if (isEmpty(vl_wide_enc.locale) || isEmpty(vl_narrow_enc.locale)) {
+	show_ctypes_f = f;
+	show_ctypes_n = 0;
+    } else {
+	show_ctypes_f = TRUE;
+	show_ctypes_n = f ? n : 0;
+    }
+    return show_CharClasses(curbp);
+}
+#endif
+
 static int
 cclass_complete(DONE_ARGS)
 {
@@ -185,7 +289,7 @@ match_charclass_regexp(int ch, REGEXVAL * exp)
     char temp[2];
     temp[0] = (char) ch;
 
-    return regexec(exp->reg, temp, temp + 1, 0, 0);
+    return nregexec(exp->reg, temp, temp + 1, 0, 0);
 }
 
 static int
@@ -204,59 +308,72 @@ update_charclasses(const char *tag, int count)
 int
 set_charclass(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    int rc = FALSE;
     int ch;
     int code;
     int count;
+    CHARTYPE ctype;
     REGEXVAL *exp;
 
-    if ((code = get_charclass_code()) > 0) {
-	code = (1 << code);
+    if ((code = get_charclass_code()) >= 0) {
+	ctype = (CHARTYPE) (1 << code);
 	if ((exp = get_charclass_regexp()) != 0) {
 	    for (count = 0, ch = 0; ch < N_chars; ch++) {
-		if (!istype(code, ch)
+		if (!istype(ctype, ch)
 		    && match_charclass_regexp(ch, exp)) {
-		    vl_chartypes_[ch] |= code;
+		    vl_ctype_set(ch, ctype);
 		    count++;
 		}
 	    }
 	    free_regexval(exp);
-	    return update_charclasses("Set", count);
+	    rc = update_charclasses("Set", count);
 	}
     }
-    return FALSE;
+    return rc;
 }
 
 /* ARGSUSED */
 int
 unset_charclass(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    int rc = FALSE;
     int ch;
     int code;
     int count;
+    CHARTYPE ctype;
     REGEXVAL *exp;
 
-    if ((code = get_charclass_code()) > 0) {
-	code = (1 << code);
+    if ((code = get_charclass_code()) >= 0) {
+	ctype = (CHARTYPE) (1 << code);
 	if ((exp = get_charclass_regexp()) != 0) {
 	    for (count = 0, ch = 0; ch < N_chars; ch++) {
-		if (istype(code, ch)
+		if (istype(ctype, ch)
 		    && match_charclass_regexp(ch, exp)) {
-		    vl_chartypes_[ch] &= ~code;
+		    vl_ctype_clr(ch, ctype);
 		    count++;
 		}
 	    }
 	    free_regexval(exp);
-	    return update_charclasses("Unset", count);
+	    rc = update_charclasses("Unset", count);
 	}
     }
-    return FALSE;
+    return rc;
+}
+
+void
+rebuild_charclasses(int print_lo, int print_hi)
+{
+    vl_ctype_init(print_lo, print_hi);
+    vl_ctype_apply();
+    update_char_classes();
 }
 
 /* ARGSUSED */
 int
 reset_charclasses(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    charinit();
+    vl_ctype_init(global_g_val(GVAL_PRINT_LOW),
+		  global_g_val(GVAL_PRINT_HIGH));
     update_char_classes();
     return TRUE;
 }
@@ -271,13 +388,13 @@ static CHARTYPE
 charclass_of(const char *arg)
 {
     CHARTYPE k;
-    if (arg == error_val) {
+    if (isErrorVal(arg)) {
 	k = 0;
     } else {
-	k = vl_chartypes_[char2int(*arg)];
+	k = vl_chartypes_[CharOf(*arg)];
 	if (*arg) {
 	    while (*++arg) {
-		k &= vl_chartypes_[char2int(*arg)];
+		k &= vl_chartypes_[CharOf(*arg)];
 	    }
 	}
     }
@@ -297,7 +414,7 @@ show_charclass(TBUFF **result, const char *arg)
 
     for (j = 0; j != vl_UNUSED; j++) {
 	if (((1 << j) & k) != 0
-	    && (s = choice_to_name(&fsm_charclass_blist, j)) != 0) {
+	    && (s = choice_to_name(&fsm_charclass_blist, (int) j)) != 0) {
 	    if (tb_length(*result))
 		tb_sappend0(result, "+");
 	    tb_sappend0(result, s);
@@ -324,14 +441,34 @@ makevarslist(int dum1 GCC_UNUSED, void *ptr)
     bpadc('-', term.cols - DOT.o);
     bputc('\n');
 
-    bprintf("%s", (char *) ptr);
+    for (p = (UVAR *) ptr; p->u_name != 0; ++p) {
+	bprintf("\n$%s = ", p->u_name);
+	bputsn_xcolor(p->u_value,
+		      (int) strlen(p->u_value),
+		      ((p->u_type == VALTYPE_ENUM ||
+			p->u_type == VALTYPE_BOOL ||
+			p->u_type == VALTYPE_MAJOR)
+		       ? XCOLOR_ENUM
+		       : ((p->u_type == VALTYPE_REGEX)
+			  ? XCOLOR_REGEX
+			  : (p->u_type == VALTYPE_STRING
+			     ? XCOLOR_STRING
+			     : (p->u_type == VALTYPE_INT
+				? XCOLOR_NUMBER
+				: (p->u_type == (char) VALTYPE_UNKNOWN
+				   ? XCOLOR_WARNING
+				   : XCOLOR_NONE))))));
+    }
+
     if (arg_stack != 0 && ((k = arg_stack->num_args) != 0)) {
 	bprintf("--- %s parameters ",
 		tb_values(arg_stack->all_args[0]));
 	bpadc('-', term.cols - DOT.o);
 	for (j = 1; j <= k; j++) {
-	    bprintf("\n$%d = %s", j,
-		    tb_values(arg_stack->all_args[j]));
+	    bprintf("\n$%d = ", j);
+	    bputsn_xcolor(tb_values(arg_stack->all_args[j]),
+			  (int) tb_length(arg_stack->all_args[j]),
+			  XCOLOR_STRING);
 	}
     }
     if (temp_vars != 0)
@@ -341,7 +478,10 @@ makevarslist(int dum1 GCC_UNUSED, void *ptr)
 	    bprintf("--- Temporary variables ");
 	    bpadc('-', term.cols - DOT.o);
 	}
-	bprintf("\n%%%s = %s", p->u_name, p->u_value);
+	bprintf("\n%%%s = ", p->u_name);
+	bputsn_xcolor(p->u_value,
+		      (int) strlen(p->u_value),
+		      XCOLOR_STRING);
     }
 }
 
@@ -364,20 +504,25 @@ is_mode_name(const char *name, int showall, VALARGS * args)
 
 /* filter out mode-names that apply only to abbreviations */
 static const char *
-get_listvalue(const char *name, int showall)
+get_listvalue(const char *name, int showall, int *typep)
 {
+    const char *result = 0;
     VALARGS args;
     int s;
     int vnum;
 
-    if ((s = is_mode_name(name, showall, &args)) == TRUE)
-	return string_mode_val(&args);
-    else if (s == SORTOFTRUE) {
+    *typep = VALTYPE_STRING;
+    if ((s = is_mode_name(name, showall, &args)) == TRUE) {
+	result = string_mode_val(&args);
+	*typep = args.names->type;
+    } else if (s == SORTOFTRUE) {
 	vnum = vl_lookup_statevar(name);
-	if (vnum != ILLEGAL_NUM)
-	    return get_statevar_val(vnum);
+	if (vnum != ILLEGAL_NUM) {
+	    result = get_statevar_val(vnum);
+	    *typep = statevar_funcs[vnum].type;
+	}
     }
-    return 0;
+    return result;
 }
 #endif /* OPT_SHOW_EVAL */
 
@@ -390,49 +535,61 @@ static int show_vars_n;
 static int
 show_VariableList(BUFFER *bp GCC_UNUSED)
 {
-    char *values;
     WINDOW *wp = curwp;
     int s;
     size_t t;
     char *v;
     const char *vv;
-    static const char fmt[] =
-    {"$%s = %.*s\n"};
     const char *const *Names;
     int showall;
+    int type;
     int rc = FALSE;
+    UVAR *list = 0;
 
     TRACE((T_CALLED "show_VariableList()\n"));
 
     Names = show_vars_f ? list_of_modes() : statevars;
     if (Names != 0) {
 	showall = show_vars_f ? (show_vars_n > 1) : FALSE;
-	/* collect data for state-variables, since some depend on window */
-	for (s = 0, t = 0; Names[s] != 0; s++) {
-	    if ((vv = get_listvalue(Names[s], showall)) != 0)
-		t += strlen(Names[s]) + strlen(fmt) + strlen(vv);
-	}
 
 	beginDisplay();
-	if ((values = typeallocn(char, t)) != 0) {
-	    for (s = 0, v = values; Names[s] != 0; s++) {
-		if ((vv = get_listvalue(Names[s], showall)) != 0) {
-		    t = strlen(vv);
-		    if (t == 0) {
-			t = 1;
-			vv = "";
-		    } else if (vv[t - 1] == '\n')
-			t--;	/* suppress trailing newline */
-		    v = lsprintf(v, fmt, Names[s], (int) t, vv);
+	/* count the variables */
+	for (t = 0; Names[t] != 0; t++) {
+	    ;
+	}
+	++t;			/* guarantee a null entry on the end */
+
+	if ((list = typecallocn(UVAR, t)) != 0) {
+	    /* collect data for state-variables, since some depend on window */
+	    for (s = 0; Names[s] != 0; s++) {
+		list[s].u_name = (char *) Names[s];
+		list[s].u_type = VALTYPE_STRING;
+		if ((vv = get_listvalue(Names[s], showall, &type)) != 0) {
+		    list[s].u_type = (char) type;
+		    if (isErrorVal(vv))
+			list[s].u_type = VALTYPE_UNKNOWN;
+		} else {
+		    vv = "";
+		}
+		list[s].u_value = v = strmalloc(vv);
+		t = strlen(v);
+		v += (t + 1);
+		if ((t != 0) && (vv[t - 1] == '\n')) {
+		    v[-2] = EOS;	/* suppress trailing newline */
 		}
 	    }
-	    rc = liststuff(VARIABLES_BufName, FALSE,
-			   makevarslist, 0, (void *) values);
-	    free(values);
 
+	    rc = liststuff(VARIABLES_BufName, FALSE,
+			   makevarslist, 0, (void *) list);
 	    /* back to the buffer whose modes we just listed */
 	    swbuffer(wp->w_bufp);
+
+	    for (s = 0; Names[s] != 0; s++) {
+		free(list[s].u_value);
+	    }
+	    FreeIfNeeded(list);
 	}
+
 	endofDisplay();
     }
     returnCode(rc);
@@ -504,7 +661,7 @@ list_amp_funcs(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
     bpadc('-', term.cols - DOT.o);
 
     for (n = 0; vl_ufuncs[n].f_name != 0; ++n) {
-	int nparams = (vl_ufuncs[n].f_code & NARGMASK);
+	int nparams = (int) (vl_ufuncs[n].f_code & NARGMASK);
 	const char *t_params = UF_PARAMS(vl_ufuncs[n].f_code);
 	const char *t_return = UF_RETURN(vl_ufuncs[n].f_code);
 
@@ -643,7 +800,7 @@ static char *
 path_head(TBUFF **result, const char *path)
 {
     char *temp = tb_values(tb_scopy(result, path));
-    if (temp != error_val && temp != 0) {
+    if (isLegalVal(temp)) {
 	*pathleaf(temp) = EOS;
 	path_trim(temp);
     }
@@ -654,7 +811,7 @@ path_head(TBUFF **result, const char *path)
  * If the given argument contains non-path characters, quote it.  Otherwise
  * simply copy it.
  */
-static void
+void
 path_quote(TBUFF **result, const char *path)
 {
 #if SYS_VMS
@@ -757,6 +914,7 @@ default_mode_value(TBUFF **result, char *name)
 static char *
 dequoted_parameter(TBUFF **tok)
 {
+    char *result = 0;
     char *previous;
     const char *newvalue;
     int strip;
@@ -787,22 +945,25 @@ dequoted_parameter(TBUFF **tok)
 	}
 
 	if (newvalue == (const char *) error_val) {
-	    return error_val;
-	} else if (strip) {
-	    if (((const char *) previous) != newvalue) {
-		TBUFF *fix = 0;
-		tb_scopy(&fix, newvalue);
-		tb_free(tok);
-		*tok = fix;
+	    result = error_val;
+	} else {
+	    if (strip) {
+		if (((const char *) previous) != newvalue) {
+		    TBUFF *fix = 0;
+		    tb_scopy(&fix, newvalue);
+		    tb_free(tok);
+		    *tok = fix;
+		}
+		tb_dequote(tok);
+	    } else if (((const char *) previous) != newvalue) {
+		tb_scopy(tok, newvalue);
 	    }
-	    tb_dequote(tok);
-	} else if (((const char *) previous) != newvalue) {
-	    tb_scopy(tok, newvalue);
+	    result = tb_values(*tok);
 	}
-	return (tb_values(*tok));
+    } else {
+	tb_free(tok);
     }
-    tb_free(tok);
-    return (0);
+    return (result);
 }
 
 #define UNI_CLASSNAME "universal"
@@ -922,6 +1083,21 @@ ufunc_get_motion(TBUFF **result, const char *name)
     return rc;
 }
 
+int
+eval_fchanged(BUFFER *bp)
+{
+    time_t current;
+    int changed;
+
+    if (valid_buffer(bp)) {
+	get_modtime(bp, &current);
+	changed = (current != bp->b_modtime);
+    } else {
+	changed = FALSE;
+    }
+    return changed;
+}
+
 #define MAXARGS 3
 
 /*
@@ -945,15 +1121,16 @@ run_func(int fnum)
     int is_error = FALSE;
     long value = 0;
     long nums[MAXARGS];
+    int old_reading = read_msgline(TRUE);
 
     TRACE((T_CALLED "run_func(%d:%s)\n", fnum, vl_ufuncs[fnum].f_name));
 
-    nargs = vl_ufuncs[fnum].f_code & NARGMASK;
-    args_numeric = vl_ufuncs[fnum].f_code & NUM;
-    args_boolean = vl_ufuncs[fnum].f_code & BOOL;
+    nargs = (int) (vl_ufuncs[fnum].f_code & NARGMASK);
+    args_numeric = (int) (vl_ufuncs[fnum].f_code & NUM);
+    args_boolean = (int) (vl_ufuncs[fnum].f_code & BOOL);
 
-    ret_numeric = vl_ufuncs[fnum].f_code & NRET;
-    ret_boolean = vl_ufuncs[fnum].f_code & BRET;
+    ret_numeric = (int) (vl_ufuncs[fnum].f_code & NRET);
+    ret_boolean = (int) (vl_ufuncs[fnum].f_code & BRET);
 
     TPRINTF(("** evaluate '%s' (0x%x), %d %s args returning %s\n",
 	     vl_ufuncs[fnum].f_name,
@@ -966,7 +1143,7 @@ run_func(int fnum)
     for (i = 0; i < nargs; i++) {
 	args[i] = 0;
 	if ((arg[i] = dequoted_parameter(&args[i])) == 0
-	    || (arg[i] == error_val)
+	    || isErrorVal(arg[i])
 	    || isTB_ERRS(args[i])) {
 	    arg[i] = error_val;
 	    is_error = TRUE;
@@ -1044,7 +1221,7 @@ run_func(int fnum)
 	break;
     case UFERROR:
 	is_error = FALSE;
-	value = (arg[0] == error_val);
+	value = isErrorVal(arg[0]);
 	break;
     case UFLT:			/* FALLTHRU */
     case UFLESS:
@@ -1084,7 +1261,7 @@ run_func(int fnum)
 	value = (strcmp(arg[0], arg[1]) >= 0);
 	break;
     case UFINDIRECT:
-	if ((sp = tokval(arg[0])) != error_val)
+	if (isLegalExp(sp, tokval(arg[0])))
 	    tb_scopy(&result, sp);
 	else
 	    is_error = TRUE;
@@ -1130,13 +1307,13 @@ run_func(int fnum)
 	if ((exp = new_regexval(arg[0], TRUE)) != 0) {
 	    int save_flag = ignorecase;
 	    ignorecase = TRUE;
-	    value = regexec(exp->reg, arg[1], (char *) 0, 0, -1);
+	    value = nregexec(exp->reg, arg[1], (char *) 0, 0, -1);
 	    ignorecase = save_flag;
 	}
 	break;
     case UFMATCH:
 	if ((exp = new_regexval(arg[0], TRUE)) != 0)
-	    value = regexec(exp->reg, arg[1], (char *) 0, 0, -1);
+	    value = nregexec(exp->reg, arg[1], (char *) 0, 0, -1);
 	break;
     case UFRANDOM:		/* FALLTHRU */
     case UFRND:
@@ -1156,7 +1333,7 @@ run_func(int fnum)
 	break;
     case UFBIND:
 	if (!is_error) {
-	    if ((sp = prc2engl(arg[0])) == error_val)
+	    if (isErrorVal(sp = prc2engl(arg[0])))
 		is_error = TRUE;
 	    else
 		tb_scopy(&result, sp);
@@ -1207,8 +1384,7 @@ run_func(int fnum)
 	break;
     case UFDQUERY:
 	if (!is_error) {
-	    if ((cp = user_reply(arg[0], arg[1])) != 0
-		&& (cp != error_val))
+	    if (isLegalExp(cp, user_reply(arg[0], arg[1])))
 		tb_scopy(&result, cp);
 	    else
 		is_error = TRUE;
@@ -1217,8 +1393,7 @@ run_func(int fnum)
     case UFQPASSWD:
 	if (!is_error) {
 	    qpasswd = TRUE;
-	    if ((cp = user_reply(arg[0], NULL)) != 0
-		&& (cp != error_val))
+	    if (isLegalExp(cp, user_reply(arg[0], NULL)))
 		tb_scopy(&result, cp);
 	    else
 		is_error = TRUE;
@@ -1227,8 +1402,7 @@ run_func(int fnum)
 	break;
     case UFQUERY:
 	if (!is_error) {
-	    if ((cp = user_reply(arg[0], error_val)) != 0
-		&& (cp != error_val))
+	    if (isLegalExp(cp, user_reply(arg[0], error_val)))
 		tb_scopy(&result, cp);
 	    else
 		is_error = TRUE;
@@ -1236,7 +1410,7 @@ run_func(int fnum)
 	break;
     case UFLOOKUP:
 	if ((i = combine_choices(&fsm_lookup_blist, arg[0])) > 0)
-	    tb_scopy(&result, SL_TO_BSL(cfg_locate(arg[1], i)));
+	    tb_scopy(&result, SL_TO_BSL(cfg_locate(arg[1], (UINT) i)));
 	break;
     case UFPATH:
 	if (!is_error) {
@@ -1316,7 +1490,8 @@ run_func(int fnum)
 	if (!is_error) {
 	    i = reg2index(*arg[0]);
 	    if ((i = index2ukb(i)) < NKREGS
-		&& i >= 0) {
+		&& i >= 0
+		&& i < NKREGS) {
 		KILL *kp = kbs[i].kbufh;
 		while (kp != 0) {
 		    tb_bappend(&result,
@@ -1328,6 +1503,22 @@ run_func(int fnum)
 		break;
 	    }
 	}
+    case UFBCHANGED:
+	if (!is_error
+	    && (bp = find_any_buffer(arg[0])) != 0) {
+	    value = b_is_changed(bp);
+	} else {
+	    is_error = TRUE;
+	}
+	break;
+    case UFFCHANGED:
+	if (!is_error
+	    && (bp = find_any_buffer(arg[0])) != 0) {
+	    value = eval_fchanged(bp);
+	} else {
+	    is_error = TRUE;
+	}
+	break;
 	/* FALLTHRU */
     case NFUNCS:
 	TRACE(("unknown function #%d\n", fnum));
@@ -1335,22 +1526,28 @@ run_func(int fnum)
 	break;
     }
 
-    if (ret_numeric)
-	render_long(&result, value);
-    else if (ret_boolean)
-	render_boolean(&result, value);
-    else
-	tb_enquote(&result);
+    if (is_error) {
+	tb_error(&result);
+    } else {
+	if (ret_numeric)
+	    render_long(&result, value);
+	else if (ret_boolean)
+	    render_boolean(&result, value);
+	else
+	    tb_enquote(&result);
+    }
 
+    TRACE2(("actual:%d:%s\n", tb_length(result), tb_visible(result)));
     TPRINTF(("-> %s%s\n",
 	     is_error ? "*" : "",
-	     is_error ? error_val : NONNULL(tb_values(result))));
+	     NONNULL(tb_values(result))));
 
     tb_free(&juggle);
     for (i = 0; i < nargs; i++) {
 	tb_free(&args[i]);
     }
-    returnString(is_error ? error_val : tb_values(result));
+    read_msgline(old_reading);
+    returnString(tb_values(result));
 }
 
 /* find a temp variable */
@@ -1394,7 +1591,7 @@ get_statevar_val(int vnum)
 
 	nested = new_level;
 	tb_init(&buffer[nested], EOS);
-	s = (*statevar_func[vnum]) (&buffer[nested], (const char *) NULL);
+	s = (*statevar_funcs[vnum].func) (&buffer[nested], (const char *) NULL);
 
 	if (s == TRUE) {
 	    tb_append(&buffer[nested], EOS);	/* trailing null, just in case... */
@@ -1586,7 +1783,10 @@ PromptAndSet(const char *name, int f, int n)
 		tb_scopy(&tmp, (vd.v_ptr->u_value) ? vd.v_ptr->u_value : "");
 	    }
 	    (void) lsprintf(prompt, "Value of %s: ", var);
-	    status = mlreply2(prompt, &tmp);
+	    status = kbd_string2(prompt, &tmp, '\n', KBD_NORMAL,
+				 (*name == '$' && name_to_choices(name + 1))
+				 ? fsm_complete
+				 : no_completion);
 	    if (status == ABORT)
 		tb_error(&tmp);
 	    else if (status != TRUE)
@@ -1608,8 +1808,7 @@ PromptAndSet(const char *name, int f, int n)
 static void
 free_UVAR_value(UVAR * p)
 {
-    if (p->u_value != error_val
-	&& p->u_value != 0) {
+    if (isLegalVal(p->u_value)) {
 	beginDisplay();
 	free(p->u_value);
 	endofDisplay();
@@ -1723,7 +1922,7 @@ unsetvar(int f GCC_UNUSED, int n GCC_UNUSED)
 static int
 set_statevar_val(int vnum, const char *value)
 {
-    return (*statevar_func[vnum]) ((TBUFF **) 0, value);
+    return (*statevar_funcs[vnum].func) ((TBUFF **) 0, value);
 }
 
 /* figure out what type of variable we're setting, and do so */
@@ -1761,7 +1960,7 @@ SetVarValue(VWRAP * var, char *name, const char *value)
     case VW_TEMPVAR:
 	beginDisplay();
 	free_UVAR_value(var->v_ptr);
-	if (value == error_val) {
+	if (isErrorVal(value)) {
 	    var->v_ptr->u_value = error_val;
 	} else if ((var->v_ptr->u_value = strmalloc(value)) == 0) {
 	    status = FALSE;
@@ -1876,9 +2075,10 @@ skip_space_tab(char *src)
 }
 
 #if OPT_COLOR
-void
+int
 set_ctrans(const char *thePalette)
 {
+    int changed = 0;
     int n = 0, value;
     char *next;
 
@@ -1891,7 +2091,10 @@ set_ctrans(const char *thePalette)
 	    if (next == thePalette)
 		break;
 	    thePalette = next;
-	    ctrans[n] = value;
+	    if (ctrans[n] != value) {
+		ctrans[n] = value;
+		++changed;
+	    }
 	    if (++n >= NCOLORS)
 		break;
 	}
@@ -1903,6 +2106,7 @@ set_ctrans(const char *thePalette)
 	}
 #endif
     }
+    return changed;
 }
 #endif
 
@@ -1969,7 +2173,7 @@ makecolorlist(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
     DOT.l = lforw(buf_head(curbp));
     DOT.o = 0;
 
-    if (getregion(&region) == TRUE)
+    if (getregion(curbp, &region) == TRUE)
 	attribute_cntl_a_seqs_in_region(&region, rgn_FULLLINE);
 }
 
@@ -2007,9 +2211,6 @@ set_palette(const char *value)
 {
     tb_curpalette = tb_scopy(&tb_curpalette, value);
 #if OPT_COLOR
-    if (term.setpal == nullterm_setpal)
-	return FALSE;
-
     relist_descolor();
     term.setpal(tb_values(tb_curpalette));
     vile_refresh(FALSE, 0);
@@ -2027,7 +2228,7 @@ render_int(TBUFF **rp, int i)
     p = tb_values(tb_alloc(rp, 32));
     q = lsprintf(p, "%d", i);
     if (rp != 0 && q != 0 && p != 0)
-	(*rp)->tb_used = (q - p + 1);
+	(*rp)->tb_used = (size_t) (q - p + 1);
     return p;
 }
 
@@ -2040,7 +2241,7 @@ render_long(TBUFF **rp, long i)
     p = tb_values(tb_alloc(rp, 32));
     q = lsprintf(p, "%ld", i);
     if (rp != 0 && q != 0 && p != 0)
-	(*rp)->tb_used = (q - p + 1);
+	(*rp)->tb_used = (size_t) (q - p + 1);
     return p;
 }
 
@@ -2055,6 +2256,19 @@ render_boolean(TBUFF **rp, int val)
     return tb_values(tb_scopy(rp, bools[val ? 1 : 0]));
 }
 
+/* represent long integer as string */
+char *
+render_ulong(TBUFF **rp, ULONG i)
+{
+    char *p, *q;
+
+    p = tb_values(tb_alloc(rp, 32));
+    q = lsprintf(p, "%lu", i);
+    if (rp != 0 && q != 0 && p != 0)
+	(*rp)->tb_used = (size_t) (q - p + 1);
+    return p;
+}
+
 #if (SYS_WINNT||SYS_VMS)
 /* unsigned to hex */
 char *
@@ -2157,7 +2371,7 @@ complete_integer(DONE_ARGS)
 
     (void) flags;
     (void) pos;
-    (void) strtol(buf, &tmp, 0);
+    IGNORE_RC(strtol(buf, &tmp, 0));
     if ((tmp != 0) && (tmp != buf) && (*tmp == 0) && isSpace(c)) {
 	if (c != NAMEC)		/* put it back (cf: kbd_complete) */
 	    unkeystroke(c);
@@ -2272,6 +2486,8 @@ complete_vars(DONE_ARGS)
 static int
 read_argument(TBUFF **paramp, const PARAM_INFO * info)
 {
+    static char empty_string[] = "";
+
     int status = TRUE;
     const char *prompt;
     TBUFF *temp;
@@ -2279,6 +2495,8 @@ read_argument(TBUFF **paramp, const PARAM_INFO * info)
     int save_clexec;
     int save_isnamed;
     char *save_execstr;
+    char fname[NFILEN];
+    char bname[NBUFN];
     KBD_OPTIONS flags = 0;	/* no expansion, etc. */
 
     if (mac_tokval(paramp) == 0) {
@@ -2349,36 +2567,40 @@ read_argument(TBUFF **paramp, const PARAM_INFO * info)
 	if (execstr == 0 || !*skip_blanks(execstr)) {
 	    clexec = FALSE;
 	    isnamedcmd = TRUE;
-	    execstr = "";
+	    execstr = empty_string;
 	}
 
-	if (info->pi_type == PT_FILE) {
-	    char fname[NFILEN];
-
-	    status = mlreply_file(tb_values(temp), (TBUFF **) 0,
-				  FILEC_UNKNOWN, fname);
-	    if (status != ABORT)
-		tb_scopy(paramp, fname);
-	} else if (info->pi_type == PT_DIR) {
-	    char fname[NFILEN];
-
-	    status = mlreply_dir(tb_values(temp), (TBUFF **) 0, fname);
-	    if (status != ABORT)
-		tb_scopy(paramp, fname);
-	} else if (info->pi_type == PT_BUFFER) {
-	    char bname[NBUFN];
-
-	    status = ask_for_bname(tb_values(temp), bname, sizeof(bname));
-	    if (status != ABORT)
-		tb_scopy(paramp, bname);
-	} else {
-	    status = kbd_reply(tb_values(temp),
-			       paramp,	/* in/out buffer */
-			       eol_history,
-			       '\n',	/* expect a newline or return */
-			       flags,	/* no expansion, etc. */
-			       complete);
+	if (info != 0) {
+	    switch (info->pi_type) {
+	    case PT_FILE:
+		status = mlreply_file(tb_values(temp), (TBUFF **) 0,
+				      FILEC_UNKNOWN, fname);
+		if (status != ABORT)
+		    tb_scopy(paramp, fname);
+		break;
+	    case PT_DIR:
+		status = mlreply_dir(tb_values(temp), (TBUFF **) 0, fname);
+		if (status != ABORT)
+		    tb_scopy(paramp, fname);
+		break;
+	    case PT_BUFFER:
+		status = ask_for_bname(tb_values(temp), bname, sizeof(bname));
+		if (status != ABORT)
+		    tb_scopy(paramp, bname);
+		break;
+	    default:
+		status = kbd_reply(tb_values(temp),
+				   paramp,	/* in/out buffer */
+				   eol_history,
+				   '\n',	/* expect a newline or return */
+				   flags,	/* no expansion, etc. */
+				   complete);
+		if (status == TRUE)
+		    tb_prequote(paramp);
+		break;
+	    }
 	}
+
 	clexec = save_clexec;
 	execstr = save_execstr;
 	isnamedcmd = save_isnamed;
@@ -2421,7 +2643,7 @@ save_arguments(BUFFER *bp)
     beginDisplay();
     if ((p = typealloc(PROC_ARGS)) == 0) {
 	status = no_memory("save_arguments");
-    } else if ((all_args = tb_allocn(max_args + 1)) == 0) {
+    } else if ((all_args = tb_allocn((size_t) (max_args + 1))) == 0) {
 	status = no_memory("save_arguments");
 	free(p);
     } else {
@@ -2460,7 +2682,7 @@ save_arguments(BUFFER *bp)
 
 		if (ok) {
 		    cp = tokval(tb_values(p->all_args[num_args]));
-		    if (cp != error_val) {
+		    if (isLegalVal(cp)) {
 			tb_scopy(&(p->all_args[num_args]), cp);
 			copied = TRUE;
 		    } else {
@@ -2642,7 +2864,7 @@ statevar_arg_eval(const char *argp)
     int vnum;
     const char *result;
 
-    if ((result = get_argument(++argp)) == error_val) {
+    if (isErrorVal(result = get_argument(++argp))) {
 	vnum = vl_lookup_statevar(argp);
 	if (vnum != ILLEGAL_NUM)
 	    result = get_statevar_val(vnum);
@@ -2819,13 +3041,13 @@ evaluate(int f, int n)
     int code = FALSE;
 
     if (read_argument(&params, &info) == TRUE) {
-	if ((tmp = tb_values(params)) != error_val && tmp != 0) {
+	if (isLegalExp(tmp, tb_values(params))) {
 	    old = execstr;
 	    execstr = tmp;
 	    TRACE(("EVAL %s\n", execstr));
 	    for (;;) {
 		if (can_evaluate()) {
-		    if ((tmp = mac_tokval(&tok)) != 0) {
+		    if ((tmp = mac_unquotedarg(&tok)) != 0) {
 			if (tb_length(cmd))
 			    tb_sappend0(&cmd, " ");
 			tb_sappend0(&cmd, tmp);
@@ -2862,7 +3084,7 @@ s2size(const char *s)
     int n = strtol(s, 0, 0);
     if (n < 0)
 	n = 0;
-    return n;
+    return (size_t) n;
 }
 
 /* use this to return a pointer to the string after the n-1 first characters */
@@ -2920,7 +3142,7 @@ label2lp(BUFFER *bp, const char *label)
 char *
 mkupper(char *s)
 {
-    if (s != error_val && s != 0) {
+    if (isLegalVal(s)) {
 	char *sp;
 
 	for (sp = s; *sp; sp++)
@@ -2935,7 +3157,7 @@ mkupper(char *s)
 char *
 mklower(char *s)
 {
-    if (s != error_val && s != 0) {
+    if (isLegalVal(s)) {
 	char *sp;
 
 	for (sp = s; *sp; sp++)
@@ -2952,7 +3174,7 @@ mktrimmed(char *str)
     char *base = str;
     char *dst = str;
 
-    if (str != error_val && str != 0) {
+    if (isLegalVal(str)) {
 	while (*str != EOS) {
 	    if (isSpace(*str)) {
 		if (dst != base)
