@@ -4,7 +4,7 @@
  *	original by Daniel Lawrence, but
  *	much modified since then.  assign no blame to him.  -pgf
  *
- * $Header: /usr/build/vile/vile/RCS/exec.c,v 1.314 2008/01/22 00:13:57 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/exec.c,v 1.341 2010/09/08 21:13:26 tom Exp $
  *
  */
 
@@ -181,6 +181,7 @@ parse_signed_number(const char *s, int *number)
 static const char *
 parse_linespec(const char *s, LINE **markptr)
 {
+    const char *base = s;
     int num;
     LINE *lp;			/* where the linespec takes us */
     int status;
@@ -233,7 +234,7 @@ parse_linespec(const char *s, LINE **markptr)
 	    last_srch_direc = (*save == '/') ? FORWARD : REVERSE;
 	    s = skip_pattern(save, &done);
 	    tb_init(&searchpat, EOS);
-	    tb_bappend(&searchpat, save + 1, s - save - 1 - done);
+	    tb_bappend(&searchpat, save + 1, (size_t) (s - save - 1 - done));
 	    tb_append(&searchpat, EOS);
 
 	    if ((gregexp = regcomp(tb_values(searchpat),
@@ -260,9 +261,7 @@ parse_linespec(const char *s, LINE **markptr)
 
 	/* if linespec was faulty, quit now */
 	if (!lp) {
-	    *markptr = lp;
-	    swapmark();
-	    return s;
+	    break;
 	}
 
 	/* maybe add an offset */
@@ -274,7 +273,17 @@ parse_linespec(const char *s, LINE **markptr)
     } while (*s == ';' || isSign(*s));
 
     *markptr = lp;
-    swapmark();
+
+    /*
+     * If we're processing a script (for example), which has no line-specifiers,
+     * we'll get a null pointer here.  But there's no reason to flag a bug in
+     * swapmark().
+     */
+    if (lp == 0 && (s == base)) {
+	DOT = MK;
+    } else {
+	swapmark();
+    }
     return s;
 }
 
@@ -400,14 +409,13 @@ making_pattern(EOL_ARGS)
 {
     int done = FALSE;
     int in_regexp = FALSE;
-    const char *src = buffer;
 
     (void) eolchar;
 
     TRACE(("making_pattern(buffer=%.*s, cpos=%d, c=%#x, eolchar=%#x)\n",
 	   (int) cpos, buffer, (int) cpos, c, eolchar));
     if (cpos != 0) {
-	src = skip_linespecs(buffer, cpos, &done);
+	(void) skip_linespecs(buffer, (int) cpos, &done);
 	in_regexp = !done || isPattern(c);
     } else if (isPattern(c)) {
 	in_regexp = TRUE;
@@ -455,13 +463,14 @@ end_of_cmd(void)
 int
 end_named_cmd(void)
 {
-    if (isnamedcmd) {
-	if (clexec)
-	    return token_ended_line;
-	else
-	    return end_of_cmd();
+    int result = FALSE;
+
+    if (clexec) {
+	result = token_ended_line;
+    } else if (isnamedcmd) {
+	result = end_of_cmd();
     }
-    return FALSE;
+    return result;
 }
 
 /* returns true iff we are in a named-command and if the user did not end the
@@ -470,13 +479,14 @@ end_named_cmd(void)
 int
 more_named_cmd(void)
 {
-    if (isnamedcmd) {
-	if (clexec)
-	    return !token_ended_line;
-	else
-	    return !end_of_cmd();
+    int result = FALSE;
+
+    if (clexec) {
+	result = !token_ended_line;
+    } else if (isnamedcmd) {
+	result = !end_of_cmd();
     }
-    return FALSE;
+    return result;
 }
 
 /*
@@ -907,7 +917,7 @@ docmd(char *cline, int execflag, int f, int n)
     execstr = cline;		/* and change execstr to the new one */
 
     do {
-	if ((token = mac_tokval(&tok)) == 0) {	/* grab first token */
+	if ((token = mac_unquotedarg(&tok)) == 0) {	/* grab first token */
 	    execstr = oldestr;
 	    returnCode(FALSE);
 	}
@@ -924,10 +934,10 @@ docmd(char *cline, int execflag, int f, int n)
      */
     if (toktyp(token) != TOK_LITSTR || isDigit(token[0])) {
 	f = TRUE;
-	n = strtol(tokval(token), 0, 0);
+	n = (int) strtol(tokval(token), 0, 0);
 
 	/* and now get the command to execute */
-	if ((token = mac_tokval(&tok)) == 0) {
+	if ((token = mac_unquotedarg(&tok)) == 0) {
 	    execstr = oldestr;
 	    returnCode(FALSE);
 	}
@@ -962,25 +972,38 @@ docmd(char *cline, int execflag, int f, int n)
 int
 call_cmdfunc(const CMDFUNC * p, int f, int n)
 {
+    int status = FALSE;
+
     switch (p->c_flags & CMD_TYPE) {
     case CMD_FUNC:		/* normal CmdFunc */
-	return (CMD_U_FUNC(p)) (f, n);
+	status = (CMD_U_FUNC(p)) (f, n);
+	break;
 
 #if OPT_NAMEBST
 #if OPT_PROCEDURES
     case CMD_PROC:		/* named procedure */
-	return dobuf(CMD_U_BUFF(p), f ? n : 1);
+	status = dobuf(CMD_U_BUFF(p), f ? n : 1, -1);
+	break;
+
+    case CMD_OPER:		/* user-defined operator */
+	regionshape = rgn_EXACT;	/* FIXME */
+	opcmd = OPOTHER;
+	status = vile_op(f, n, user_operator, CMD_U_BUFF(p)->b_bname);
+	break;
 #endif
 
 #if OPT_PERL
     case CMD_PERL:		/* perl subroutine */
-	return perl_call_sub(CMD_U_PERL(p), p->c_flags & OPER, f, n);
+	status = perl_call_sub(CMD_U_PERL(p), p->c_flags & OPER, f, n);
+	break;
 #endif
 #endif /* OPT_NAMEBST */
+    default:
+	mlforce("BUG: invalid CMDFUNC type");
+	status = FALSE;
+	break;
     }
-
-    mlforce("BUG: invalid CMDFUNC type");
-    return FALSE;
+    return status;
 }
 
 /*
@@ -1004,7 +1027,7 @@ execute(const CMDFUNC * execfunc, int f, int n)
     }
 
     TRACE((T_CALLED "execute(execfunc=%p(%s:%s), f=%d, n=%d)\n",
-	   execfunc,
+	   (const void *) execfunc,
 	   TRACE_CMDFUNC(execfunc),
 	   f, n));
 
@@ -1016,8 +1039,9 @@ execute(const CMDFUNC * execfunc, int f, int n)
 	/* but if we're in insertmode, it's okay, since we must
 	   be executing a function key, like an arrow key,
 	   that the user will want to have replayed later */
-	if ((curwp == NULL || !insertmode) && (flags & REDO) == 0)
+	if ((!insertmode) && (flags & REDO) == 0)
 	    dotcmdstop();
+
 	if (flags & UNDO) {
 	    /* undoable command can't be permitted when read-only */
 	    if (!(flags & VIEWOK)) {
@@ -1088,6 +1112,68 @@ get_token(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *a
     return get_token2(src, tok, endfunc, eolchar, actual);
 }
 
+/*
+ * This is used for the "map" command, which doesn't honor anything except ^V.
+ */
+char *
+get_token1(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *actual)
+{
+    int c, chr;
+
+    if (actual != 0)
+	*actual = EOS;
+    if (src == 0)
+	return src;
+
+    /* first scan past any whitespace in the source string */
+    src = skip_space_tab(src);
+
+    /* scan through the source string, which may be quoted */
+    while ((c = *src) != EOS) {
+	if (c == quotec) {
+	    if (*++src == EOS)
+		break;
+	}
+#if OPT_MODELINE
+	/*
+	 * Multiple settings on the line may be separated by colons.
+	 */
+	else if (c == ':' && in_modeline) {
+	    if (actual != 0)
+		*actual = ' ';
+	    src++;
+	    break;
+	} else
+#endif
+	if (c == eolchar) {
+	    if (actual != 0)
+		*actual = *src;
+	    if (!isBlank(c))
+		src++;
+	    break;
+	} else if (endfunc(tb_values(*tok), tb_length(*tok), c, eolchar)) {
+	    if (actual != 0)
+		*actual = *src;
+	    break;
+	} else if (isBlank(c)) {
+	    if (actual != 0)
+		*actual = *src;
+	    break;
+	}
+
+	chr = *src++;		/* record the character */
+	tb_append(tok, chr);
+    }
+
+    /* scan past any whitespace remaining in the source string */
+    src = skip_space_tab(src);
+    token_ended_line = isreturn(*src) || *src == EOS;
+
+    tb_append(tok, EOS);
+
+    return src;
+}
+
 /*
  * This is split-out from get_token() so we can handle the special case of
  * re-gluing a "!" to a quoted string from kbd_reply().
@@ -1096,6 +1182,7 @@ char *
 get_token2(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *actual)
 {
     int quotef = EOS;		/* nonzero iff the current string quoted */
+    int quotes = FALSE;
     int c, i, d, chr;
     int shell = tb_length(*tok) && isShellOrPipe(tb_values(*tok));
     int first = TRUE;
@@ -1114,7 +1201,7 @@ get_token2(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *
 	if (quotef == SQUOTE) {
 	    if (c == SQUOTE) {
 		if (*++src == SQUOTE) {
-		    /* double a quote to insert one */
+		    /* double a single-quote to insert one */
 		    chr = *src++;
 		} else {
 		    quotef = EOS;
@@ -1125,11 +1212,18 @@ get_token2(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *
 	    }
 	} else if (quotef == EOS && c == SQUOTE) {
 	    quotef = SQUOTE;
+	    if (first)
+		quotes = TRUE;
 	    chr = *src++;
 	} else if (c == BACKSLASH) {	/* process special characters */
+	    /*
+	     * FIXME - if quotef is EOS, we should implicitly double-quote.
+	     */
 	    src++;
-	    if (*src == EOS)
+	    if (*src == EOS) {
+		/* FIXME - if leading but no trailing quote, is error */
 		break;
+	    }
 	    switch (c = *src++) {
 	    case 'r':
 		chr = '\r';
@@ -1221,8 +1315,12 @@ get_token2(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *
 		    break;
 		} else if (c == DQUOTE) {
 		    quotef = c;
-		    /* note that leading quote
-		       is included */
+		    if (first)
+			quotes = TRUE;
+		    /*
+		     * Note that a leading quote is included in the result from
+		     * this function.
+		     */
 		} else if (isBlank(c)) {
 		    if (actual != 0)
 			*actual = *src;
@@ -1246,6 +1344,17 @@ get_token2(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *
     token_ended_line = isreturn(*src) || *src == EOS;
 
     tb_append(tok, EOS);
+    /*
+     * Check for case where the input may have had a leading backslash to
+     * introduce a quote.
+     *
+     * FIXME: we could have several adjacent strings such as
+     *   \'xx\"yy\'
+     * It may be better to process those in chunks.
+     */
+    if (tb_length(*tok) && !quotes)
+	tb_prequote(tok);
+
     /* both double and single-quote strings are reduced to a single-quote
      * string with no trailing quote, to simplify use in the caller.
      */
@@ -1277,7 +1386,7 @@ macroize(TBUFF **p, TBUFF *src, int skip)
 
 		TRACE(("macroizing %s\n", tb_visible(src)));
 		(void) tb_append(p, DQUOTE);
-		for (n = skip; n < len; n++) {
+		for (n = (size_t) skip; n < len; n++) {
 		    c = txt[n];
 		    if (multi && count++)
 			(void) tb_sappend(p, "\" \"");
@@ -1316,6 +1425,7 @@ mac_tokens(void)
 int
 mac_token(TBUFF **tok)
 {
+    int result;
     int savcle;
     const char *oldstr = execstr;
 
@@ -1324,9 +1434,15 @@ mac_token(TBUFF **tok)
 
     /* get and advance past token */
     execstr = get_token(execstr, tok, eol_null, EOS, (int *) 0);
+    result = (execstr != oldstr);
+
+    if (clexec && result) {
+	if (execstr == 0 || *skip_blanks(execstr) == EOS)
+	    set_end_string('\n');
+    }
 
     clexec = savcle;
-    return (execstr != oldstr);
+    return result;
 }
 
 /* fetch and isolate and evaluate the next token from execstr */
@@ -1336,15 +1452,15 @@ mac_tokval(TBUFF **tok)
     if (mac_token(tok) != 0) {
 	char *previous = tb_values(*tok);
 	const char *newvalue = tokval(previous);
+	int old_type = toktyp(previous);
 	/* evaluate the token */
 	if (previous != 0
 	    && newvalue != 0
+	    && old_type != TOK_QUOTSTR
 	    && (const char *) previous != newvalue) {
 	    /*
 	     * Check for the special case where we're just shifting the result
-	     * down by one since we're stripping a leading quote.  I don't want
-	     * to use memmove() in tb_buff.c just for this instance since it's
-	     * less portable -TD
+	     * down by one since we're stripping a leading quote.
 	     */
 	    if (((const char *) previous) + 1 == newvalue) {
 		TBUFF *fix = 0;
@@ -1362,6 +1478,21 @@ mac_tokval(TBUFF **tok)
     return (0);
 }
 
+/*
+ * The result of mac_tokval() could be a literal string with quotes.
+ * Dequote that level.
+ */
+char *
+mac_unquotedarg(TBUFF **tok)
+{
+    char *result = mac_tokval(tok);
+    if (toktyp(result) == TOK_QUOTSTR) {
+	tb_dequote(tok);
+	result = tb_values(*tok);
+    }
+    return result;
+}
+
 /*
  * get a macro line argument
  */
@@ -1414,7 +1545,7 @@ decode_parameter_info(TBUFF *tok, PARAM_INFO * result)
 #endif
 
 static int
-setup_macro_buffer(TBUFF *name, int flag)
+setup_macro_buffer(TBUFF *name, int bufnum, UINT flags)
 {
 #if OPT_MACRO_ARGS || OPT_ONLINEHELP
     static TBUFF *temp;
@@ -1428,10 +1559,10 @@ setup_macro_buffer(TBUFF *name, int flag)
     BUFFER *bp;
 
     /* construct the buffer name for the macro */
-    if (flag < 0)
+    if (bufnum < 0)
 	(void) add_brackets(bname, tb_values(name));
     else
-	(void) lsprintf(bname, MACRO_N_BufName, flag);
+	(void) lsprintf(bname, MACRO_N_BufName, bufnum);
 
     /* set up the buffer for the new macro */
     if ((bp = bfind(bname, BFINVS)) == NULL) {
@@ -1450,7 +1581,7 @@ setup_macro_buffer(TBUFF *name, int flag)
 
     /* save this into the list of : names */
 #if OPT_NAMEBST
-    if (flag < 0) {
+    if (bufnum < 0) {
 	CMDFUNC *cf = typecalloc(CMDFUNC);
 
 	if (!cf)
@@ -1461,7 +1592,7 @@ setup_macro_buffer(TBUFF *name, int flag)
 #else
 	cf->cu.c_buff = bp;
 #endif
-	cf->c_flags = UNDO | REDO | CMD_PROC | VIEWOK;
+	cf->c_flags = UNDO | REDO | VIEWOK | flags;
 
 #if OPT_MACRO_ARGS || OPT_ONLINEHELP
 	if (limit != 0) {
@@ -1515,32 +1646,41 @@ setup_macro_buffer(TBUFF *name, int flag)
     return TRUE;
 }
 
+/* can't store macros interactively */
+static int
+can_store_macro(const char *name)
+{
+    int status = (clexec != 0);
+
+    if (!status) {
+	mlforce("[Cannot store %s interactively]", name);
+    }
+    return status;
+}
+
 /*
  * Set up a macro buffer and flag to store all executed command lines there. 
  * 'n' is the macro number to use.
  */
 int
-storemac(int f, int n)
+store_macro(int f, int n)
 {
-    /* can't store macros interactively */
-    if (!clexec) {
-	mlforce("[Cannot store macros interactively]");
-	return FALSE;
-    }
+    int status;
 
-    /* the numeric arg is our only way of distinguishing macros */
-    if (!f) {
+    if (!can_store_macro("macros")) {
+	status = FALSE;
+    } else if (!f) {
+	/* the numeric arg is our only way of distinguishing macros */
 	mlforce("[No macro specified]");
-	return FALSE;
-    }
-
-    /* range check the macro number */
-    if (n < 1 || n > OPT_EXEC_MACROS) {
+	status = FALSE;
+    } else if (n < 1 || n > OPT_EXEC_MACROS) {
+	/* range check the macro number */
 	mlforce("[Macro number out of range]");
-	return FALSE;
+	status = FALSE;
+    } else {
+	status = setup_macro_buffer((TBUFF *) 0, n, (UINT) CMD_PROC);
     }
-
-    return setup_macro_buffer((TBUFF *) 0, n);
+    return status;
 }
 
 #if OPT_PROCEDURES
@@ -1549,28 +1689,51 @@ storemac(int f, int n)
  * there.  'n' is the macro number to use.
  */
 int
-storeproc(int f, int n)
+store_proc(int f, int n)
 {
     static TBUFF *name;		/* procedure name */
     int status;			/* return status */
 
-    /* if a number was given, then they must mean macro N */
-    if (f)
-	return storemac(f, n);
+    if (f) {
+	/* if a number was given, then they must mean macro N */
+	status = store_macro(f, n);
+    } else if (!can_store_macro("procedures")) {
+	status = FALSE;
+    } else {
 
-    /* can't store procedures interactively */
-    if (!clexec) {
-	mlforce("[Cannot store procedures interactively]");
-	return FALSE;
+	/* get the name of the procedure */
+	tb_scopy(&name, "");
+	if ((status = kbd_reply("Procedure name: ", &name,
+				eol_command, ' ',
+				KBD_NORMAL, no_completion)) == TRUE) {
+	    status = setup_macro_buffer(name, -1, (UINT) CMD_PROC);
+	}
     }
+    return status;
+}
 
-    /* get the name of the procedure */
-    tb_scopy(&name, "");
-    if ((status = kbd_reply("Procedure name: ", &name,
-			    eol_command, ' ', KBD_NORMAL, no_completion)) != TRUE)
-	return status;
+int
+store_op(int f, int n)
+{
+    static TBUFF *name;		/* procedure name */
+    int status;			/* return status */
 
-    return setup_macro_buffer(name, -1);
+    (void) f;
+    (void) n;
+
+    if (!can_store_macro("operators")) {
+	status = FALSE;
+    } else {
+
+	/* get the name of the procedure */
+	tb_scopy(&name, "");
+	if ((status = kbd_reply("Operator name: ", &name,
+				eol_command, ' ',
+				KBD_NORMAL, no_completion)) == TRUE) {
+	    status = setup_macro_buffer(name, -1, (UINT) CMD_OPER);
+	}
+    }
+    return status;
 }
 
 /*
@@ -1586,7 +1749,7 @@ execproc(int f, int n)
     char bufn[NBUFN];		/* name of buffer to execute */
 
     if ((status = mlreply("Execute procedure: ",
-			  name, sizeof(name))) != TRUE) {
+			  name, (int) sizeof(name))) != TRUE) {
 	return status;
     }
 
@@ -1597,7 +1760,7 @@ execproc(int f, int n)
     if ((bp = find_b_name(bufn)) == NULL) {
 	return FALSE;
     }
-    return dobuf(bp, need_a_count(f, n, 1));
+    return dobuf(bp, need_a_count(f, n, 1), -1);
 
 }
 
@@ -1624,7 +1787,7 @@ execbuf(int f, int n)
 	return FALSE;
     }
 
-    return dobuf(bp, need_a_count(f, n, 1));
+    return dobuf(bp, need_a_count(f, n, 1), -1);
 }
 #endif
 
@@ -1692,7 +1855,7 @@ push_variable(char *name)
     p->name = strmalloc(name);
     p->m_value = 0;
     p->p_value = tokval(name);
-    if (p->p_value == error_val) {
+    if (isErrorVal(p->p_value)) {
 	/* special case: so we can delete vars */
 	if (toktyp(name) != TOK_TEMPVAR) {
 	    free(p->name);
@@ -1871,7 +2034,7 @@ begin_directive(char **const cmdpp,
 
     case D_WHILE:
 	if (!ifstk.disabled) {
-	    if ((value = mac_tokval(&argtkn)) == 0) {
+	    if ((value = mac_unquotedarg(&argtkn)) == 0) {
 		status = DDIR_INCOMPLETE;
 		break;
 	    } else if (scan_bool(value) != TRUE) {
@@ -1902,7 +2065,7 @@ begin_directive(char **const cmdpp,
 	if (!ifstk.disabled) {
 	    ifstk.fired = FALSE;
 	    ifstk.disabled = ifstk.level;
-	    if ((value = mac_tokval(&argtkn)) == 0)
+	    if ((value = mac_unquotedarg(&argtkn)) == 0)
 		status = DDIR_INCOMPLETE;
 	    else if (scan_bool(value) == TRUE) {
 		ifstk.disabled = 0;
@@ -1918,7 +2081,7 @@ begin_directive(char **const cmdpp,
 	    if (ifstk.fired) {
 		if (!ifstk.disabled)
 		    ifstk.disabled = ifstk.level;
-	    } else if ((value = mac_tokval(&argtkn)) == 0) {
+	    } else if ((value = mac_unquotedarg(&argtkn)) == 0) {
 		status = DDIR_INCOMPLETE;
 	    } else if (!ifstk.fired
 		       && ifstk.disabled == ifstk.level
@@ -2029,7 +2192,7 @@ begin_directive(char **const cmdpp,
 
     case D_TRACE:
 	if (!ifstk.disabled) {
-	    if ((value = mac_tokval(&argtkn)) == 0) {
+	    if ((value = mac_unquotedarg(&argtkn)) == 0) {
 		mlforce("[Trace is %s]", tracemacros ? "on" : "off");
 	    } else {
 		tracemacros = scan_bool(value);
@@ -2089,7 +2252,7 @@ setup_dobuf(BUFFER *bp, WHLOOP ** result)
 	if (i <= 0)
 	    continue;
 
-	dirnum = dname_to_dirnum(&cmdp, i);
+	dirnum = dname_to_dirnum(&cmdp, (size_t) i);
 	if (dirnum == D_WHILE ||
 	    dirnum == D_BREAK ||
 	    dirnum == D_ENDWHILE) {
@@ -2171,7 +2334,7 @@ setup_dobuf(BUFFER *bp, WHLOOP ** result)
 
 #if OPT_TRACE && !SMALLER
 static const char *
-TraceIndent(int level, char *cmdp, int length)
+TraceIndent(int level, char *cmdp, size_t length)
 {
     static const char indent[] = ".  .  .  .  .  .  .  .  ";
     switch (dname_to_dirnum(&cmdp, length)) {
@@ -2184,7 +2347,7 @@ TraceIndent(int level, char *cmdp, int length)
     default:
 	break;
     }
-    level = strlen(indent) - (3 * level);
+    level = (int) strlen(indent) - (3 * level);
     if (level < 0)
 	level = 0;
     return &indent[level];
@@ -2228,7 +2391,7 @@ compute_indent(char *cmd, size_t length, int *indent, int *indstate)
 	int next = 0;
 
 	cmd = skip_blanks(cmd);
-	dirnum = dname_to_dirnum(&cmd, (length - (cmd - base)));
+	dirnum = dname_to_dirnum(&cmd, (length - (size_t) (cmd - base)));
 	switch (dirnum) {
 	case D_IF:		/* FALLTHRU */
 	case D_WHILE:		/* FALLTHRU */
@@ -2251,8 +2414,10 @@ compute_indent(char *cmd, size_t length, int *indent, int *indstate)
 	    break;
 	}
 	*indstate = next;
-	if ((*indent += here) < 0)
-	    *indent = 0;
+	if (indent != 0) {
+	    if ((*indent += here) < 0)
+		*indent = 0;
+	}
     }
 }
 
@@ -2274,7 +2439,7 @@ handle_endm(void)
 	macrobuffer->b_dot.o = 0;
 	macrobuffer = 0;
 #if !SMALLER
-	compute_indent((char *) 0, 0, (int *) 0, (int *) 0);
+	compute_indent((char *) 0, (size_t) 0, (int *) 0, (int *) 0);
 #endif
     }
 }
@@ -2301,7 +2466,7 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
     static BUFFER *dobuferrbp = 0;
 
     TRACE((T_CALLED "perform_dobuf(bp=%p, whlist=%p) buffer '%s'\n",
-	   bp, whlist, bp->b_bname));
+	   (void *) bp, (void *) whlist, bp->b_bname));
 
     bp->b_inuse++;
 
@@ -2314,10 +2479,11 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 	if (llength(lp) <= 0)
 	    linlen = 0;
 	else
-	    linlen = llength(lp);
+	    linlen = (size_t) llength(lp);
 
 	if (glue) {
-	    if ((linebuf = castrealloc(char, linebuf, glue + linlen + 1)) == 0) {
+	    linebuf = castrealloc(char, linebuf, (size_t) glue + linlen + 1);
+	    if (linebuf == 0) {
 		status = no_memory("during macro execution");
 		break;
 	    }
@@ -2328,7 +2494,8 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 #if !SMALLER
 	    compute_indent(lvalue(lp), linlen, &indent, &indstate);
 #endif
-	    if ((linebuf = cmdp = castalloc(char, indent + linlen + 1)) == 0) {
+	    linebuf = cmdp = castalloc(char, (size_t) indent + linlen + 1);
+	    if (linebuf == 0) {
 		status = no_memory("during macro execution");
 		break;
 	    }
@@ -2369,7 +2536,7 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 	if (lforw(lp) != buf_head(bp)
 	    && linlen != 0
 	    && isEscaped(cmdp + linlen)) {
-	    glue = (int) (linlen + (cmdp - linebuf) - 1);
+	    glue = (int) (linlen + (size_t) (cmdp - linebuf) - 1);
 	    continue;
 	}
 	cmdp = skip_space_tab(linebuf);
@@ -2555,7 +2722,7 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
  * execute the contents of a buffer
  */
 int
-dobuf(BUFFER *bp, int limit)
+dobuf(BUFFER *bp, int limit, int real_cmd_count)
 {
     TBUFF *macro_result = 0;
     int status = FALSE;
@@ -2579,7 +2746,9 @@ dobuf(BUFFER *bp, int limit)
 
 	    for (counter = 1; counter <= limit; counter++) {
 		if (dobufnesting == 1)
-		    cmd_count = counter;
+		    cmd_count = ((real_cmd_count >= 0)
+				 ? real_cmd_count
+				 : counter);
 
 #if ! SMALLER
 		if (setup_dobuf(bp, &whlist) != TRUE) {
@@ -2656,6 +2825,16 @@ dobuf(BUFFER *bp, int limit)
     returnCode(status);
 }
 
+/*
+ * Dummy function used to tell vile_op() that we are passing a user-defined
+ * operator's name in the last parameter.
+ */
+int
+user_operator(void)
+{
+    return TRUE;
+}
+
 /*
  * Common function for startup-file, and for :so command.
  */
@@ -2777,7 +2956,7 @@ dofile(char *fname)
 
 	/* go execute it! */
 	bp->b_refcount += 1;
-	status = dobuf(bp, 1);
+	status = dobuf(bp, 1, -1);
 	bp->b_refcount -= 1;
 
 	/*
@@ -2816,7 +2995,7 @@ cbuf(int f, int n, int bufnum)
 	return FALSE;
     }
 
-    return dobuf(bp, need_a_count(f, n, 1));
+    return dobuf(bp, need_a_count(f, n, 1), -1);
 }
 
 #include "neexec.h"
