@@ -5,14 +5,14 @@
  * reading and writing of the disk are
  * in "fileio.c".
  *
- * $Header: /usr/build/vile/vile/RCS/file.c,v 1.415 2008/04/15 00:10:33 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/file.c,v 1.446 2010/09/06 18:21:27 tom Exp $
  */
 
 #include "estruct.h"
 #include "edef.h"
 #include "nefsms.h"
 
-#if SYS_WINNT
+#if SYS_WINNT || SYS_MINGW
 #include <io.h>			/* for mktemp */
 #include <direct.h>		/* for mkdir */
 #define vl_mkdir(path,mode) mkdir(path)
@@ -186,7 +186,7 @@ check_file_changed(BUFFER *bp, char *fn)
 #if OPT_LCKFILES
 	/* release own lock before read the file again */
 	if (global_g_val(GMDUSEFILELOCK)) {
-	    if (!b_val(curbp, MDLOCKED) && !b_val(curbp, MDVIEW))
+	    if (!b_val(bp, MDLOCKED) && !b_val(bp, MDVIEW))
 		release_lock(fn);
 	}
 #endif
@@ -281,6 +281,7 @@ CleanAfterPipe(int Wrote)
     if (must_clean_pipe) {
 #if SYS_UNIX || SYS_MSDOS
 	term.unclean();		/* may clear the screen as a side-effect */
+	term.open();		/* (re)open */
 	term.kopen();
 	term.flush();
 	if (Wrote)
@@ -496,6 +497,8 @@ fileuid_compare(FUID * fuid1, FUID * fuid2)
 	    fuid1->ino == fuid2->ino &&
 	    fuid1->dev == fuid2->dev);
 #else
+    (void) fuid1;
+    (void) fuid2;
     return (FALSE);		/* Doesn't really matter */
 #endif
 }
@@ -506,6 +509,8 @@ fileuid_same(BUFFER *bp, FUID * fuid)
 #ifdef CAN_CHECK_INO
     return fileuid_compare(&bp->b_fileuid, fuid);
 #else
+    (void) bp;
+    (void) fuid;
     return (FALSE);		/* Doesn't really matter */
 #endif
 }
@@ -576,6 +581,7 @@ reset_to_unnamed(BUFFER *bp)
 int
 fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    BUFFER *bp = curbp;
     int status;
     char fname[NFILEN];
     W_VALUES *save_wvals;
@@ -586,18 +592,18 @@ fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 	if ((status = mlreply_file("Replace with file: ", (TBUFF **) 0,
 				   FILEC_REREAD, fname)) != TRUE)
 	    returnCode(status);
-    } else if (b_is_temporary(curbp)) {
+    } else if (b_is_temporary(bp)) {
 	returnCode(cannot_reread());
-    } else if (!(global_g_val(GMDWARNREREAD) || curbp->b_fname[0] == EOS)
-	       || ((status = mlyesno("Reread current buffer")) == TRUE)) {
-	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
+    } else if (!(global_g_val(GMDWARNREREAD) || bp->b_fname[0] == EOS)
+	       || (mlyesno("Reread current buffer") == TRUE)) {
+	(void) vl_strncpy(fname, bp->b_fname, sizeof(fname));
 	/* Check if we are rereading an unnamed-buffer if it is not
 	 * associated with a file.
 	 */
-	if (is_internalname(curbp->b_fname)) {
-	    if (eql_bname(curbp, STDIN_BufName)
-		|| eql_bname(curbp, UNNAMED_BufName)) {
-		status = bclear(curbp);
+	if (is_internalname(bp->b_fname)) {
+	    if (eql_bname(bp, STDIN_BufName)
+		|| eql_bname(bp, UNNAMED_BufName)) {
+		status = bclear(bp);
 		if (status == TRUE)
 		    mlerase();
 		curwp->w_flag |= (WFMODE | WFHARD);
@@ -612,31 +618,31 @@ fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 #if OPT_LCKFILES
     /* release own lock before read the replaced file */
     if (global_g_val(GMDUSEFILELOCK)) {
-	if (!b_val(curbp, MDLOCKED) && !b_val(curbp, MDVIEW))
-	    release_lock(curbp->b_fname);
+	if (!b_val(bp, MDLOCKED) && !b_val(bp, MDVIEW))
+	    release_lock(bp->b_fname);
     }
 #endif
 
     /* we want no errors or complaints, so mark it unchanged */
-    b_clr_changed(curbp);
+    b_clr_changed(bp);
 
-    save_wvals = save_window_modes(curbp);
-    status = readin(fname, TRUE, curbp, TRUE);
+    save_wvals = save_window_modes(bp);
+    status = readin(fname, TRUE, bp, TRUE);
     if (status == ABORT) {
-	reset_to_unnamed(curbp);
+	reset_to_unnamed(bp);
     } else {
-	set_buffer_name(curbp);
+	set_buffer_name(bp);
 #if OPT_FINDERR
 	/*
 	 * Doing ":e!" on [Output] will reset the buffer name to match the
 	 * command.  Even if it was not [Output], it is still useful to make
 	 * the error-buffer match the last-read shell/pipe output.
 	 */
-	if (isShellOrPipe(curbp->b_fname))
-	    set_febuff(curbp->b_bname);
+	if (isShellOrPipe(bp->b_fname))
+	    set_febuff(bp->b_bname);
 #endif
     }
-    restore_window_modes(curbp, save_wvals);
+    restore_window_modes(bp, save_wvals);
 
     returnCode(status);
 }
@@ -818,12 +824,12 @@ writelinesmsg(char *fn, int nline, B_COUNT nchar)
 	chars_len = (int) strlen(strchars);
 	outlen = (term.cols - 1) -
 	    (
-		(sizeof(WRITE_FILE_FMT) - 13) +
-		strlen(action) +
+		(int) (sizeof(WRITE_FILE_FMT) - 13) +
+		(int) strlen(action) +
 		lines_len +
 		chars_len +
-		strlen(PLURAL(nline)) +
-		strlen(PLURAL(nchar))
+		(int) strlen(PLURAL(nline)) +
+		(int) strlen(PLURAL(nchar))
 	    );
 	mlforce(WRITE_FILE_FMT,
 		action,
@@ -831,7 +837,7 @@ writelinesmsg(char *fn, int nline, B_COUNT nchar)
 		PLURAL(nline),
 		strchars,
 		PLURAL(nchar),
-		path_trunc(fn, outlen, tmp, sizeof(tmp)));
+		path_trunc(fn, outlen, tmp, (int) sizeof(tmp)));
     } else {
 	mlforce("[%d lines]", nline);
     }
@@ -857,7 +863,7 @@ set_rdonly_modes(BUFFER *bp, int always)
 static void
 ff_read_directory(BUFFER *bp, char *fname)
 {
-    int count = fill_directory_buffer(bp, fname, 0);
+    int count = fill_directory_buffer(bp, fname, (size_t) 0);
 
     mlwrite("[Read %d line%s]", count, PLURAL(count));
     b_set_flags(bp, BFDIRS);
@@ -898,6 +904,9 @@ ff_load_directory(char *fname)
 static int
 kifile(char *fname)
 {
+#if OPT_ENCRYPT
+    BUFFER *bp = curbp;
+#endif
     int s;
     int nline;
     size_t i;
@@ -918,7 +927,7 @@ kifile(char *fname)
 
 	nline = 0;
 #if OPT_ENCRYPT
-	if ((s = vl_resetkey(curbp, fname)) == TRUE)
+	if ((s = vl_resetkey(bp, fname)) == TRUE)
 #endif
 	{
 	    mlwrite("[Reading...]");
@@ -1011,8 +1020,9 @@ getfile2bp(const char *fname,	/* file name to find */
 	    if (have_fuid) {
 		for_each_buffer(bp) {
 		    /* is the same unique file */
-		    if (fileuid_same(bp, &fuid))
+		    if (fileuid_same(bp, &fuid)) {
 			return bp;
+		    }
 		}
 	    }
 	}
@@ -1028,7 +1038,9 @@ getfile2bp(const char *fname,	/* file name to find */
 	makename(bname, nfname);	/* New buffer name.     */
 	/* make sure the buffer name doesn't exist */
 	while ((bp = find_b_name(bname)) != NULL) {
-	    if (!b_is_changed(bp) && is_empty_buf(bp) &&
+	    if (!b_is_argument(bp) &&
+		!b_is_changed(bp) &&
+		is_empty_buf(bp) &&
 		!ffexists(bp->b_fname)) {
 		/* empty, unmodified, and non-existent --
 		   then it's okay to re-use this buffer */
@@ -1041,7 +1053,7 @@ getfile2bp(const char *fname,	/* file name to find */
 	    if (!ok_to_ask || !global_g_val(GMDWARNRENAME))
 		continue;
 	    hst_glue(' ');
-	    s = mlreply("Will use buffer name: ", bname, sizeof(bname));
+	    s = mlreply("Will use buffer name: ", bname, (int) sizeof(bname));
 	    if (s == ABORT)
 		return 0;
 	    if (s == FALSE || bname[0] == EOS)
@@ -1158,7 +1170,7 @@ strip_if_dosmode(BUFFER *bp)
 		if ((chr == '\r')
 		    || (len == 1 && chr == '\032')) {
 		    if (b_val(bp, MDUNDO_DOS_TRIM)) {
-			copy_for_undo(lp);
+			CopyForUndo(lp);
 		    }
 		    llength(lp)--;
 		    flag = TRUE;
@@ -1217,17 +1229,17 @@ explicit_dosmode(BUFFER *bp, RECORD_SEP record_sep)
  * Recompute the buffer size, redisplay the [Settings] buffer.
  */
 static int
-modified_record_sep(RECORD_SEP record_sep)
+modified_record_sep(BUFFER *bp, RECORD_SEP record_sep)
 {
     TRACE((T_CALLED "modified_record_sep(%s)\n",
 	   choice_to_name(&fsm_recordsep_blist, record_sep)));
 
-    explicit_dosmode(curbp, record_sep);
-    guess_dosmode(curbp);
-    explicit_dosmode(curbp, record_sep);	/* ignore the guess - only want to strip CR's */
+    explicit_dosmode(bp, record_sep);
+    guess_dosmode(bp);
+    explicit_dosmode(bp, record_sep);	/* ignore the guess - only want to strip CR's */
 
-    b_clr_counted(curbp);
-    bsizes(curbp);
+    b_clr_counted(bp);
+    bsizes(bp);
 
     returnCode(TRUE);
 }
@@ -1239,7 +1251,7 @@ modified_record_sep(RECORD_SEP record_sep)
 int
 set_rs_crlf(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return modified_record_sep(RS_CRLF);
+    return modified_record_sep(curbp, RS_CRLF);
 }
 
 /* as above, but forces unix-mode instead */
@@ -1247,7 +1259,7 @@ set_rs_crlf(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 set_rs_lf(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return modified_record_sep(RS_LF);
+    return modified_record_sep(curbp, RS_LF);
 }
 
 /* as above, but forces macintosh-mode instead */
@@ -1255,7 +1267,7 @@ set_rs_lf(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 set_rs_cr(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return modified_record_sep(RS_CR);
+    return modified_record_sep(curbp, RS_CR);
 }
 #endif /* OPT_DOSFILES */
 
@@ -1337,30 +1349,37 @@ guess_recordseparator(BUFFER *bp, UCHAR * buffer, B_COUNT length, L_NUM * lines)
 	}
     }
 
-    TRACE(("guess_recordseparator assume %s CR:%ld, LF:%ld, CRLF:%ld\n",
+    TRACE(("guess_recordseparator assume %s, rs=%s CR:%ld, LF:%ld, CRLF:%ld\n",
 	   global_b_val(MDDOS) ? "dos" : "unix",
+	   choice_to_name(&fsm_recordsep_blist, global_b_val(VAL_RECORD_SEP)),
 	   (long) (count_cr - count_dos), (long) count_unix, (long) count_dos));
 
     if (count_lf != 0) {
 	result = RS_LF;
-	if (global_b_val(MDDOS)) {
+	if (global_b_val(VAL_RECORD_SEP) == RS_AUTO
+	    && (count_dos == 0 || count_dos == count_lf)) {
+	    result = count_dos ? RS_CRLF : RS_LF;
+	} else if (global_b_val(MDDOS)) {
 	    if ((bp->b_flag & BFEXEC) != 0) {
 		result = (count_dos && !count_unix) ? RS_CRLF : RS_LF;
-	    } else if (check_percent_crlf(bp, count_dos, count_unix)) {
+	    } else if (check_percent_crlf(bp, (int) count_dos, (int) count_unix)) {
 		result = RS_CRLF;
 	    }
+	} else if (global_b_val(VAL_RECORD_SEP) == RS_CR
+		   && (count_cr || !count_unix)) {
+	    result = RS_CR;
 	}
 	if (buffer[length - 1] != '\n')
 	    ++count_lf;
-	*lines = count_lf;
+	*lines = (L_NUM) ((result == RS_CR) ? count_cr : count_lf);
     } else if (count_cr != 0) {
 	result = RS_CR;
 	if (buffer[length - 1] != '\r')
 	    ++count_cr;
-	*lines = count_cr;
+	*lines = (L_NUM) count_cr;
     } else {
 	*lines = 1;		/* we still need a line, to allocate data */
-	result = RS_DEFAULT;
+	result = RS_SYS_DEFAULT;
     }
 
     TRACE(("...line count = %ld, format=%s\n", (long) *lines,
@@ -1428,11 +1447,11 @@ quickreadf(BUFFER *bp, int *nlinep)
     } else if (request == 0) {
 	/* avoid malloc(0) problems down below; let slowreadf() do the work */
 	rc = FIONUL;
-    } else if ((buffer = castalloc(UCHAR, request)) == NULL) {
+    } else if ((buffer = castalloc(UCHAR, request + 1)) == NULL) {
 	rc = FIOMEM;
     }
 #if OPT_ENCRYPT
-    else if ((rc = vl_resetkey(curbp, (const char *) buffer)) != TRUE) {
+    else if ((rc = vl_resetkey(bp, (const char *) buffer)) != TRUE) {
 	free(buffer);
     }
 #endif
@@ -1454,7 +1473,7 @@ quickreadf(BUFFER *bp, int *nlinep)
 #if OPT_ENCRYPT
 	if (b_val(bp, MDCRYPT)
 	    && bp->b_cryptkey[0]) {	/* decrypt the file */
-	    vl_setup_encrypt(bp->b_cryptkey);
+	    vl_setup_encrypt(bp->b_cryptkey, seed);
 	    vl_encrypt_blok((char *) buffer, (UINT) length);
 	}
 #endif
@@ -1471,12 +1490,13 @@ quickreadf(BUFFER *bp, int *nlinep)
 	/*
 	 * Modify readin()'s setting for newline mode if needed:
 	 */
-	if (buffer[length - 1] != (UCHAR) (rscode == RS_CR ? '\r' : '\n')) {
+	if (length == 0
+	    || (buffer[length - 1] != (UCHAR) (rscode == RS_CR ? '\r' : '\n'))) {
 	    set_b_val(bp, MDNEWLINE, FALSE);
 	}
 
 	/* allocate all of the line structs we'll need */
-	if ((bp->b_LINEs = typeallocn(LINE, nlines)) == NULL) {
+	if ((bp->b_LINEs = typeallocn(LINE, (unsigned) nlines)) == NULL) {
 	    free(buffer);
 	    ffrewind();
 	    rc = FIOMEM;
@@ -1489,17 +1509,19 @@ quickreadf(BUFFER *bp, int *nlinep)
 
 	    /* loop through the buffer again, creating
 	       line data structure for each line */
+	    memset(bp->b_LINEs, 0, sizeof(LINE));
 	    for (lp = bp->b_LINEs, offset = 0; lp != bp->b_LINEs_end; ++lp) {
 		B_COUNT next = next_recordseparator(buffer, length, rscode, offset);
-		if (next == offset)
+		if (next == offset) {
 		    break;
+		}
 #if !SMALLER
 		lp->l_number = ++lineno;
 #endif
-		llength(lp) = next - offset - 1;
+		llength(lp) = (C_NUM) (next - offset - 1);
 		if (!b_val(bp, MDNEWLINE) && next == length)
 		    llength(lp) += 1;
-		lp->l_size = llength(lp) + 1;
+		lp->l_size = (size_t) (llength(lp) + 1);
 		lvalue(lp) = (char *) (buffer + offset);
 		set_lforw(lp, lp + 1);
 		if (lp != bp->b_LINEs)
@@ -1512,7 +1534,8 @@ quickreadf(BUFFER *bp, int *nlinep)
 		decode_charset(bp, lp);
 		offset = next;
 	    }
-	    lp--;		/* point at last line again */
+	    if (lp != bp->b_LINEs)
+		lp--;		/* point at last line again */
 
 	    /* connect the end of the list */
 	    set_lforw(lp, buf_head(bp));
@@ -1558,7 +1581,7 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 #endif
 
     TRACE((T_CALLED "readin(fname=%s, lockfl=%d, bp=%p, mflg=%d)\n",
-	   NONNULL(fname), lockfl, bp, mflg));
+	   fname, lockfl, (void *) bp, mflg));
 
 #if SYS_WINNT && DISP_NTWIN
     ffshadow = file_is_closed;	/* an assumption */
@@ -1620,10 +1643,10 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 	    int outlen;
 	    char tmp[NFILEN];
 
-	    outlen = (term.cols - 1) -
-		(sizeof(READING_FILE_FMT) - 3);
+	    outlen = ((term.cols - 1) -
+		      (int) (sizeof(READING_FILE_FMT) - 3));
 	    mlforce(READING_FILE_FMT,
-		    path_trunc(fname, outlen, tmp, sizeof(tmp)));
+		    path_trunc(fname, outlen, tmp, (int) sizeof(tmp)));
 #undef READING_FILE_FMT
 	}
 #if OPT_VMS_PATH
@@ -1762,7 +1785,7 @@ slowreadf(BUFFER *bp, int *nlinep)
     TRACE((T_CALLED "slowreadf(buffer=%s, file=%s)\n", bp->b_bname, bp->b_fname));
 
 #if OPT_ENCRYPT
-    if ((s = vl_resetkey(curbp, curbp->b_fname)) != TRUE)
+    if ((s = vl_resetkey(bp, bp->b_fname)) != TRUE)
 	returnCode(s);
 #endif
     b_set_counted(bp);		/* make 'addline()' do the counting */
@@ -1863,7 +1886,7 @@ slowreadf(BUFFER *bp, int *nlinep)
 	TRACE(("...try looking for UTF-8\n"));
 	for_each_line(lp, bp) {
 	    if (llength(lp) > 0) {
-		check = check_utf8((UCHAR *) lvalue(lp), llength(lp));
+		check = check_utf8((UCHAR *) lvalue(lp), (B_COUNT) llength(lp));
 		if (check == FALSE) {
 		    found = FALSE;
 		} else if (check == TRUE) {
@@ -1887,6 +1910,57 @@ slowreadf(BUFFER *bp, int *nlinep)
     returnCode(s);
 }
 
+/*
+ * Append to the buffer-name characters up to the allowed length, which will
+ * make a well-formed name.  The 'blanks' option is true to allow embedded
+ * blanks, e.g., for a shell command.  A trailing null is added after the
+ * 'len' bytes have been appended.
+ */
+static char *
+add_to_name(char *target, const char *source, size_t len, int blanks)
+{
+    while (len-- != 0) {
+	UINT ch;
+#if OPT_MULTIBYTE
+	if (vl_encoding >= enc_UTF8) {
+	    int used;
+	    int check;
+	    int code;
+
+	    used = vl_conv_to_utf32(&ch, source, len + 1);
+	    if (used <= 0)
+		break;
+	    source += used;
+
+	    if ((code = vl_ucs_to_8bit(&check, (int) ch)) != 0) {
+		if (ch <= 127)
+		    ch = (UINT) check;
+		else
+		    code = 0;
+	    }
+	    if (!code) {
+		if (isPrint(ch)) {
+		    ch = 'X';
+		} else {
+		    ch = '?';
+		}
+	    }
+	} else {
+	    ch = CharOf(*source++);
+	}
+#else
+	ch = CharOf(*source++);
+#endif
+	if (!isPrint(ch))
+	    break;
+	if (isSpace(ch))
+	    ch = (UINT) (blanks ? ' ' : '-');
+	*target++ = (char) ch;
+    }
+    *target = '\0';		/* buffer names are always null-terminated */
+    return target;
+}
+
 /*
  * Take a (null-terminated) file name, and from it fabricate a buffer name.
  * This routine knows about the syntax of file names on the target system.  I
@@ -1898,7 +1972,6 @@ makename(char *bname, const char *fname)
 {
     char *fcp;
     char *bcp;
-    int j;
     char temp[NFILEN];
 
     TRACE(("makename(%s)\n", fname));
@@ -1940,44 +2013,19 @@ makename(char *bname, const char *fname)
 		++fcp;
 	    } while (isSpace(*fcp));
 
-	    (void) vl_strncpy(bcp, fcp, NBUFN - 2);
-
-	    for (j = 4; (j < NBUFN) && isPrint(*bcp); j++) {
-		bcp++;
-	    }
+	    bcp = add_to_name(bcp, fcp, (size_t) (NBUFN - 4), TRUE);
 	    (void) strcpy(bcp, SCRTCH_RIGHT);
 
 	} else {
 
-	    (void) vl_strncpy(bcp, pathleaf(fcp), NBUFN);
+	    (void) add_to_name(bcp, pathleaf(fcp), (size_t) (NBUFN - 1), FALSE);
 
-#if SYS_UNIX
-	    /* UNIX filenames can have any characters (other than EOS!).
-	     * Refuse (rightly) to deal with leading/trailing blanks, but allow
-	     * embedded blanks.  For this special case, ensure that the buffer
-	     * name has no blanks, otherwise it is difficult to reference from
-	     * commands.
-	     */
-	    for (j = 0; j < NBUFN; j++) {
-		if (*bcp == EOS)
-		    break;
-		if (isSpace(*bcp))
-		    *bcp = '-';
-		bcp++;
-	    }
-#endif
 	}
 
 #else /* !(SYS_UNIX||SYS_VMS||SYS_MSDOS) */
 
-	bcp = skip_string(fcp);
-	{
-	    char *cp2 = bname;
-	    vl_strncpy(bname, bcp, NBUFN);
-	    cp2 = strchr(bname, ':');
-	    if (cp2)
-		*cp2 = EOS;
-	}
+	(void) add_to_name(bcp, fcp, NBUFN - 1, FALSE);
+
 #endif
     }
     if (*bname == EOS)
@@ -2002,7 +2050,7 @@ unqname(char *name)
 	j = strlen(strcpy(name, NO_BNAME));
 
     /* check to see if this name is in use */
-    vl_strncpy(newname, name, NBUFN);
+    add_to_name(newname, name, (size_t) (NBUFN - 1), TRUE);
     adjust = is_scratchname(newname);
     strcpy(suffixbuf, "-");
     while (find_b_name(newname) != NULL) {
@@ -2013,8 +2061,11 @@ unqname(char *name)
 	/* that is, put suffix at end if it fits, or else
 	   overwrite some of the name to make it fit */
 	/* the suffix is in "base 36" */
-	suffixlen = format_int(suffixbuf + 1, ++i, 36) + 1;
-	k = NBUFN - 1 - suffixlen;
+	suffixlen = format_int(suffixbuf + 1, (UINT)++ i, 36) + 1;
+	if ((suffixlen + 2) >= NBUFN)
+	    break;
+
+	k = (size_t) (NBUFN - 1 - suffixlen);
 	if (j < k)
 	    k = j;
 	if (adjust) {
@@ -2023,7 +2074,7 @@ unqname(char *name)
 	} else
 	    strcpy(&newname[k], suffixbuf);
     }
-    strncpy0(name, newname, NBUFN);
+    strncpy0(name, newname, (size_t) NBUFN);
     TRACE(("unqname ->%s\n", name));
 }
 
@@ -2037,17 +2088,20 @@ filewrite(int f, int n)
     int s;
     char fname[NFILEN];
     int forced = (f && n == SPECIAL_BANG_ARG);
+    BUFFER *bp = curbp;
 
     if (more_named_cmd()) {
 	if ((s = mlreply_file("Write to file: ", (TBUFF **) 0,
-			      forced ? FILEC_WRITE2
-			      : FILEC_WRITE, fname)) != TRUE)
+			      (UINT) (forced
+				      ? FILEC_WRITE2
+				      : FILEC_WRITE),
+			      fname)) != TRUE)
 	    return s;
     } else
-	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
+	(void) vl_strncpy(fname, bp->b_fname, sizeof(fname));
 
-    if ((s = writeout(fname, curbp, forced, TRUE)) == TRUE)
-	unchg_buff(curbp, 0);
+    if ((s = writeout(fname, bp, forced, TRUE)) == TRUE)
+	unchg_buff(bp, 0);
     return s;
 }
 
@@ -2063,12 +2117,13 @@ filesave(int f, int n)
 {
     int s;
     int forced = (f && n == SPECIAL_BANG_ARG);	/* then it was :w! */
+    BUFFER *bp = curbp;
 
-    if (no_file_name(curbp->b_fname))
+    if (no_file_name(bp->b_fname))
 	return FALSE;
 
-    if ((s = writeout(curbp->b_fname, curbp, forced, TRUE)) == TRUE)
-	unchg_buff(curbp, 0);
+    if ((s = writeout(bp->b_fname, bp, forced, TRUE)) == TRUE)
+	unchg_buff(bp, 0);
     return s;
 }
 
@@ -2091,7 +2146,16 @@ write_encoded_text(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 
     if (b_val(bp, VAL_FILE_ENCODING) > enc_UTF8) {
 	int len = encode_charset(bp, buf, nbuf, ending);
-	rc = ffputline(bp->encode_utf_buf, len, 0);
+	/*
+	 * UTF-16 and UTF-32 buffers are generally bigger than UTF-8.
+	 */
+	if (nbuf > (2 * len)) {
+	    rc = FIOABRT;
+	    mlforce("Error writing buffer as %s (encoding error)!",
+		    encoding2s(b_val(bp, VAL_FILE_ENCODING)));
+	} else {
+	    rc = ffputline(bp->encode_utf_buf, len, 0);
+	}
     } else {
 	rc = ffputline(buf, nbuf, ending);
     }
@@ -2102,77 +2166,29 @@ write_encoded_text(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 #define FFPUTLINE(bp,buf,len,end) ffputline(buf,len,end)
 #endif
 
-static int
-actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int encoded)
+int
+write_region(BUFFER *bp, REGION * rp, int encoded, int *nline, B_COUNT * nchar)
 {
     int s;
     LINE *lp;
-    int nline;
-    B_COUNT nchar;
     const char *ending = get_record_sep(bp);
-    int len_rs = len_record_sep(bp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(bp);
     C_NUM offset = rp->r_orig.o;
 
-    /* this is adequate as long as we cannot write parts of lines */
-    int whole_file = ((rp->r_orig.l == lforw(buf_head(bp)))
-		      && (rp->r_end.l == buf_head(bp)));
-
-    TRACE((T_CALLED " actually_write\n"));
-#if OPT_HOOKS
-    if (run_a_hook(&writehook)) {
-	/*
-	 * The write-hook may have modified the buffer.  Assume
-	 * the worst, and reconstruct the region.
-	 */
-	(void) bsizes(bp);
-	if (whole_file
-	    || line_no(bp, rp->r_orig.l) > bp->b_linecount
-	    || line_no(bp, rp->r_end.l) > bp->b_linecount) {
-	    setup_file_region(bp, rp);
-	} else {
-	    DOT = rp->r_orig;
-	    MK = rp->r_end;
-	    (void) getregion(rp);
-	}
-	offset = rp->r_orig.o;
-    }
-#endif
-
-#if OPT_ENCRYPT
-    if ((s = vl_resetkey(curbp, fn)) != TRUE)
-	returnCode(s);
-#endif
-
-    /* open writes error message, if needed */
-    if ((s = ffwopen(fn, forced)) != FIOSUC)
-	returnCode(FALSE);
-
-    /* disable "working..." while we are writing - not reading - a pipe, since
-     * it may be a quasi-interactive process that we don't want to modify its
-     * display.
-     */
-    if (isShellOrPipe(fn)) {
-	beginDisplay();
-    }
-
-    /* tell us we're writing */
-    if (msgf == TRUE)
-	mlwrite("[Writing...]");
-
-    CleanToPipe(TRUE);
-
     lp = rp->r_orig.l;
-    nline = 0;			/* Number of lines     */
-    nchar = 0;			/* Number of chars     */
+    *nline = 0;			/* Number of lines     */
+    *nchar = 0;			/* Number of chars     */
 
 #if OPT_MULTIBYTE
     if ((s = write_bom(bp)) != FIOSUC) {
 	goto out;
     }
+#else
+    s = FIOSUC;
 #endif
 
     /* first (maybe partial) line and succeeding whole lines */
-    while ((rp->r_size + offset) >= line_length(lp)) {
+    while ((rp->r_size + (B_COUNT) offset) >= (B_COUNT) line_length(lp)) {
 	C_NUM len = llength(lp) - offset;
 	char *text = lvalue(lp) + offset;
 
@@ -2204,21 +2220,97 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
 	if ((s = FFPUTLINE(bp, text, len, ending)) != FIOSUC)
 	    goto out;
 
-	++nline;
-	nchar += line_length(lp);
+	*nline += 1;
+	*nchar += line_length(lp);
 	offset = 0;
 	lp = lforw(lp);
     }
 
     /* last line (fragment) */
     if (rp->r_size > 0) {
-	if ((s = FFPUTLINE(bp, lvalue(lp), rp->r_size, NULL)) != FIOSUC)
+	C_NUM len = llength(lp);
+	char *text = lvalue(lp);
+
+#if OPT_SELECTIONS
+	if (encoded) {
+	    TBUFF *temp;
+	    if ((temp = encode_attributes(lp, bp, rp)) != 0) {
+		text = tb_values(temp);
+		len = (int) tb_length(temp);
+	    }
+	    if ((s = FFPUTLINE(bp, text, len, NULL)) != FIOSUC) {
+		tb_free(&temp);
+		goto out;
+	    }
+	    tb_free(&temp);
+	} else
+#endif
+	if ((s = FFPUTLINE(bp, lvalue(lp), (int) rp->r_size, NULL)) != FIOSUC)
 	    goto out;
-	nchar += rp->r_size;
-	++nline;		/* it _looks_ like a line */
+	*nchar += rp->r_size;
+	*nline += 1;		/* it _looks_ like a line */
     }
 
   out:
+    return s;
+}
+
+static int
+actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int encoded)
+{
+    int s;
+    int nline;
+    B_COUNT nchar;
+
+    /* this is adequate as long as we cannot write parts of lines */
+    int whole_file = ((rp->r_orig.l == lforw(buf_head(bp)))
+		      && (rp->r_end.l == buf_head(bp)));
+
+    TRACE((T_CALLED "actually_write %s\n",
+	   encoded ? "encoded" : "not encoded"));
+#if OPT_HOOKS
+    if (run_a_hook(&writehook)) {
+	/*
+	 * The write-hook may have modified the buffer.  Assume
+	 * the worst, and reconstruct the region.
+	 */
+	(void) bsizes(bp);
+	if (whole_file
+	    || line_no(bp, rp->r_orig.l) > bp->b_linecount
+	    || line_no(bp, rp->r_end.l) > bp->b_linecount) {
+	    setup_file_region(bp, rp);
+	} else {
+	    DOT = rp->r_orig;
+	    MK = rp->r_end;
+	    (void) getregion(bp, rp);
+	}
+    }
+#endif
+
+#if OPT_ENCRYPT
+    if ((s = vl_resetkey(bp, fn)) != TRUE)
+	returnCode(s);
+#endif
+
+    /* open writes error message, if needed */
+    if (ffwopen(fn, forced) != FIOSUC)
+	returnCode(FALSE);
+
+    /* disable "working..." while we are writing - not reading - a pipe, since
+     * it may be a quasi-interactive process that we don't want to modify its
+     * display.
+     */
+    if (isShellOrPipe(fn)) {
+	beginDisplay();
+    }
+
+    /* tell us we're writing */
+    if (msgf == TRUE)
+	mlwrite("[Writing...]");
+
+    CleanToPipe(TRUE);
+
+    s = write_region(bp, rp, encoded, &nline, &nchar);
     if (s == FIOSUC) {		/* No write error.      */
 #if OPT_VMS_PATH
 	if (same_fname(fn, bp, FALSE))
@@ -2275,7 +2367,7 @@ file_protection(char *fn)
     if (!isShellOrPipe(fn)) {
 	if (file_stat(fn, &sb) == 0) {
 	    if (isFileMode(sb.st_mode))
-		result = sb.st_mode & 0777;
+		result = (int) (sb.st_mode & 0777);
 	}
     }
     return result;
@@ -2339,12 +2431,12 @@ writereg(REGION * rp,
 #endif
 			if (!ffaccess(fn, FL_WRITEABLE)
 			    && (protection = file_protection(fn)) >= 0) {
-			    chmod(SL_TO_BSL(fn), protection | 0600);
+			    chmod(SL_TO_BSL(fn), (mode_t) (protection | 0600));
 			}
 		    }
 		    status = actually_write(rp, fn, msgf, bp, forced, encoded);
 		    if (protection > 0)
-			chmod(SL_TO_BSL(fn), protection);
+			chmod(SL_TO_BSL(fn), (mode_t) protection);
 #if SYS_WINNT
 		    if (write_acl_added)
 			(void) w32_remove_write_acl(fn, prev_access_mask);
@@ -2389,6 +2481,7 @@ writeout(const char *fn, BUFFER *bp, int forced, int msgf)
 static int
 prompt_and_write_region(int encoded)
 {
+    BUFFER *bp = curbp;
     REGION region;
     int status;
     char fname[NFILEN];
@@ -2398,15 +2491,15 @@ prompt_and_write_region(int encoded)
 	    mlwrite("Range not written");
 	    return FALSE;
 	}
-	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
+	(void) vl_strncpy(fname, bp->b_fname, sizeof(fname));
     } else {
 	if ((status = mlreply_file("Write region to file: ",
 				   (TBUFF **) 0, FILEC_WRITE | FILEC_PROMPT,
 				   fname)) != TRUE)
 	    return status;
     }
-    if ((status = getregion(&region)) == TRUE)
-	status = writereg(&region, fname, TRUE, curbp, FALSE, encoded);
+    if ((status = getregion(bp, &region)) == TRUE)
+	status = writereg(&region, fname, TRUE, bp, FALSE, encoded);
     return status;
 }
 
@@ -2434,12 +2527,15 @@ write_enc_region(void)
 int
 kwrite(char *fn, int msgf)
 {
+#if OPT_ENCRYPT
+    BUFFER *bp = curbp;
+#endif
     KILL *kp;			/* pointer into kill register */
     int nline;
     B_COUNT nchar;
     int s;
     int c;
-    int i;
+    unsigned i;
     char *sp;			/* pointer into string to insert */
 
     /* make sure there is something to put */
@@ -2449,7 +2545,7 @@ kwrite(char *fn, int msgf)
 	return FALSE;		/* not an error, just nothing */
     }
 #if OPT_ENCRYPT
-    if ((s = vl_resetkey(curbp, fn)) != TRUE)
+    if ((s = vl_resetkey(bp, fn)) != TRUE)
 	return s;
 #endif
     if ((s = ffwopen(fn, FALSE)) != FIOSUC) {	/* Open writes message. */
@@ -2500,6 +2596,7 @@ kwrite(char *fn, int msgf)
 int
 vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    BUFFER *bp = curbp;
     int s;
     char fname[NFILEN];
 
@@ -2515,19 +2612,19 @@ vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
 
 #if OPT_LCKFILES
     if (global_g_val(GMDUSEFILELOCK)) {
-	if (!b_val(curbp, MDLOCKED) && !b_val(curbp, MDVIEW))
-	    release_lock(curbp->b_fname);
-	ch_fname(curbp, fname);
-	make_global_b_val(curbp, MDLOCKED);
-	make_global_b_val(curbp, VAL_LOCKER);
-	grab_lck_file(curbp, fname);
+	if (!b_val(bp, MDLOCKED) && !b_val(bp, MDVIEW))
+	    release_lock(bp->b_fname);
+	ch_fname(bp, fname);
+	make_global_b_val(bp, MDLOCKED);
+	make_global_b_val(bp, VAL_LOCKER);
+	grab_lck_file(bp, fname);
     } else
 #endif
-	ch_fname(curbp, fname);
+	ch_fname(bp, fname);
 
 #if SYS_WINNT && DISP_NTWIN
     /* remember new filename if later written out to disk */
-    b_clr_registered(curbp);
+    b_clr_registered(bp);
 #endif
 
     curwp->w_flag |= WFMODE;
@@ -2542,6 +2639,7 @@ vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 ifile(char *fname, int belowthisline, FILE *haveffp)
 {
+    BUFFER *bp = curbp;
     LINE *prevp;
     LINE *newlp;
     LINE *nextp;
@@ -2550,9 +2648,9 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     int nline;
 
     TRACE((T_CALLED "ifile(fname=%s, belowthisline=%d, haveffp=%p)\n",
-	   NONNULL(fname), belowthisline, haveffp));
+	   NONNULL(fname), belowthisline, (void *) haveffp));
 
-    b_clr_invisible(curbp);	/* we are not temporary */
+    b_clr_invisible(bp);	/* we are not temporary */
     if (haveffp == 0) {
 	if ((status = ffropen(fname)) == FIOERR)	/* Hard file open */
 	    goto out;
@@ -2561,13 +2659,13 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 #if COMPLETE_FILES || COMPLETE_DIRS
 	else if ((status = ff_load_directory(fname)) == FIOERR)
 	    goto out;
-	if (b_is_directory(curbp)) {
+	if (b_is_directory(bp)) {
 	    /* special case: contents are already added */
 	    goto out;
 	}
 #endif
 #if OPT_ENCRYPT
-	if ((status = vl_resetkey(curbp, fname)) != TRUE)
+	if ((status = vl_resetkey(bp, fname)) != TRUE)
 	    returnCode(status);
 #endif
 	mlwrite("[Inserting...]");
@@ -2584,7 +2682,7 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     nextp = 0;
     while ((status = ffgetline(&nbytes)) <= FIOSUC) {
 #if OPT_DOSFILES
-	if (b_val(curbp, MDDOS)
+	if (b_val(bp, MDDOS)
 	    && (nbytes != 0)
 	    && fflinebuf[nbytes - 1] == '\r')
 	    nbytes--;
@@ -2595,19 +2693,21 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 	}
 
 	beginDisplay();
-	if (add_line_at(curbp, prevp, fflinebuf, (int) nbytes) != TRUE) {
+	if (add_line_at(bp, prevp, fflinebuf, (int) nbytes) != TRUE) {
 	    status = FIOMEM;
 	    newlp = 0;
 	} else {
 	    newlp = lforw(prevp);
-	    if ((tag_for_undo(newlp)) == ABORT) {
-		status = FIOMEM;
-		/*
-		 * De-link the line added for which we cannot undo.
-		 * If we don't do this, undoworker() won't find a match
-		 * against the buffer, and will be deadlocked.
-		 */
-		lremove(curbp, prevp);
+	    if (OkUndo(bp)) {
+		if ((tag_for_undo(newlp)) == ABORT) {
+		    status = FIOMEM;
+		    /*
+		     * De-link the line added for which we cannot undo.
+		     * If we don't do this, undoworker() won't find a match
+		     * against the buffer, and will be deadlocked.
+		     */
+		    lremove(bp, prevp);
+		}
 	    }
 	}
 	endofDisplay();
@@ -2629,17 +2729,17 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     /* mark the window for changes.  could this be moved up to
      * where we actually insert a line? */
     if (nline)
-	chg_buff(curbp, WFHARD);
+	chg_buff(bp, WFHARD);
 
   out:
     /* copy window parameters back to the buffer structure */
-    copy_traits(&(curbp->b_wtraits), &(curwp->w_traits));
+    copy_traits(&(bp->b_wtraits), &(curwp->w_traits));
 
     if (status < FIOERR)
 	imply_alt(fname, FALSE, FALSE);
 
 #if COMPLETE_FILES || COMPLETE_DIRS
-    if (b_is_directory(curbp)) {
+    if (b_is_directory(bp)) {
 	DOT.l = lback(DOT.l);
     } else
 #endif
@@ -2684,7 +2784,7 @@ create_save_dir(char *dirnam)
 
     for (; n < TABLESIZE(tbl); n++) {
 	if (is_directory(tbl[n])) {
-	    int omask = vl_umask(0077);
+	    mode_t omask = vl_umask(0077);
 	    (void) pathcat(dirnam, tbl[n], "vileDXXXXXX");
 
 	    /* on failure, keep going */
@@ -2844,13 +2944,17 @@ imdying(int ACTUAL_SIG_ARGS)
 			      "echo errors:  check in current",
 			      "directory too.;");
 
-	    cp = lsprintf(cp, ") | %s %s", *mailcmdp, np);
+	    (void) lsprintf(cp, ") | %s %s", *mailcmdp, np);
 
-	    (void) system(cmd);
+	    IGNORE_RC(system(cmd));
 	}
     }
     term.cursorvis(TRUE);	/* ( this might work ;-) */
-    if (signo != SIGHUP && signo != SIGINT) {
+    if (
+#ifdef SIGHUP
+	   signo != SIGHUP &&
+#endif
+	   signo != SIGINT) {
 	term.clean(FALSE);
 #ifdef VILE_ERROR_ABORT
 	ExitProgram(signo);
