@@ -5,7 +5,7 @@
  * Written by T.E.Dickey for vile (march 1993).
  *
  *
- * $Header: /usr/build/vile/vile/RCS/filec.c,v 1.126 2007/08/29 00:44:55 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/filec.c,v 1.129 2010/09/05 17:38:50 tom Exp $
  *
  */
 
@@ -180,7 +180,7 @@ makeString(BUFFER *bp, LINE *lp, char *text, size_t len)
 	text = add_backslashes(text);
 #endif
 	(void) strcpy(lvalue(np), text);
-	lvalue(np)[len + extra - 1] = 0;	/* clear scan indicator */
+	lvalue(np)[len + (size_t) extra - 1] = 0;	/* clear scan indicator */
 	llength(np) -= extra;	/* hide the null and scan indicator */
 
 	set_lforw(lback(lp), np);
@@ -693,10 +693,12 @@ fillMyBuff(BUFFER *bp, char *name)
 {
     size_t dots = 0;
     int count = 0;
-    int trim_leaf = FALSE;
     char *s;
     char path[NFILEN + 2];
     char temp[NFILEN + 2];
+#if OPT_VMS_PATH
+    int trim_leaf = FALSE;
+#endif
 
     TRACE(("fillMyBuff '%s'\n", name));
 
@@ -765,7 +767,9 @@ fillMyBuff(BUFFER *bp, char *name)
 	strip_dots(path, &dots);	/* ignore trailing /. or /.. */
 #endif
 	if (!is_environ(path) && !is_directory(path)) {
+#if OPT_VMS_PATH
 	    trim_leaf = TRUE;
+#endif
 	    *pathleaf(path) = EOS;
 	    if (!is_directory(path))
 		return 1;
@@ -799,7 +803,7 @@ fillMyBuff(BUFFER *bp, char *name)
 		need = strlen(temp);
 		want = strlen(path);
 		for_each_line(lp, bp) {
-		    size_t have = llength(lp);
+		    size_t have = (size_t) llength(lp);
 		    if (have == need
 			&& !memcmp(lvalue(lp), temp, need))
 			count = -1;
@@ -838,7 +842,7 @@ makeMyList(BUFFER *bp, char *name)
 	len++;
 
     (void) bsizes(bp);
-    need = bp->b_linecount + 2;
+    need = (size_t) bp->b_linecount + 2;
     if (bp->b_index_size < need) {
 	bp->b_index_size = need * 2;
 	if (bp->b_index_list == 0)
@@ -923,7 +927,7 @@ path_completion(DONE_ARGS)
 #endif
 
     TRACE(("path_completion(%#x '%c' %d:\"%s\")\n",
-	   flags, c, *pos, visible_buff(buf, (int) *pos, TRUE)));
+	   flags, c, (int) *pos, visible_buff(buf, (int) *pos, TRUE)));
     (void) flags;
 
     if (buf == 0)
@@ -1198,7 +1202,7 @@ mlreply_file(const char *prompt, TBUFF **buffer, UINT flag, char *result)
     int do_prompt = (clexec || isnamedcmd || (flag & FILEC_PROMPT));
     int ok_expand = (flag & FILEC_EXPAND);
 
-    flag &= ~(FILEC_PROMPT | FILEC_EXPAND);
+    flag &= (UINT) (~(FILEC_PROMPT | FILEC_EXPAND));
 
 #if COMPLETE_FILES
     if (do_prompt && !clexec) {
