@@ -1,7 +1,7 @@
 /*
  * Common utility functions for vile syntax/highlighter programs
  *
- * $Header: /usr/build/vile/vile/filters/RCS/filters.c,v 1.115 2008/05/25 23:06:02 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/filters.c,v 1.154 2010/07/13 13:27:07 tom Exp $
  *
  */
 
@@ -33,13 +33,12 @@
 #define HASH_LENGTH 256
 
 struct _keyword {
-#if USE_TSEARCH || !defined(HAVE_TDESTROY)
-    KEYWORD *next;
-#endif
-    char *kw_name;
-    char *kw_attr;
-    unsigned kw_size;		/* strlen(kw_name) */
-    unsigned short kw_flag;	/* nonzero for classes */
+    KEYWORD *kw_next;
+    char *kw_name;		/* the keyword name */
+    char *kw_attr;		/* its color attributes, if any */
+    char *kw_flag;		/* optional flags for the syntax parser */
+    size_t kw_size;		/* strlen(kw_name) */
+    unsigned short kw_type;	/* nonzero for classes */
     short kw_used;		/* nonzero for classes */
 };
 
@@ -50,14 +49,22 @@ struct _classes {
     char *name;
 #if USE_TSEARCH
     void *data;
-#if !defined(HAVE_TDESTROY)
     KEYWORD *keywords;
-#endif
 #else
     KEYWORD **data;
 #endif
 };
 
+typedef struct {
+    int zero_or_more;
+    int zero_or_all;
+    int meta_ch;
+    int eqls_ch;
+    char *default_table;
+    char *default_attr;
+} FLTCHARS;
+
+char *default_table;
 char *default_attr;
 int zero_or_more = '*';		/* zero or more of the following */
 int zero_or_all = '?';		/* zero or all of the following */
@@ -79,32 +86,82 @@ static CLASS *current_class;
  * flt_bfr_*() function data
  */
 static char *flt_bfr_text = 0;
-static char *flt_bfr_attr = "";
-static unsigned flt_bfr_used = 0;
-static unsigned flt_bfr_size = 0;
+static const char *flt_bfr_attr = "";
+static size_t flt_bfr_used = 0;
+static size_t flt_bfr_size = 0;
 
 /*
  * OpenKeywords() function data
  */
 static char *str_keyword_name = 0;
 static char *str_keyword_file = 0;
-static unsigned len_keyword_name = 0;
-static unsigned len_keyword_file = 0;
+static size_t len_keyword_name = 0;
+static size_t len_keyword_file = 0;
 
 /******************************************************************************
  * Private functions                                                          *
  ******************************************************************************/
 
-#define KW_FLAG(p) ((p) ? ((p)->kw_flag ? "class" : "keyword") : "?")
+#define KW_FLAG(p) ((p) ? ((p)->kw_type ? "class" : "keyword") : "?")
+
+#define COPYIT(name) save->name = name
+#define SAVEIT(name) if (name) save->name = strmalloc(name); else save->name = 0
+
+static void
+flt_save_chars(FLTCHARS * save)
+{
+    COPYIT(zero_or_more);
+    COPYIT(zero_or_all);
+    COPYIT(meta_ch);
+    COPYIT(eqls_ch);
+
+    SAVEIT(default_table);
+    SAVEIT(default_attr);
+}
+
+#undef COPYIT
+#undef SAVEIT
+
+static void
+flt_init_chars(void)
+{
+    zero_or_more = '*';
+    zero_or_all = '?';
+    meta_ch = '.';
+    eqls_ch = ':';
+}
+
+#define COPYIT(name) name = save->name
+#define SAVEIT(name) if (save->name) { FreeIfNeeded(name); name = save->name; }
+
+static void
+flt_restore_chars(FLTCHARS * save)
+{
+    COPYIT(zero_or_more);
+    COPYIT(zero_or_all);
+    COPYIT(meta_ch);
+    COPYIT(eqls_ch);
+
+    SAVEIT(default_table);
+    SAVEIT(default_attr);
+}
+
+#undef COPYIT
+#undef SAVEIT
 
 /* FIXME */
 static void
-init_data(KEYWORD * data, const char *name, const char *attribute, int classflag)
+init_data(KEYWORD * data,
+	  const char *name,
+	  const char *attribute,
+	  int classflag,
+	  char *flag)
 {
     data->kw_name = strmalloc(name);
     data->kw_size = strlen(data->kw_name);
     data->kw_attr = strmalloc(attribute);
-    data->kw_flag = classflag;
+    data->kw_flag = (flag != 0) ? strmalloc(flag) : 0;
+    data->kw_type = (unsigned short) classflag;
     data->kw_used = 2;
 }
 
@@ -116,6 +173,8 @@ free_data(KEYWORD * data)
 	    free(data->kw_name);
 	if (data->kw_attr)
 	    free(data->kw_attr);
+	if (data->kw_flag)
+	    free(data->kw_flag);
 	free(data);
     }
 }
@@ -136,9 +195,6 @@ compare_data(const void *a, const void *b)
     return strcmp(p->kw_name, q->kw_name);
 }
 
-#ifdef HAVE_TDESTROY
-#define flt_tdestroy tdestroy
-#else
 static void
 flt_tdestroy(void *root, void (*pFreeNode) (void *nodep))
 {
@@ -146,18 +202,22 @@ flt_tdestroy(void *root, void (*pFreeNode) (void *nodep))
     if (root != 0) {
 	for (p = classes; p != 0; p = p->next) {
 	    if (p->data == root) {
+#ifdef HAVE_TDESTROY
+		tdestroy(root, pFreeNode);
+		p->keywords = 0;
+#else
 		while (p->keywords != 0) {
 		    KEYWORD *q = p->keywords;
-		    p->keywords = q->next;
+		    p->keywords = q->kw_next;
 		    if (tdelete(q, &(p->data), compare_data) != 0)
 			pFreeNode(q);
 		}
+#endif
 		break;
 	    }
 	}
     }
 }
-#endif
 #endif /* USE_TSEARCH */
 
 static void
@@ -174,61 +234,89 @@ AttrsOnce(KEYWORD * entry)
 }
 
 static void
-ExecAbbrev(char *param)
+ExecAbbrev(const char *param)
 {
     zero_or_more = *param;
     VERBOSE(1, ("set abbrev '%c'\n", zero_or_more));
 }
 
 static void
-ExecBrief(char *param)
+ExecBrief(const char *param)
 {
     zero_or_all = *param;
     VERBOSE(1, ("set brief '%c'\n", zero_or_all));
 }
 
 static void
-ExecClass(char *param)
+ExecClass(const char *param)
 {
-    parse_keyword(param, 1);
+    char *temp = strmalloc(param);
+    parse_keyword(temp, 1);
+    free(temp);
 }
 
 static void
-ExecDefault(char *param)
+ExecDefault(const char *param)
 {
-    char *s = skip_ident(param);
+    char *temp = strmalloc(param);
+    char *s = skip_ident(temp);
     int save = *s;
+    int isClass;
 
     *s = 0;
-    if (!*param)
-	param = NAME_KEYWORD;
-    if (is_class(param)) {
-	free(default_attr);
-	default_attr = strmalloc(param);
+    if (!*temp) {
+	free(temp);
+	temp = strmalloc(NAME_KEYWORD);
+	isClass = (is_class(temp) != 0);
+    } else {
+	isClass = (is_class(temp) != 0);
+	*s = (char) save;
+    }
+
+    if (isClass) {
+	flt_init_attr(temp);
 	VERBOSE(1, ("set default_attr '%s' %p\n", default_attr, default_attr));
     } else {
-	*s = save;
-	VERBOSE(1, ("not a class:%s", param));
+	VERBOSE(1, ("not a class:%s", temp));
     }
+
+    free(temp);
 }
 
 static void
-ExecEquals(char *param)
+ExecEquals(const char *param)
 {
     eqls_ch = *param;
 }
 
 static void
-ExecMeta(char *param)
+ExecMeta(const char *param)
 {
     meta_ch = *param;
 }
 
 /*
- * Include a symbol table from another key-file.
+ * Include a symbol table from another key-file, merging it into the current
+ * table.
  */
 static void
-ExecSource(char *param)
+ExecMerge(const char *param)
+{
+    int save_meta = meta_ch;
+    int save_eqls = eqls_ch;
+
+    flt_read_keywords(param);
+
+    meta_ch = save_meta;
+    eqls_ch = save_eqls;
+}
+
+/*
+ * Include a symbol table from another key-file, creating it in a table named
+ * according to the parameter.
+ */
+static void
+ExecSource(const char *param)
 {
     int save_meta = meta_ch;
     int save_eqls = eqls_ch;
@@ -247,20 +335,23 @@ ExecSource(char *param)
  * performance.
  */
 static void
-ExecTable(char *param)
+ExecTable(const char *param)
 {
-    flt_make_symtab(param);
-    flt_read_keywords(MY_NAME);	/* provide default values for this table */
+    if (*param) {
+	flt_make_symtab(param);
+	flt_read_keywords(MY_NAME);	/* provide default values for this table */
+    } else {
+	set_symbol_table(default_table);
+    }
 }
 
 static KEYWORD *
 FindIdentifier(const char *name)
 {
-    unsigned size;
     KEYWORD *result = 0;
 
-    if (name != 0 && (size = strlen(name)) != 0) {
 #if USE_TSEARCH
+    if (name != 0 && strlen(name) != 0) {
 	KEYWORD find;
 	void *pp;
 
@@ -268,7 +359,10 @@ FindIdentifier(const char *name)
 	if ((pp = tfind(&find, &(current_class->data), compare_data)) != 0) {
 	    result = *(KEYWORD **) pp;
 	}
+    }
 #else
+    size_t size;
+    if (name != 0 && (size = strlen(name)) != 0) {
 	int Index = hash_function(name);
 
 	result = my_table[Index];
@@ -277,10 +371,10 @@ FindIdentifier(const char *name)
 		&& strcmp(result->kw_name, name) == 0) {
 		break;
 	    }
-	    result = result->next;
+	    result = result->kw_next;
 	}
-#endif /* TSEARCH */
     }
+#endif /* TSEARCH */
     return result;
 }
 
@@ -302,7 +396,7 @@ Free(char *ptr)
  * hides them.
  */
 static FILE *
-OpenKeywords(char *classname)
+OpenKeywords(const char *table_name)
 {
 #define OPEN_IT(p) if ((fp = fopen(p, "r")) != 0) { \
 			VERBOSE(1,("Opened %s", p)); return fp; } else { \
@@ -312,17 +406,17 @@ OpenKeywords(char *classname)
     static char suffix[] = KEYFILE_SUFFIX;
 
     FILE *fp;
-    char *path;
-    unsigned need;
+    const char *path;
+    size_t need;
     char myLeaf[20];
 
-    need = sizeof(suffix) + strlen(classname) + 2;
+    need = sizeof(suffix) + strlen(table_name) + 2;
     str_keyword_file = do_alloc(str_keyword_file, need, &len_keyword_file);
     if (str_keyword_file == 0) {
 	CannotAllocate("OpenKeywords");
 	return 0;
     }
-    sprintf(str_keyword_file, "%s%s", classname, suffix);
+    sprintf(str_keyword_file, "%s%s", table_name, suffix);
 
     if (strchr(str_keyword_file, PATHSEP) != 0) {
 	OPEN_IT(str_keyword_file);
@@ -380,7 +474,7 @@ ParseDirective(char *line)
 {
     static struct {
 	const char *name;
-	void (*func) (char *param);
+	void (*func) (const char *param);
     } table[] = {
 	/* *INDENT-OFF* */
 	{ "abbrev",  ExecAbbrev   },
@@ -389,7 +483,7 @@ ParseDirective(char *line)
 	{ "default", ExecDefault  },
 	{ "equals",  ExecEquals   },
 	{ "include", flt_read_keywords },
-	{ "merge",   ExecSource   },
+	{ "merge",   ExecMerge    },
 	{ "meta",    ExecMeta     },
 	{ "source",  ExecSource   },
 	{ "table",   ExecTable    },
@@ -400,7 +494,7 @@ ParseDirective(char *line)
 
     if (*(line = skip_blanks(line)) == meta_ch) {
 	line = skip_blanks(line + 1);
-	if ((len = (skip_ident(line) - line)) != 0) {
+	if ((len = (unsigned) (skip_ident(line) - line)) != 0) {
 	    for (n = 0; n < sizeof(table) / sizeof(table[0]); n++) {
 		if (!strncmp(line, table[n].name, len)) {
 		    (*table[n].func) (skip_blanks(line + len));
@@ -413,10 +507,10 @@ ParseDirective(char *line)
     return 0;
 }
 
-static unsigned
+static size_t
 TrimBlanks(char *src)
 {
-    unsigned len = strlen(src);
+    size_t len = strlen(src);
 
     while (len != 0
 	   && isspace(CharOf(src[len - 1])))
@@ -428,14 +522,20 @@ TrimBlanks(char *src)
  * Public functions                                                           *
  ******************************************************************************/
 
-char *
-ci_keyword_attr(char *text)
+const char *
+ci_keyword_attr(const char *text)
 {
     return keyword_attr(lowercase_of(text));
 }
 
+const char *
+ci_keyword_flag(const char *text)
+{
+    return keyword_flag(lowercase_of(text));
+}
+
 char *
-class_attr(char *name)
+class_attr(const char *name)
 {
     KEYWORD *data;
     char *result = 0;
@@ -450,7 +550,7 @@ class_attr(char *name)
 }
 
 void *
-flt_alloc(void *ptr, unsigned need, unsigned *have, unsigned size)
+flt_alloc(void *ptr, size_t need, size_t *have, size_t size)
 {
     need += (2 * size);		/* allow for trailing null, etc */
     if ((need > *have) || (ptr == 0)) {
@@ -469,28 +569,28 @@ flt_alloc(void *ptr, unsigned need, unsigned *have, unsigned size)
  * that will be written as one attributed region.
  */
 void
-flt_bfr_append(char *text, int length)
+flt_bfr_append(const char *text, int length)
 {
-    flt_bfr_text = do_alloc(flt_bfr_text, flt_bfr_used + length, &flt_bfr_size);
+    flt_bfr_text = do_alloc(flt_bfr_text, flt_bfr_used + (size_t) length, &flt_bfr_size);
     if (flt_bfr_text != 0) {
-	strncpy(flt_bfr_text + flt_bfr_used, text, length);
-	flt_bfr_used += length;
+	strncpy(flt_bfr_text + flt_bfr_used, text, (unsigned) length);
+	flt_bfr_used += (unsigned) length;
     } else {
 	CannotAllocate("flt_bfr_append");
     }
 }
 
 void
-flt_bfr_begin(char *attr)
+flt_bfr_begin(const char *attr)
 {
     flt_bfr_finish();
     flt_bfr_attr = attr;
 }
 
 void
-flt_bfr_embed(char *text, int length, char *attr)
+flt_bfr_embed(const char *text, int length, const char *attr)
 {
-    char *save = flt_bfr_attr;
+    const char *save = flt_bfr_attr;
 
     if ((save == 0 && attr == 0) ||
 	(save != 0 && attr != 0 && !strcmp(save, attr))) {
@@ -516,7 +616,7 @@ void
 flt_bfr_finish(void)
 {
     if (flt_bfr_used) {
-	flt_puts(flt_bfr_text, flt_bfr_used,
+	flt_puts(flt_bfr_text, (int) flt_bfr_used,
 		 (flt_bfr_attr
 		  ? flt_bfr_attr
 		  : ""));
@@ -528,9 +628,160 @@ flt_bfr_finish(void)
 int
 flt_bfr_length(void)
 {
-    return flt_bfr_used;
+    return (int) flt_bfr_used;
 }
 
+static CLASS *
+find_symtab(const char *table_name)
+{
+    CLASS *p;
+    for (p = classes; p != 0; p = p->next) {
+	if (!strcmp(p->name, table_name)) {
+	    break;
+	}
+    }
+    return p;
+}
+
+/*----------------------------------------------------------------------------*/
+
+static int dump_meta_ch;
+static int dump_eqls_ch;
+
+static void
+dump_init(void)
+{
+    dump_meta_ch = meta_ch;
+    dump_eqls_ch = eqls_ch;
+
+    if (dump_meta_ch != '.')
+	flt_message(".meta %c\n", dump_meta_ch);
+    if (dump_eqls_ch != ':')
+	flt_message("%cequals %c\n", dump_meta_ch, dump_eqls_ch);
+}
+
+static int
+dump_update2(const char *name, int keep)
+{
+    int result;
+
+    for (result = 32; result < 127; ++result) {
+	if (result != keep && ispunct(result) && strchr(name, result) == 0) {
+	    break;
+	}
+    }
+
+    return result;
+}
+
+/*
+ * emit "meta" and "equals" directives if the name contains the current meta or
+ * equals characters.
+ */
+static void
+dump_update(const char *name)
+{
+    int newc;
+
+    if (strchr(name, dump_meta_ch) != 0) {
+	newc = dump_update2(name, dump_eqls_ch);
+	flt_message("%cmeta %c\n", dump_meta_ch, newc);
+	dump_meta_ch = newc;
+    }
+    if (strchr(name, dump_eqls_ch) != 0) {
+	dump_eqls_ch = dump_update2(name, dump_meta_ch);
+	flt_message("%cequals %c\n", dump_meta_ch, dump_eqls_ch);
+    }
+}
+
+static void
+dump_class(const KEYWORD * const q)
+{
+    if (q->kw_type) {
+	dump_update(q->kw_name);
+	flt_message("%cclass %s%c%s",
+		    dump_meta_ch,
+		    q->kw_name,
+		    dump_eqls_ch,
+		    NonNull(q->kw_attr));
+	if (q->kw_flag)
+	    flt_message("%c%s",
+			dump_eqls_ch,
+			q->kw_flag);
+	flt_message("\n");
+    }
+}
+
+static void
+dump_keyword(const KEYWORD * const q)
+{
+    if (!q->kw_type) {
+	dump_update(q->kw_name);
+	flt_message("%s%c%s",
+		    q->kw_name,
+		    dump_eqls_ch,
+		    NonNull(q->kw_attr));
+	if (q->kw_flag)
+	    flt_message("%c%s",
+			dump_eqls_ch,
+			q->kw_flag);
+	flt_message("\n");
+    }
+}
+
+typedef void DumpFunc(const KEYWORD * const);
+
+#if USE_TSEARCH
+static DumpFunc *dump_func;
+
+static void
+dump_walker(const void *nodep, const VISIT which, const int depth)
+{
+    const KEYWORD *const p = *(KEYWORD * const *) nodep;
+    (void) depth;
+    if (which == postorder || which == leaf)
+	dump_func(p);
+}
+#endif
+
+static void
+dump_symtab(CLASS * p, DumpFunc dumpFunc)
+{
+#if USE_TSEARCH
+    dump_func = dumpFunc;
+    twalk(p->data, dump_walker);
+#else
+    int i;
+    KEYWORD *q;
+
+    for (i = 0; i < HASH_LENGTH; i++) {
+	for (q = p->data[i]; q != 0; q = q->kw_next) {
+	    dumpFunc(q);
+	}
+    }
+#endif
+}
+
+void
+flt_dump_symtab(const char *table_name)
+{
+    CLASS *p;
+
+    if (table_name == 0) {
+	dump_init();
+	for (p = classes; p != 0; p = p->next) {
+	    flt_dump_symtab(p->name);
+	}
+    } else if ((p = find_symtab(table_name)) != 0) {
+	dump_update(p->name);
+	flt_message("%ctable %s\n", dump_meta_ch, p->name);
+	dump_symtab(p, dump_class);
+	dump_symtab(p, dump_keyword);
+    }
+}
+
+/*----------------------------------------------------------------------------*/
+
 void
 flt_free(char **p, unsigned *len)
 {
@@ -539,7 +790,7 @@ flt_free(char **p, unsigned *len)
 }
 
 void
-flt_free_keywords(char *classname)
+flt_free_keywords(const char *table_name)
 {
     CLASS *p, *q;
 #if !USE_TSEARCH
@@ -547,16 +798,16 @@ flt_free_keywords(char *classname)
     int i;
 #endif
 
-    VERBOSE(1, ("flt_free_keywords(%s)", classname));
+    VERBOSE(1, ("flt_free_keywords(%s)", table_name));
     for (p = classes, q = 0; p != 0; q = p, p = p->next) {
-	if (!strcmp(classname, p->name)) {
+	if (!strcmp(table_name, p->name)) {
 #if USE_TSEARCH
 	    flt_tdestroy(p->data, free_node);
 #else
 	    my_table = p->data;
 	    for (i = 0; i < HASH_LENGTH; i++) {
 		while ((ptr = my_table[i]) != 0) {
-		    my_table[i] = ptr->next;
+		    my_table[i] = ptr->kw_next;
 		    free_data(ptr);
 		}
 	    }
@@ -583,31 +834,50 @@ flt_free_symtab(void)
 	flt_free_keywords(classes->name);
 }
 
+void
+flt_init_table(const char *table_name)
+{
+    if (default_table != 0)
+	free(default_table);
+    default_table = strmalloc(table_name);
+
+    VERBOSE(3, ("flt_init_table:%s", table_name));
+}
+
+void
+flt_init_attr(const char *attr_name)
+{
+    if (default_attr != 0)
+	free(default_attr);
+    default_attr = strmalloc(attr_name);
+
+    VERBOSE(3, ("flt_init_attr:%s", attr_name));
+}
+
 /*
  * We drop the old symbol table each time we read the data, to allow keywords
  * to be removed from the table.  Also, it is necessary for some filters such
  * as m4 to be able to restart to a known state.
  */
 void
-flt_initialize(void)
+flt_initialize(const char *table_name)
 {
-    if (default_attr != 0)
-	free(default_attr);
-    default_attr = strmalloc(NAME_KEYWORD);
-
-    zero_or_more = '*';
-    zero_or_all = '?';
-    meta_ch = '.';
-    eqls_ch = ':';
-    FltOptions('v') = 0;
+    flt_init_table(table_name);
+    flt_init_attr(NAME_KEYWORD);
+    flt_init_chars();
 
     flt_free_symtab();
+    flt_make_symtab(table_name);
 }
 
 void
-flt_make_symtab(char *classname)
+flt_make_symtab(const char *table_name)
 {
-    if (!set_symbol_table(classname)) {
+    if (*table_name == '\0') {
+	table_name = default_table;
+    }
+
+    if (!set_symbol_table(table_name)) {
 	CLASS *p;
 
 	if ((p = typecallocn(CLASS, 1)) == 0) {
@@ -615,7 +885,7 @@ flt_make_symtab(char *classname)
 	    return;
 	}
 
-	p->name = strmalloc(classname);
+	p->name = strmalloc(table_name);
 	if (p->name == 0) {
 	    free(p);
 	    CannotAllocate("flt_make_symtab");
@@ -636,7 +906,7 @@ flt_make_symtab(char *classname)
 	my_table = p->data;
 	current_class = p;
 
-	VERBOSE(1, ("flt_make_symtab(%s)", classname));
+	VERBOSE(1, ("flt_make_symtab(%s)", table_name));
 
 	/*
 	 * Mark all of the standard predefined classes when we first create a
@@ -659,15 +929,32 @@ flt_make_symtab(char *classname)
 }
 
 void
-flt_read_keywords(char *classname)
+flt_setup_symbols(const char *table_name)
+{
+    if (!set_symbol_table(table_name)) {
+	flt_make_symtab(table_name);
+	flt_read_keywords(MY_NAME);
+	flt_read_keywords(table_name);
+	/* set back to the name we asked for, in case reader switched */
+	set_symbol_table(table_name);
+    }
+}
+
+void
+flt_read_keywords(const char *table_name)
 {
     FILE *kwfile;
     char *line = 0;
     char *name;
-    unsigned line_len = 0;
+    size_t line_len = 0;
+    FLTCHARS fltchars;
 
-    VERBOSE(1, ("flt_read_keywords(%s)", classname));
-    if ((kwfile = OpenKeywords(classname)) != NULL) {
+    VERBOSE(1, ("flt_read_keywords(%s)", table_name));
+
+    flt_save_chars(&fltchars);
+    flt_init_chars();
+
+    if ((kwfile = OpenKeywords(table_name)) != NULL) {
 	int linenum = 0;
 	while (readline(kwfile, &line, &line_len) != 0) {
 
@@ -683,9 +970,11 @@ flt_read_keywords(char *classname)
 	fclose(kwfile);
     }
     Free(line);
+
+    flt_restore_chars(&fltchars);
 }
 
-char *
+const char *
 get_symbol_table(void)
 {
 #if USE_TSEARCH
@@ -715,10 +1004,9 @@ hash_function(const char *id)
 }
 
 static KEYWORD *
-alloc_keyword(const char *ident, const char *attribute, int classflag)
+alloc_keyword(const char *ident, const char *attribute, int classflag, char *flag)
 {
     KEYWORD *nxt;
-    int Index;
 
     if ((nxt = FindIdentifier(ident)) != 0) {
 	Free(nxt->kw_attr);
@@ -728,9 +1016,8 @@ alloc_keyword(const char *ident, const char *attribute, int classflag)
 	}
     } else {
 	nxt = NULL;
-	Index = hash_function(ident);
 	if ((nxt = typecallocn(KEYWORD, 1)) != NULL) {
-	    init_data(nxt, ident, attribute, classflag);
+	    init_data(nxt, ident, attribute, classflag, flag);
 
 	    if (nxt->kw_name != 0
 		&& nxt->kw_attr != 0) {
@@ -742,12 +1029,13 @@ alloc_keyword(const char *ident, const char *attribute, int classflag)
 		} else {
 		    nxt = 0;
 		}
-#ifndef HAVE_TDESTROY
-		nxt->next = current_class->keywords;
-		current_class->keywords = nxt;
-#endif
+		if (nxt != 0) {
+		    nxt->kw_next = current_class->keywords;
+		    current_class->keywords = nxt;
+		}
 #else
-		nxt->next = my_table[Index];
+		int Index = hash_function(ident);
+		nxt->kw_next = my_table[Index];
 		my_table[Index] = nxt;
 #endif
 	    } else {
@@ -759,17 +1047,18 @@ alloc_keyword(const char *ident, const char *attribute, int classflag)
     return nxt;
 }
 
-void
-insert_keyword(const char *ident, const char *attribute, int classflag)
+static void
+insert_keyword2(const char *ident, const char *attribute, int classflag, char *flag)
 {
     KEYWORD *nxt;
     char *mark;
     char *temp;
 
-    VERBOSE(2, ("insert_keyword(%s, %s, %d)\n",
+    VERBOSE(2, ("insert_keyword(%s, %s, %d, %s)\n",
 		ident,
-		attribute,
-		classflag));
+		NONNULL(attribute),
+		classflag,
+		NONNULL(flag)));
 
     if (zero_or_more
 	&& (mark = strchr(ident, zero_or_more)) != 0
@@ -779,9 +1068,9 @@ insert_keyword(const char *ident, const char *attribute, int classflag)
 	    mark = temp + (mark - ident);
 	    while (*mark == zero_or_more) {
 		*mark = 0;
-		insert_keyword(temp, attribute, classflag);
+		insert_keyword2(temp, attribute, classflag, flag);
 		if ((mark[0] = mark[1]) != 0) {
-		    *(++mark) = zero_or_more;
+		    *(++mark) = (char) zero_or_more;
 		}
 	    }
 	    free(temp);
@@ -796,16 +1085,16 @@ insert_keyword(const char *ident, const char *attribute, int classflag)
 	    mark = temp + (mark - ident);
 	    if (*mark == zero_or_all) {
 		*mark = 0;
-		insert_keyword(temp, attribute, classflag);
+		insert_keyword2(temp, attribute, classflag, flag);
 		while ((mark[0] = mark[1]) != 0)
 		    ++mark;
-		insert_keyword(temp, attribute, classflag);
+		insert_keyword2(temp, attribute, classflag, flag);
 	    }
 	    free(temp);
 	} else {
 	    CannotAllocate("insert_keyword");
 	}
-    } else if ((nxt = alloc_keyword(ident, attribute, classflag)) == 0) {
+    } else if ((nxt = alloc_keyword(ident, attribute, classflag, flag)) == 0) {
 	CannotAllocate("insert_keyword");
     } else {
 	VERBOSE(3, ("...\tname \"%s\"\tattr \"%s\"",
@@ -814,12 +1103,18 @@ insert_keyword(const char *ident, const char *attribute, int classflag)
     }
 }
 
+void
+insert_keyword(const char *ident, const char *attribute, int classflag)
+{
+    insert_keyword2(ident, attribute, classflag, 0);
+}
+
 KEYWORD *
-is_class(char *name)
+is_class(const char *name)
 {
     KEYWORD *result = FindIdentifier(name);
     if (result != 0) {
-	if (result->kw_flag == 0) {
+	if (result->kw_type == 0) {
 	    result = 0;
 	}
     }
@@ -827,38 +1122,66 @@ is_class(char *name)
 }
 
 KEYWORD *
-is_keyword(char *name)
+is_keyword(const char *name)
 {
     KEYWORD *result;
     if ((result = FindIdentifier(name)) != 0
-	&& result->kw_flag == 0) {
+	&& result->kw_type == 0) {
 	return result;
     }
     return 0;
 }
 
-char *
-keyword_attr(char *name)
+const char *
+keyword_attr(const char *name)
 {
-    KEYWORD *data = is_keyword(name);
-    char *result = 0;
+    KEYWORD *data = keyword_data(name);
+    const char *result = 0;
 
     if (data != 0) {
 	result = data->kw_attr;
-	while ((data = is_class(result)) != 0)
-	    result = data->kw_attr;
     }
     VERBOSE(1, ("keyword_attr(%s) = %p %s\n", name, result, NONNULL(result)));
     return result;
 }
 
-char *
-lowercase_of(char *text)
+KEYWORD *
+keyword_data(const char *name)
+{
+    KEYWORD *data = is_keyword(name);
+    KEYWORD *result = 0;
+
+    if (data != 0) {
+	result = data;
+	name = data->kw_attr;
+	while ((data = is_class(name)) != 0) {
+	    result = data;
+	    name = data->kw_attr;
+	}
+    }
+    return result;
+}
+
+const char *
+keyword_flag(const char *name)
+{
+    KEYWORD *data = is_keyword(name);
+    const char *result = 0;
+
+    if (data != 0) {
+	result = data->kw_flag;
+    }
+    VERBOSE(1, ("keyword_flag(%s) = %p %s\n", name, result, NONNULL(result)));
+    return result;
+}
+
+const char *
+lowercase_of(const char *text)
 {
     static char *name;
-    static unsigned used;
+    static size_t used;
     unsigned n;
-    char *result;
+    const char *result;
 
 #if NO_LEAKS
     if (text == 0) {
@@ -869,7 +1192,7 @@ lowercase_of(char *text)
     if ((name = do_alloc(name, strlen(text), &used)) != 0) {
 	for (n = 0; text[n] != 0; n++) {
 	    if (isalpha(CharOf(text[n])) && isupper(CharOf(text[n])))
-		name[n] = tolower(CharOf(text[n]));
+		name[n] = (char) tolower(CharOf(text[n]));
 	    else
 		name[n] = text[n];
 	}
@@ -886,9 +1209,11 @@ void
 parse_keyword(char *name, int classflag)
 {
     char *args = 0;
+    char *flag = 0;
     char *s, *t;
     int quoted = 0;
 
+    VERBOSE(1, ("parse_keyword(%s, %d)\n", name, classflag));
     if ((s = strchr(name, eqls_ch)) != 0) {
 	*s++ = 0;
 	s = skip_blanks(s);
@@ -899,13 +1224,30 @@ parse_keyword(char *name, int classflag)
 		    if (*s == QUOTE) {
 			if (*++s != QUOTE)
 			    quoted = 0;
+			if (*s == 0)
+			    continue;
 		    }
 		} else {
 		    if (*s == QUOTE) {
 			quoted = 1;
 			s++;
+			continue;
 		    } else if (!isalnum(CharOf(*s))) {
-			args = 0;	/* error: ignore */
+			if (!*name) {
+			    args = 0;	/* comment: ignore */
+			} else if (*s == eqls_ch) {
+			    flag = s;
+			    *flag++ = '\0';
+			    VERBOSE(2, ("found flags \"%s\" for %s\n", flag, name));
+			    /* an empty color field inherits the default */
+			    TrimBlanks(args);
+			    if (*args == 0)
+				args = default_attr;
+			} else {
+			    VERBOSE(1, ("unexpected:%s%c%s\n", name,
+					eqls_ch, s));
+			    args = 0;	/* error: ignore */
+			}
 			break;
 		    }
 		}
@@ -916,20 +1258,22 @@ parse_keyword(char *name, int classflag)
 	TrimBlanks(name);
     }
 
-    VERBOSE(2, ("parsed\tname \"%s\"\tattr \"%s\"%s",
+    VERBOSE(2, ("parsed\tname \"%s\"\tattr \"%s\" flag \"%s\"%s",
 		name,
 		NONNULL(args),
+		NONNULL(flag),
 		(classflag || is_class(name)) ? " - class" : ""));
 
     if (*name && args) {
-	insert_keyword(name, args, classflag);
+	insert_keyword2(name, args, classflag, flag);
     } else if (*name) {
-	KEYWORD *data;
 	if (args == 0) {
 	    args = default_attr;
 	    VERBOSE(2, ("using attr \"%s\"", args));
 	}
-	if ((data = FindIdentifier(args)) != 0) {
+	if (args != 0
+	    && strcmp(name, args)
+	    && FindIdentifier(args) != 0) {
 	    /*
 	     * Insert the classname rather than the data->kw_attr value,
 	     * since insert_keyword makes a copy of the string we pass to it.
@@ -938,16 +1282,16 @@ parse_keyword(char *name, int classflag)
 	     * runtime to compare pointers from keyword_attr() to determine if
 	     * two tokens have the same class.  sql-filt.l uses this feature.
 	     */
-	    insert_keyword(name, args, classflag);
+	    insert_keyword2(name, args, classflag, flag);
 	}
     }
 }
 
 char *
-readline(FILE *fp, char **ptr, unsigned *len)
+readline(FILE *fp, char **ptr, size_t *len)
 {
     char *buf = *ptr;
-    unsigned used = 0;
+    size_t used = 0;
 
     if (buf == 0) {
 	*len = BUFSIZ;
@@ -964,7 +1308,7 @@ readline(FILE *fp, char **ptr, unsigned *len)
 	    if (buf == 0)
 		return 0;
 	}
-	buf[used++] = ch;
+	buf[used++] = (char) ch;
 	if (ch == '\n')
 	    break;
     }
@@ -974,14 +1318,14 @@ readline(FILE *fp, char **ptr, unsigned *len)
 }
 
 int
-set_symbol_table(const char *classname)
+set_symbol_table(const char *table_name)
 {
     CLASS *p;
     for (p = classes; p != 0; p = p->next) {
-	if (!strcmp(classname, p->name)) {
+	if (!strcmp(table_name, p->name)) {
 	    my_table = p->data;
 	    current_class = p;
-	    VERBOSE(3, ("set_symbol_table:%s", classname));
+	    VERBOSE(3, ("set_symbol_table:%s", table_name));
 	    return 1;
 	}
     }
@@ -1017,5 +1361,7 @@ filters_leaks(void)
 
     flt_free(&flt_bfr_text, &flt_bfr_size);
     (void) lowercase_of(0);
+    FreeAndNull(default_table);
+    FreeAndNull(default_attr);
 }
 #endif
