@@ -1,5 +1,5 @@
 /*
- * $Header: /usr/build/vile/vile/filters/RCS/pl-filt.c,v 1.97 2008/06/01 14:23:52 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/pl-filt.c,v 1.102 2010/07/13 13:35:01 tom Exp $
  *
  * Filter to add vile "attribution" sequences to perl scripts.  This is a
  * translation into C of an earlier version written for LEX/FLEX.
@@ -48,7 +48,7 @@ typedef struct {
     int has_no_pattern;
 } AfterKey;
 
-static char *put_embedded(char *, int, char *);
+static char *put_embedded(char *, int, const char *);
 
 static char *Action_attr;
 static char *Comment_attr;
@@ -114,7 +114,7 @@ count_blanks(char *s)
 	}
 	++s;
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /******************************************************************************
@@ -186,15 +186,15 @@ is_STRINGS(char *s, int *err, int delim)
 	    }
 	    s++;
 	}
-	found = s - base;
+	found = (int) (s - base);
     }
     return found;
 }
 
 static int
-is_KEYWORD(char *s)
+is_KEYWORD(const char *s)
 {
-    char *base = s;
+    const char *base = s;
     int ch;
     int quote = 0;
 
@@ -223,7 +223,7 @@ is_KEYWORD(char *s)
     }
     if (quote == (s - base))
 	s = base;
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -256,7 +256,7 @@ is_QIDENT(char *s)
 		break;
 	}
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -264,9 +264,9 @@ is_QIDENT(char *s)
  * If not double-quoted, look also for syntax such as $foo'bar
  */
 static int
-is_NORMALVARS(char *s, int dquoted)
+is_NORMALVARS(const char *s, int dquoted)
 {
-    char *base = s;
+    const char *base = s;
     int ch;
     int squoted = 0;
     int part1 = 0;
@@ -298,13 +298,13 @@ is_NORMALVARS(char *s, int dquoted)
 	}
 	s++;
     }
-    return (part1 && (dquoted || (squoted == part2))) ? (s - base) : 0;
+    return (part1 && (dquoted || (squoted == part2))) ? (int) (s - base) : 0;
 }
 
 static int
-is_OTHERVARS(char *s)
+is_OTHERVARS(const char *s)
 {
-    char *base = s;
+    const char *base = s;
     int ch;
     int part1 = 0;
     int part2 = 0;
@@ -337,7 +337,7 @@ is_OTHERVARS(char *s)
 	}
 	s++;
     }
-    return (part1 || part2) ? (s - base) : 0;
+    return (part1 || part2) ? (int) (s - base) : 0;
 }
 
 static int
@@ -428,7 +428,7 @@ is_NUMBER(char *s, int *err)
 	s++;
     }
 
-    return value ? (s - base) : 0;
+    return value ? (int) (s - base) : 0;
 }
 
 static char *
@@ -460,7 +460,7 @@ is_ELLIPSIS(char *s)
 }
 
 static int
-is_IDENT(char *s, int quoted)
+is_IDENT(const char *s, int quoted)
 {
     int found;
 
@@ -498,7 +498,7 @@ is_COMMENT(char *s)
 	}
 	s = t;
     }
-    return (s - base);
+    return (int) (s - base);
 }
 
 /* preprocessor stuff:
@@ -530,18 +530,18 @@ is_PREPROC(char *s)
 	    return 0;
     }
     s += is_BLANK(s);
-    return (s - base);
+    return (int) (s - base);
 }
 
 /******************************************************************************
  ******************************************************************************/
 
 static int
-end_marker(char *s, char *marker, int only)
+end_marker(char *s, const char *marker, int only)
 {
-    int len = strlen(marker);
+    size_t len = strlen(marker);
 
-    return (ATLEAST(s, len)
+    return (ATLEAST(s, (int) len)
 	    && !strncmp(s, marker, len)
 	    && (!only || (s[len] == '\n')));
 }
@@ -569,7 +569,7 @@ begin_POD(char *s)
 	&& ATLEAST(s, 2)
 	&& s[0] == '='
 	&& isalpha(CharOf(s[1]))) {
-	result = s + 1 - base;
+	result = (int) (s + 1 - base);
 	while (base != s) {
 	    flt_putc(*base++);	/* skip the newlines */
 	}
@@ -601,8 +601,8 @@ begin_HERE(char *s, int *quoted)
 	s += count_blanks(s);
 	base = s;
 	if ((ok = is_QIDENT(s)) != 0) {
-	    unsigned temp = 0;
-	    char *marker = do_alloc((char *) 0, ok + 1, &temp);
+	    size_t temp = 0;
+	    char *marker = do_alloc((char *) 0, (size_t) (ok + 1), &temp);
 	    char *d = marker;
 
 	    s += ok;
@@ -665,7 +665,7 @@ line_size(char *s)
 	    break;
 	s++;
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -697,7 +697,7 @@ is_QUOTE(char *s, int *delims)
     while (MORE(s) && isIdent(*s)) {
 	++s;
     }
-    if ((len = (s - base)) != 0) {
+    if ((len = (size_t) (s - base)) != 0) {
 	switch (len) {
 	case 1:
 	    if (*base == 'm' || *base == 'q') {
@@ -732,7 +732,7 @@ is_QUOTE(char *s, int *delims)
 		 (s != base) ? (int) (s - base) : 1,
 		 (s != base) ? base : ""));
     }
-    return (s - base);
+    return (int) (s - base);
 }
 
 static int
@@ -849,7 +849,7 @@ add_to_PATTERN(char *s)
 	    s++;
 	}
 	DPRINTF(("*finally\n->%.*s\n", line_size(s), s));
-	return (s - base);
+	return (int) (s - base);
     }
     return 0;
 }
@@ -880,7 +880,7 @@ write_PATTERN(char *s, int len)
 	len -= skip;
     }
 
-    skip = skip_BLANKS(s) - s;
+    skip = (int) (skip_BLANKS(s) - s);
     if (skip) {
 	s += skip;
 	len -= skip;
@@ -955,7 +955,7 @@ write_PATTERN(char *s, int len)
 	}
     }
     if (comment)
-	first += (skip_BLANKS(s) - s);
+	first += (int) (skip_BLANKS(s) - s);
     if (comment) {
 	flt_puts(s + first, (len - first), Comment_attr);
     } else {
@@ -1003,7 +1003,7 @@ is_Subscript(char *s, int len, char *delimp)
 {
     if (delimp != 0) {
 	int delim = *delimp;
-	int delim2 = delim;
+	int delim2;
 	char *next;
 
 	if (delim != 0 && (next = strchr(LOOKUP_TERM, delim)) != 0) {
@@ -1025,7 +1025,7 @@ put_newline(char *s)
 }
 
 static int
-var_embedded(char *s)
+var_embedded(const char *s)
 {
     if (*s == '$') {
 	if (s[1] == L_PAREN
@@ -1038,7 +1038,7 @@ var_embedded(char *s)
 }
 
 static char *
-put_embedded(char *s, int len, char *attr)
+put_embedded(char *s, int len, const char *attr)
 {
     int id;
     int j, k;
@@ -1112,37 +1112,39 @@ put_document(char *s)
 static void
 check_keyword(char *s, int ok, AfterKey * state)
 {
+    size_t len = (size_t) ok;
+
     state->may_have_pattern = 0;
     state->has_no_pattern = 0;
 
     switch (ok) {
     case 2:
-	state->may_have_pattern = (!strncmp(s, "if", ok)
-				   || !strncmp(s, "eq", ok)
-				   || !strncmp(s, "ge", ok)
-				   || !strncmp(s, "gt", ok)
-				   || !strncmp(s, "le", ok)
-				   || !strncmp(s, "lt", ok)
-				   || !strncmp(s, "ne", ok)
-				   || !strncmp(s, "or", ok));
+	state->may_have_pattern = (!strncmp(s, "if", len)
+				   || !strncmp(s, "eq", len)
+				   || !strncmp(s, "ge", len)
+				   || !strncmp(s, "gt", len)
+				   || !strncmp(s, "le", len)
+				   || !strncmp(s, "lt", len)
+				   || !strncmp(s, "ne", len)
+				   || !strncmp(s, "or", len));
 	break;
     case 3:
-	state->has_no_pattern = !strncmp(s, "sub", ok);
-	state->may_have_pattern = (!strncmp(s, "and", ok)
-				   || !strncmp(s, "cmp", ok)
-				   || !strncmp(s, "not", ok)
-				   || !strncmp(s, "xor", ok));
+	state->has_no_pattern = !strncmp(s, "sub", len);
+	state->may_have_pattern = (!strncmp(s, "and", len)
+				   || !strncmp(s, "cmp", len)
+				   || !strncmp(s, "not", len)
+				   || !strncmp(s, "xor", len));
 	break;
     case 4:
-	state->may_have_pattern = !strncmp(s, "grep", ok);
+	state->may_have_pattern = !strncmp(s, "grep", len);
 	break;
     case 5:
-	state->may_have_pattern = (!strncmp(s, "split", ok)
-				   || !strncmp(s, "until", ok)
-				   || !strncmp(s, "while", ok));
+	state->may_have_pattern = (!strncmp(s, "split", len)
+				   || !strncmp(s, "until", len)
+				   || !strncmp(s, "while", len));
 	break;
     case 6:
-	state->may_have_pattern = !strncmp(s, "unless", ok);
+	state->may_have_pattern = !strncmp(s, "unless", len);
 	break;
     }
 }
@@ -1153,7 +1155,7 @@ check_keyword(char *s, int ok, AfterKey * state)
 static char *
 put_IDENT(char *s, int ok, AfterKey * if_wrd)
 {
-    char *attr = 0;
+    const char *attr = 0;
     char save = s[ok];
 
     s[ok] = '\0';
@@ -1170,13 +1172,13 @@ put_IDENT(char *s, int ok, AfterKey * if_wrd)
 static char *
 put_NOKEYWORD(char *s, int ok, AfterKey * if_wrd)
 {
-    char *attr = 0;
+    const char *attr = 0;
     char save = s[ok];
 
     s[ok] = '\0';
     attr = keyword_attr(s);
     s[ok] = save;
-    flt_puts(s, ok, "");
+    flt_puts(s, ok, attr);
     check_keyword(s, ok, if_wrd);
     return s + ok;
 }
@@ -1197,7 +1199,7 @@ end_BACKTIC(char *s)
 	}
 	++s;
     }
-    return found ? (s - base) : 0;
+    return found ? (int) (s - base) : 0;
 }
 
 /*
@@ -1225,7 +1227,7 @@ is_GLOB(char *s)
 	    }
 	}
     }
-    return (wild && both) ? (s - base) : 0;
+    return (wild && both) ? (int) (s - base) : 0;
 }
 
 /*
@@ -1235,7 +1237,7 @@ is_GLOB(char *s)
 static int
 is_FORMAT(char *s, int len)
 {
-    if (len == 6 && !strncmp(s, "format", len)) {
+    if (len == 6 && !strncmp(s, "format", (size_t) len)) {
 	s += len;
 	s += is_BLANK(s);
 	s += is_NAME(s);
@@ -1264,7 +1266,7 @@ init_filter(int before GCC_UNUSED)
 static void
 do_filter(FILE *input GCC_UNUSED)
 {
-    static unsigned used;
+    static size_t used;
     static char *line;
 
     AfterKey if_old;
@@ -1283,8 +1285,8 @@ do_filter(FILE *input GCC_UNUSED)
     int quoted = 0;
     int save;
     size_t request = 0;
-    unsigned actual = 0;
-    unsigned mark_len = 0;
+    size_t actual = 0;
+    size_t mark_len = 0;
 
     (void) input;
 
@@ -1322,7 +1324,6 @@ do_filter(FILE *input GCC_UNUSED)
 	the_last = the_file + the_size;
 
 	s = the_file;
-	if_old = nullKey;
 	if_wrd = nullKey;
 	while (MORE(s)) {
 	    if (*s == '\n' || s == the_file) {
@@ -1388,7 +1389,7 @@ do_filter(FILE *input GCC_UNUSED)
 		    flt_puts(s, ok, Preproc_attr);
 		    s += ok;
 		} else if ((ok = is_COMMENT(s)) != 0) {
-		    int skip = (skip_BLANKS(s) - s);
+		    int skip = (int) (skip_BLANKS(s) - s);
 		    ok -= skip;
 		    s += skip;
 		    flt_puts(s, ok, Comment_attr);
@@ -1457,7 +1458,7 @@ do_filter(FILE *input GCC_UNUSED)
 			clearOp();
 			flt_puts(s, ok, keyword_attr(s));
 			check_keyword(s, ok, &if_wrd);
-			s[ok] = save;
+			s[ok] = (char) save;
 			s += ok;
 		    }
 		} else if ((ok = is_Option(s)) != 0) {
