@@ -1,5 +1,5 @@
 /*
- * $Header: /usr/build/vile/vile/filters/RCS/rubyfilt.c,v 1.45 2008/05/26 00:53:35 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/rubyfilt.c,v 1.51 2010/07/13 13:37:15 tom Exp $
  *
  * Filter to add vile "attribution" sequences to ruby scripts.  This is a
  * translation into C of an earlier version written for LEX/FLEX.
@@ -21,7 +21,7 @@ DefineFilter("ruby");
 #define isIdent0(c)  (isalpha(CharOf(c)) || c == '_')
 
 #define MORE(s)	     ((s) < the_last)
-#define ATLEAST(s,n) (the_last - (s) > (n))
+#define ATLEAST(s,n) ((the_last - (s)) > (n))
 
 #ifdef DEBUG
 #define DPRINTF(params) if(FltOptions('d'))printf params
@@ -73,6 +73,19 @@ static char *the_file;
 static char *the_last;
 static size_t the_size;
 
+/*
+ * Stack of here-document tags
+ */
+#define HERE_TAGS struct _here_tags
+HERE_TAGS {
+    HERE_TAGS *next;
+    char *value;
+    int strip;
+    int quote;
+};
+
+static HERE_TAGS *here_tags;
+
 #ifdef DEBUG
 static char *
 stateName(States state)
@@ -215,7 +228,7 @@ is_STRINGS(char *s, int *err, int left_delim, int right_delim, int single)
 	    }
 	    s++;
 	}
-	found = s - base;
+	found = (int) (s - base);
 	DPRINTF(("...found(%d)\n", found));
     }
     return found;
@@ -246,7 +259,7 @@ is_QIDENT(char *s)
 	    break;
 	}
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -327,7 +340,7 @@ is_INSTANCE(char *s)
 	if (*s == '@')
 	    s++;
 	if ((found = is_KEYWORD(s)) != 0)
-	    found += (s - base);
+	    found += (int) (s - base);
     }
     return found;
 }
@@ -474,7 +487,7 @@ is_NUMBER(char *s, int *err)
 	s++;
     }
 
-    return value ? (s - base) : 0;
+    return value ? (int) (s - base) : 0;
 }
 
 /*
@@ -498,7 +511,7 @@ is_COMMENT(char *s)
 	}
 	s = t;
     }
-    return (s - base);
+    return (int) (s - base);
 }
 
 /******************************************************************************
@@ -510,12 +523,12 @@ is_COMMENT(char *s)
  * after the keyword.
  */
 static int
-end_marker(char *s, char *marker, int only)
+end_marker(char *s, const char *marker, int only)
 {
-    int len = strlen(marker);
+    int len = (int) strlen(marker);
 
     return (ATLEAST(s, len)
-	    && !strncmp(s, marker, len)
+	    && !strncmp(s, marker, (size_t) len)
 	    && isspace(CharOf(s[len]))
 	    && (!only || (s[len] == '\n')));
 }
@@ -539,15 +552,58 @@ end_POD(char *s)
 }
 
 /*
- * FIXME: we're only implementing a single here-doc (documentation says they
- * can be stacked), and don't check for unbalanced quotes).
+ * Ruby allows more than one here-document to begin on a line.  They "stack",
+ * and are processed in succession.
  */
-static char *
-begin_HERE(char *s, int *quoted, int *strip_here)
+static void
+make_here_tag(char *value, int quote, int strip)
 {
-    char *base;
-    char *marker;
+    size_t size = 0;
+    HERE_TAGS *data = type_alloc(HERE_TAGS, (char *) 0, 1, &size);
+
+    if (data != 0) {
+	HERE_TAGS *p = here_tags;
+	HERE_TAGS *q = 0;
+
+	while (p != 0) {
+	    q = p;
+	    p = p->next;
+	}
+	if (q != 0)
+	    q->next = data;
+	else
+	    here_tags = data;
+	data->next = p;
+	data->value = value;
+	data->quote = quote;
+	data->strip = strip;
+    }
+}
+
+static char *
+free_here_tag(void)
+{
+    HERE_TAGS *next = here_tags->next;
+    char *result = next ? next->value : 0;
+
+    free(here_tags->value);
+    free(here_tags);
+    here_tags = next;
+
+    return result;
+}
+
+/*
+ * FIXME: we don't check for unbalanced quotes.
+ */
+static int
+begin_HERE(char *s)
+{
+    char *base = s;
+    char *first;
+    char *marker = 0;
     int ok;
+    int quote = 1;
     int strip = 0;
 
     if (ATLEAST(s, 3)
@@ -555,33 +611,32 @@ begin_HERE(char *s, int *quoted, int *strip_here)
 	&& s[1] == '<'
 	&& !isBlank(s[2])) {
 	s += 2;
-	base = s;
+	first = s;
 	if (*s == '-') {
 	    strip = 1;
 	    ++s;
 	}
 	if ((ok = is_QIDENT(s)) != 0) {
-	    unsigned temp = 0;
+	    size_t temp = 0;
 
 	    s += ok;
-	    *quoted = 0;
-	    *strip_here = 1;
+	    quote = 0;
 
-	    if ((marker = do_alloc((char *) 0, ok + 1, &temp)) != 0) {
+	    if ((marker = do_alloc((char *) 0, (size_t) (ok + 1), &temp)) != 0) {
 		char *d = marker;
-		while (base != s) {
-		    if (isIdent(*base))
-			*d++ = *base;
+		while (first != s) {
+		    if (isIdent(*first))
+			*d++ = *first;
 		    else
-			*quoted = (*base != DQUOTE);
-		    base++;
+			quote = (*first != DQUOTE);
+		    first++;
 		}
 		*d = 0;
+		make_here_tag(marker, quote, strip);
 	    }
-	    return marker;
 	}
     }
-    return 0;
+    return (marker ? (int) (s - base) : 0);
 }
 
 static char *
@@ -596,7 +651,7 @@ skip_BLANKS(char *s)
 	++s;
     }
     if (s != base) {
-	flt_puts(base, s - base, "");
+	flt_puts(base, (int) (s - base), "");
     }
     return s;
 }
@@ -672,7 +727,7 @@ is_OPERATOR(char *s)
 	for (n = 0; n < TABLESIZE(table); ++n) {
 	    if (ATLEAST(s, table[n].len)
 		&& table[n].ops[0] == *s
-		&& !memcmp(s, table[n].ops, table[n].len)) {
+		&& !memcmp(s, table[n].ops, (size_t) table[n].len)) {
 		found = table[n].len;
 		break;
 	    }
@@ -723,7 +778,7 @@ is_REGEXP(char *s, int left_delim, int right_delim)
 	    while (MORE(s) && isalpha(CharOf(*s))) {
 		++s;
 	    }
-	    found = s - base;
+	    found = (int) (s - base);
 	    break;
 	} else {
 	    ++s;
@@ -879,7 +934,7 @@ line_size(char *s)
 	    break;
 	s++;
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 static char *
@@ -967,7 +1022,7 @@ put_remainder(char *s, char *attr, int quoted)
 static char *
 put_COMMENT(char *s, int ok)
 {
-    int skip = (skip_BLANKS(s) - s);
+    int skip = (int) (skip_BLANKS(s) - s);
     ok -= skip;
     s += skip;
     flt_puts(s, ok, Comment_attr);
@@ -977,13 +1032,13 @@ put_COMMENT(char *s, int ok)
 static char *
 put_KEYWORD(char *s, int ok, int *had_op)
 {
-    char *attr = 0;
+    const char *attr = 0;
     char save = s[ok];
 
     s[ok] = '\0';
     attr = keyword_attr(s);
     s[ok] = save;
-    if ((attr == 0 || *attr == '\0') && isupper(CharOf(s[0])))
+    if (isEmpty(attr) && isupper(CharOf(s[0])))
 	attr = Type_attr;
     flt_puts(s, ok, attr);
     *had_op = (attr == Keyword_attr);
@@ -1002,7 +1057,7 @@ put_OPERATOR(char *s, int ok, int *had_op)
 static char *
 put_VARIABLE(char *s, int ok)
 {
-    char *attr = 0;
+    const char *attr = 0;
     char save = s[ok];
 
     s[ok] = '\0';
@@ -1048,7 +1103,7 @@ put_REGEXP(char *s, int length, int delim)
 	    while (MORE(s) && isalpha(CharOf(*s))) {
 		++s;
 	    }
-	    flt_bfr_embed(last, s - last, Keyword_attr);
+	    flt_bfr_embed(last, (int) (s - last), Keyword_attr);
 	    break;
 	} else {
 	    flt_bfr_append(s, 1);
@@ -1071,10 +1126,10 @@ init_filter(int before GCC_UNUSED)
 static void
 do_filter(FILE *input GCC_UNUSED)
 {
-    static unsigned used;
+    static size_t used;
     static char *line;
 
-    unsigned actual = 0;
+    size_t actual = 0;
     size_t request = 0;
     States state = eCODE;
     TType this_tok = tNULL;
@@ -1085,8 +1140,6 @@ do_filter(FILE *input GCC_UNUSED)
     int ok;
     int err;
     int delim;
-    int strip_here = 0;		/* strip leading blanks in here-document */
-    int quote_here = 0;		/* tokens in here-document are quoted */
     int had_op = 1;		/* true to allow regex */
 
     (void) input;
@@ -1126,6 +1179,8 @@ do_filter(FILE *input GCC_UNUSED)
 	s = the_file;
 	while (MORE(s)) {
 	    if (*s == '\n') {
+		if (marker != 0)
+		    state = eHERE;
 		in_line = -1;
 		if (state == eCODE)
 		    had_op = 1;
@@ -1194,11 +1249,11 @@ do_filter(FILE *input GCC_UNUSED)
 		break;
 	    case eCODE:
 		if ((last_tok == tKEYWORD || last_tok == tOPERATOR)
-		    && (marker = begin_HERE(s, &quote_here, &strip_here)) != 0) {
-		    ok = 1;	/* FIXME: should allow other tokens on line */
+		    && (ok = begin_HERE(s)) != 0) {
 		    Parsed(this_tok, tHERE);
-		    state = eHERE;
-		    s = put_remainder(s, String_attr, quote_here);
+		    flt_puts(s, ok, String_attr);
+		    s += ok;
+		    marker = here_tags->value;
 		} else if ((in_line < 0 && begin_POD(s + 1))
 			   || ((s == the_file && begin_POD(s)))) {
 		    DPRINTF(("...POD\n"));
@@ -1237,18 +1292,19 @@ do_filter(FILE *input GCC_UNUSED)
 		    s += ok;
 		} else if ((ok = is_MKEYWORD(s, 1)) != 0) {
 		    Parsed(this_tok, tKEYWORD);
-		    if (ok == 5 && !strncmp(s, "alias", ok))
+		    if (ok == 5 && !strncmp(s, "alias", (size_t) ok))
 			state = eALIAS;
-		    else if (ok == 5 && !strncmp(s, "class", ok))
+		    else if (ok == 5 && !strncmp(s, "class", (size_t) ok))
 			state = eCLASS;
-		    else if (ok == 3 && !strncmp(s, "def", ok))
+		    else if (ok == 3 && !strncmp(s, "def", (size_t) ok))
 			state = eDEF;
 		    s = put_KEYWORD(s, ok, &had_op);
 		} else if ((ok = is_VARIABLE(s)) != 0) {
 		    Parsed(this_tok, tVARIABLE);
 		    s = put_VARIABLE(s, ok);
 		    had_op = 0;
-		} else if (ATLEAST(s, (ok = 2)) && !strncmp(s, "?\"", ok)) {
+		} else if (ATLEAST(s, (ok = 2))
+			   && !strncmp(s, "?\"", (size_t) ok)) {
 		    Parsed(this_tok, tVARIABLE);
 		    s = put_VARIABLE(s, ok);	/* csv.rb uses it, undocumented */
 		    had_op = 0;
@@ -1286,12 +1342,19 @@ do_filter(FILE *input GCC_UNUSED)
 		}
 		break;
 	    case eHERE:
-		if (end_marker(s + is_BLANK(s), marker, 1)) {
+		if (here_tags == 0) {
 		    state = eCODE;
-		    free(marker);
-		    marker = 0;
+		} else if (end_marker(s + (here_tags->strip
+					   ? is_BLANK(s)
+					   : 0),
+				      marker, 1)) {
+		    if ((marker = free_here_tag()) == 0)
+			state = eCODE;
 		}
-		s = put_remainder(s, String_attr, quote_here);
+		s = put_remainder(s, String_attr,
+				  (here_tags
+				   ? here_tags->quote
+				   : 0));
 		break;
 	    case ePOD:
 		if (end_POD(s))
@@ -1325,6 +1388,10 @@ do_filter(FILE *input GCC_UNUSED)
 	}
 	free(the_file);
     }
+    while (here_tags != 0) {
+	flt_error("expected tag:%s", here_tags->value);
+	(void) free_here_tag();
+    }
 }
 
 #if NO_LEAKS
