@@ -13,7 +13,7 @@
  *
  *	modify (ifdef-style) 'expand_leaf()' to allow ellipsis.
  *
- * $Header: /usr/build/vile/vile/RCS/glob.c,v 1.90 2007/08/08 22:50:30 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/glob.c,v 1.94 2010/09/05 17:41:52 tom Exp $
  *
  */
 
@@ -121,13 +121,17 @@ string_has_globs(const char *item)
 	if (*item == GLOB_SINGLE || *item == GLOB_MULTI)
 	    return TRUE;
 #if OPT_GLOB_ELLIPSIS || SYS_VMS
-	if (!strncmp(item, GLOB_ELLIPSIS, sizeof(GLOB_ELLIPSIS) - 1))
-	    return TRUE;
+	if (vl_glob_opts & vl_GLOB_ELLIPSIS) {
+	    if (!strncmp(item, GLOB_ELLIPSIS, sizeof(GLOB_ELLIPSIS) - 1))
+		return TRUE;
+	}
 #endif
 #if !OPT_VMS_PATH
 #if OPT_GLOB_RANGE
-	if (*item == GLOB_RANGE[0])
-	    return TRUE;
+	if (vl_glob_opts & vl_GLOB_RANGE) {
+	    if (*item == GLOB_RANGE[0])
+		return TRUE;
+	}
 #endif
 #endif
 	item++;
@@ -157,10 +161,12 @@ string_has_wildcards(const char *item)
 #endif
 #if !OPT_VMS_PATH
 #if OPT_GLOB_ENVIRON
-	if (*item == '$'
+	if ((vl_glob_opts & vl_GLOB_ENVIRON)
+	    && (*item == GLOB_ENVIRON[0])
 	    && (item == base || !isEscaped(item))
-	    && (isname(item[1]) || isdelim(item[1])))
+	    && (isname(item[1]) || isdelim(item[1]))) {
 	    return TRUE;
+	}
 #endif
 #endif
 	item++;
@@ -251,7 +257,8 @@ glob_match_leaf(char *leaf, char *pattern)
 	    if (!multi && *leaf != EOS)
 		return FALSE;
 #if OPT_GLOB_RANGE
-	} else if (*pattern == GLOB_RANGE[0]) {
+	} else if ((vl_glob_opts & vl_GLOB_RANGE)
+		   && (*pattern == GLOB_RANGE[0])) {
 	    int found = FALSE;
 	    char *first = ++pattern;
 	    int negate = 0;
@@ -375,7 +382,7 @@ expand_leaf(char *path,		/* built-up pathname, top-level */
 	    (void) strcpy(path, ".");
 	leaf = skip_string(path) + 1;
     } else {
-	len = (int) (wild - pattern) - 1;
+	len = (size_t) (wild - pattern) - 1;
 	if (*(s = path) != EOS) {
 	    s += strlen(s);
 	    *s++ = SLASHC;
@@ -599,11 +606,12 @@ compar(const void *a, const void *b)
 static int
 glob_from_pipe(const char *pattern)
 {
+    static char only_echo[] = "!echo %s";
 #ifdef GVAL_GLOB
     char *cmd = global_g_val_ptr(GVAL_GLOB);
     int single;
 #else
-    static char cmd[] = "!echo %s";
+    char *cmd = only_echo;
     static int single = TRUE;
 #endif
     FILE *cf;
@@ -620,7 +628,7 @@ glob_from_pipe(const char *pattern)
      * and break up the output with newlines.
      */
     if (!isShellOrPipe(cmd)) {
-	cmd = "!echo %s";
+	cmd = only_echo;
 	single = TRUE;
 	d = cmd + 1;
     } else {
@@ -697,54 +705,56 @@ expand_environ(char *pattern)
     const char *s;
     char save[NFILEN];
 
-    for (j = 0; pattern[j] != EOS; j++) {
-	if (j != 0 && isEscaped(pattern + j))
-	    continue;
-
-	k = j + 1;
-	if (pattern[j] == '$' && pattern[k] != EOS) {
-	    if (pattern[k] == '(')
-		delim = ')';
-	    else if (pattern[k] == '{')
-		delim = '}';
-	    else if (isAlnum(pattern[k]))
-		delim = EOS;
-	    else
+    if (vl_glob_opts & vl_GLOB_ENVIRON) {
+	for (j = 0; pattern[j] != EOS; j++) {
+	    if (j != 0 && isEscaped(pattern + j))
 		continue;
 
-	    if (delim != EOS)
-		k++;
-	    left =
-		right = k;
+	    k = j + 1;
+	    if (pattern[j] == GLOB_ENVIRON[0] && pattern[k] != EOS) {
+		if (pattern[k] == '(')
+		    delim = ')';
+		else if (pattern[k] == '{')
+		    delim = '}';
+		else if (isAlnum(pattern[k]))
+		    delim = EOS;
+		else
+		    continue;
 
-	    while (pattern[k] != EOS) {
-		right = k;
-		if (delim != EOS) {
-		    if (pattern[k++] == delim)
-			break;
-		} else if (isname(pattern[k])) {
+		if (delim != EOS)
 		    k++;
-		} else {
-		    break;
+		left =
+		    right = k;
+
+		while (pattern[k] != EOS) {
+		    right = k;
+		    if (delim != EOS) {
+			if (pattern[k++] == delim)
+			    break;
+		    } else if (isname(pattern[k])) {
+			k++;
+		    } else {
+			break;
+		    }
 		}
-	    }
-	    if (delim == EOS)
-		right = k;
+		if (delim == EOS)
+		    right = k;
 
-	    (void) strcpy(save, pattern + k);
-	    if (right != left) {
-		pattern[right] = EOS;
+		(void) strcpy(save, pattern + k);
+		if (right != left) {
+		    pattern[right] = EOS;
 #if SYS_MSDOS || SYS_OS2
-		mkupper(pattern + left);
+		    mkupper(pattern + left);
 #endif
-		if ((s = getenv(pattern + left)) == 0)
+		    if ((s = getenv(pattern + left)) == 0)
+			s = "";
+		} else
 		    s = "";
-	    } else
-		s = "";
 
-	    (void) strcpy(pattern + j, s);
-	    (void) strcat(pattern, save);
-	    j += (int) strlen(s) - 1;
+		(void) strcpy(pattern + j, s);
+		(void) strcat(pattern, save);
+		j += (int) strlen(s) - 1;
+	    }
 	}
     }
 }
@@ -847,6 +857,10 @@ expand_pattern(char *item)
 #endif
 	expand_environ(pattern);
 	if (string_has_globs(pattern)) {
+	    /*
+	     * FIXME - stripping escapes is done too early, in case we have
+	     * mixed [pattern] and \[xxx\] token.
+	     */
 	    strip_escapes(pattern);
 	    if ((result = expand_leaf(builtup, pattern)) != FALSE
 		&& (myLen - first > 1)) {
