@@ -55,7 +55,7 @@
  *	not (yet) correspond to :-commands.  Before implementing, probably will
  *	have to make TESTC a settable mode.
  *
- * $Header: /usr/build/vile/vile/RCS/history.c,v 1.86 2007/08/29 00:46:27 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/history.c,v 1.90 2015/01/22 23:53:22 tom Exp $
  *
  */
 
@@ -67,6 +67,7 @@
 #define HST_QUOTES 1
 
 #define	tb_args(p)	tb_values(p), (int)tb_length(p)
+#define	tb_args2(p)	tb_values(p), tb_length(p)
 #define	lp_args(p)	lvalue(p), llength(p)
 
 typedef struct {
@@ -200,7 +201,7 @@ endOfParm(HST * parm, char *src, int offset, int limit)
 		    }
 		} else
 #endif
-		if ((parm->endfunc) (src, n, src[n], parm->eolchar))
+		if ((parm->endfunc) (src, (size_t) n, src[n], parm->eolchar))
 		    break;
 	    }
 	}
@@ -310,7 +311,7 @@ glueBufferToResult(TBUFF **dst, TBUFF *src)
 	&& needQuotes(src)) {
 	append_quoted_token(dst, tb_values(src), tb_length(src));
     } else {
-	(void) tb_bappend(dst, tb_args(src));
+	(void) tb_bappend(dst, tb_args2(src));
     }
 }
 
@@ -361,14 +362,14 @@ hst_append(TBUFF *cmd, int glue, int can_extend)
 	   (int) tb_length(cmd),
 	   tb_visible(cmd),
 	   glue));
-    TRACE(("...MyText        :%d:%s\n", tb_length(MyText), tb_visible(MyText)));
+    TRACE(("...MyText        :%d:%s\n", (int) tb_length(MyText), tb_visible(MyText)));
     if (can_extend && willExtend(tb_args(cmd))
 	&& tb_length(cmd) > (size_t) skip) {
 	kbd_pushback(cmd, skip);
     }
 
     glueBufferToResult(&MyText, cmd);
-    TRACE(("...MyText        :%d:%s\n", tb_length(MyText), tb_visible(MyText)));
+    TRACE(("...MyText        :%d:%s\n", (int) tb_length(MyText), tb_visible(MyText)));
     MyGlue = glue;
 }
 
@@ -402,6 +403,12 @@ hst_flush(void)
     WINDOW *wp;
     LINE *lp;
 
+    TRACE(("hst_flush %d:%s\n",
+	   MyLevel,
+	   visible_buff(tb_values(MyText),
+			tb_length(MyText),
+			FALSE)));
+
     if (MyLevel <= 0)
 	return;
     if (MyLevel-- != 1)
@@ -486,7 +493,8 @@ hst_find(HST * parm, BUFFER *bp, LINE *lp, int direction)
 	    return 0;		/* empty or no matches */
 
 	if (!lisreal(lp)
-	    || ((size_t) llength(lp) <= tb_length(MyText) + willGlue())
+	    || ((size_t) llength(lp) <= (tb_length(MyText)
+					 + (size_t) willGlue()))
 	    || (sameLine(lp, tb_args(MyText)) < 0))
 	    continue;		/* prefix mismatches */
 
@@ -546,23 +554,31 @@ hst_display(HST * parm, char *src, int srclen)
 	int n = endOfParm(parm, src, offset, srclen) - offset;
 
 	src += offset;
+	srclen -= offset;
 	stripped = src;
 #if HST_QUOTES
 	TRACE(("...offset=%d, n=%d, MyText=%s\n", offset, n, tb_visible(MyText)));
-	if (tb_length(MyText) != 0 && isShellOrPipe(tb_values(MyText)))
+	if (tb_length(MyText) != 0 &&
+	    isShellOrPipe(tb_values(MyText))) {
 	    TRACE(("...MyText is a shell command\n"));
-	else if (tb_length(MyText) == 0 && offset == 0 && isShellOrPipe(src))
+	} else if (srclen == 0) {
+	    TRACE(("...OOPS - no source?\n"));
+	} else if (tb_length(MyText) == 0 && offset == 0 && isShellOrPipe(src)) {
 	    TRACE(("...src is a shell command\n"));
-	else if (*src == DQUOTE || isSpace(parm->eolchar))
+	} else if (tb_length(MyText) == 0 && offset == 0 && *src == SQUOTE) {
+	    TRACE(("...src is a position command\n"));
+	    n = srclen;
+	} else if (*src == DQUOTE || isSpace(parm->eolchar)) {
 	    stripped = stripQuotes(src, n,
 				   isSpace(parm->eolchar) ? ' ' : parm->eolchar,
 				   &n);
+	}
 #endif
 	TRACE(("...hst_display offset=%d, string='%.*s'\n", offset, n, stripped));
-	*parm->position = kbd_show_response(parm->buffer,
-					    stripped,
-					    n,
-					    parm->eolchar, parm->options);
+	*parm->position = (size_t) kbd_show_response(parm->buffer,
+						     stripped,
+						     (size_t) n,
+						     parm->eolchar, parm->options);
 #if HST_QUOTES
 	if (stripped != 0 && stripped != src)
 	    free(stripped);
@@ -601,29 +617,33 @@ static LINE *
 hst_scroll(LINE *lp1, HST * parm)
 {
     BUFFER *bp = makeMyBuff();
-    LINE *lp0 = buf_head(bp);
-    LINE *lp2 = hst_find(parm, bp, lp1, h_direction);
+    LINE *result = 0;
 
-    if (lp1 != lp2) {
-	if (lp2 == 0) {
-	    if (h_direction + h_distance == 0) {
-		lp1 = lp0;
-		h_distance = 0;
-		display_TBUFF(parm, h_original);
-	    } else {
-		if (lp1 == lp0)	/* nothing to scroll for */
+    if (bp != 0) {
+	LINE *lp0 = buf_head(bp);
+	LINE *lp2 = hst_find(parm, bp, lp1, h_direction);
+
+	if (lp1 != lp2) {
+	    if (lp2 == 0) {
+		if (h_direction + h_distance == 0) {
+		    lp1 = lp0;
 		    h_distance = 0;
-		kbd_alarm();
+		    display_TBUFF(parm, h_original);
+		} else {
+		    if (lp1 == lp0)	/* nothing to scroll for */
+			h_distance = 0;
+		    kbd_alarm();
+		}
+		result = lp1;
+	    } else {
+		h_distance += h_direction;
+		display_LINE(parm, lp2);
+		h_was_edited++;
+		result = lp2;
 	    }
-	    return lp1;
-	} else {
-	    h_distance += h_direction;
-	    display_LINE(parm, lp2);
-	    h_was_edited++;
-	    return lp2;
 	}
     }
-    return 0;
+    return result;
 }
 
 /*
