@@ -44,7 +44,7 @@
  *	tgetc_avail()     true if a key is avail from tgetc() or below.
  *	keystroke_avail() true if a key is avail from keystroke() or below.
  *
- * $Header: /usr/build/vile/vile/RCS/input.c,v 1.316 2007/12/27 19:18:51 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/input.c,v 1.336 2010/07/25 09:29:16 tom Exp $
  *
  */
 
@@ -126,7 +126,7 @@ shell_complete(DONE_ARGS)
     size_t base;
     size_t first = 0;
 
-    TRACE(("shell_complete %d:'%s'\n", *pos, buf));
+    TRACE(("shell_complete %d:'%s'\n", (int) *pos, buf));
     if (buf != 0) {
 	if (isShellOrPipe(buf))
 	    first++;
@@ -151,6 +151,14 @@ shell_complete(DONE_ARGS)
 }
 #endif /* COMPLETE_FILES */
 
+int
+read_msgline(int flag)
+{
+    int old_value = reading_msg_line;
+    reading_msg_line = flag;
+    return old_value;
+}
+
 /*
  * Ask a yes or no question in the message line. Return either TRUE, FALSE, or
  * ABORT. The ABORT status is returned if the user bumps out of the question
@@ -164,7 +172,7 @@ mlyesno(const char *prompt)
 
     /* in case this is right after a shell escape */
     if (update(TRUE)) {
-	reading_msg_line = TRUE;
+	int old_reading = read_msgline(TRUE);
 	for_ever {
 	    mlforce("%s [y/n]? ", prompt);
 	    c = (char) keystroke();	/* get the response */
@@ -183,7 +191,7 @@ mlyesno(const char *prompt)
 		break;
 	    }
 	}
-	reading_msg_line = FALSE;
+	read_msgline(old_reading);
     }
     return result;
 }
@@ -198,7 +206,7 @@ mlquickask(const char *prompt, const char *respchars, int *cp)
     int result = ABORT;
 
     if (update(TRUE)) {
-	reading_msg_line = TRUE;
+	int old_reading = read_msgline(TRUE);
 	for_ever {
 	    mlforce("%s ", prompt);
 	    *cp = keystroke();	/* get the response */
@@ -214,7 +222,7 @@ mlquickask(const char *prompt, const char *respchars, int *cp)
 
 	    kbd_alarm();
 	}
-	reading_msg_line = FALSE;
+	read_msgline(old_reading);
     }
     return result;
 }
@@ -562,7 +570,7 @@ tgetc(int quoted)
 	    } while (c == -1);
 	}
 	(void) im_waiting(FALSE);
-	if (quoted || ((UINT) c != kcod2key(intrc)))
+	if (quoted || (c != kcod2key(intrc)))
 	    record_char(c);
     }
 
@@ -570,10 +578,10 @@ tgetc(int quoted)
 #if OPT_TRACE
     {
 	char temp[NSTRING];
-	TRACE(("tgetc(%s) = %s\n", (quoted
-				    ? "quoted"
-				    : "unquoted"),
-	       kcod2prc(c, temp)));
+	TRACE(("tgetc(%s) = %s(%#x)\n", (quoted
+					 ? "quoted"
+					 : "unquoted"),
+	       kcod2prc(c, temp), c));
     }
 #endif
     return c;
@@ -770,25 +778,26 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
 {
     int i = 0;
     MARK save_dot;
-    int first = -1;
-    int last = -1;
+    int first_ch = -1;
+    int last_ch = -1;
+    int first_off = -1;
+    int last_off = -1;
+    int at_end = 0;
 
     TRACE((T_CALLED "vl_ctype2tbuff(incl=%#lx)\n", (ULONG) inclchartype));
 
     save_dot = DOT;
 
-    tb_init(result, EOS);
-
     /*
      * If we're processing the whole line, find the beginning of the token
      * by searching to the left.
      */
     if (whole_line
 	&& DOT.o > 0
-	&& istype(inclchartype, char_at(DOT))) {
+	&& HasCType(inclchartype, CharAtDot())) {
 	while (DOT.o > 0) {
 	    backchar_to_bol(TRUE, 1);
-	    if (!istype(inclchartype, char_at(DOT))) {
+	    if (!HasCType(inclchartype, CharAtDot())) {
 		forwchar_to_eol(TRUE, 1);
 		break;
 	    }
@@ -798,18 +807,19 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
     /*
      * Search for the start of the token, and then the end of the token.
      */
+    first_off = DOT.o;
     while (!is_at_end_of_line(DOT)) {
-	last = CharOf(char_at(DOT));
-	if (first < 0)
-	    first = last;
+	last_ch = CharAtDot();
+	if (first_ch < 0)
+	    first_ch = last_ch;
 #if OPT_WIDE_CTYPES
 	if (i == 0) {
 	    if (inclchartype & vl_scrtch) {
-		if (first != SCRTCH_LEFT[0])
+		if (first_ch != SCRTCH_LEFT[0])
 		    inclchartype &= ~vl_scrtch;
 	    }
 	    if (inclchartype & vl_shpipe) {
-		if (first != SHPIPE_LEFT[0])
+		if (first_ch != SHPIPE_LEFT[0])
 		    inclchartype &= ~vl_shpipe;
 	    }
 	}
@@ -817,32 +827,38 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
 	/* allow "[!command]" */
 	if ((inclchartype & vl_scrtch)
 	    && (i == 1)
-	    && (last == SHPIPE_LEFT[0])) {
+	    && (last_ch == SHPIPE_LEFT[0])) {
 	    /*EMPTY */ ;
-	    /* guard against things like "[Buffer List]" on VMS */
 	} else if ((inclchartype & vl_pathn)
-		   && !ispath(last)
+		   && !HasCType(vl_pathn, last_ch)
 		   && (inclchartype == vl_pathn)) {
+	    /* guard against things like "[Buffer List]" on VMS */
 	    break;
 	} else
 #endif
-	if (inclchartype && !istype(inclchartype, last))
+	if (inclchartype && !HasCType(inclchartype, last_ch)) {
 	    break;
-	tb_append(result, last);
-	if (!forwchar_to_eol(TRUE, 1))
+	}
+	if (!forwchar_to_eol(TRUE, 1)) {
+	    at_end = 1;
 	    break;
+	}
 	i++;
 #if OPT_WIDE_CTYPES
 	if (inclchartype & vl_scrtch) {
 	    if ((inclchartype & vl_pathn)
-		&& ispath(char_at(DOT)))
+		&& HasCType(vl_pathn, CharAtDot())) {
 		continue;
-	    if (last == SCRTCH_RIGHT[0])
+	    }
+	    if (last_ch == SCRTCH_RIGHT[0]) {
 		break;
+	    }
 	}
 #endif
     }
-    tb_append(result, EOS);
+    last_off = DOT.o + at_end;
+    if (!inclchartype)
+	last_off += len_record_sep(curbp);
 
 #if OPT_WIDE_CTYPES
 #if OPT_VMS_PATH
@@ -851,11 +867,18 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
     } else
 #endif
     if (inclchartype & vl_scrtch) {
-	if (last != SCRTCH_RIGHT[0])
-	    tb_init(result, EOS);
+	if (last_ch != SCRTCH_RIGHT[0]) {
+	    last_off = first_off - 1;
+	}
     }
 #endif
 
+    tb_init(result, EOS);
+    while (first_off < last_off && first_off < llength(DOT.l)) {
+	tb_append(result, lgetc(DOT.l, first_off));
+	++first_off;
+    }
+    tb_append(result, EOS);
     DOT = save_dot;
 
     returnCode(tb_length(*result) != 0);
@@ -955,7 +978,7 @@ kbd_delimiter(void)
  * pathnames and buffer names.  Escape those by doubling them.
  */
 char *
-add_backslashes2(char *text, char *find)
+add_backslashes2(char *text, const char *find)
 {
     static TBUFF *temp = 0;
     int n;
@@ -1224,7 +1247,7 @@ kbd_kill_response(TBUFF *buffer, size_t *position, int c)
     if (buf != 0) {
 	tmp = tb_copy(&tmp, buffer);
 	while (cpos > 0) {
-	    cpos--;
+	    cpos -= TbBytesBefore(tmp, cpos);
 	    kbd_erase();
 	    if (c == wkillc) {
 		if (!isSpace(buf[cpos])) {
@@ -1418,7 +1441,7 @@ user_reply(const char *prompt, const char *dft_val)
 	ptb = &passwdbuf;
     }
 
-    if (dft_val != error_val) {
+    if (isLegalVal(dft_val)) {
 	TRACE(("user_reply, given default value %s\n", dft_val));
 	tb_scopy(ptb, dft_val);
     }
@@ -1464,17 +1487,9 @@ shiftMiniBuffer(int offs)
     static const int slack = 5;
     int shift = w_val(wminip, WVAL_SIDEWAYS);
     int adjust;
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    KBD_DATA save;
 
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    savemk = MK;
-
-    curbp = bminip;
-    curwp = wminip;
+    kbd_start(&save);
 
     adjust = offs - (shift + LastMsgCol - 1);
     if (adjust >= 0)
@@ -1482,12 +1497,9 @@ shiftMiniBuffer(int offs)
     else if (shift > 0 && (slack + adjust) < 0)
 	mvleftwind(TRUE, slack - adjust);
 
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
+    kbd_finish(&save);
 
     kbd_flush();
-    endofDisplay();
 }
 
 static int
@@ -1499,36 +1511,54 @@ fakeKeyCode(const CMDFUNC * f)
 }
 
 static int
-editMiniBuffer(TBUFF **buf, size_t *cpos, int c, int margin, int quoted)
+reallyEditMiniBuffer(TBUFF **buf,
+		     size_t *cpos,
+		     int c,
+#if OPT_MULTIBYTE
+		     int recur,
+#endif
+		     int margin,
+		     int quoted)
 {
+#if OPT_MULTIBYTE
+#define editMiniBuffer(buf,cpos,c,margin,quoted) \
+  reallyEditMiniBuffer(buf,cpos,c,FALSE,margin,quoted)
+#else
+#define editMiniBuffer(buf,cpos,c,margin,quoted) \
+  reallyEditMiniBuffer(buf,cpos,c,margin,quoted)
+#endif
     int edited = FALSE;
     const CMDFUNC *cfp;
     int save_insertmode = insertmode;
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    KBD_DATA save;
 
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    curbp = bminip;
-    curwp = wminip;
-    savemk = MK;
+    kbd_start(&save);
     ++no_minimsgs;
 
     TRACE((T_CALLED
 	   "editMiniBuffer(%d:%s) called with c=%#x, miniedit=%d, dot=%d\n",
-	   *cpos, tb_visible(*buf),
+	   (int) *cpos, tb_visible(*buf),
 	   c, miniedit, DOT.o));
 
 #if OPT_MULTIBYTE
-    if (char2int(c) >= (int) iBIT(MinCBits)) {
+    /*
+     * Anything that does not fit into 8 bits has to be chopped up to pass it
+     * into the command-interpreting and data-insertion.  Do this by converting
+     * to UTF-8 and recurring.
+     *
+     * As a special case, codes in 128..255 have to be converted to UTF-8, but
+     * we want to do this only for the case where input is assumed to be in
+     * UTF-8.  That makes it still usable for ASCII and Latin-1 editing.
+     */
+    if (!recur
+	&& ((char2int(c) >= (int) iBIT(MinCBits))
+	    || (char2int(c) >= 128 && b_is_utfXX(bminip)))) {
 	UCHAR temp[10];
 	int used = vl_conv_to_utf8(temp, c, sizeof(temp));
 	int n;
 
 	for (n = 0; n < used; ++n) {
-	    edited |= editMiniBuffer(buf, cpos, temp[n], margin, quoted);
+	    edited |= reallyEditMiniBuffer(buf, cpos, temp[n], TRUE, margin, quoted);
 	}
     } else
 #endif
@@ -1566,7 +1596,7 @@ editMiniBuffer(TBUFF **buf, size_t *cpos, int c, int margin, int quoted)
 	    DOT.o = first;
 	    MK = DOT;
 	    curwp->w_line = DOT;
-	    regionshape = EXACT;	/* operdel(), etc., do not set this */
+	    regionshape = rgn_EXACT;	/* operdel(), etc., do not set this */
 	    (void) execute(cfp, FALSE, 1);
 	    insertmode = save_insertmode;
 	    edited = TRUE;
@@ -1648,16 +1678,14 @@ editMiniBuffer(TBUFF **buf, size_t *cpos, int c, int margin, int quoted)
 	    *cpos = tb_length(*buf);
 
 	TRACE(("...editMiniBuffer(%d:%s) returns %d, miniedit=%d, dot=%d\n",
-	       *cpos, tb_visible(*buf),
+	       (int) *cpos, tb_visible(*buf),
 	       edited, miniedit, DOT.o));
     }
 
     --no_minimsgs;
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
+
+    kbd_finish(&save);
     kbd_flush();
-    endofDisplay();
 
     returnCode(edited);
 }
@@ -1699,7 +1727,13 @@ read_quoted(int count, int inscreen)
 	    c = '0';
 	    digs = 5;		/* 4 plus the fake '0' */
 	    base = 16;
-	    str = "unicode";
+	    str = "unicode (hex)";
+	    limit = 0xffff;
+	} else if (c == 'U') {
+	    c = '0';
+	    digs = 6;		/* 5 plus the fake '0' */
+	    base = 10;
+	    str = "unicode (dec)";
 	    limit = 0xffff;
 	} else
 #endif
@@ -1834,6 +1868,9 @@ kbd_reply(const char *prompt,	/* put this out first */
     const char *result;
     const char *old_bname = curbp->b_bname;
     WINDOW *old_wp = curwp;
+    int old_reading;
+
+    assert(extbuf != 0);
 
     TRACE((T_CALLED "kbd_reply(prompt=%s, extbuf=%s, options=%#x)\n",
 	   TRACE_NULL(prompt),
@@ -1850,8 +1887,8 @@ kbd_reply(const char *prompt,	/* put this out first */
      * Allowing the user to change the string to something like "ERROR_VALUE"
      * would be confusing and is not really necessary.
      */
-    if ((extbuf != 0 && *extbuf != 0)
-	&& (tb_values(*extbuf) == error_val
+    if (*extbuf != 0
+	&& (isErrorVal(tb_values(*extbuf))
 	    || isTB_ERRS(*extbuf))) {
 	tb_init(extbuf, EOS);
     }
@@ -1875,9 +1912,11 @@ kbd_reply(const char *prompt,	/* put this out first */
 	}
 
 	TRACE(("...getting token from: %s\n", str_visible0(execstr)));
-	execstr = ((options & KBD_REGLUE) != 0 && pushed_back)
-	    ? get_token2(execstr, extbuf, endfunc, eolchar, &actual)
-	    : get_token(execstr, extbuf, endfunc, eolchar, &actual);
+	execstr = (dontmap
+		   ? get_token1(execstr, extbuf, endfunc, eolchar, &actual)
+		   : (((options & KBD_REGLUE) != 0 && pushed_back)
+		      ? get_token2(execstr, extbuf, endfunc, eolchar, &actual)
+		      : get_token(execstr, extbuf, endfunc, eolchar, &actual)));
 	StrToBuff(*extbuf);	/* FIXME: get_token should use TBUFF */
 	TRACE(("...got token, result:  %s\n", tb_visible(*extbuf)));
 	status = (tb_length(*extbuf) != 0);
@@ -1906,7 +1945,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 		 * contents.  Dequote it if we started with a function and got
 		 * a quoted string.
 		 */
-		if (result == error_val) {
+		if (isErrorVal(result)) {
 		    status = ABORT;
 		    (void) tb_scopy(extbuf, "");
 		} else if (toktyp(tb_values(buf)) == TOK_FUNCTION &&
@@ -1964,7 +2003,7 @@ kbd_reply(const char *prompt,	/* put this out first */
     }
 
     quotef = FALSE;
-    reading_msg_line = TRUE;
+    old_reading = read_msgline(TRUE);
 
     /* ask for the input string with the prompt */
     if (prompt != 0)
@@ -2219,7 +2258,7 @@ kbd_reply(const char *prompt,	/* put this out first */
     popdown_completions(old_bname, old_wp);
 #endif
     miniedit = FALSE;
-    reading_msg_line = FALSE;
+    read_msgline(old_reading);
     tb_free(&buf);
     shiftMiniBuffer(0);
 
