@@ -4,7 +4,7 @@
  * Most code probably by Dan Lawrence or Dave Conroy for MicroEMACS
  * Extensions for vile by Paul Fox
  *
- * $Header: /usr/build/vile/vile/RCS/insert.c,v 1.166 2008/05/26 15:17:59 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/insert.c,v 1.174 2010/02/06 00:02:10 tom Exp $
  */
 
 #include	"estruct.h"
@@ -29,6 +29,7 @@ static int shiftwidth(int f, int n);
 static int tab(int f, int n);
 #if !SMALLER
 static int istring(int f, int n, int mode);
+static int inumber(int f, int n, int mode);
 #endif
 
 /* value of insertmode maintained through subfuncs */
@@ -426,10 +427,11 @@ replacechar(int f, int n)
     else {
 	int vi_fix = (!DOT_ARGUMENT || (dotcmdrep <= 1));
 
-	(void) ldel_chars(n, FALSE);
+	set_insertmode(INSMODE_OVR);
 	if (c == quotec) {
 	    t = s = quote_next(f, n);
 	} else {
+	    (void) ldel_chars((B_COUNT) n, FALSE);
 	    if (isreturn(c)) {
 		if (vi_fix)
 		    s = lnewline();
@@ -565,9 +567,9 @@ ins_anytime(int playback, int cur_count, int max_count, int *splicep)
 	     */
 	    if (global_g_val(GMDINSEXEC)) {
 		if (c == cntl_a) {
-		    c = CTLA | keystroke();
+		    c = (int) (CTLA | keystroke());
 		} else if (c == cntl_x) {
-		    c = CTLX | keystroke();
+		    c = (int) (CTLX | keystroke());
 		}
 	    }
 #if OPT_MOUSE
@@ -723,7 +725,7 @@ blanks_on_line(void)
     int list = w_val(curwp, WMDLIST);
 
     for (DOT.o = 0; DOT.o < llength(DOT.l); DOT.o += BytesAt(DOT.l, DOT.o)) {
-	if (isSpace(char_at(DOT))
+	if (isSpace(CharAtDot())
 	    && getccol(list) >= indentwas) {
 	    code = TRUE;
 	    break;
@@ -745,6 +747,26 @@ is_cindent_char(BUFFER *bp, int ch)
 	&& (strchr(b_val_ptr(bp, VAL_CINDENT_CHARS), ch) != 0);
 }
 
+/*
+ * Check if we're replacing text in a UTF-8 buffer.  We'll want to only delete
+ * the old characters at the beginning of each UTF-8 sequence.
+ */
+#if OPT_MULTIBYTE
+static int
+is_utf8_continuation(BUFFER *bp, int ch)
+{
+    int result = FALSE;
+    if (b_is_utfXX(bp) && (ch >= 0x80)) {
+	if ((ch & 0xc0) != 0x80) {
+	    result = TRUE;
+	}
+    }
+    return result;
+}
+#else
+#define is_utf8_continuation(bp, ch) FALSE
+#endif
+
 int
 inschar(int c, int *backsp_limit_p)
 {
@@ -873,9 +895,10 @@ inschar(int c, int *backsp_limit_p)
        and next char is not a tab or we are at a tab stop,
        delete a char forword                        */
     if ((insertmode == INSMODE_OVR)
+	&& !is_utf8_continuation(curbp, c)
 	&& (!DOT_ARGUMENT || (dotcmdrep <= 1))
 	&& (DOT.o < llength(DOT.l))
-	&& (char_at(DOT) != '\t'
+	&& (CharAtDot() != '\t'
 	    || DOT.o % tabstop_val(curbp) == tabstop_val(curbp) - 1)) {
 	autoindented = -1;
 	(void) ldel_chars(1, FALSE);
@@ -929,33 +952,89 @@ istring(int f, int n, int mode)
     static char tstring[NPAT + 1];	/* string to add */
 
     /* ask for string to insert */
-    status = mlreply("String to insert: ", tstring, NPAT);
-    if (status != TRUE)
-	return (status);
+    if ((status = mlreply("String to insert: ", tstring, NPAT)) == TRUE) {
+	n = need_a_count(f, n, 1);
 
-    n = need_a_count(f, n, 1);
+	if (n < 0)
+	    n = -n;
 
-    if (n < 0)
-	n = -n;
+	set_insertmode(mode);
 
-    set_insertmode(mode);
+	backsp_limit = BackspaceLimit();
 
-    backsp_limit = BackspaceLimit();
-
-    /* insert it */
-    while (n--) {
-	tp = tstring;
-	while (*tp) {
-	    status = inschar(*tp++, &backsp_limit);
-	    if (status != TRUE) {
-		set_insertmode(FALSE);
-		return (status);
+	/* insert it */
+	while (n--) {
+	    tp = tstring;
+	    while (*tp) {
+		if ((status = inschar(*tp++, &backsp_limit)) != TRUE) {
+		    n = 0;
+		    break;
+		}
 	    }
 	}
-    }
 
-    set_insertmode(FALSE);
-    return (TRUE);
+	set_insertmode(FALSE);
+    }
+    return (status);
+}
+#endif
+
+#if ! SMALLER
+int
+appnumber(int f, int n)
+{
+    TRACE((T_CALLED "appnumber(f=%d, n=%d)\n", f, n));
+    advance_one_char();
+    returnCode(inumber(f, n, INSMODE_INS));
+}
+
+int
+insnumber(int f, int n)
+{
+    TRACE((T_CALLED "insnumber(f=%d, n=%d)\n", f, n));
+    returnCode(inumber(f, n, INSMODE_INS));
+}
+
+int
+overwnumber(int f, int n)
+{
+    TRACE((T_CALLED "overwnumber(f=%d, n=%d)\n", f, n));
+    returnCode(inumber(f, n, INSMODE_OVR));
+}
+
+/* ask for and insert or overwrite a character-number into the current */
+/* buffer at the current point */
+static int
+inumber(int f, int n, int mode)
+{
+    int status;			/* status return code */
+    int backsp_limit;
+    int value;
+    static char tnumber[NPAT + 1];	/* number to add */
+
+    /* ask for number to insert */
+    if ((status = mlreply("Number to insert: ", tnumber, NPAT)) == TRUE) {
+	if ((status = string_to_number(tnumber, &value)) == TRUE) {
+
+	    n = need_a_count(f, n, 1);
+
+	    if (n < 0)
+		n = -n;
+
+	    set_insertmode(mode);
+
+	    backsp_limit = BackspaceLimit();
+
+	    /* insert it */
+	    while (n--) {
+		if ((status = inschar(value, &backsp_limit)) != TRUE)
+		    break;
+	    }
+
+	    set_insertmode(FALSE);
+	}
+    }
+    return (status);
 }
 #endif
 
@@ -1275,7 +1354,7 @@ shiftwidth(int f GCC_UNUSED, int n GCC_UNUSED)
     space_count = 0;
     all_white = TRUE;
     while (DOT.o < char_index) {
-	c = char_at(DOT);
+	c = CharAtDot();
 	if (c == ' ') {
 	    space_count++;
 	} else {
@@ -1351,15 +1430,19 @@ quote_next(int f, int n)
     c = read_quoted(n, TRUE);
 
     if (c < 0) {
-	return ABORT;
-    } else if (c == '\n') {
-	do {
-	    s = lnewline();
-	} while ((s == TRUE) && (--n != 0));
-	return s;
+	s = ABORT;
     } else {
-	return lins_chars(n, c);
+	if (insertmode == INSMODE_OVR)
+	    (void) ldel_chars((B_COUNT) n, FALSE);
+	if (c == '\n') {
+	    do {
+		s = lnewline();
+	    } while ((s == TRUE) && (--n != 0));
+	} else {
+	    s = lins_chars(n, c);
+	}
     }
+    return s;
 }
 
 #if OPT_EVAL
