@@ -10,7 +10,7 @@
  * editing must be being displayed, which means that "b_nwnd" is non zero,
  * which means that the dot and mark values in the buffer headers are nonsense.
  *
- * $Header: /usr/build/vile/vile/RCS/line.c,v 1.188 2007/10/31 23:55:39 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/line.c,v 1.216 2010/09/07 00:29:28 tom Exp $
  *
  */
 
@@ -21,10 +21,12 @@
 #define poison(p,s)
 #endif
 
-#include	"estruct.h"
-#include	"edef.h"
+#include "estruct.h"
+#include "edef.h"
 
-#define roundlenup(n) ((n+NBLOCK-1) & (UINT)~(NBLOCK-1))
+#include <assert.h>
+
+#define roundlenup(n) (((size_t) (n) + NBLOCK - 1) & (size_t)~(NBLOCK-1))
 
 static int doput(int f, int n, int after, REGIONSHAPE shape);
 static int ldelnewline(void);
@@ -234,21 +236,20 @@ lremove(BUFFER *bp, LINE *lp)
     {
 	AREGION *ap = bp->b_attribs;
 	while (ap != NULL) {
+	    AREGION *next = ap->ar_next;
 	    int samestart = (ap->ar_region.r_orig.l == lp);
 	    int sameend = (ap->ar_region.r_end.l == lp);
+
 	    if (samestart && sameend) {
 		AREGION *tofree = ap;
-		ap = ap->ar_next;
 		free_attrib(bp, tofree);
 	    } else if (samestart) {
 		ap->ar_region.r_orig = mark_after;
-		ap = ap->ar_next;
 	    } else if (sameend) {
 		ap->ar_region.r_end.l = lback(lp);
 		ap->ar_region.r_end.o = llength(ap->ar_region.r_end.l);
-		ap = ap->ar_next;
-	    } else
-		ap = ap->ar_next;
+	    }
+	    ap = next;
 	}
     }
 #endif /* OPT_VIDEO_ATTRS */
@@ -320,7 +321,7 @@ lreplc(LINE *lp, C_NUM off, int c)
 	return FALSE;
 
     if (lvalue(lp)[off] != (char) c) {
-	copy_for_undo(lp);
+	CopyForUndo(lp);
 	lvalue(lp)[off] = (char) c;
 
 	chg_buff(curbp, WFEDIT);
@@ -329,30 +330,40 @@ lreplc(LINE *lp, C_NUM off, int c)
     return TRUE;
 }
 
+#define BumpOff(name, lp, n, doto) if (name.l == lp && name.o > doto) name.o += n
+
 /*
  * Adjust DOT and MK in the given window, which may be the minibuffer.
  */
-static void
-after_linsert(WINDOW *wp, LINE *lp, int n, int doto)
-{
-    if (wp->w_dot.l == lp) {
-	if (wp == curwp || wp->w_dot.o > doto)
-	    wp->w_dot.o += n;
-    }
 #if WINMARK
-    if (wp->w_mark.l == lp) {
-	if (wp->w_mark.o > doto)
-	    wp->w_mark.o += n;
-    }
-#endif
-    if (wp->w_lastdot.l == lp) {
-	if (wp->w_lastdot.o > doto)
-	    wp->w_lastdot.o += n;
-    }
+#define after_linsert(wp, lp, n, doto) \
+{ \
+    if (wp == curwp) { \
+	if (wp->w_dot.l == lp) { \
+	    wp->w_dot.o += n; \
+	} \
+    } else { \
+	BumpOff (wp->w_dot, lp, n, doto); \
+    } \
+    BumpOff(wp->w_mark, lp, n, doto); \
+    BumpOff(wp->w_lastdot, lp, n, doto); \
 }
+#else
+#define after_linsert(wp, lp, n, doto) \
+{ \
+    if (wp == curwp) { \
+	if (wp->w_dot.l == lp) { \
+	    wp->w_dot.o += n; \
+	} \
+    } else { \
+	BumpOff (wp->w_dot, lp, n, doto); \
+    } \
+    BumpOff(wp->w_lastdot, lp, n, doto); \
+}
+#endif
 
 /*
- * Insert "n" copies of the character "c" at the current location of dot. In
+ * Insert "n" copies of the byte "c" at the current location of dot. In
  * the easy case all that happens is the text is stored in the line. In the
  * hard case, the line has to be reallocated. When the window list is updated,
  * take special care; I screwed it up once. You always update dot in the
@@ -365,7 +376,6 @@ lins_bytes(int n, int c)
 {
     char *cp1;
     char *cp2;
-    LINE *tmp;
     LINE *lp1;
     LINE *lp2;
     LINE *lp3;
@@ -375,9 +385,12 @@ lins_bytes(int n, int c)
     char *ntext;
     size_t nsize;
     int rc = TRUE;
+    USHORT changed = 0;
 
     beginDisplay();
 
+    assert(n >= 0);
+
     lp1 = DOT.l;		/* Current line         */
     if (n == 0) {
 	;
@@ -395,37 +408,40 @@ lins_bytes(int n, int c)
 	    set_lback(lp2, lp3);
 	    (void) memset(lvalue(lp2), c, (size_t) n);
 
-	    tag_for_undo(lp2);
+	    TagForUndo(lp2);
 
 	    /* don't move DOT until after tagging for undo */
 	    /*  (it's important in an empty buffer) */
 	    DOT.l = lp2;
 	    DOT.o = n;
-	    chg_buff(curbp, WFINS | WFEDIT);
+	    changed = (WFINS | WFEDIT);
 	}
     } else {
 	doto = DOT.o;		/* Save for later.      */
-	tmp = lp1;
-	nsize = llength(tmp) + n;
-	if (nsize > tmp->l_size) {	/* Hard: reallocate     */
+	lp1 = lp1;
+	nsize = (size_t) (llength(lp1) + 1 + doto + n);
+	if (nsize > lp1->l_size) {	/* Hard: reallocate     */
 	    /* first, create the new image */
-	    nsize = roundlenup((int) nsize);
-	    copy_for_undo(lp1);
+	    nsize = roundlenup(nsize);
+	    CopyForUndo(lp1);
 	    if ((ntext = castalloc(char, nsize)) == NULL) {
 		rc = FALSE;
 	    } else {
-		if (lvalue(lp1) && doto)	/* possibly NULL if l_size == 0 */
+		if (lvalue(lp1) && doto) {	/* possibly NULL if l_size == 0 */
+		    assert((size_t) doto < nsize);
 		    (void) memcpy(&ntext[0], &lvalue(lp1)[0], (size_t) doto);
+		}
 		(void) memset(&ntext[doto], c, (size_t) n);
 		if (lvalue(lp1)) {
 #if OPT_LINE_ATTRS
 		    UCHAR *l_attrs = lp1->l_attrs;
 		    lp1->l_attrs = 0;	/* momentarily detach */
 #endif
-		    if (llength(lp1) != doto)
+		    if (llength(lp1) > doto) {
 			(void) memcpy(&ntext[doto + n],
 				      &lvalue(lp1)[doto],
 				      (size_t) (llength(lp1) - doto));
+		    }
 		    ltextfree(lp1, curbp);
 #if OPT_LINE_ATTRS
 		    lp1->l_attrs = l_attrs;	/* reattach */
@@ -434,24 +450,29 @@ lins_bytes(int n, int c)
 		lvalue(lp1) = ntext;
 		lp1->l_size = nsize;
 		llength(lp1) += n;
+		assert((size_t) llength(lp1) <= lp1->l_size);
 	    }
 	} else {		/* Easy: in place       */
-	    copy_for_undo(lp1);
-	    chg_buff(curbp, WFEDIT);
-	    tmp = lp1;
+	    CopyForUndo(lp1);
+	    changed = (WFEDIT);
 	    /* don't use memcpy:  overlapping regions.... */
-	    llength(tmp) += n;
-	    if (llength(tmp) - n > doto) {
-		cp2 = &lvalue(tmp)[llength(tmp)];
+	    llength(lp1) += n;
+	    assert((size_t) llength(lp1) <= lp1->l_size);
+	    if (llength(lp1) > (n + doto)) {
+		cp2 = &lvalue(lp1)[llength(lp1)];
 		cp1 = cp2 - n;
-		while (cp1 != &lvalue(tmp)[doto])
+		while (cp1 != &lvalue(lp1)[doto])
 		    *--cp2 = *--cp1;
 	    }
-	    for (i = 0; i < n; ++i)	/* Add the characters       */
-		lvalue(tmp)[doto + i] = (char) c;
+	    assert(llength(lp1) >= (n + doto));
+	    for (i = 0; i < n; ++i) {	/* Add the characters       */
+		lvalue(lp1)[doto + i] = (char) c;
+	    }
 	}
 	if (rc != FALSE) {
-	    chg_buff(curbp, WFEDIT);
+	    int did_wminip = FALSE;
+
+	    changed = (WFEDIT);
 #if ! WINMARK
 	    if (MK.l == lp1) {
 		if (MK.o > doto)
@@ -460,8 +481,12 @@ lins_bytes(int n, int c)
 #endif
 	    for_each_window(wp) {	/* Update windows       */
 		after_linsert(wp, lp1, n, doto);
+		if (wp == wminip)
+		    did_wminip = TRUE;
+	    }
+	    if (!did_wminip) {
+		after_linsert(wminip, lp1, n, doto);
 	    }
-	    after_linsert(wminip, lp1, n, doto);
 	    do_mark_iterate(mp, {
 		if (mp->l == lp1) {
 		    if (mp->o > doto)
@@ -477,6 +502,10 @@ lins_bytes(int n, int c)
 #endif
 	}
     }
+
+    if (changed)
+	chg_buff(curbp, changed);
+
     endofDisplay();
     return (rc);
 }
@@ -494,11 +523,17 @@ lins_chars(int n, int c)
     UCHAR target[10];
     int nbytes;
     int nn;
+    int mapped;
 
-    if (c > 127 && b_is_utfXX(curbp)) {
-	nbytes = vl_conv_to_utf8(target, c, sizeof(target));
-    } else if (!vl_mb_is_8bit(c)) {
-	nbytes = vl_conv_to_utf8(target, c, sizeof(target));
+    if ((c > 127) && b_is_utfXX(curbp)) {
+	nbytes = vl_conv_to_utf8(target, (UINT) c, sizeof(target));
+    } else if (okCTYPE2(vl_wide_enc) && !vl_mb_is_8bit(c)) {
+	nbytes = 1;
+	if (b_is_utfXX(curbp)) {
+	    nbytes = vl_conv_to_utf8(target, (UINT) c, sizeof(target));
+	} else if (vl_ucs_to_8bit(&mapped, c)) {
+	    c = mapped;
+	}
     } else {
 	nbytes = 1;
     }
@@ -513,7 +548,7 @@ lins_chars(int n, int c)
 		break;
 	}
     } else {
-	rc = lins_bytes(n, c);
+	rc = lins_bytes(1, c);
     }
     return rc;
 }
@@ -530,12 +565,14 @@ lins_chars(int n, int c)
 int
 lnewline(void)
 {
+    int rc = FALSE;
     char *cp1;
     char *cp2;
     LINE *lp1;
     LINE *lp2;
     int doto;
     WINDOW *wp;
+    USHORT changed = 0;
 
     lp1 = DOT.l;		/* Get the address and  */
     doto = DOT.o;		/* offset of "."        */
@@ -544,100 +581,124 @@ lnewline(void)
 	&& lforw(lp1) == lp1) {
 	/* empty buffer -- just  create empty line */
 	lp2 = lalloc(doto, curbp);
-	if (lp2 == 0)
-	    return (FALSE);
-	/* put lp2 in below lp1 */
-	set_lforw(lp2, lforw(lp1));
-	set_lforw(lp1, lp2);
-	set_lback(lforw(lp2), lp2);
-	set_lback(lp2, lp1);
+	if (lp2 != 0) {
+	    /* put lp2 in below lp1 */
+	    set_lforw(lp2, lforw(lp1));
+	    set_lforw(lp1, lp2);
+	    set_lback(lforw(lp2), lp2);
+	    set_lback(lp2, lp1);
 
-	tag_for_undo(lp2);
+	    TagForUndo(lp2);
 
-	for_each_window(wp) {
-	    if (wp->w_line.l == lp1)
-		wp->w_line.l = lp2;
-	    if (wp->w_dot.l == lp1)
-		wp->w_dot.l = lp2;
+	    for_each_window(wp) {
+		if (wp->w_line.l == lp1)
+		    wp->w_line.l = lp2;
+		if (wp->w_dot.l == lp1)
+		    wp->w_dot.l = lp2;
+	    }
+
+	    changed = (WFHARD | WFINS);
+
+	    rc = lnewline();	/* vi really makes _2_ lines */
 	}
+    } else {
+	lp2 = lalloc(doto, curbp);	/* New first half line */
+	if (lp2 != 0) {
+	    if (doto > 0) {
+		CopyForUndo(lp1);
+#if OPT_LINE_ATTRS
+		if (lp1->l_attrs != 0) {
+		    if (doto == llength(lp1)) {
+			lp2->l_attrs = lp1->l_attrs;
+			lp1->l_attrs = 0;
+		    } else {
+			UCHAR *newattr
+			= typeallocn(UCHAR, (unsigned) (llength(lp1) + 1));
+			int n;
 
-	chg_buff(curbp, WFHARD | WFINS);
+			if (newattr != 0) {
+			    lp2->l_attrs = newattr;
+			    lp2->l_attrs[doto] = EOS;
+			    for (n = 0; (lp1->l_attrs[n] =
+					 lp1->l_attrs[doto + n]) != EOS; ++n) {
+				;
+			    }
+			} else {
+			    FreeAndNull(lp1->l_attrs);
+			}
+		    }
+		}
+#endif
+		cp1 = lvalue(lp1);	/* Shuffle text around  */
+		cp2 = lvalue(lp2);
+		while (cp1 < &lvalue(lp1)[doto])
+		    *cp2++ = *cp1++;
+		cp2 = lvalue(lp1);
+		while (cp1 < &lvalue(lp1)[llength(lp1)])
+		    *cp2++ = *cp1++;
+		llength(lp1) -= doto;
+	    }
+	    /* put lp2 in above lp1 */
+	    set_lback(lp2, lback(lp1));
+	    set_lback(lp1, lp2);
+	    set_lforw(lback(lp2), lp2);
+	    set_lforw(lp2, lp1);
 
-	return lnewline();	/* vi really makes _2_ lines */
-    }
-
-    lp2 = lalloc(doto, curbp);	/* New first half line */
-    if (lp2 == 0)
-	return (FALSE);
-
-    if (doto > 0) {
-	LINE *tmp;
-
-	copy_for_undo(lp1);
-	tmp = lp1;
-	cp1 = lvalue(tmp);	/* Shuffle text around  */
-	cp2 = lvalue(lp2);
-	while (cp1 < &lvalue(tmp)[doto])
-	    *cp2++ = *cp1++;
-	cp2 = lvalue(tmp);
-	while (cp1 < &lvalue(tmp)[llength(tmp)])
-	    *cp2++ = *cp1++;
-	llength(tmp) -= doto;
-    }
-    /* put lp2 in above lp1 */
-    set_lback(lp2, lback(lp1));
-    set_lback(lp1, lp2);
-    set_lforw(lback(lp2), lp2);
-    set_lforw(lp2, lp1);
-
-    tag_for_undo(lp2);
-    dumpuline(lp1);
+	    TagForUndo(lp2);
+	    dumpuline(lp1);
 
 #if ! WINMARK
-    if (MK.l == lp1) {
-	if (MK.o < doto)
-	    MK.l = lp2;
-	else
-	    MK.o -= doto;
-    }
+	    if (MK.l == lp1) {
+		if (MK.o < doto)
+		    MK.l = lp2;
+		else
+		    MK.o -= doto;
+	    }
 #endif
-    for_each_window(wp) {
-	if (wp->w_line.l == lp1)
-	    wp->w_line.l = lp2;
-	if (wp->w_dot.l == lp1) {
-	    if (wp->w_dot.o < doto)
-		wp->w_dot.l = lp2;
-	    else
-		wp->w_dot.o -= doto;
-	}
+	    for_each_window(wp) {
+		if (wp->w_line.l == lp1)
+		    wp->w_line.l = lp2;
+		if (wp->w_dot.l == lp1) {
+		    if (wp->w_dot.o < doto)
+			wp->w_dot.l = lp2;
+		    else
+			wp->w_dot.o -= doto;
+		}
 #if WINMARK
-	if (wp->w_mark.l == lp1) {
-	    if (wp->w_mark.o < doto)
-		wp->w_mark.l = lp2;
-	    else
-		wp->w_mark.o -= doto;
-	}
+		if (wp->w_mark.l == lp1) {
+		    if (wp->w_mark.o < doto)
+			wp->w_mark.l = lp2;
+		    else
+			wp->w_mark.o -= doto;
+		}
 #endif
-	if (wp->w_lastdot.l == lp1) {
-	    if (wp->w_lastdot.o < doto)
-		wp->w_lastdot.l = lp2;
-	    else
-		wp->w_lastdot.o -= doto;
+		if (wp->w_lastdot.l == lp1) {
+		    if (wp->w_lastdot.o < doto)
+			wp->w_lastdot.l = lp2;
+		    else
+			wp->w_lastdot.o -= doto;
+		}
+	    }
+	    do_mark_iterate(mp, {
+		if (mp->l == lp1) {
+		    if ((mp->o < doto) ||
+			(doto == llength(lp2) && llength(lp1) == 0)) {
+			mp->l = lp2;
+		    } else {
+			mp->o -= doto;
+		    }
+		}
+	    });
+	    changed = (WFHARD | WFINS);
+	    rc = TRUE;
 	}
     }
-    do_mark_iterate(mp, {
-	if (mp->l == lp1) {
-	    if (mp->o < doto)
-		mp->l = lp2;
-	    else
-		mp->o -= doto;
-	}
-    });
-#if OPT_LINE_ATTRS
-    FreeAndNull(lp1->l_attrs);
-#endif
-    chg_buff(curbp, WFHARD | WFINS);
-    return (TRUE);
+
+    if (changed) {
+	chg_buff(curbp, changed);
+    }
+
+    return rc;
 }
 
 /*
@@ -656,13 +717,13 @@ ldel_bytes(B_COUNT nbytes, int kflag)
     char *cp2;
     LINE *dotp;
     LINE *nlp;
-    int doto;
+    C_NUM doto;
     B_COUNT uchunk;
-    int schunk;
+    long schunk;
     WINDOW *wp;
-    int i;
+    C_NUM i;
     int status = TRUE;
-    B_COUNT len_rs = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
     lines_deleted = 0;
     while (nbytes != 0) {
@@ -672,11 +733,14 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	    status = FALSE;
 	    break;
 	}
-	uchunk = llength(dotp) - doto;	/* Size of chunk.    */
+	if (doto > llength(dotp))
+	    uchunk = 0;
+	else
+	    uchunk = (B_COUNT) (llength(dotp) - doto);	/* Size of chunk.  */
 	if (uchunk > nbytes)
 	    uchunk = nbytes;
 
-	schunk = uchunk;
+	schunk = (long) uchunk;
 	if (schunk < 0) {
 	    status = FALSE;
 	    break;		/* oops: sign-extension */
@@ -686,7 +750,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	    /* first take out any whole lines below this one */
 	    nlp = lforw(dotp);
 	    while (nlp != buf_head(curbp)
-		   && line_length(nlp) < nbytes) {
+		   && (line_length(nlp) < nbytes)) {
 		if (kflag) {
 		    status = kinsert('\n');
 		    for (i = 0; i < llength(nlp) &&
@@ -697,8 +761,12 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 		    break;
 		lremove(curbp, nlp);
 		lines_deleted++;
-		toss_to_undo(nlp);
-		nbytes -= line_length(nlp);
+		TossToUndo(nlp);
+		assert((B_COUNT) line_length(nlp) <= nbytes);
+		if (line_length(nlp) > nbytes)
+		    nbytes = 0;
+		else
+		    nbytes -= line_length(nlp);
 		nlp = lforw(dotp);
 	    }
 	    if (status != TRUE)
@@ -713,7 +781,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	    lines_deleted++;
 	    continue;
 	}
-	copy_for_undo(DOT.l);
+	CopyForUndo(DOT.l);
 	chg_buff(curbp, WFEDIT);
 
 	cp1 = lvalue(dotp) + doto;	/* Scrunch text.     */
@@ -730,10 +798,10 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	}
 	while (cp2 != lvalue(dotp) + llength(dotp))
 	    *cp1++ = *cp2++;
-	llength(dotp) -= schunk;
+	llength(dotp) -= (int) schunk;
 #if ! WINMARK
 	if (MK.l == dotp && MK.o > doto) {
-	    MK.o -= schunk;
+	    MK.o -= (C_NUM) schunk;
 	    if (MK.o < doto)
 		MK.o = doto;
 	}
@@ -741,7 +809,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	for_each_window(wp) {	/* Fix windows          */
 	    if (wp->w_dot.l == dotp
 		&& wp->w_dot.o > doto) {
-		wp->w_dot.o -= schunk;
+		wp->w_dot.o -= (C_NUM) schunk;
 		if (wp->w_dot.o < doto)
 		    wp->w_dot.o = doto;
 	    }
@@ -755,7 +823,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 #endif
 	    if (wp->w_lastdot.l == dotp
 		&& wp->w_lastdot.o > doto) {
-		wp->w_lastdot.o -= schunk;
+		wp->w_lastdot.o -= (C_NUM) schunk;
 		if (wp->w_lastdot.o < doto)
 		    wp->w_lastdot.o = doto;
 	    }
@@ -763,17 +831,19 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	do_mark_iterate(mp, {
 	    if (mp->l == dotp
 		&& mp->o > doto) {
-		mp->o -= schunk;
+		mp->o -= (C_NUM) schunk;
 		if (mp->o < doto)
 		    mp->o = doto;
 	    }
 	});
 #if OPT_LINE_ATTRS
-	if (!lattr_shift(curbp, dotp, doto, -schunk)) {
+	if (!lattr_shift(curbp, dotp, doto, (int) -schunk)) {
 	    status = FALSE;
 	    break;
 	}
 #endif
+	assert(uchunk != 0);
+	assert(uchunk <= nbytes);
 	nbytes -= uchunk;
     }
     return (status);
@@ -787,7 +857,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 int
 ldel_chars(B_COUNT nchars, int kflag)
 {
-    return ldel_bytes(count_bytes(DOT.l, DOT.o, nchars), kflag);
+    return ldel_bytes((B_COUNT) count_bytes(DOT.l, DOT.o, (int) nchars), kflag);
 }
 #endif
 
@@ -807,21 +877,21 @@ lrepl_ctype(CHARTYPE type, const char *np, int length)
 
     TRACE((T_CALLED "lrepl_ctype:%s%lx:%.*s\n",
 	   type ? "word" : "line",
-	   (ULONG) type, length, NONNULL(np)));
+	   (ULONG) type, length, np));
 
     if ((status = check_editable(curbp)) == TRUE) {
 
 	mayneedundo();
 
 	if (type != 0) {	/* it's an exact region */
-	    regionshape = EXACT;
+	    regionshape = rgn_EXACT;
 	    while (DOT.o < llength(DOT.l)
-		   && istype(type, char_at(DOT))) {
+		   && HasCType(type, CharAtDot())) {
 		if ((status = forwdelchar(FALSE, 1)) != TRUE)
 		    break;
 	    }
 	} else {		/* it's the full line */
-	    regionshape = FULLLINE;
+	    regionshape = rgn_FULLLINE;
 	    DOT.o = w_left_margin(curwp);
 	    if ((status = deltoeol(TRUE, 1)) == TRUE)
 		DOT.o = w_left_margin(curwp);
@@ -870,9 +940,9 @@ lrepl_regex(REGEXVAL * rexp, const char *np, int length)
 	mayneedundo();
 
 	if (lregexec(exp, DOT.l, DOT.o, llength(DOT.l))) {
-	    int old = exp->endp[0] - exp->startp[0];
+	    int old = (int) (exp->endp[0] - exp->startp[0]);
 	    if (old > 0) {
-		regionshape = EXACT;
+		regionshape = rgn_EXACT;
 		status = forwdelchar(TRUE, old);
 	    }
 	} else {
@@ -914,7 +984,7 @@ ldelnewline(void)
     len = llength(lp1);
     /* if the current line is empty, remove it */
     if (len == 0) {		/* Blank line.          */
-	toss_to_undo(lp1);
+	TossToUndo(lp1);
 	lremove(curbp, lp1);
 	return (TRUE);
     }
@@ -926,13 +996,13 @@ ldelnewline(void)
 	return (TRUE);
     else if ((add = llength(lp2)) == 0) {
 	/* next line blank? */
-	toss_to_undo(lp2);
+	TossToUndo(lp2);
 	lremove(curbp, lp2);
 	return (TRUE);
     }
 
     beginDisplay();
-    copy_for_undo(DOT.l);
+    CopyForUndo(DOT.l);
 
     /* no room in line above, make room */
     if (add > (C_NUM) lp1->l_size - len) {
@@ -943,13 +1013,13 @@ ldelnewline(void)
 	if ((ntext = castalloc(char, nsize)) == NULL)
 	      return (FALSE);
 	if (lvalue(lp1)) {	/* possibly NULL if l_size == 0 */
-	    (void) memcpy(&ntext[0], &lvalue(lp1)[0], len);
+	    (void) memcpy(&ntext[0], &lvalue(lp1)[0], (size_t) len);
 	    ltextfree(lp1, curbp);
 	}
 	lvalue(lp1) = ntext;
 	lp1->l_size = nsize;
     }
-    (void) memcpy(lvalue(lp1) + len, lvalue(lp2), add);
+    (void) memcpy(lvalue(lp1) + len, lvalue(lp2), (size_t) add);
 #if ! WINMARK
     if (MK.l == lp2) {
 	MK.l = lp1;
@@ -985,10 +1055,11 @@ ldelnewline(void)
     FreeAndNull(lp2->l_attrs);
 #endif
     llength(lp1) += add;
+    assert((size_t) llength(lp1) <= lp1->l_size);
     set_lforw(lp1, lforw(lp2));
     set_lback(lforw(lp2), lp1);
     dumpuline(lp1);
-    toss_to_undo(lp2);
+    TossToUndo(lp2);
     endofDisplay();
 
     return (TRUE);
@@ -1050,7 +1121,7 @@ kdone(void)
 	kregwidth = 0;
 	kcharpending = -1;
     }
-    kregflag &= ~KNEEDCLEAN;
+    kregflag = (USHORT) (kregflag & ~KNEEDCLEAN);
     kbs[ukb].kbflag = kregflag;
     relist_registers();
 }
@@ -1111,7 +1182,7 @@ kinsert(int c)
 
     if (rc != FALSE) {
 	/* and now insert the character */
-	kbp->kbufp->d_chunk[kbp->kused++] = (char) c;
+	kbp->kbufp->d_chunk[kbp->kused++] = (UCHAR) c;
 	kchars++;
 	if (c == '\n') {
 	    klines++;
@@ -1238,7 +1309,7 @@ usekreg(int f, int n)
 
     if (clexec) {
 	TBUFF *tok = 0;
-	char *name = mac_tokval(&tok);	/* get the next token */
+	char *name = mac_unquotedarg(&tok);	/* get the next token */
 	if (name != 0) {
 	    status = execute(engl2fnc(name), f, n);
 	    tb_free(&tok);
@@ -1319,18 +1390,6 @@ static struct {
     KILLREG killreg;
 } undo_kill;
 
-static int
-kb_size(int ii, KILL * kp)
-{
-    int size = 0;
-    if (kp != 0) {
-	if (kp->d_next != 0)
-	    size = kb_size(ii, kp->d_next);
-	size += KbSize(ii, kp);
-    }
-    return size;
-}
-
 /*
  * Kill a region, saving it into the kill-buffer.  If the buffer that we're
  * modifying is not-undoable, save information so that we can restore the
@@ -1371,37 +1430,37 @@ end_kill(void)
 int
 putbefore(int f, int n)
 {
-    return doput(f, n, FALSE, EXACT);
+    return doput(f, n, FALSE, rgn_EXACT);
 }
 
 int
 putafter(int f, int n)
 {
-    return doput(f, n, TRUE, EXACT);
+    return doput(f, n, TRUE, rgn_EXACT);
 }
 
 int
 lineputbefore(int f, int n)
 {
-    return doput(f, n, FALSE, FULLLINE);
+    return doput(f, n, FALSE, rgn_FULLLINE);
 }
 
 int
 lineputafter(int f, int n)
 {
-    return doput(f, n, TRUE, FULLLINE);
+    return doput(f, n, TRUE, rgn_FULLLINE);
 }
 
 int
 rectputbefore(int f, int n)
 {
-    return doput(f, n, FALSE, RECTANGLE);
+    return doput(f, n, FALSE, rgn_RECTANGLE);
 }
 
 int
 rectputafter(int f, int n)
 {
-    return doput(f, n, TRUE, RECTANGLE);
+    return doput(f, n, TRUE, rgn_RECTANGLE);
 }
 
 static int
@@ -1419,16 +1478,16 @@ doput(int f, int n, int after, REGIONSHAPE shape)
 	return (FALSE);
     }
 
-    if (shape == EXACT) {
+    if (shape == rgn_EXACT) {
 	if ((kbs[ukb].kbflag & (KLINES | KAPPEND)))
-	    shape = FULLLINE;
+	    shape = rgn_FULLLINE;
 	else if (kbs[ukb].kbflag & KRECT)
-	    shape = RECTANGLE;
+	    shape = rgn_RECTANGLE;
 	else
-	    shape = EXACT;
+	    shape = rgn_EXACT;
     }
 
-    if (shape == FULLLINE) {
+    if (shape == rgn_FULLLINE) {
 	if (after && !is_header_line(DOT, curbp))
 	    DOT.l = lforw(DOT.l);
 	DOT.o = 0;
@@ -1449,7 +1508,7 @@ doput(int f, int n, int after, REGIONSHAPE shape)
     if (is_header_line(DOT, curbp)) {
 	DOT.l = lback(DOT.l);
     }
-    if (shape == FULLLINE) {
+    if (shape == rgn_FULLLINE) {
 	(void) firstnonwhite(FALSE, 1);
     }
     ukb = 0;
@@ -1528,14 +1587,14 @@ PutChar(int n, REGIONSHAPE shape)
     /* for each time.... */
     while (n--) {
 	kp = kbs[ukb].kbufh;
-	if (shape == RECTANGLE) {
+	if (shape == rgn_RECTANGLE) {
 	    width = kbs[ukb].kbwidth;
 	    col = getcol(DOT, FALSE);
 	}
 	while (kp != NULL) {
-	    i = KbSize(ukb, kp);
+	    i = (int) KbSize(ukb, kp);
 	    sp = (char *) kp->d_chunk;
-	    if (shape == RECTANGLE) {
+	    if (shape == rgn_RECTANGLE) {
 		while (i-- > 0) {
 		    c = *sp++;
 		    if (width == 0 || c == '\n') {
@@ -1637,7 +1696,7 @@ PutChar(int n, REGIONSHAPE shape)
 		}
 	    }
 	} else {
-	    if (shape == FULLLINE && !suppressnl) {
+	    if (shape == rgn_FULLLINE && !suppressnl) {
 		if (lnewline() != TRUE) {
 		    status = FALSE;
 		    break;
@@ -1689,6 +1748,8 @@ execkreg(int f, int n)
     /* make sure there is something to execute */
     do {
 	jj = index2ukb(j);
+	if (jj < 0 || jj >= NKREGS)
+	    return FALSE;
 	kp = kbs[jj].kbufh;
 	if (kp == NULL)
 	    return TRUE;	/* not an error, just nothing */
@@ -1706,10 +1767,10 @@ execkreg(int f, int n)
 	    tkp = kp;
 	    while (--whichkb != 0)
 		tkp = tkp->d_next;
-	    i = KbSize(jj, tkp);
+	    i = (int) KbSize(jj, tkp);
 	    sp = (char *) tkp->d_chunk + i - 1;
 	    while (i--) {
-		mapungetc((*sp--) | YESREMAP);
+		mapungetc((int) ((UINT) (*sp--) | YESREMAP));
 	    }
 	    kbcount--;
 	}
@@ -1727,7 +1788,7 @@ loadkreg(int f, int n GCC_UNUSED)
     ksetup();
     *respbuf = EOS;
     s = mlreply_no_opts("Load register with: ",
-			respbuf, sizeof(respbuf));
+			respbuf, (UINT) sizeof(respbuf));
     if (s != TRUE)
 	return FALSE;
     if (f)
@@ -1786,7 +1847,7 @@ makereglist(
 	    }
 	    bpadc(' ', REGS_PREFIX - DOT.o);
 	    do {
-		j = KbSize(ii, kp);
+		j = (int) KbSize(ii, kp);
 		p = kp->d_chunk;
 
 		while (j-- > 0) {
@@ -1799,7 +1860,7 @@ makereglist(
 			bputc(c);
 		    } else if (c != '\n') {
 			bputc('^');
-			bputc(toalpha(c));
+			bputc((int) toalpha(c));
 		    }
 		    if (c == '\n') {
 			first = TRUE;
