@@ -22,7 +22,7 @@
  */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/main.c,v 1.641 2008/05/30 19:01:28 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/main.c,v 1.692 2010/08/16 09:49:46 tom Exp $
  */
 
 #define realdef			/* Make global definitions not external */
@@ -37,35 +37,6 @@
 
 #if OPT_LOCALE
 #include	<locale.h>
-
-/*
- * On Linux, the normal/wide ctypes give comparable results in the range 0-255,
- * reflecting the fact that codes 128-255 in Unicode are the "same" as
- * Latin-1.  However, Solaris' wide ctypes give only "non-space" results for
- * 128-255.  Since we're using these functions in vile 9.6 only for the normal
- * ctypes (the narrow 8-bit locale), just use the normal ctype functions.
- */
-#if 0				/* def HAVE_WCTYPE */
-#include	<wctype.h>
-#define sys_iscntrl(n)  iswcntrl(n)
-#define sys_isdigit(n)  iswdigit(n)
-#define sys_islower(n)  iswlower(n)
-#define sys_isprint(n)  iswprint(n)
-#define sys_ispunct(n)  iswpunct(n)
-#define sys_isspace(n)  iswspace(n)
-#define sys_isupper(n)  iswupper(n)
-#define sys_isxdigit(n) iswxdigit(n)
-#else
-#define sys_iscntrl(n)  iscntrl(n)
-#define sys_isdigit(n)  isdigit(n)
-#define sys_islower(n)  islower(n)
-#define sys_isprint(n)  isprint(n)
-#define sys_ispunct(n)  ispunct(n)
-#define sys_isspace(n)  isspace(n)
-#define sys_isupper(n)  isupper(n)
-#define sys_isxdigit(n) isxdigit(n)
-#endif
-
 #endif /* OPT_LOCALE */
 
 #if CC_NEWDOSCC
@@ -80,6 +51,19 @@
 #include <processes.h>
 #endif
 
+#if OPT_FILTER
+#include <filters.h>
+#endif
+
+#if OPT_PERL
+#if defined(DECL_ENVIRON)
+extern char **environ;
+#elif !defined(HAVE_ENVIRON)
+static char **my_environ = 0;
+#define environ my_environ
+#endif
+#endif
+
 extern char *exec_pathname;
 extern const char *const pathname[];	/* startup file path/name array */
 
@@ -92,9 +76,13 @@ static int cmd_mouse_motion(const CMDFUNC * cfp);
 static void get_executable_dir(void);
 static void global_val_init(void);
 static void main_loop(void);
-static void make_startup_file(char *name);
+static void make_startup_file(const char *name);
 static void siginit(int enabled);
 
+#if OPT_MULTIBYTE
+static void pre_init_ctype(void);
+#endif
+
 extern const int nametblsize;
 
 /*--------------------------------------------------------------------------*/
@@ -169,7 +157,7 @@ run_startup_commands(BUFFER *cmds)
     set_rdonly(cmds, cmds->b_fname, MDVIEW);
 
     /* go execute it! */
-    if (dobuf(cmds, 1) != TRUE) {
+    if (dobuf(cmds, 1, -1) != TRUE) {
 	result = FALSE;
     }
     /* if we managed to load a buffer, don't go back to the unnamed buffer */
@@ -236,7 +224,7 @@ default_fill(void)
 static char *
 get_set_locale(const char *value)
 {
-    char *result = setlocale(LC_ALL, value);
+    char *result = setlocale(LC_CTYPE, value);
     if (result != 0)
 	result = strmalloc(result);
     return result;
@@ -255,11 +243,69 @@ set_posix_locale(void)
 #endif
 #endif /* OPT_LOCALE */
 
+/*
+ * For command-line use, run the appropriate syntax filter on the current
+ * buffer, writing the attributed text to the standard output.  This assumes
+ * that if filters are built-in, that the appropriate one is available.
+ */
+static void
+filter_to_stdio(FILE *fp GCC_UNUSED)
+{
+    static char *my_macro;
+    int s;
+
+    TRACE(("SyntaxFilter-only:\n"));
+    TRACE(("\tcurfname:%s\n", curbp->b_fname));
+
+#if OPT_FILTER
+    if (curbp->majr != 0 && flt_lookup(curbp->majr->shortname)) {
+	const CMDFUNC *cmd = &f_operattrdirect;
+	clexec = TRUE;
+	filter_only = TRUE;
+	gotobob(FALSE, 0);
+	havemotion = &f_gotoeob;
+	(void) CMD_U_FUNC(cmd) (FALSE, 0);
+    } else
+#endif
+    {
+	/*
+	 * Rely upon the macro to construct the proper filter path.
+	 */
+	if (my_macro == 0)
+	    my_macro = strmalloc("HighlightFilter");
+
+	/*
+	 * If we used a built-in filter, its output is written as a side effect
+	 * of calling flt_putc, etc.  Otherwise we have to write it ourselves
+	 * now.
+	 */
+	s = docmd(my_macro, TRUE, FALSE, 0);
+	if (s == TRUE) {
+	    int nlines;
+	    B_COUNT nchars;
+	    REGION region;
+
+	    gotoeob(FALSE, 1);
+	    swapmark();
+	    gotobob(FALSE, 1);
+
+	    regionshape = rgn_FULLLINE;
+	    (void) getregion(curbp, &region);
+
+	    ffp = fp;
+	    ffstatus = file_is_pipe;
+
+	    write_region(curbp, &region, TRUE, &nlines, &nchars);
+	}
+    }
+}
+
 /*--------------------------------------------------------------------------*/
 
 int
 MainProgram(int argc, char *argv[])
 {
+    static char dft_vileinit[] = "vileinit.rc";
     int tt_opened;
     BUFFER *bp;
     int carg;			/* current arg to scan */
@@ -284,9 +330,16 @@ MainProgram(int argc, char *argv[])
     *startkey = EOS;
 #endif
 
+    /*
+     * This is needed for a fallback case in locales.
+     */
+#if OPT_MULTIBYTE
+    pre_init_ctype();
+#endif
+
 #if OPT_LOCALE
     {
-	char *env = "";
+	const char *env = "";
 	char *old_locale = get_set_locale("");
 	char *old_encoding = 0;
 
@@ -306,16 +359,10 @@ MainProgram(int argc, char *argv[])
 	    && (((env = getenv("LC_ALL")) != 0 && *env != 0) ||
 		((env = getenv("LC_CTYPE")) != 0 && *env != 0) ||
 		((env = getenv("LANG")) != 0 && *env != 0))) {
-	    char *utf;
 	    char *tmp;
 
 	    TRACE(("Checking for UTF-8 suffix of '%s'\n", env));
-	    if (((utf = strstr(env, ".UTF-8")) != 0
-		 || (utf = strstr(env, ".utf-8")) != 0
-		 || (utf = strstr(env, ".UTF8")) != 0
-		 || (utf = strstr(env, ".utf8")) != 0)
-		&& (tmp = strmalloc(env)) != 0) {
-		tmp[utf - env] = EOS;
+	    if ((tmp = vl_narrowed(env)) != 0) {
 		vl_init_8bit(env, tmp);
 		env = tmp;
 	    } else {
@@ -382,10 +429,15 @@ MainProgram(int argc, char *argv[])
     }
 #endif /* OPT_LOCALE */
 
+#if OPT_PERL
+    perl_init(&argc, &argv, &environ);
+#endif
+
 #if OPT_NAMEBST
     build_namebst(nametbl, 0, nametblsize - 1);
 #endif
-    charinit();			/* character types -- we need these early  */
+    vl_ctype_init(global_g_val(GVAL_PRINT_LOW),
+		  global_g_val(GVAL_PRINT_HIGH));
     global_val_init();		/* global buffer values */
     winit(FALSE);		/* command-line */
 #if !SYS_UNIX
@@ -393,11 +445,12 @@ MainProgram(int argc, char *argv[])
     expand_wild_args(&argc, &argv);
 #endif
     vl_glob_opts = vl_GLOB_ALL;
-    prog_arg = argv[0];		/* this contains our only clue to exec-path */
+    prog_arg = strmalloc(argv[0]);	/* our only clue to exec-path */
 #if SYS_MSDOS || SYS_OS2 || SYS_OS2_EMX || SYS_WINNT
     if (strchr(pathleaf(prog_arg), '.') == 0) {
 	char *t = typeallocn(char, strlen(prog_arg) + 5);
 	lsprintf(t, "%s.exe", prog_arg);
+	free(prog_arg);
 	prog_arg = t;
     }
 #endif
@@ -432,7 +485,15 @@ MainProgram(int argc, char *argv[])
      */
     siginit(TRUE);
 #if OPT_DUMBTERM
-    if (isatty(fileno(stdin))
+    /* try to avoid staircasing if -F option goes directly to terminal */
+    for (carg = 1; carg < argc; ++carg) {
+	if (!strcmp(argv[carg], "-F")) {
+	    filter_only = -TRUE;
+	    break;
+	}
+    }
+    if (!filter_only
+	&& isatty(fileno(stdin))
 	&& isatty(fileno(stdout))) {
 	tt_opened = open_terminal(&dumb_term);
     } else
@@ -505,6 +566,9 @@ MainProgram(int argc, char *argv[])
 		case 'E':
 		    b2printf(opts_bp, "set noview\n");
 		    break;
+		case 'F':
+		    filter_only = -TRUE;
+		    break;
 		case 'g':	/* -g for initial goto */
 		case 'G':
 		    b2printf(opts_bp, "%s goto-line\n", GetArgVal(param));
@@ -516,7 +580,7 @@ MainProgram(int argc, char *argv[])
 
 		case 'i':
 		case 'I':
-		    vileinit = "vileinit.rc";
+		    vileinit = dft_vileinit;
 		    /*
 		     * If the user has no startup file, make a simple
 		     * one that points to this one.
@@ -607,7 +671,7 @@ MainProgram(int argc, char *argv[])
      * in this case).
      */
 #if DISP_TERMCAP || DISP_CURSES || DISP_ANSI
-    if (!isatty(fileno(stdout))) {
+    if (!filter_only && !isatty(fileno(stdout))) {
 	fprintf(stderr, "vile: ex mode is not implemented\n");
 	ExitProgram(BADEXIT);
     }
@@ -639,7 +703,7 @@ MainProgram(int argc, char *argv[])
 #if !SYS_WINNT
 #if !DISP_X11
 #if SYS_UNIX
-	char *tty = 0;
+	const char *tty = 0;
 #else
 	FILE *in;
 	int fd;
@@ -747,25 +811,37 @@ MainProgram(int argc, char *argv[])
     }
 #endif
 
-    if (!tt_opened)
-	siginit(TRUE);
-    (void) open_terminal((TERM *) 0);
-    term.kopen();		/* open the keyboard */
-    term.rev(FALSE);
-
-    /*
-     * The terminal driver sets default colors and $palette during the
-     * open_terminal function.  Save that information as the 'default'
-     * color scheme.  We have to do this before reading .vilerc
-     */
+    if (filter_only) {
 #if OPT_COLOR_SCHEMES
-    init_scheme();
+	init_scheme();
+#endif
+    } else {
+	if (!tt_opened)
+	    siginit(TRUE);
+	(void) open_terminal((TERM *) 0);
+	term.kopen();		/* open the keyboard */
+	term.rev(FALSE);
+
+	/*
+	 * The terminal driver sets default colors and $palette during the
+	 * open_terminal function.  Save that information as the 'default'
+	 * color scheme.  We have to do this before reading .vilerc
+	 */
+#if OPT_COLOR_SCHEMES
+	init_scheme();
 #endif
 
-    if (vtinit() != TRUE)	/* allocate display memory */
-	tidy_exit(BADEXIT);
+	if (vtinit() != TRUE)	/* allocate display memory */
+	    tidy_exit(BADEXIT);
+
+	winit(TRUE);		/* windows */
+
+#if OPT_MENUS
+	if (delay_menus)
+	    vlmenu_load(FALSE, 1);
+#endif
+    }
 
-    winit(TRUE);		/* windows */
     set_global_b_val(VAL_FILL, default_fill());
 
     /* Create an unnamed buffer, so that the initialization-file will have
@@ -874,7 +950,7 @@ MainProgram(int argc, char *argv[])
     purge_msgs();
 #endif
     if (startstat == TRUE)	/* else there's probably an error message */
-	mlforce(msg);
+	mlforce("%s", msg);
 
   begin:
 #if SYS_VMS
@@ -886,48 +962,72 @@ MainProgram(int argc, char *argv[])
 	(void) term.setdescrip(init_descrip);
     }
 #endif
-    (void) update(FALSE);
+    if (filter_only) {
+#ifndef filter_to_stdio
+
+	/*
+	 * Process files in the order they appear on the command-line.
+	 */
+	set_global_g_val(GMDABUFF, FALSE);
+	sortlistbuffers();
+	firstbuffer(FALSE, 0);
+
+	do {
+	    filter_only = TRUE;
+	    filter_to_stdio(stdout);
+	    filter_only = FALSE;	/* ...so nextbuffer doesn't write stdout */
+	    fflush(stdout);
+	} while (nextbuffer(FALSE, 0) == TRUE);
+	ExitProgram(GOODEXIT);
+#else
+	fprintf(stderr,
+		"The -F option is not supported in this configuration\n");
+#endif
+
+    } else {
+	(void) update(FALSE);
 
 #if DISP_NTWIN
-    /* Draw winvile's main window, for the very first time. */
-    winvile_start();
+	/* Draw winvile's main window, for the very first time. */
+	winvile_start();
 #endif
 
 #if OPT_POPUP_MSGS
-    if (global_g_val(GMDPOPUP_MSGS) && (startstat != TRUE)) {
-	bp = bfind(MESSAGES_BufName, BFINVS);
-	bsizes(bp);
-	TRACE(("Checking size of popup messages: %d\n", bp->b_linecount));
-	if (bp->b_linecount > 1) {
-	    int save = global_g_val(GMDPOPUP_MSGS);
-	    set_global_g_val(GMDPOPUP_MSGS, TRUE);
-	    popup_msgs();
-	    tb_init(&mlsave, EOS);
-	    set_global_g_val(GMDPOPUP_MSGS, save);
+	if (global_g_val(GMDPOPUP_MSGS) && (startstat != TRUE)) {
+	    bp = bfind(MESSAGES_BufName, BFINVS);
+	    bsizes(bp);
+	    TRACE(("Checking size of popup messages: %d\n", bp->b_linecount));
+	    if (bp->b_linecount > 1) {
+		int save = global_g_val(GMDPOPUP_MSGS);
+		set_global_g_val(GMDPOPUP_MSGS, TRUE);
+		popup_msgs();
+		tb_init(&mlsave, EOS);
+		set_global_g_val(GMDPOPUP_MSGS, save);
+	    }
 	}
-    }
-    if (global_g_val(GMDPOPUP_MSGS) == -TRUE)
-	set_global_g_val(GMDPOPUP_MSGS, FALSE);
+	if (global_g_val(GMDPOPUP_MSGS) == -TRUE)
+	    set_global_g_val(GMDPOPUP_MSGS, FALSE);
 #endif
 
 #ifdef GMDCD_ON_OPEN
-    /* reset a one-shot, e.g., for winvile's drag/drop code */
-    if (global_g_val(GMDCD_ON_OPEN) < 0) {
-	set_directory_from_file(curbp);
-	set_global_g_val(GMDCD_ON_OPEN, FALSE);
-    }
+	/* reset a one-shot, e.g., for winvile's drag/drop code */
+	if (global_g_val(GMDCD_ON_OPEN) < 0) {
+	    set_directory_from_file(curbp);
+	    set_global_g_val(GMDCD_ON_OPEN, FALSE);
+	}
 #endif
 
-    /* We won't always be able to show messages before the screen is
-     * initialized.  Give it one last chance.
-     */
-    if ((startstat != TRUE) && tb_length(mlsave))
-	mlforce("%.*s", (int) tb_length(mlsave), tb_values(mlsave));
+	/* We won't always be able to show messages before the screen is
+	 * initialized.  Give it one last chance.
+	 */
+	if ((startstat != TRUE) && tb_length(mlsave))
+	    mlforce("%.*s", (int) tb_length(mlsave), tb_values(mlsave));
 
-    /* process commands */
-    main_loop();
+	/* process commands */
+	main_loop();
 
-    /* NOTREACHED */
+	/* NOTREACHED */
+    }
     return BADEXIT;
 }
 
@@ -942,7 +1042,7 @@ main_loop(void)
 
     for_ever {
 
-#if OPT_WORKING && DOALLOC
+#if OPT_WORKING && DOALLOC && (DISP_TERMCAP || DISP_CURSES)
 	assert(allow_working_msg());
 #endif
 	hst_reset();
@@ -1067,7 +1167,7 @@ main_loop(void)
 static void
 get_executable_dir(void)
 {
-#if SYS_UNIX || SYS_VMS
+#if SYS_UNIX || SYS_VMS || SYS_MSDOS || SYS_OS2
     char temp[NFILEN];
     char *s, *t;
 
@@ -1092,9 +1192,16 @@ get_executable_dir(void)
 	 * 't' points to first char after the last ':' or ']' in
 	 * the exe's path.
 	 */
+	free(prog_arg);
 	prog_arg = strmalloc(t);
 	*t = EOS;
 	exec_pathname = strmalloc(lengthen_path(strcpy(temp, s)));
+#ifndef VILE_STARTUP_PATH
+	/* workaround for DJGPP, to add executable's directory to startup path */
+	if (vile_getenv("VILE_STARTUP_PATH") == 0) {
+	    append_to_path_list(&startup_path, exec_pathname);
+	}
+#endif
     }
     free(s);
 #elif SYS_WINNT
@@ -1138,17 +1245,15 @@ tidy_exit(int code)
 #if DISP_NTWIN
     winvile_cleanup();
 #endif
+#if DISP_NTCONS
+    term.close();
+#endif
 #if DISP_X11
     term.close();		/* need this if $xshell left subprocesses */
 #endif
     term.clean(TRUE);
-#if SYS_UNIX
+#ifdef SIGHUP
     setup_handler(SIGHUP, SIG_IGN);
-#endif
-#if NO_LEAKS
-#ifdef HAVE__NC_FREEALL
-    _nc_freeall();
-#endif
 #endif
     ExitProgram(code);
 }
@@ -1214,6 +1319,36 @@ no_memory(const char *s)
 
 #define DFT_IDENTIFIER_EXPR    "\\w\\+"
 
+#if OPT_MULTIBYTE
+/* default value assumes POSIX locale, since it is used before vl_ctyle init */
+#define DFT_LATIN1_EXPR	"\\(\
+aa\\|\
+af\\|\
+br\\|\
+de\\|\
+en\\|\
+es\\|\
+fr\\|\
+fi\\|\
+ga\\|\
+gb\\|\
+gl\\|\
+gv\\|\
+id\\|\
+it\\|\
+kl\\|\
+ms\\|\
+nl\\|\
+nn\\|\
+om\\|\
+pt\\|\
+so\\|\
+sv\\|\
+tl\\)\
+_[A-Za-z]\\+$\
+"
+#endif
+
 			/* where do paragraphs start? */
 #define DFT_PARAGRAPHS  "^\\.[ILPQ]P\\>\\|^\\.P\\>\\|\
 ^\\.LI\\>\\|^\\.[plinb]p\\>\\|^\\.\\?\\s*$"
@@ -1292,12 +1427,17 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(GMDERRORBELLS, TRUE);	/* alarms are noticeable */
 	    setINT(GMDEXPAND_PATH, FALSE);
 	    setINT(GMDIMPLYBUFF, FALSE);	/* imply-buffer */
+	    setINT(GMDINSEXEC, FALSE);	/* allow ^F/^B, etc., to be interpreted during insert mode */
+	    setINT(GMDMAPLONGER, FALSE);	/* favor longer maps */
 	    setINT(GMDMULTIBEEP, TRUE);		/* multiple beeps for multiple motion failures */
+	    setINT(GMDPIN_TAGSTACK, FALSE);	/* tagstack push/pop ops are pinned to curwin */
 	    setINT(GMDREMAP, TRUE);	/* allow remapping by default */
+	    setINT(GMDREMAPFIRST, FALSE);	/* should 1st char of a map be remapped? */
 	    setINT(GMDRONLYRONLY, FALSE);	/* Set readonly-on-readonly */
 	    setINT(GMDRONLYVIEW, FALSE);	/* Set view-on-readonly */
+	    setINT(GMDSAMEBANGS, FALSE);	/* use same "!!" data for ^X-! */
 	    setINT(GMDSMOOTH_SCROLL, FALSE);
-	    setINT(GMDSPACESENT, TRUE);		/* add two spaces after each sentence */
+	    setINT(GMDUNIQ_BUFS, FALSE);	/* only one buffer per file */
 	    setINT(GMDWARNRENAME, TRUE);	/* warn before renaming a buffer */
 	    setINT(GMDWARNREREAD, TRUE);	/* warn before rereading a buffer */
 	    setINT(GMDWARNUNREAD, TRUE);	/* warn if quitting without looking at all buffers */
@@ -1338,6 +1478,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GMDHISTORY
 	    setINT(GMDHISTORY, TRUE);
 #endif
+#ifdef GMDMOUSE
+	    setINT(GMDMOUSE, FALSE);
+#endif
 #ifdef GMDPOPUP_CHOICES
 	    setINT(GMDPOPUP_CHOICES, TRUE);
 #endif
@@ -1347,9 +1490,6 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GMDPOPUP_MSGS
 	    setINT(GMDPOPUP_MSGS, -TRUE);	/* popup-msgs */
 #endif
-#ifdef GVAL_READER_POLICY
-	    setINT(GVAL_READER_POLICY, RP_BOTH);
-#endif
 #ifdef GMDRESOLVE_LINKS
 	    setINT(GMDRESOLVE_LINKS, FALSE);	/* set noresolve-links by default in case we've got NFS problems */
 #endif
@@ -1386,6 +1526,12 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GVAL_CCOLOR
 	    setINT(GVAL_CCOLOR, ENUM_UNKNOWN);	/* cursor color */
 #endif
+#ifdef GVAL_CHK_ACCESS
+	    setINT(GVAL_CHK_ACCESS, FL_CDIR);	/* access-check */
+#endif
+#ifdef GVAL_COLOR_SCHEME
+	    setPAT(GVAL_COLOR_SCHEME, 0);	/* default scheme is 0 */
+#endif
 #ifdef GVAL_CSUFFIXES
 	    setPAT(GVAL_CSUFFIXES, DFT_CSUFFIX);
 #endif
@@ -1410,6 +1556,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GVAL_POPUP_CHOICES
 	    setINT(GVAL_POPUP_CHOICES, POPUP_CHOICES_DELAYED);
 #endif
+#ifdef GVAL_READER_POLICY
+	    setINT(GVAL_READER_POLICY, RP_BOTH);
+#endif
 #ifdef GVAL_REDIRECT_KEYS
 	    setTXT(GVAL_REDIRECT_KEYS,
 		   "F5::S,F10::S,F11::S,F7::F,F5:C:,F9::Y");
@@ -1417,12 +1566,15 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GVAL_RECENT_FILES
 	    setINT(GVAL_RECENT_FILES, 0);
 #endif
-#ifdef GVAL_RECENT_FOLDERS
-	    setINT(GVAL_RECENT_FOLDERS, 0);
+#ifdef GVAL_RECENT_FLDRS
+	    setINT(GVAL_RECENT_FLDRS, 0);
 #endif
 #ifdef GVAL_SCROLLPAUSE
 	    setINT(GVAL_SCROLLPAUSE, 0);
 #endif
+#ifdef GVAL_VIDEO
+	    setINT(GVAL_VIDEO, 0);
+#endif
 #ifdef GVAL_VTFLASH
 	    setINT(GVAL_VTFLASH, VTFLASH_OFF);	/* hardwired flash off */
 #endif
@@ -1448,6 +1600,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(MDREADONLY, FALSE);	/* readonly */
 	    setINT(MDSHOWMAT, FALSE);	/* show-match */
 	    setINT(MDSHOWMODE, TRUE);	/* show-mode */
+	    setINT(MDSPACESENT, TRUE);	/* add two spaces after each sentence */
 	    setINT(MDSWRAP, TRUE);	/* scan wrap */
 	    setINT(MDTABINSERT, TRUE);	/* allow tab insertion */
 	    setINT(MDTAGSRELTIV, FALSE);	/* path relative tag lookups */
@@ -1461,7 +1614,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(VAL_ASAVECNT, 256);	/* autosave count */
 	    setINT(VAL_FILL, default_fill());	/* column for paragraph reformat */
 	    setINT(VAL_PERCENT_CRLF, 50);	/* threshold for crlf conversion */
-	    setINT(VAL_RECORD_SEP, RS_DEFAULT);
+	    setINT(VAL_RECORD_SEP, RS_SYS_DEFAULT);
 	    setINT(VAL_SHOW_FORMAT, SF_FOREIGN);
 	    setINT(VAL_SWIDTH, 8);	/* shiftwidth */
 	    setINT(VAL_TAB, 8);	/* tab stop */
@@ -1512,6 +1665,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef MDMODELINE
 	    setINT(MDMODELINE, FALSE);
 #endif
+#ifdef MDREUSE_POS
+	    setINT(MDREUSE_POS, FALSE);		/* reuse position */
+#endif
 #ifdef MDTAGIGNORECASE
 	    setINT(MDTAGIGNORECASE, FALSE);	/* ignore case in tags */
 #endif
@@ -1548,6 +1704,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef VAL_MODELINES
 	    setINT(VAL_MODELINES, 5);
 #endif
+#ifdef VAL_PERCENT_AUTOCOLOR
+	    setINT(VAL_PERCENT_AUTOCOLOR, 50);
+#endif
 #ifdef VAL_PERCENT_UTF8
 	    setINT(VAL_PERCENT_UTF8, 90);
 #endif
@@ -1633,19 +1792,19 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
  Char %{$curchar}d of %{$bchars}d\
  (%P%%) char is 0x%{$char}x or 0%{$char}o"
 
-static char *
+static const char *
 default_help_file(void)
 {
-    char *result;
+    const char *result;
     if ((result = vile_getenv("VILE_HELP_FILE")) == 0)
 	result = DFT_HELP_FILE;
     return result;
 }
 
-static char *
+static const char *
 default_libdir_path(void)
 {
-    char *result;
+    const char *result;
     if ((result = vile_getenv("VILE_LIBDIR_PATH")) == 0)
 	result = DFT_LIBDIR_PATH;
     return result;
@@ -1672,10 +1831,10 @@ default_menu_file(void)
 }
 #endif
 
-static char *
+static const char *
 default_startup_file(void)
 {
-    char *result;
+    const char *result;
     if ((result = vile_getenv("VILE_STARTUP_FILE")) == 0)
 	result = DFT_STARTUP_FILE;
     return result;
@@ -1695,20 +1854,16 @@ default_startup_path(void)
 	append_to_path_list(&result, "/usr/bin");
 	append_to_path_list(&result, "/bin");
 	append_to_path_list(&result, "/");
-#else
-#if SYS_VMS
+#elif SYS_VMS
 	append_to_path_list(&result, "sys$login");
 	append_to_path_list(&result, "sys$sysdevice:[vmstools]");
 	append_to_path_list(&result, "sys$library");
-#else /* SYS_UNIX */
-#if defined(VILE_STARTUP_PATH)
+#elif defined(VILE_STARTUP_PATH)
 	append_to_path_list(&result, VILE_STARTUP_PATH);
 #else
 	append_to_path_list(&result, "/usr/local/lib");
 	append_to_path_list(&result, "/usr/local");
 	append_to_path_list(&result, "/usr/lib");
-#endif /* VILE_STARTUP_PATH */
-#endif /* SYS_VMS... */
 #endif /* SYS_MSDOS... */
     }
 #ifdef HELP_LOC
@@ -1717,6 +1872,14 @@ default_startup_path(void)
     return result;
 }
 
+#if OPT_MULTIBYTE
+static void
+pre_init_ctype(void)
+{
+    tb_scopy(&latin1_expr, DFT_LATIN1_EXPR);
+}
+#endif
+
 #if OPT_EVAL
 /*
  * Returns a string representing the default/initial value of the given
@@ -1734,9 +1897,17 @@ init_state_value(int which)
 	value = DFT_FILENAME_EXPR;
 	break;
 #endif
+    case VAR_FONT:
+	value = default_font;
+	break;
     case VAR_HELPFILE:
 	value = default_help_file();
 	break;
+#if OPT_MULTIBYTE
+    case VAR_LATIN1_EXPR:
+	value = DFT_LATIN1_EXPR;
+	break;
+#endif
     case VAR_LIBDIR_PATH:
 	value = default_libdir_path();
 	break;
@@ -1811,58 +1982,19 @@ copy_kbindtbl(BINDINGS * dst)
 }
 #endif
 
-#if OPT_FILTER && defined(WIN32)
+#if OPT_FILTER && defined(WIN32) && !SYS_MINGW
 extern void flt_array(void);
 #endif
 
+/*
+ * Initialize compiled-in state variable values.
+ */
 static void
-global_val_init(void)
+state_val_init(void)
 {
+#if OPT_EVAL
     int i;
 
-    TRACE((T_CALLED "global_val_init()\n"));
-#if OPT_FILTER && defined(WIN32)
-    flt_array();
-#endif
-    /* set up so the global value pointers point at the global
-       values.  we never actually use the global pointers
-       directly, but when buffers get a copy of the
-       global_b_values structure, the pointers will still point
-       back at the global values, which is what we want */
-    for (i = 0; i <= NUM_G_VALUES; i++) {
-	make_local_val(global_g_values.gv, i);
-	init_mode_value(&global_g_values.gv[i], UNI_MODE, i);
-    }
-
-    for (i = 0; i <= NUM_B_VALUES; i++) {
-	make_local_val(global_b_values.bv, i);
-	init_mode_value(&global_b_values.bv[i], BUF_MODE, i);
-    }
-
-    for (i = 0; i <= NUM_W_VALUES; i++) {
-	make_local_val(global_w_values.wv, i);
-	init_mode_value(&global_w_values.wv[i], WIN_MODE, i);
-    }
-
-#if OPT_MAJORMODE
-    /*
-     * Built-in majormodes (see 'modetbl')
-     */
-    alloc_mode("c", TRUE);
-    alloc_mode("vile", TRUE);
-
-    set_submode_val("c", VAL_SWIDTH, 8);	/* C file shiftwidth */
-    set_submode_val("c", VAL_TAB, 8);	/* C file tab stop */
-    set_submode_val("c", MDCINDENT, TRUE);	/* C-style indent */
-    set_submode_txt("c", VAL_CINDENT_CHARS, DFT_CINDENT_CHARS);		/* C-style indent flags */
-
-    set_majormode_rexp("c", MVAL_MODE_SUFFIXES, DFT_CSUFFIX);
-#ifdef MVAL_MODE_WILDCARD
-    set_majormode_rexp("c", MVAL_MODE_WILDCARD, DFT_WILDCARD);
-#endif
-#endif /* OPT_MAJORMODE */
-
-#if OPT_EVAL
     /*
      * Note that not all state variables can be initialized explicitly, many
      * are either read-only, or have too many side-effects to set blindly.
@@ -1885,13 +2017,65 @@ global_val_init(void)
 #if OPT_MLFORMAT
     modeline_format = strmalloc(DFT_MLFORMAT);
 #endif
+#if OPT_MULTIBYTE
+    tb_scopy(&latin1_expr, DFT_LATIN1_EXPR);
+#endif
 #if OPT_POSFORMAT
     position_format = strmalloc(DFT_POSFORMAT);
 #endif
 #if OPT_FINDERR
-    filename_expr = strmalloc(DFT_FILENAME_EXPR);
+    tb_scopy(&filename_expr, DFT_FILENAME_EXPR);
 #endif
 #endif
+}
+
+static void
+global_val_init(void)
+{
+    int i;
+
+    TRACE((T_CALLED "global_val_init()\n"));
+#if OPT_FILTER && defined(WIN32) && !SYS_MINGW
+    flt_array();
+#endif
+    /* set up so the global value pointers point at the global
+       values.  we never actually use the global pointers
+       directly, but when buffers get a copy of the
+       global_b_values structure, the pointers will still point
+       back at the global values, which is what we want */
+    for (i = 0; i <= NUM_G_VALUES; i++) {
+	make_local_val(global_g_values.gv, i);
+	init_mode_value(&global_g_values.gv[i], UNI_MODE, i);
+    }
+
+    for (i = 0; i <= NUM_B_VALUES; i++) {
+	make_local_val(global_b_values.bv, i);
+	init_mode_value(&global_b_values.bv[i], BUF_MODE, i);
+    }
+
+    for (i = 0; i <= NUM_W_VALUES; i++) {
+	make_local_val(global_w_values.wv, i);
+	init_mode_value(&global_w_values.wv[i], WIN_MODE, i);
+    }
+    state_val_init();
+
+#if OPT_MAJORMODE
+    /*
+     * Built-in majormodes (see 'modetbl')
+     */
+    alloc_mode("c", TRUE);
+    alloc_mode("vile", TRUE);
+
+    set_submode_val("c", VAL_SWIDTH, 8);	/* C file shiftwidth */
+    set_submode_val("c", VAL_TAB, 8);	/* C file tab stop */
+    set_submode_val("c", MDCINDENT, TRUE);	/* C-style indent */
+    set_submode_txt("c", VAL_CINDENT_CHARS, DFT_CINDENT_CHARS);		/* C-style indent flags */
+
+    set_majormode_rexp("c", MVAL_MODE_SUFFIXES, DFT_CSUFFIX);
+#ifdef MVAL_MODE_WILDCARD
+    set_majormode_rexp("c", MVAL_MODE_WILDCARD, DFT_WILDCARD);
+#endif
+#endif /* OPT_MAJORMODE */
 
     /*
      * Initialize the "normal" bindings for insert and command mode to
@@ -2029,12 +2213,16 @@ siginit(int enabled)
 #endif
 	{ SIGSEGV, imdying },
 	{ SIGTERM, imdying },
+#ifdef SIGQUIT
 #ifdef VILE_DEBUG
 	{ SIGQUIT, imdying },
 #else
 	{ SIGQUIT, SIG_IGN },
 #endif
+#endif
+#ifdef SIGPIPE
 	{ SIGPIPE, SIG_IGN },
+#endif
 #if defined(SIGWINCH) && ! DISP_X11
 	{ SIGWINCH, sizesignal },
 #endif
@@ -2315,77 +2503,6 @@ quit(int f, int n GCC_UNUSED)
     siguninit();
 #if OPT_WORKING
     setup_handler(SIGALRM, SIG_IGN);
-#endif
-#if NO_LEAKS
-    {
-	beginDisplay();		/* ...this may take a while... */
-
-	/* free all of the global data structures */
-	onel_leaks();
-	path_leaks();
-	kbs_leaks();
-	bind_leaks();
-	map_leaks();
-	tags_leaks();
-	wp_leaks();
-	bp_leaks();
-#if !SMALLER
-	ev_leaks();
-#endif
-	mode_leaks();
-	vars_leaks();
-	fileio_leaks();
-#if DISP_X11
-	x11_leaks();
-#endif
-
-	free_local_vals(g_valnames, global_g_values.gv, global_g_values.gv);
-	free_local_vals(b_valnames, global_b_values.bv, global_b_values.bv);
-	free_local_vals(w_valnames, global_w_values.wv, global_w_values.wv);
-
-	FreeAndNull(libdir_path);
-	FreeAndNull(startup_path);
-	FreeAndNull(gregexp);
-	tb_free(&tb_matched_pat);
-#if OPT_LOCALE
-	FreeAndNull(vl_real_enc.locale);
-	FreeAndNull(vl_real_enc.encoding);
-#endif /* OPT_LOCALE */
-#if OPT_MLFORMAT
-	FreeAndNull(modeline_format);
-#endif
-#if OPT_POSFORMAT
-	FreeAndNull(position_format);
-#endif
-	FreeAndNull(helpfile);
-	FreeAndNull(startup_file);
-#if SYS_UNIX
-	if (exec_pathname != 0
-	    && strcmp(exec_pathname, "."))
-	    FreeAndNull(exec_pathname);
-#endif
-#if OPT_CURTOKENS
-	free_regexval(buf_fname_expr.v.r);
-#endif
-#if OPT_FILTER
-	flt_leaks();
-	filters_leaks();
-#endif
-	/* do these last, e.g., for tb_matched_pat */
-	itb_leaks();
-	tb_leaks();
-
-	term.close();
-#if OPT_LOCALE
-	eightbit_leaks();
-#endif
-	vt_leaks();
-
-	/* whatever is left over must be a leak */
-	show_alloc();
-	show_elapsed();
-	trace_leaks();
-    }
 #endif
     tidy_exit(GOODEXIT);
     /* NOTREACHED */
@@ -2527,268 +2644,6 @@ reptc_func(int f GCC_UNUSED, int n GCC_UNUSED)
 
 /*----------------------------------------------------------------------------*/
 
-/* initialize our version of the "chartypes" stuff normally in ctypes.h */
-/* also called later, if charset-affecting modes change, for instance */
-void
-charinit(void)
-{
-    int c;
-
-    TRACE((T_CALLED "charinit() lo=%d, hi=%d\n",
-	   global_g_val(GVAL_PRINT_LOW),
-	   global_g_val(GVAL_PRINT_HIGH)));
-
-    /* If we're using the locale functions, set our flags based on its
-     * tables.  Note that just because you have 'setlocale()' doesn't mean
-     * that the tables are present or correct.  But this is a start.
-     *
-     * NOTE:  Solaris8 and some versions of M$ incorrectly classify tab as a
-     * printable character (ANSI C says control characters are not printable).
-     * Ignore that (the former fixes it in Solaris9).
-     */
-#if OPT_LOCALE
-    TRACE(("wide_locale:%s\n", NonNull(vl_wide_enc.locale)));
-    TRACE(("narrow_locale:%s\n", NonNull(vl_narrow_enc.locale)));
-    for (c = 0; c < N_chars; c++) {
-	if (okCTYPE2(vl_narrow_enc)) {
-	    vlCTYPE(c) = 0;
-	    if (sys_iscntrl(c))
-		vlCTYPE(c) |= vl_cntrl;
-	    if (sys_isdigit(c))
-		vlCTYPE(c) |= vl_digit;
-	    if (sys_islower(c))
-		vlCTYPE(c) |= vl_lower;
-	    if (sys_isprint(c) && c != '\t')
-		vlCTYPE(c) |= vl_print;
-	    if (sys_ispunct(c))
-		vlCTYPE(c) |= vl_punct;
-	    if (sys_isspace(c))
-		vlCTYPE(c) |= vl_space;
-	    if (sys_isupper(c))
-		vlCTYPE(c) |= vl_upper;
-#ifdef vl_xdigit
-	    if (sys_isxdigit(c))
-		vlCTYPE(c) |= vl_xdigit;
-#endif
-	    vl_uppercase[c] = (char) toupper(c);
-	    vl_lowercase[c] = (char) tolower(c);
-	} else {
-	    /* fallback to built-in character tables */
-	    if (okCTYPE2(vl_wide_enc)) {
-		vlCTYPE(c) = vl_ctype_latin1[c];
-	    } else {
-		vlCTYPE(c) = vl_ctype_ascii[c];
-	    }
-	    vl_uppercase[c] = (char) c;
-	    vl_lowercase[c] = (char) c;
-	    if (isAlpha(c)) {
-		if (isUpper(c)) {
-		    vl_lowercase[c] = (char) (c ^ DIFCASE);
-		} else {
-		    vl_uppercase[c] = (char) (c ^ DIFCASE);
-		}
-	    }
-	}
-    }
-#else /* ! OPT_LOCALE */
-    (void) memset((char *) vl_chartypes_, 0, sizeof(vl_chartypes_));
-
-    /* control characters */
-    for (c = 0; c < ' '; c++)
-	vlCTYPE(c) |= vl_cntrl;
-    vlCTYPE(127) |= vl_cntrl;
-
-    /* lowercase */
-    for (c = 'a'; c <= 'z'; c++)
-	vlCTYPE(c) |= vl_lower;
-#if OPT_ISO_8859
-    for (c = 0xc0; c <= 0xd6; c++)
-	vlCTYPE(c) |= vl_lower;
-    for (c = 0xd8; c <= 0xde; c++)
-	vlCTYPE(c) |= vl_lower;
-#endif
-    /* uppercase */
-    for (c = 'A'; c <= 'Z'; c++)
-	vlCTYPE(c) |= vl_upper;
-#if OPT_ISO_8859
-    for (c = 0xdf; c <= 0xf6; c++)
-	vlCTYPE(c) |= vl_upper;
-    for (c = 0xf8; c <= 0xff; c++)
-	vlCTYPE(c) |= vl_upper;
-#endif
-
-    /*
-     * If you want to do this properly, compile-in locale support.
-     */
-    for (c = 0; c < N_chars; c++) {
-	vl_uppercase[c] = (char) c;
-	vl_lowercase[c] = (char) c;
-	if (isAlpha(c)) {
-	    if (isUpper(c)) {
-		vl_lowercase[c] = (char) (c ^ DIFCASE);
-	    } else {
-		vl_uppercase[c] = (char) (c ^ DIFCASE);
-	    }
-	}
-    }
-
-    /* digits */
-    for (c = '0'; c <= '9'; c++)
-	vlCTYPE(c) |= vl_digit;
-#ifdef vl_xdigit
-    /* hex digits */
-    for (c = '0'; c <= '9'; c++)
-	vlCTYPE(c) |= vl_xdigit;
-    for (c = 'a'; c <= 'f'; c++)
-	vlCTYPE(c) |= vl_xdigit;
-    for (c = 'A'; c <= 'F'; c++)
-	vlCTYPE(c) |= vl_xdigit;
-#endif
-
-    /* punctuation */
-    for (c = '!'; c <= '/'; c++)
-	vlCTYPE(c) |= vl_punct;
-    for (c = ':'; c <= '@'; c++)
-	vlCTYPE(c) |= vl_punct;
-    for (c = '['; c <= '`'; c++)
-	vlCTYPE(c) |= vl_punct;
-    for (c = L_CURLY; c <= '~'; c++)
-	vlCTYPE(c) |= vl_punct;
-#if OPT_ISO_8859
-    for (c = 0xa1; c <= 0xbf; c++)
-	vlCTYPE(c) |= vl_punct;
-#endif
-
-    /* printable */
-    for (c = ' '; c <= '~'; c++)
-	vlCTYPE(c) |= vl_print;
-
-    /* whitespace */
-    vlCTYPE(' ') |= vl_space;
-#if OPT_ISO_8859
-    vlCTYPE(0xa0) |= vl_space;
-#endif
-    vlCTYPE('\t') |= vl_space;
-    vlCTYPE('\r') |= vl_space;
-    vlCTYPE('\n') |= vl_space;
-    vlCTYPE('\f') |= vl_space;
-
-#endif /* OPT_LOCALE */
-
-    /* legal in pathnames */
-    vlCTYPE('.') |= vl_pathn;
-    vlCTYPE('_') |= vl_pathn;
-    vlCTYPE('~') |= vl_pathn;
-    vlCTYPE('-') |= vl_pathn;
-    vlCTYPE('/') |= vl_pathn;
-
-    /* legal in "identifiers" */
-    vlCTYPE('_') |= vl_ident | vl_qident;
-    vlCTYPE(':') |= vl_qident;
-#if SYS_VMS
-    vlCTYPE('$') |= vl_ident | vl_qident;
-#endif
-
-    c = global_g_val(GVAL_PRINT_LOW);
-
-    /*
-     * Guard against setting printing-high before printing-low while we have a
-     * buffer which may be repainted and possibly trashing the display.
-     */
-    if (c == 0
-	&& global_g_val(GVAL_PRINT_HIGH) >= 254)
-	c = 160;
-
-    if (c < HIGHBIT)
-	c = HIGHBIT;
-    TRACE(("Forcing printable for [%d..min(%d,%d)]\n",
-	   c, global_g_val(GVAL_PRINT_HIGH) - 1, N_chars - 1));
-    while (c <= global_g_val(GVAL_PRINT_HIGH) && c < N_chars)
-	vlCTYPE(c++) |= vl_print;
-
-#if DISP_X11
-    for (c = 0; c < N_chars; c++) {
-	if (isPrint(c) && !gui_isprint(c)) {
-	    vlCTYPE(c) &= ~vl_print;
-	}
-    }
-#endif
-    /* backspacers: ^H, rubout */
-    vlCTYPE('\b') |= vl_bspace;
-    vlCTYPE(127) |= vl_bspace;
-
-    /* wildcard chars for most shells */
-    vlCTYPE('*') |= vl_wild;
-    vlCTYPE('?') |= vl_wild;
-#if !OPT_VMS_PATH
-#if SYS_UNIX
-    vlCTYPE('~') |= vl_wild;
-#endif
-    vlCTYPE(L_BLOCK) |= vl_wild;
-    vlCTYPE(R_BLOCK) |= vl_wild;
-    vlCTYPE(L_CURLY) |= vl_wild;
-    vlCTYPE(R_CURLY) |= vl_wild;
-    vlCTYPE('$') |= vl_wild;
-    vlCTYPE('`') |= vl_wild;
-#endif
-
-    /* ex mode line specifiers */
-    vlCTYPE(',') |= vl_linespec;
-    vlCTYPE('%') |= vl_linespec;
-    vlCTYPE('-') |= vl_linespec;
-    vlCTYPE('+') |= vl_linespec;
-    vlCTYPE(';') |= vl_linespec;
-    vlCTYPE('.') |= vl_linespec;
-    vlCTYPE('$') |= vl_linespec;
-    vlCTYPE('\'') |= vl_linespec;
-
-    /* fences */
-    vlCTYPE(L_CURLY) |= vl_fence;
-    vlCTYPE(R_CURLY) |= vl_fence;
-    vlCTYPE(L_PAREN) |= vl_fence;
-    vlCTYPE(R_PAREN) |= vl_fence;
-    vlCTYPE(L_BLOCK) |= vl_fence;
-    vlCTYPE(R_BLOCK) |= vl_fence;
-
-#if OPT_VMS_PATH
-    vlCTYPE(L_BLOCK) |= vl_pathn;
-    vlCTYPE(R_BLOCK) |= vl_pathn;
-    vlCTYPE(L_ANGLE) |= vl_pathn;
-    vlCTYPE(R_ANGLE) |= vl_pathn;
-    vlCTYPE('$') |= vl_pathn;
-    vlCTYPE(':') |= vl_pathn;
-    vlCTYPE(';') |= vl_pathn;
-#endif
-
-#if OPT_MSDOS_PATH
-    vlCTYPE(BACKSLASH) |= vl_pathn;
-    vlCTYPE(':') |= vl_pathn;
-#endif
-
-#if OPT_WIDE_CTYPES
-    /* scratch-buffer-names (usually superset of vl_pathn) */
-    vlCTYPE(SCRTCH_LEFT[0]) |= vl_scrtch;
-    vlCTYPE(SCRTCH_RIGHT[0]) |= vl_scrtch;
-    vlCTYPE(' ') |= vl_scrtch;	/* ...to handle "[Buffer List]" */
-#endif
-
-    for (c = 0; c < N_chars; c++) {
-	if (!(isSpace(c)))
-	    vlCTYPE(c) |= vl_nonspace;
-	if (isDigit(c))
-	    vlCTYPE(c) |= vl_linespec;
-	if (isAlpha(c) || isDigit(c))
-	    vlCTYPE(c) |= vl_ident | vl_pathn | vl_qident;
-#if OPT_WIDE_CTYPES
-	if (isSpace(c) || isPrint(c))
-	    vlCTYPE(c) |= vl_shpipe;
-	if (ispath(c))
-	    vlCTYPE(c) |= vl_scrtch;
-#endif
-    }
-    returnVoid();
-}
-
 #if OPT_HEAPSIZE
 
 /* track overall heap usage */
@@ -3067,7 +2922,7 @@ cmd_mouse_motion(const CMDFUNC * cfp)
 }
 
 static void
-make_startup_file(char *name)
+make_startup_file(const char *name)
 {
     FILE *fp;
     char temp[NFILEN];
@@ -3114,7 +2969,7 @@ int
 valid_window(WINDOW *test)
 {
     beginDisplay();
-    if (test != NULL && test != wminip) {
+    if (test != NULL && test != wminip && test != wnullp) {
 	WINDOW *wp;
 	int valid = FALSE;
 
@@ -3172,6 +3027,13 @@ valid_line_wp(LINE *test, WINDOW *wp)
 }
 #endif
 
+void
+exit_program(int code)
+{
+    free_all_leaks();
+    exit(code);
+}
+
 #ifdef VILE_ERROR_ABORT
 #undef ExitProgram		/* in case it is oleauto_exit() */
 
@@ -3185,6 +3047,91 @@ ExitProgram(int code)
 	siginit(FALSE);
 	abort();
     }
-    exit(code);
+    exit_program(code);
 }
 #endif
+
+#if NO_LEAKS
+void
+free_all_leaks(void)
+{
+    beginDisplay();		/* ...this may take a while... */
+
+    TRACE((T_CALLED "free_all_leaks\n"));
+
+    /* free all of the global data structures */
+    onel_leaks();
+    path_leaks();
+    kbs_leaks();
+    bind_leaks();
+    map_leaks();
+    tags_leaks();
+    wp_leaks();
+    bp_leaks();
+#if !SMALLER
+    ev_leaks();
+#endif
+    mode_leaks();
+    vars_leaks();
+    fileio_leaks();
+#if DISP_X11
+    x11_leaks();
+#endif
+
+    free_local_vals(g_valnames, global_g_values.gv, global_g_values.gv);
+    free_local_vals(b_valnames, global_b_values.bv, global_b_values.bv);
+    free_local_vals(w_valnames, global_w_values.wv, global_w_values.wv);
+
+    FreeAndNull(prog_arg);
+    FreeAndNull(libdir_path);
+    FreeAndNull(startup_path);
+    FreeAndNull(gregexp);
+    tb_free(&tb_matched_pat);
+#if OPT_LOCALE
+    FreeAndNull(vl_real_enc.locale);
+    FreeAndNull(vl_real_enc.encoding);
+#endif /* OPT_LOCALE */
+#if OPT_MLFORMAT
+    FreeAndNull(modeline_format);
+#endif
+#if OPT_POSFORMAT
+    FreeAndNull(position_format);
+#endif
+    FreeAndNull(helpfile);
+    FreeAndNull(startup_file);
+#if SYS_UNIX
+    if (exec_pathname != 0
+	&& strcmp(exec_pathname, "."))
+	FreeAndNull(exec_pathname);
+#endif
+#if OPT_CURTOKENS
+    free_regexval(buf_fname_expr.v.r);
+#endif
+#if OPT_FILTER
+    flt_leaks();
+    filters_leaks();
+#endif
+    vl_ctype_leaks();
+    /* do these last, e.g., for tb_matched_pat */
+    itb_leaks();
+    tb_leaks();
+
+    /* only do this for drivers which run inside a terminal */
+#if DISP_TERMCAP || DISP_CURSES
+    term.close();
+#endif
+
+#if OPT_LOCALE
+    eightbit_leaks();
+#endif
+    vt_leaks();
+
+    /* whatever is left over must be a leak */
+    show_alloc();
+    show_elapsed();
+    trace_leaks();
+#ifdef HAVE__NC_FREEALL
+    _nc_freeall();
+#endif
+}
+#endif /* NO_LEAKS */
