@@ -3,7 +3,7 @@
  *	Original interface by Otto Lind, 6/3/93
  *	Additional map and map! support by Kevin Buettner, 9/17/94
  *
- * $Header: /usr/build/vile/vile/RCS/map.c,v 1.112 2007/08/31 22:53:03 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/map.c,v 1.121 2010/09/06 18:29:22 tom Exp $
  *
  */
 
@@ -72,6 +72,8 @@ MAPREC {
      */
 };
 
+#define STRIP_REMAPFLAGS(c) ((c) & (int) ~REMAPFLAGS)
+
 #define MAPF_SYSTIMER	0x01
 #define MAPF_USERTIMER	0x02
 #define MAPF_TIMERS	0x03
@@ -160,10 +162,11 @@ makemaplist(int dummy GCC_UNUSED, void *mapp)
 			remapnote = "   ";
 		    }
 		    bprintf("%s ", remapnote);
-		    for (i = 0; i < depth; i++)
-			bputc(the_lhs_string[i]);
+		    bputsn_xcolor(the_lhs_string, (int) depth, XCOLOR_STRING);
 		}
-		bprintf("\t%s\n", mapstr);
+		bputc('\t');
+		bputsn_xcolor(mapstr, -1, XCOLOR_STRING);
+		bputc('\n');
 	    }
 	    mp = mp->dlink;
 	} else if (depth != 0)
@@ -227,10 +230,11 @@ addtomap(struct maprec **mpp,
 
     if (ks != 0 && kslen != 0) {
 
+	TRACE2(("addtomap %s\n", visible_buff(ks, kslen, FALSE)));
 	while (*mpp && kslen) {
 	    mp = *mpp;
 	    mp->flags |= flags;
-	    if (char2int(*ks) == mp->ch) {
+	    if (CharOf(*ks) == mp->ch) {
 		mpp = &mp->dlink;
 		ks++;
 		kslen--;
@@ -246,7 +250,8 @@ addtomap(struct maprec **mpp,
 	    }
 	    *mpp = mp;
 	    mp->dlink = mp->flink = NULL;
-	    mp->ch = char2int(*ks++);
+	    TRACE2(("...adding %#x\n", CharOf(*ks)));
+	    mp->ch = CharOf(*ks++);
 	    mp->srv = NULL;
 	    mp->flags = flags;
 	    mp->irv = -1;
@@ -261,6 +266,7 @@ addtomap(struct maprec **mpp,
 		if (mp->srv)
 		    free(mp->srv);
 		mp->srv = strmalloc(srv);
+		TRACE2(("...addtomap %s\n", visible_buff(srv, strlen(srv), FALSE)));
 	    }
 	    mp->flags = flags;
 	}
@@ -287,7 +293,7 @@ delfrommap(struct maprec **mpp, const char *ks, int length)
 	while (*mpp && n < length) {
 	    if (pass)
 		mstk[depth] = mpp;
-	    if ((*mpp)->ch == char2int(ks[n])) {
+	    if ((*mpp)->ch == CharOf(ks[n])) {
 		mpp = &(*mpp)->dlink;
 		n++;
 		depth++;
@@ -298,7 +304,7 @@ delfrommap(struct maprec **mpp, const char *ks, int length)
 	if (n < length)
 	    return FALSE;	/* not in map */
 	if (!pass) {
-	    mstk = typecallocn(struct maprec **, depth + 1);
+	    mstk = typecallocn(struct maprec **, (size_t) depth + 1);
 	    if (mstk == 0)
 		return no_memory("delfrommap");
 	}
@@ -368,8 +374,10 @@ maplookup(int c,
     itb_append(&unmatched, c);
     count++;
 
+    TRACE2(("maplookup unmatched:%s\n", itb_visible(unmatched)));
     matchedcnt = 0;
     while (mp != 0) {
+	TRACE2(("... c=%#x, mp->ch=%#x\n", c, mp->ch));
 	if (c == mp->ch) {
 	    if (mp->irv != -1 || mp->srv != NULL) {
 		rmp = mp;
@@ -439,7 +447,7 @@ maplookup(int c,
 	    if ((c = (*get) ()) < 0)
 		break;
 
-	    itb_append(&unmatched, c & ~REMAPFLAGS);
+	    itb_append(&unmatched, STRIP_REMAPFLAGS(c));
 	    count++;
 
 	} else
@@ -451,8 +459,10 @@ maplookup(int c,
 	no_memory("maplookup");
     } else if (had_start && (rmp != 0)) {
 	/* unget the unmatched suffix */
-	while (suffix && (count > 0))
+	while (suffix && (count > 0)) {
 	    (void) itb_append(outp, itb_values(unmatched)[--count]);
+	    TRACE2(("...1 ungetting %#x\n", itb_values(unmatched)[count]));
+	}
 	/* unget the mapping and elide correct number of recorded chars */
 	if (rmp->srv) {
 	    UINT remapflag;
@@ -463,17 +473,23 @@ maplookup(int c,
 		remapflag = NOREMAP;
 	    else
 		remapflag = 0;
-	    while (cp > rmp->srv)
-		(void) itb_append(outp, char2int(*--cp) | remapflag);
+	    while (cp > rmp->srv) {
+		(void) itb_append(outp, CharOf(*--cp) | (int) remapflag);
+		TRACE2(("...2 ungetting %#x|%#x\n", CharOf(*cp), remapflag));
+	    }
 	} else {
 	    (void) itb_append(outp, rmp->irv);
+	    TRACE2(("...3 ungetting %#x\n", rmp->irv));
 	}
     } else {			/* didn't find a match */
-	while (count > 0)
+	while (count > 0) {
 	    (void) itb_append(outp, itb_values(unmatched)[--count]);
+	    TRACE2(("...4 ungetting %#x\n", itb_values(unmatched)[count]));
+	}
 	matchedcnt = 0;
     }
     itb_free(&unmatched);
+    TRACE2(("...maplookup %d:%s\n", matchedcnt, itb_visible(*outp)));
     return matchedcnt;
 }
 
@@ -503,32 +519,32 @@ map_common(struct maprec **mpp, const char *bufname, UINT remapflag)
 
 #if OPT_SHOW_MAPS
     if (end_named_cmd()) {
-	return show_mapped_chars(bufname);
-    }
+	status = show_mapped_chars(bufname);
+    } else
 #endif
-    tb_scopy(&kbuf, "");
-    status = kbd_reply("change this string: ", &kbuf, eol_history,
-		       ' ', KBD_NOMAP | KBD_NOEVAL, no_completion);
-    if (status != TRUE)
-	return status;
+    {
+	tb_scopy(&kbuf, "");
+	status = kbd_reply("change this string: ", &kbuf, eol_history,
+			   ' ', KBD_NOMAP | KBD_NOEVAL, no_completion);
+	if (status == TRUE) {
+	    hst_glue(' ');
+	    tb_scopy(&val, "");
+	    if (!clexec) {
+		status = kbd_reply("to this new string: ", &val, eol_history,
+				   '\n', KBD_NOMAP, no_completion);
+	    } else {
+		(void) mac_literalarg(&val);	/* consume to end of line */
+		status = tb_length(val) > 1;
+	    }
+	    if (status == TRUE) {
+		reverse_abbr(mpp, bufname, kbuf);
 
-    hst_glue(' ');
-    tb_scopy(&val, "");
-    if (!clexec) {
-	status = kbd_reply("to this new string: ", &val, eol_history,
-			   '\n', KBD_NOMAP, no_completion);
-    } else {
-	(void) mac_literalarg(&val);	/* consume to end of line */
-	status = tb_length(val) > 1;
+		status = addtomap(mpp, tb_values(kbuf), tb_length0(kbuf),
+				  MAPF_USERTIMER | remapflag, -1, tb_values(val));
+		relist_mappings(bufname);
+	    }
+	}
     }
-    if (status != TRUE)
-	return status;
-
-    reverse_abbr(mpp, bufname, kbuf);
-
-    status = addtomap(mpp, tb_values(kbuf), tb_length0(kbuf),
-		      MAPF_USERTIMER | remapflag, -1, tb_values(val));
-    relist_mappings(bufname);
     return status;
 }
 
@@ -694,7 +710,7 @@ save_keystroke(int c)
 	if (kr->kused >= NUMKEYSTR * 2) {	/* time to dump the oldest half */
 	    (void) memcpy((kp->d_chunk),
 			  (&kp->d_chunk[NUMKEYSTR / 2]),
-			  NUMKEYSTR / 2);
+			  (size_t) (NUMKEYSTR / 2));
 	    kr->kused = NUMKEYSTR / 2;
 	}
     }
@@ -767,6 +783,7 @@ sysmapped_c_avail(void)
 void
 mapungetc(int c)
 {
+    TRACE2(("mapungetc %#x\n", c));
     if (tgetc_avail()) {
 	tungetc(c);
     } else {
@@ -796,10 +813,10 @@ mapgetc(void)
 	    mlforce("[Infinite loop detected in %s sequence]",
 		    (insertmode) ? "map!" : "map");
 	    catnap(1000, FALSE);	/* FIXME: be sure message gets out */
-	    return esc_c | NOREMAP;
+	    return esc_c | (int) NOREMAP;
 	}
 	mapgetc_ungotcnt--;
-	return itb_last(mapgetc_ungottenchars) | remapflag;
+	return itb_last(mapgetc_ungottenchars) | (int) remapflag;
     }
     infloopcount = 0;
     return tgetc(mapgetc_raw_flag);
@@ -848,10 +865,10 @@ mapped_c(int remap, int raw)
     mapgetc_raw_flag = raw;
     c = mapgetc();
 
-    if ((c & YESREMAP) == 0 && (!remap || (c & NOREMAP)))
-	return (c & ~REMAPFLAGS);
+    if (((UINT) c & YESREMAP) == 0 && (!remap || ((UINT) c & NOREMAP)))
+	return STRIP_REMAPFLAGS(c);
 
-    c &= ~REMAPFLAGS;
+    c = STRIP_REMAPFLAGS(c);
 
     if (reading_msg_line)
 	mp = 0;
@@ -862,9 +879,9 @@ mapped_c(int remap, int raw)
 
     /* if we got a function key from the lower layers, turn it into '#c'
        and see if the user remapped that */
-    if ((c & SPEC)
+    if (((UINT) c & SPEC)
 #if OPT_KEY_MODIFY
-	&& !(c & mod_KEY)	/* we don't do this special case */
+	&& !((UINT) c & mod_KEY)	/* we don't do this special case */
 #endif
 	) {
 	mapungetc(kcod2key(c));
@@ -883,17 +900,18 @@ mapped_c(int remap, int raw)
 			    mapped_c_start,
 			    TRUE);
 
-	while (itb_more(mappedchars))
+	while (itb_more(mappedchars)) {
 	    mapungetc(itb_next(mappedchars));
+	}
 
 	/* if the user has not mapped '#c', we return the wide code we got
 	   in the first place.  unless they wanted it quoted.  then we
 	   leave it as is */
 	if (!raw && speckey && !matched) {
-	    c = mapgetc() & ~REMAPFLAGS;
+	    c = STRIP_REMAPFLAGS(mapgetc());
 	    if (c != poundc)
 		dbgwrite("BUG: # problem in mapped_c");
-	    return (mapgetc() & ~REMAPFLAGS) | SPEC;
+	    return (STRIP_REMAPFLAGS(mapgetc()) | (int) SPEC);
 	}
 
 	c = mapgetc();
@@ -904,9 +922,9 @@ mapped_c(int remap, int raw)
 	speckey = FALSE;
 
     } while (matched &&
-	     ((remap && !(c & NOREMAP)) || (c & YESREMAP)));
+	     ((remap && !((UINT) c & NOREMAP)) || ((UINT) c & YESREMAP)));
 
-    return c & ~REMAPFLAGS;
+    return STRIP_REMAPFLAGS(c);
 
 }
 
