@@ -15,28 +15,10 @@
  * by Tom Dickey, 1993.    -pgf
  *
  *
- * $Header: /usr/build/vile/vile/RCS/mktbls.c,v 1.150 2007/11/25 22:03:51 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/mktbls.c,v 1.168 2010/05/01 14:46:32 tom Exp $
  *
  */
 
-/*
- * gcc (at least through 4.04) allows offsets, but g++ does not.
- *
- * Visual C++ warns about cases for 64-bit compiles that do not work with
- * the default USE_OFFSETS=1, and that works with modern C compilers, but
- * that does not work with some legacy C compilers, e.g., Solaris.  Since
- * the Unix compilers mostly follow LP64 model, it is not (yet) necessary
- * to make a configure script check to distinguish between the two - just
- * support that USE_OFFSET=2 for Visual C++.
- */
-#if defined(__GNUG__) || (defined(__OS2__) && defined(__IBMC__) && (__IBMC__ >= 200))
-#define USE_OFFSETS 0
-#elif defined(_MSC_VER)
-#define USE_OFFSETS 2
-#else
-#define USE_OFFSETS 1
-#endif
-
 /* stuff borrowed/adapted from estruct.h */
 
 #ifdef HAVE_CONFIG_H
@@ -56,6 +38,10 @@
 
 #endif /* !defined(HAVE_CONFIG_H) */
 
+#if defined(WIN32)
+#include "w32vile.h"
+#endif
+
 #ifndef SYS_VMS
 #define SYS_VMS 0
 #endif
@@ -103,6 +89,8 @@ extern void free(char *ptr);
 #define BADEXIT		1
 #endif
 
+#define ENUM_PREFIX	"vile_"
+
 #define	TABLESIZE(v)	(sizeof(v)/sizeof(v[0]))
 
 #ifndef OPT_EXEC_MACROS
@@ -113,7 +101,7 @@ extern void free(char *ptr);
 
 #define MAX_BIND        4	/* maximum # of key-binding types */
 #define	MAX_PARSE	5	/* maximum # of tokens on line */
-#define	LEN_BUFFER	60	/* nominal buffer-length */
+#define	LEN_BUFFER	120	/* nominal buffer-length */
 #define	MAX_BUFFER	(LEN_BUFFER*10)
 #define	LEN_CHRSET	256	/* total # of chars in set (ascii) */
 
@@ -157,9 +145,11 @@ typedef struct stringl {
     struct stringl *nst;
 } LIST;
 
-static char *Blank = "";
+static char Blank[] =
+{0};
 
 static LIST *all_names;
+static LIST *all_aliases;
 static LIST *all_kbind;		/* data for kbindtbl[] */
 static LIST *all_w32bind;	/* w32 data for kbindtbl[] */
 static LIST *all_funcs;		/* data for extern-lines in neproto.h */
@@ -178,7 +168,7 @@ static LIST *all_bmodes;	/* data for BUFFER modes */
 static LIST *all_wmodes;	/* data for WINDOW modes */
 
 static void save_all_modes(const char *type, char *normal, const char
-			   *abbrev, char *cond);
+			   *abbrev, const char *cond);
 static void save_all_submodes(const char *type, char *normal, const char
 			      *abbrev, char *cond);
 
@@ -193,6 +183,7 @@ typedef enum {
     ,SECT_BUFF
     ,SECT_WIND
     ,SECT_FSMS
+    ,SECT_ABBR
 } SECTIONS;
 
 	/* definitions for indices to 'asciitbl[]' vs 'kbindtbl[]' */
@@ -330,6 +321,7 @@ free_LIST(LIST ** p)
 	    FreeIfNeeded(q->Note);
 	free((char *) q);
     }
+    *p = 0;
 }
 
 /******************************************************************************/
@@ -532,7 +524,7 @@ Parse(char *input, char **vec)
     int expecting = TRUE, count = 0, quote = 0, n, c;
 
     for (c = 0; c < MAX_PARSE; c++)
-	vec[c] = "";
+	vec[c] = Blank;
     for (c = (int) strlen(input); c > 0 && isSpace(input[c - 1]); c--)
 	input[c - 1] = EOS;
 
@@ -682,13 +674,22 @@ CheckModes(char *name)
 	badfmt("illegal mode-name");
 }
 
+static char *
+AllocKey(char *normal, int type, char *abbrev)
+{
+    char *tmp = Alloc((unsigned) (4 + strlen(normal) + strlen(abbrev)));
+    Sprintf(tmp, "%s\n%c\n%s", normal, type, abbrev);
+    return tmp;
+}
+
 /* make a sort-key for mode-name */
 static char *
 Mode2Key(char *type, char *name, char *cond, int submode)
 {
     int c;
-    char *abbrev = AbbrevMode(name), *normal = NormalMode(name), *tmp =
-    Alloc((unsigned) (4 + strlen(normal) + strlen(abbrev)));
+    char *abbrev = AbbrevMode(name);
+    char *normal = NormalMode(name);
+    char *tmp;
 
     CheckModes(normal);
     CheckModes(abbrev);
@@ -707,7 +708,43 @@ Mode2Key(char *type, char *name, char *cond, int submode)
     if (submode)
 	save_all_submodes(type, normal, abbrev, cond);
 
-    Sprintf(tmp, "%s\n%c\n%s", normal, c, abbrev);
+    tmp = AllocKey(normal, c, abbrev);
+#if NO_LEAKS
+    free(normal);
+    free(abbrev);
+#endif
+    return tmp;
+}
+
+/* make a sort-key for state-variables */
+static char *
+Vars2Key(char *type, char *name, char *cond)
+{
+    int c;
+    char *abbrev = AbbrevMode(name);
+    char *normal = NormalMode(name);
+    char *tmp;
+
+    CheckModes(normal);
+    CheckModes(abbrev);
+
+    switch (c = *type) {
+    case 'b':
+    case 'e':
+    case 'i':
+    case 's':
+	break;
+    case 'r':
+	c = 'x';		/* make this sort after strings */
+    }
+
+    /* insert into 'all_modes' to provide for common name-completion
+     * table, and into 'all_statevars' to get name/index correspondence.
+     */
+    InsertSorted(&all_modes, name, "state", "", formcond("OPT_EVAL", cond),
+		 "");
+
+    tmp = AllocKey(normal, c, abbrev);
 #if NO_LEAKS
     free(normal);
     free(abbrev);
@@ -749,7 +786,7 @@ Name2Address(char *name, char *type)
     char *temp;
 
     temp = Name2Symbol(name);
-    if (strlen(temp) + 1 + (isboolean(*type) ? 4 : 0) > len)
+    if (strlen(temp) + 1 + (size_t) (isboolean(*type) ? 4 : 0) > len)
 	badfmt("bug: buffer overflow in Name2Address");
 
     (void) strcpy(base, temp);
@@ -759,63 +796,24 @@ Name2Address(char *name, char *type)
     return base;
 }
 
-/* define Member_Offset macro, used in index-definitions */
-static void
-DefineOffset(FILE *fp)
-{
-    Fprintf(fp, "#ifndef\tMember_Offset\n");
-#if USE_OFFSETS == 2
-    Fprintf(fp, "\
-#define\tMember_Offset(T,M) ((int)(((long)(&(((T*)0)->M) - (char *)0))/\\\n\
-\t\t\t\t ((long)(&(((T*)0)->Q1) - &(((T*)0)->s_MAX)))))\n");
-#elif USE_OFFSETS == 1
-    Fprintf(fp, "\
-#define\tMember_Offset(T,M) ((int)(((long)&(((T*)0)->M))/\\\n\
-\t\t\t\t ((long)&(((T*)0)->Q1) - (long)&(((T*)0)->s_MAX))))\n");
-#else
-    Fprintf(fp, "\
-#define\tMember_Offset(T,M) ((vile_ ## M)-1)\n");
-#endif
-    Fprintf(fp, "#endif\n");
-}
-
 /* generate the index-struct (used for deriving ifdef-able index definitions) */
 static void
 WriteIndexStruct(FILE *fp, LIST * p, const char *ppref)
 {
-#if USE_OFFSETS
-    char *s, temp[MAX_BUFFER], line[MAX_BUFFER], *vec[MAX_PARSE];
-
-    BeginIf();
-    Fprintf(fp, "typedef\tstruct\t%c\n", L_CURL);
-    while (p != 0) {
-	WriteIf(fp, p->Cond);
-	(void) Parse(strcpy(line, p->Name), vec);
-	Sprintf(temp, "\tchar\t%s;", s = Name2Symbol(vec[1]));
-	free(s);
-	if (p->Note[0]) {
-	    (void) PadTo(32, temp);
-	    Sprintf(temp + strlen(temp), "/* %s */", p->Note);
-	}
-	Fprintf(fp, "%s\n", temp);
-	p = p->nst;
-    }
-
-    FlushIf(fp);
-    Fprintf(fp, "\tchar\ts_MAX;\n");
-    Fprintf(fp, "\tchar\tQ1;\n");
-    Fprintf(fp, "\t%c Index%s;\n\n", R_CURL, ppref);
-#else
     char *s, temp[MAX_BUFFER], line[MAX_BUFFER], *vec[MAX_PARSE];
     int count = 0;
 
     BeginIf();
     Fprintf(fp, "typedef\tenum\t%c\n", L_CURL);
-    Fprintf(fp, "\tMIN_%c_VALUES = 0\n", *ppref);
+    /*
+     * The first symbol is unused; it is a placeholder in case the first value
+     * is ifdef'd out.
+     */
+    Fprintf(fp, "\tDUMMY_%c_VALUES = -1\n", *ppref);
     for (; p != 0; p = p->nst, count++) {
 	WriteIf(fp, p->Cond);
 	(void) Parse(strcpy(line, p->Name), vec);
-	Sprintf(temp, "\t,vile_%s", s = Name2Symbol(vec[1]));
+	Sprintf(temp, "\t,%s%s", ENUM_PREFIX, s = Name2Symbol(vec[1]));
 	free(s);
 	if (p->Note[0]) {
 	    (void) PadTo(32, temp);
@@ -832,7 +830,6 @@ WriteIndexStruct(FILE *fp, LIST * p, const char *ppref)
     Fprintf(nemode, "%s", temp);
 
     Fprintf(fp, "\t%c Index%s;\n\n", R_CURL, ppref);
-#endif
 }
 
 /* generate the index-definitions */
@@ -852,8 +849,8 @@ WriteModeDefines(LIST * p, const char *ppref)
 		p->Func);
 	(void) PadTo(24, temp);
 	WriteIf(nemode, p->Cond);
-	Sprintf(temp + strlen(temp), "Member_Offset(Index%s, %s)",
-		ppref, s = Name2Symbol(vec[1]));
+	Sprintf(temp + strlen(temp), "%s%s",
+		ENUM_PREFIX, s = Name2Symbol(vec[1]));
 	free(s);
 	Fprintf(nemode, "%s\n", temp);
     }
@@ -861,15 +858,6 @@ WriteModeDefines(LIST * p, const char *ppref)
     Fprintf(nemode, "\n");
     FlushIf(nemode);
 
-#if USE_OFFSETS
-    Sprintf(temp, "#define NUM_%c_VALUES\tMember_Offset(Index%s, s_MAX)",
-	    *ppref, ppref);
-    (void) PadTo(32, temp);
-    Sprintf(temp + strlen(temp), "/* TABLESIZE(%c_valnames) -- %s */\n",
-	    toLower(*ppref), ppref);
-    Fprintf(nemode, "%s", temp);
-#endif
-
     Fprintf(nemode, "#define MAX_%c_VALUES\t%d\n\n", *ppref, count);
 }
 
@@ -891,7 +879,9 @@ WriteModeSymbols(LIST * p)
 	s = 0;
 
 	Sprintf(temp + strlen(temp), "%s,",
-		*vec[3] ? (s = Name2Address(vec[3], vec[2])) : "\"X\"");
+		(*vec[3]
+		 ? (s = Name2Address(vec[3], vec[2]))
+		 : "\"X\""));
 	(void) PadTo(48, temp);
 	if (s != 0)
 	    free(s);
@@ -915,20 +905,22 @@ save_all_modes(
 		  const char *type,
 		  char *normal,
 		  const char *abbrev,
-		  char *cond)
+		  const char *cond)
 {
     if (isboolean(*type)) {
 	char t_normal[LEN_BUFFER], t_abbrev[LEN_BUFFER];
-	save_all_modes("Bool",
-		       strcat(strcpy(t_normal, "no"), normal),
-		       *abbrev
-		       ? strcat(strcpy(t_abbrev, "no"), abbrev)
-		       : "",
-		       cond);
+	strcat(strcpy(t_normal, "no"), normal);
+	if (*abbrev) {
+	    strcat(strcpy(t_abbrev, "no"), abbrev);
+	} else {
+	    strcpy(t_abbrev, "");
+	}
+	save_all_modes("Bool", t_normal, t_abbrev, cond);
     }
     InsertSorted(&all_modes, normal, type, "", cond, "");
-    if (*abbrev)
+    if (*abbrev && strcmp(normal, abbrev)) {
 	InsertSorted(&all_modes, abbrev, type, "", cond, "");
+    }
 }
 
 static void
@@ -1125,7 +1117,7 @@ dump_majors(void)
     LIST *p, *q;
     char norm[LEN_BUFFER], abbr[LEN_BUFFER], type[LEN_BUFFER];
     char normal[LEN_BUFFER], abbrev[LEN_BUFFER];
-    char *my_cond = "OPT_MAJORMODE";
+    const char *my_cond = "OPT_MAJORMODE";
 
     for (q = all_majors; q; q = q->nst) {
 	Sprintf(normal, "%smode", q->Name);	/* FIXME */
@@ -1194,6 +1186,55 @@ dump_mmodes(void)
     write_lines(nemode, bottom);
 }
 
+/******************************************************************************/
+static void
+save_abbr_in(LIST ** q, LIST * p, char **vec)
+{
+    char full_name[LEN_BUFFER];
+    char part_name[LEN_BUFFER];
+    char type_name[LEN_BUFFER];
+    int count;
+
+    for (; p; p = p->nst) {
+	*full_name = EOS;
+	*type_name = EOS;
+	*part_name = EOS;
+	count = sscanf(p->Name, "%s\n%s\n%s", full_name, type_name, part_name);
+	if (count == 2 || (count == 3 && *part_name == EOS)) {
+	    if (!strcmp(full_name, vec[1])) {
+		free(p->Name);
+		p->Name = AllocKey(full_name, *type_name, vec[2]);
+		*q = p;
+	    }
+	} else if (count == 3) {
+	    if (!strcmp(full_name, vec[1])) {
+		fprintf(stderr, "\"%s\" already has alias \"%s\"\n",
+			full_name,
+			vec[2]);
+		badfmt("naming conflict");
+	    }
+	}
+    }
+}
+
+static void
+save_abbrs(char **vec)
+{
+    LIST *p = 0;
+
+    save_abbr_in(&p, all_gmodes, vec);
+    save_abbr_in(&p, all_qmodes, vec);
+    save_abbr_in(&p, all_bmodes, vec);
+    save_abbr_in(&p, all_wmodes, vec);
+    save_abbr_in(&p, all_mmodes, vec);
+    if (p) {
+	InsertSorted(&all_modes, vec[2], p->Data, "", p->Cond, "");
+    } else {
+	fprintf(stderr, "Did not find \"%s\"\n", vec[1]);
+	longjmp(my_top, 1);
+    }
+}
+
 /******************************************************************************/
 static void
 save_all_submodes(
@@ -1204,16 +1245,18 @@ save_all_submodes(
 {
     if (isboolean(*type)) {
 	char t_normal[LEN_BUFFER], t_abbrev[LEN_BUFFER];
-	save_all_submodes("Bool",
-			  strcat(strcpy(t_normal, "no"), normal),
-			  *abbrev
-			  ? strcat(strcpy(t_abbrev, "no"), abbrev)
-			  : "",
-			  cond);
+	strcat(strcpy(t_normal, "no"), normal);
+	if (*abbrev) {
+	    strcat(strcpy(t_abbrev, "no"), abbrev);
+	} else {
+	    strcpy(t_abbrev, "");
+	}
+	save_all_submodes("Bool", t_normal, t_abbrev, cond);
     }
     InsertSorted(&all_submodes, normal, type, "", cond, "");
-    if (*abbrev)
+    if (*abbrev && strcmp(normal, abbrev)) {
 	InsertSorted(&all_submodes, abbrev, type, "", cond, "");
+    }
 }
 
 static void
@@ -1266,8 +1309,9 @@ predefine_submodes(char **vec, int len)
 		break;
 	    }
 	}
-	if (!found)
+	if (!found) {
 	    InsertSorted(&all_majors, vec[2], "", "", "", "");
+	}
 	if (len > 2) {
 	    for (p = all_bmodes, found = FALSE; p; p = p->nst) {
 		if (sscanf(p->Name, "%s\n%s\n%s",
@@ -1411,6 +1455,7 @@ start_vars_h(char **argv)
 	"\tstruct UVAR *next;",
 	"\tchar *u_name;\t\t/* name of temp variable */",
 	"\tchar *u_value;\t\t/* value (string) */",
+	"\tchar u_type;\t\t/* type (normally string) */",
 	"} UVAR;",
 	"",
 	"decl_uninit( UVAR *temp_vars );\t/* temporary variables */",
@@ -1439,6 +1484,8 @@ init_statevars(void)
 	"",
 	"/*\tlist of recognized state variables\t*/",
 	"",
+	"extern BLIST blist_statevars;",
+	"",
 	"#ifdef realdef",
 	"DECL_EXTERN_CONST(char *const statevars[]) = {"
     };
@@ -1448,15 +1495,23 @@ init_statevars(void)
 	write_lines(nevars, head);
 }
 
+#define NonNull(s) ((s) ? (s) : "")
+/*
+ * vec[0] is empty
+ * vec[1] is variable name
+ * vec[2] is symbol
+ * vec[3] is condition
+ * vec[4] is comment (unused so far)
+ */
 static void
-save_statevars(char **vec)
+save_statevars(char *type, char **vec)
 {
-    /* insert into 'all_modes' to provide for common name-completion
-     * table, and into 'all_statevars' to get name/index correspondence.
-     */
-    InsertSorted(&all_modes, vec[1], "state", "", formcond("OPT_EVAL",
-							   vec[3]), "");
-    InsertSorted(&all_statevars, vec[1], vec[2], "", vec[3], vec[0]);
+    char *name = vec[1];
+    char *vars = vec[2];
+    char *cond = vec[3];
+    char *key = Vars2Key(type, name, cond);
+
+    InsertSorted(&all_statevars, key, vars, "", cond, vec[0]);
 }
 
 static void
@@ -1469,7 +1524,6 @@ dump_statevars(void)
 	"BLIST blist_statevars = init_blist(statevars);",
 	"#else",
 	"extern const char *const statevars[];",
-	"extern BLIST blist_statevars;",
 	"#endif",
 	"",
 	""
@@ -1487,29 +1541,36 @@ dump_statevars(void)
 	"",
 	"typedef int (StateFunc)(TBUFF **resultp, const char *valuep);",
 	"",
+	"typedef struct { StateFunc *func; char type; } StateVars;",
 	"",
 	"#ifdef realdef",
-	"DECL_EXTERN(StateFunc *statevar_func[]) = {",
+	"DECL_EXTERN(const StateVars statevar_funcs[]) = {",
 	""
     };
     static const char *const tail[] =
     {
-	"\t(StateFunc *)NULL",
+	"\t{ (StateFunc *)NULL, 0 }",
 	"};",
 	"#else",
-	"extern StateFunc *statevar_func[];",
+	"extern const StateVars statevar_funcs[];",
 	"#endif /* realdef */",
 	"",
     };
+    char temp[LEN_BUFFER], norm[LEN_BUFFER], type[LEN_BUFFER];
     LIST *p;
     char *s;
+    int first = 1;
 
     BeginIf();
     for (p = all_statevars; p != 0; p = p->nst) {
-	if (p == all_statevars)
+	if (sscanf(p->Name, "%s\n%s", norm, type) != 2)
+	    continue;
+	if (first) {
 	    init_statevars();
+	    first = 0;
+	}
 	WriteIf(nevars, p->Cond);
-	Fprintf(nevars, "\t%s,\n", s = Name2Symbol(p->Name));
+	Fprintf(nevars, "\t%s,\n", s = Name2Symbol(norm));
 	free(s);
     }
     FlushIf(nevars);
@@ -1532,8 +1593,13 @@ dump_statevars(void)
     write_lines(nevars, middle2);
     /* emit the variable get/set routine table */
     for (p = all_statevars; p != 0; p = p->nst) {
+	if (sscanf(p->Name, "%s\n%s", norm, type) != 2)
+	    continue;
 	WriteIf(nevars, p->Cond);
-	Fprintf(nevars, "\tvar_%s,\n", p->Func);
+	Sprintf(temp, "\t{ var_%s,", p->Func);
+	(void) PadTo(32, temp);
+	Sprintf(temp + strlen(temp), "VALTYPE_%s },", c2TYPE(*type));
+	Fprintf(nevars, "%s\n", temp);
     }
     FlushIf(nevars);
     write_lines(nevars, tail);
@@ -1551,19 +1617,22 @@ init_fsms(void)
 {
     int n;
 
-    if (fsm_uc_name == 0)
+    if (fsm_uc_name == 0) {
 	badfmt("Missing table name");
+    } else {
 
-    for (n = 0; fsm_uc_name[n] != '\0'; n++)
-	fsm_uc_name[n] = (char) toUpper(fsm_uc_name[n]);
-    Fprintf(nefsms, "\n");
-    Fprintf(nefsms, "#if OPT_%s_CHOICES\n", fsm_uc_name);
-    Fprintf(nefsms, "#ifndef realdef\n");
-    Fprintf(nefsms, "extern const FSM_CHOICES fsm_%s_choices[];\n", fsm_lc_name);
-    Fprintf(nefsms, "extern FSM_BLIST fsm_%s_blist;\n", fsm_lc_name);
-    Fprintf(nefsms, "#else\n");
-    Fprintf(nefsms, "DECL_EXTERN_CONST(FSM_CHOICES fsm_%s_choices[]) = %c\n",
-	    fsm_lc_name, L_CURL);
+	for (n = 0; fsm_uc_name[n] != '\0'; n++)
+	    fsm_uc_name[n] = (char) toUpper(fsm_uc_name[n]);
+	Fprintf(nefsms, "\n");
+	Fprintf(nefsms, "extern FSM_BLIST fsm_%s_blist;\n\n", fsm_lc_name);
+	Fprintf(nefsms, "#if OPT_%s_CHOICES\n", fsm_uc_name);
+	Fprintf(nefsms, "#ifndef realdef\n");
+	Fprintf(nefsms, "extern const FSM_CHOICES fsm_%s_choices[];\n", fsm_lc_name);
+	Fprintf(nefsms, "#else\n");
+	Fprintf(nefsms,
+		"DECL_EXTERN_CONST(FSM_CHOICES fsm_%s_choices[]) = %c\n",
+		fsm_lc_name, L_CURL);
+    }
 }
 
 static void
@@ -1662,6 +1731,7 @@ save_funcs(
 {
     char temp[MAX_BUFFER];
     char *s;
+    LIST *p;
 
     if (strcmp(cond, old_cond)) {
 	if (*old_cond) {
@@ -1680,27 +1750,51 @@ save_funcs(
     Sprintf(temp, "extern int %s ( int f, int n );", func);
     InsertOnEnd(&all_funcs, temp);
 
-    s = strcpy(temp, "\t");
+    s = strcpy(temp, "\n");
+
+    s = append(s, "static const char *syn_");
+    s = append(s, func);
+    s = append(s, "[] = \n{\n");
+    for (p = all_aliases; p != 0; p = p->nst) {
+	if (p->Cond[0]) {
+	    s = append(s, "#if ");
+	    s = append(s, p->Cond);
+	    s = append(s, "\n");
+	}
+	s = append(s, "\t\"");
+	s = append(s, p->Name);
+	s = append(s, "\",\n");
+	if (p->Cond[0]) {
+	    s = append(s, "#endif\n");
+	}
+    }
+    s = append(s, "\t0\n};\n");
+    s = append(s, "\n");
+
     s = append(s, "DECL_EXTERN_CONST(CMDFUNC f_");
     s = append(s, func);
     s = append(s, ")");
     (void) PadTo(32, temp);
-    s = append(s, "= {\n\tINIT_UNION(");
+    s = append(s, "=\n{\n\tINIT_UNION(");
     s = append(s, func);
     s = append(s, "),\n\t");
     s = append(s, flags);
-    s = append(s, "\n#if OPT_MACRO_ARGS\n\t\t,0\n#endif");
-    s = append(s, "\n#if OPT_TRACE\n\t\t,\"");
+    s = append(s, "\n\t,syn_");
+    s = append(s, func);
+    s = append(s, "\n#if OPT_MACRO_ARGS\n\t,0\n#endif");
+    s = append(s, "\n#if OPT_TRACE\n\t,\"");
     s = append(s, func);
     s = append(s, "\"\n#endif");
-    s = append(s, "\n#if OPT_ONLINEHELP\n\t\t,\"");
+    s = append(s, "\n#if OPT_ONLINEHELP\n\t,\"");
     s = append(s, help);
-    (void) append(s, "\"\n#endif\n };");
+    (void) append(s, "\"\n#endif\n};");
     InsertOnEnd(&all__FUNCs, temp);
 
     s = append(strcpy(temp, "extern const CMDFUNC f_"), func);
     (void) append(s, ";");
     InsertOnEnd(&all__CMDFs, temp);
+
+    free_LIST(&all_aliases);
 }
 
 static void
@@ -1768,6 +1862,12 @@ save_names(char *name, char *func, char *cond)
     InsertSorted(&all_names, name, func, "", cond, "");
 }
 
+static void
+save_aliases(char *name, char *func, char *cond)
+{
+    InsertSorted(&all_aliases, name, func, "", cond, "");
+}
+
 static void
 dump_names(void)
 {
@@ -1777,6 +1877,7 @@ dump_names(void)
     Fprintf(nename, "\n/* if you maintain this by hand, keep it in */\n");
     Fprintf(nename, "/* alphabetical order!!!! */\n\n");
     Fprintf(nename, "#include <blist.h>\n\n");
+    Fprintf(nename, "extern BLIST blist_nametbl;\n\n");
     Fprintf(nename, "#ifdef real_NAMETBL\n\n");
     Fprintf(nename, "EXTERN_CONST NTAB nametbl[] = {\n");
 
@@ -1790,7 +1891,6 @@ dump_names(void)
     Fprintf(nename, "\t{ NULL, NULL }\n};\n\n");
     Fprintf(nename, "BLIST blist_nametbl = init_blist(nametbl);\n\n");
     Fprintf(nename, "#else\n\n");
-    Fprintf(nename, "extern BLIST blist_nametbl;\n\n");
     Fprintf(nename, "#endif\n");
 }
 
@@ -1861,6 +1961,11 @@ save_ufuncs(char **vec)
 static void
 dump_ufuncs(void)
 {
+    static const char *const head[] =
+    {
+	"extern BLIST blist_ufuncs;",
+	"",
+    };
     static const char *const middle[] =
     {
 	"\tDATA(NULL, 0, \"\"),",
@@ -1869,7 +1974,6 @@ dump_ufuncs(void)
 	"BLIST blist_ufuncs = init_blist(vl_ufuncs);",
 	"#else",
 	"extern const UFUNC vl_ufuncs[];",
-	"extern BLIST blist_ufuncs;",
 	"#endif",
 	"",
     };
@@ -1877,6 +1981,7 @@ dump_ufuncs(void)
     LIST *p;
     int count;
 
+    write_lines(nevars, head);
     for (p = all_ufuncs, count = 0; p != 0; p = p->nst) {
 	if (!count++)
 	    init_ufuncs();
@@ -2030,6 +2135,7 @@ free_mktbls(void)
     int k;
 
     free_LIST(&all_names);
+    free_LIST(&all_aliases);
     free_LIST(&all_funcs);
     free_LIST(&all__FUNCs);
     free_LIST(&all__CMDFs);
@@ -2122,6 +2228,9 @@ main(int argc, char *argv[])
 
 	case '.':		/* a new section */
 	    switch (col1) {
+	    case 'a':
+		section = SECT_ABBR;
+		break;
 	    case 'c':
 		section = SECT_CMDS;
 		break;
@@ -2160,6 +2269,11 @@ main(int argc, char *argv[])
 
 	case '\t':		/* a new function */
 	    switch (section) {
+	    case SECT_ABBR:
+		if (r != 2)
+		    badfmt("looking for ABBR assignments");
+		save_abbrs(vec);
+		break;
 	    case SECT_CMDS:
 		switch (col1) {
 		case '"':	/* then it's an english name */
@@ -2167,6 +2281,7 @@ main(int argc, char *argv[])
 			badfmt("looking for english name");
 
 		    save_names(vec[1], func, formcond(fcond, vec[2]));
+		    save_aliases(vec[1], func, formcond(fcond, vec[2]));
 		    break;
 
 		case '\'':	/* then it's a key */
@@ -2229,6 +2344,12 @@ main(int argc, char *argv[])
 		save_wmodes(modetype, vec);
 		break;
 
+	    case SECT_VARS:
+		if (r < 2 || r > 4)
+		    badfmt("looking for state variables");
+		save_statevars(modetype, vec);
+		break;
+
 	    default:
 		badfmt("did not expect a tab");
 	    }
@@ -2244,8 +2365,7 @@ main(int argc, char *argv[])
 		   a help line for it.  save the previous
 		   one now, and hang onto this one */
 		if (func[0]) {	/* flush the old one */
-		    save_funcs(func, flags, fcond,
-			       old_fcond, funchelp);
+		    save_funcs(func, flags, fcond, old_fcond, funchelp);
 		    funchelp[0] = EOS;
 		}
 		(void) strcpy(func, vec[1]);
@@ -2258,9 +2378,14 @@ main(int argc, char *argv[])
 		break;
 
 	    case SECT_VARS:
-		if (r < 2 || r > 4)
-		    badfmt("looking for char *statevars[]");
-		save_statevars(vec);
+		if (r != 1
+		    || (!strcmp(vec[1], "bool")
+			&& !strcmp(vec[1], "enum")
+			&& !strcmp(vec[1], "int")
+			&& !strcmp(vec[1], "string")
+			&& !strcmp(vec[1], "regex")))
+		    badfmt("looking for mode datatype");
+		(void) strcpy(modetype, vec[1]);
 		break;
 
 	    case SECT_FUNC:
@@ -2329,7 +2454,6 @@ main(int argc, char *argv[])
 
     if (all_wmodes || all_bmodes) {
 	nemode = OpenHeader("nemode.h", argv);
-	DefineOffset(nemode);
 	dump_majors();
 	dump_all_modes();
 	dump_all_submodes();
