@@ -2,7 +2,7 @@
  *		The routines in this file handle the conversion of pathname
  *		strings.
  *
- * $Header: /usr/build/vile/vile/RCS/path.c,v 1.156 2007/09/23 23:43:59 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/path.c,v 1.169 2010/09/06 17:29:59 tom Exp $
  *
  *
  */
@@ -12,8 +12,10 @@
 
 #if SYS_UNIX
 #include <sys/types.h>
+#if !SYS_MINGW
 #include <pwd.h>
 #endif
+#endif
 
 #if SYS_VMS
 #include <starlet.h>
@@ -28,27 +30,6 @@
 
 #include "dirstuff.h"
 
-/* C Set did not define S_IFMT */
-#if !defined(S_IFMT)
-#define S_IFMT (S_IFDIR|S_IFREG)
-#endif
-
-#if !defined(S_ISDIR)
-# define S_ISDIR(m)  (((m) & S_IFMT) == S_IFDIR)
-#endif
-
-#if !defined(S_ISREG)
-# define S_ISREG(m)  (((m) & S_IFMT) == S_IFREG)
-#endif
-
-#ifndef S_IFIFO
-#define S_IFIFO S_IFREG
-#endif
-
-#if !defined(S_ISFIFO)
-# define S_ISFIFO(m)  (((m) & S_IFMT) == S_IFIFO)
-#endif
-
 #if SYS_WINNT
 # include <direct.h>
 # define curdrive() (_getdrive() + ('A' - 1))
@@ -78,7 +59,7 @@ curr_dir_on_drive(int d)
 #endif
 #endif
 
-static int is_absolute_pathname(char *path);
+static char empty_string[] = "";
 
 /*
  * Fake directory-routines for system where we cannot otherwise figure out how
@@ -306,26 +287,27 @@ pathleaf(char *path)
  * must be no longer than NFILEN.
  */
 char *
-pathcat(char *dst, const char *path, char *leaf)
+pathcat(char *dst, const char *path, const char *cleaf)
 {
     char save_path[NFILEN];
     char save_leaf[NFILEN];
+    char *leaf;
     char *s;
     size_t have;
 
     if (dst == 0)
 	return 0;
 
-    if (leaf == 0)
-	leaf = "";
+    if (cleaf == 0)
+	cleaf = "";
 
-    leaf = vl_strncpy(save_leaf, leaf, NFILEN);		/* leaf may be in dst */
+    leaf = vl_strncpy(save_leaf, cleaf, (size_t) NFILEN);	/* leaf may be in dst */
 
     if (isEmpty(path)) {
 	(void) strcpy(dst, leaf);
     } else {
 
-	path = vl_strncpy(save_path, path, NFILEN);	/* path may be in dst */
+	path = vl_strncpy(save_path, path, (size_t) NFILEN);	/* path may be in dst */
 
 #if OPT_MSDOS_PATH
 	if (is_msdos_drive(save_path) != 0
@@ -339,7 +321,7 @@ pathcat(char *dst, const char *path, char *leaf)
 	    }
 	} else
 #endif
-	if (is_absolute_pathname(leaf)) {
+	if (is_abs_pathname(leaf)) {
 	    (void) strcpy(dst, leaf);
 	} else {
 	    have = strlen(strcpy(dst, path));
@@ -437,7 +419,7 @@ find_user(const char *name)
 		return q->path;
 	    }
 	}
-#if SYS_UNIX
+#if SYS_UNIX && !SYS_MINGW
 	/* not-found, do a lookup.
 	 * First try getpwnam with the specified name,
 	 * which will use ~ or whatever was passed
@@ -449,7 +431,7 @@ find_user(const char *name)
 	     * so lookup the current uid, and then lookup
 	     * the name based on that.
 	     */
-	    p = getpwuid((int) getuid());
+	    p = getpwuid((uid_t) getuid());
 	}
 
 	/* if either of the above lookups worked
@@ -1212,6 +1194,12 @@ canonpath(char *ss)
 
 	    p++;
 	    pp++;		/* leave the leading slash */
+#if OPT_UNC_PATH
+	    if (is_slashc(*pp)) {
+		p++;
+		pp++;
+	    }
+#endif
 	    while (*pp) {
 		if (is_slashc(*pp)) {
 		    pp++;
@@ -1493,7 +1481,7 @@ lengthen_path(char *path)
 	return (lengthen_path(f) != 0) ? path : 0;
 
     if ((f = path) == 0
-	|| (f == error_val))
+	|| isErrorVal(f))
 	return path;
 
     if (*path != EOS && isInternalName(path)) {
@@ -1625,7 +1613,7 @@ lengthen_path(char *path)
 #endif
 	len = (int) strlen(temp);
 	temp[len++] = SLASHC;
-	(void) vl_strncpy(temp + len, f, sizeof(temp) - len);
+	(void) vl_strncpy(temp + len, f, sizeof(temp) - (size_t) len);
 	(void) strcpy(path, temp);
     }
 #if OPT_MSDOS_PATH
@@ -1650,8 +1638,8 @@ lengthen_path(char *path)
  * Returns true if the argument looks like an absolute pathname (e.g., on
  * unix, begins with a '/').
  */
-static int
-is_absolute_pathname(char *path)
+int
+is_abs_pathname(char *path)
 {
     char *f;
     if ((f = is_appendname(path)) != 0)
@@ -1666,7 +1654,7 @@ is_absolute_pathname(char *path)
 
 #if OPT_MSDOS_PATH
     if ((f = is_msdos_drive(path)) != 0)
-	return is_absolute_pathname(f);
+	return is_abs_pathname(f);
 #endif
 
 #if SYS_UNIX || OPT_MSDOS_PATH || SYS_VMS
@@ -1718,7 +1706,7 @@ int
 is_pathname(char *path)
 {
     return is_relative_pathname(path)
-	|| is_absolute_pathname(path);
+	|| is_abs_pathname(path);
 }
 
 /*
@@ -1796,7 +1784,7 @@ is_scratchname(const char *fn)
  * Test if the given path is a directory
  */
 int
-is_directory(char *path)
+is_directory(const char *path)
 {
 #if OPT_VMS_PATH
     char *s;
@@ -1814,7 +1802,7 @@ is_directory(char *path)
     /* If the name doesn't look like a directory, there's no point in
      * wasting time doing a 'stat()' call.
      */
-    s = vms_pathleaf(path);
+    s = vms_pathleaf((char *) path);
     if ((s = strchr(s, '.')) != 0) {
 	char ftype[NFILEN];
 	(void) mkupper(strcpy(ftype, s));
@@ -1835,7 +1823,7 @@ is_directory(char *path)
      * as a dir.
      */
     if (is_slashc(path[0]) && is_slashc(path[1])) {
-	char *end = skip_string(path);
+	const char *end = skip_cstring(path);
 	int slashes = 0;
 	if (end > path && is_slashc(end[-1]))
 	    end--;
@@ -1904,7 +1892,7 @@ parse_pathlist(const char *list, char *result)
     return list;
 }
 
-#if SYS_WINNT && !CC_TURBO
+#if SYS_WINNT && !CC_TURBO && !CC_MINGW
 /********                                               \\  opendir  //
  *                                                        ===========
  * opendir
@@ -2163,7 +2151,7 @@ prepend_to_path_list(char **path_list, char *path)
 	if ((t = *path_list) != 0) {
 	    need += strlen(t);
 	} else {
-	    t = "";
+	    t = empty_string;
 	}
 	if ((s = typeallocn(char, need)) != 0) {
 	    lsprintf(s, "%s%c%s", find, vl_pathchr, t);
@@ -2180,7 +2168,7 @@ prepend_to_path_list(char **path_list, char *path)
  * Append the given path to a path-list
  */
 void
-append_to_path_list(char **path_list, char *path)
+append_to_path_list(char **path_list, const char *path)
 {
     char *s, *t;
     size_t need;
@@ -2199,7 +2187,7 @@ append_to_path_list(char **path_list, char *path)
 	if ((t = *path_list) != 0) {
 	    need += strlen(t);
 	} else {
-	    t = "";
+	    t = empty_string;
 	}
 	if ((s = typeallocn(char, need)) != 0) {
 	    if (*t != EOS)
@@ -2258,7 +2246,7 @@ path_trunc(char *path, int max_path_len, char *trunc_buf, int trunc_buf_len)
 	max_len -= sizeof(">>>") - 1;
 #else
 	strcpy(trunc_buf, "...");
-	max_len -= sizeof("...") - 1;
+	max_len -= (int) sizeof("...") - 1;
 #endif
 	strcat(trunc_buf, path + path_len - max_len);
     }
