@@ -2,7 +2,7 @@
  * This file contains the command processing functions for a number of random
  * commands. There is no functional grouping here, for sure.
  *
- * $Header: /usr/build/vile/vile/RCS/random.c,v 1.313 2007/12/27 19:50:05 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/random.c,v 1.351 2015/02/01 19:41:09 tom Exp $
  *
  */
 
@@ -99,9 +99,11 @@ set_rdonly(BUFFER *bp, const char *name, int mode)
     fix_cmode(bp, FALSE);
 }
 
-/* generic "lister", which takes care of popping a window/buffer pair under
-	the given name, and calling "func" with a couple of args to fill in
-	the buffer */
+/*
+ * This is a generic "lister", which takes care of popping a window/buffer pair
+ * under the given name, and calling "func" with a couple of args to fill in
+ * the buffer.
+ */
 int
 liststuff(const char *name,
 	  int appendit,
@@ -109,56 +111,85 @@ liststuff(const char *name,
 	  int iarg,
 	  void *vargp)
 {
+    int rc = FALSE;
     BUFFER *bp;
-    int s;
     WINDOW *wp;
     int alreadypopped;
     BUFFER *ocurbp = curbp;
+    L_NUM save_line = -1;
+    C_NUM save_off = -1;
+    int save_top = -1;
 
     TRACE((T_CALLED
 	   "liststuff(name=%s, appendit=%d, func=%p, iarg=%d, vargp=%p)\n",
-	   TRACE_NULL(name), appendit, func, iarg, vargp));
+	   TRACE_NULL(name), appendit, func, iarg, (void *) vargp));
 
     /* create the buffer list buffer   */
     bp = bfind(name, BFSCRTCH);
-    if (bp == NULL)
-	returnCode(FALSE);
+    if (bp != NULL) {
 
-    if (!appendit && (s = bclear(bp)) != TRUE)	/* clear old text (?) */
-	returnCode(s);
-    b_set_scratch(bp);
-    alreadypopped = (bp->b_nwnd != 0);
-    if (popupbuff(bp) == FALSE) {
-	(void) zotbuf(bp);
-	returnCode(FALSE);
-    }
+#ifdef MDREUSE_POS
+	if ((appendit <= 0)
+	    && b_val(bp, MDREUSE_POS)
+	    && (wp = bp2any_wp(bp)) != NULL) {
+	    save_line = line_no(bp, wp->w_dot.l);
+	    save_off = wp->w_dot.o;
+	    save_top = line_no(bp, wp->w_line.l);
+	}
+#endif
 
-    if ((wp = bp2any_wp(bp)) != NULL) {
-	make_local_w_val(wp, WMDNUMBER);
-	set_w_val(wp, WMDNUMBER, FALSE);
-    }
-    if (appendit) {
-	(void) gotoeob(FALSE, 1);
-	MK = DOT;
-    }
-    /* call the passed in function, giving it both the integer and
-       character pointer arguments */
-    (*func) (iarg, vargp);
-    if (alreadypopped && appendit) {
-	(void) gomark(FALSE, 1);
-	(void) forwbline(FALSE, 1);
-	(void) reposition(FALSE, 1);
-    } else {			/* if we're not appending, go to the top */
-	(void) gotobob(FALSE, 1);
-    }
-    set_rdonly(bp, non_filename(), MDVIEW);
+	/* clear old text (?) */
+	if ((appendit > 0) || (rc = bclear(bp)) == TRUE) {
 
-    if (alreadypopped)		/* don't switch to the popup if it wasn't there */
-	swbuffer(ocurbp);
-    else
-	shrinkwrap();		/* only resize if it's fresh */
+	    b_set_scratch(bp);
+	    alreadypopped = (bp->b_nwnd != 0);
+	    if (popupbuff(bp) == FALSE) {
+		(void) zotbuf(bp);
+	    } else {
 
-    returnCode(TRUE);
+		if ((wp = bp2any_wp(bp)) != NULL) {
+		    make_local_w_val(wp, WMDNUMBER);
+		    set_w_val(wp, WMDNUMBER, FALSE);
+		}
+		if (appendit > 0) {
+		    (void) gotoeob(FALSE, 1);
+		    MK = DOT;
+		}
+		/*
+		 * call the passed in function, giving it both the integer and
+		 * character pointer arguments
+		 */
+		(*func) (iarg, vargp);
+		if (alreadypopped && (appendit > 0)) {
+		    (void) gomark(FALSE, 1);
+		    (void) forwbline(FALSE, 1);
+		    (void) reposition(FALSE, 1);
+		} else if (wp != 0 && (appendit < 0 && save_line >= 0)) {
+		    (void) vl_gotoline(save_top);
+		    wp->w_line.l = wp->w_dot.l;
+		    wp->w_line.o = 0;
+		    if (save_line != save_top)
+			(void) vl_gotoline(save_line);
+		    gocol(save_off);
+		} else {
+		    /* if we're not appending, go to the top */
+		    (void) gotobob(FALSE, 1);
+		}
+		set_rdonly(bp, non_filename(), MDVIEW);
+
+		if (alreadypopped) {
+		    /* don't switch to the popup if it wasn't there */
+		    swbuffer(ocurbp);
+		} else {
+		    /* only resize if it's fresh */
+		    shrinkwrap();
+		}
+
+		rc = TRUE;
+	    }
+	}
+    }
+    returnCode(rc);
 }
 
 /*
@@ -209,7 +240,7 @@ showcpos(int f GCC_UNUSED, int n GCC_UNUSED)
 	if (lp == DOT.l) {
 	    predlines = numlines;
 	    predchars = numchars + DOT.o;
-	    curchar = char_at_mark(DOT);
+	    curchar = CharAtDot();
 	}
 	/* on to the next line */
 	++numlines;
@@ -346,14 +377,14 @@ char_no(BUFFER *the_buffer, MARK the_mark)
 {
     LINE *lp;
     B_COUNT curchar = 0;
-    int rslen = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
     for_each_line(lp, the_buffer) {
 	if (lp == the_mark.l) {
-	    curchar += the_mark.o;
+	    curchar += (B_COUNT) the_mark.o;
 	    break;
 	}
-	curchar += llength(lp) + rslen;
+	curchar += line_length(lp);
     }
 
     return curchar;
@@ -382,19 +413,19 @@ gotochr(int f, int n)
 {
     LINE *lp;
     B_COUNT len;
-    B_COUNT len_rs = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
     if (!f) {
 	DOT.l = lback(buf_head(curbp));
 	DOT.o = llength(DOT.l);
     } else {
-	B_COUNT goal = n;
+	B_COUNT goal = (B_COUNT) n;
 	for_each_line(lp, curbp) {
 	    len = line_length(lp);
 	    if (goal <= len) {
 		DOT.l = lp;
 		/* we cannot move into the middle of the line-separator */
-		if ((DOT.o = goal - 1) >= llength(lp))
+		if ((DOT.o = (C_NUM) (goal - 1)) >= llength(lp))
 		    DOT.o = llength(lp) - 1;
 		goal = 0;
 		break;
@@ -426,6 +457,60 @@ gotopct(int f, int n)
 }
 #endif
 
+/*
+ * Get the (possibly wide) character at the given line and offset.
+ *
+ * If this is a wide character and the buffer is encoded for UTF-8,
+ * replace the whole character at that offset, possibly changing the
+ * length of the line.
+ */
+void
+set_char2(LINE *lp, int offset, int ch)
+{
+#if OPT_MULTIBYTE
+    if (b_is_utfXX(curbp)) {
+	UCHAR buffer[10];
+	int old_len = bytes_at(lp, offset);
+	int new_len = vl_conv_to_utf8(buffer, (UINT) ch, sizeof(buffer));
+	int n;
+
+	if (new_len > old_len) {
+	    lins_bytes(new_len - old_len, ' ');
+	} else if (new_len < old_len) {
+	    ldel_bytes((B_COUNT) (old_len - new_len), FALSE);
+	}
+	for (n = 0; n < new_len; ++n)
+	    lputc(lp, offset + n, buffer[n]);
+    } else
+#endif
+	lputc(lp, offset, (char) ch);
+}
+
+/*
+ * Get the (possibly wide) character at the given line and offset.
+ */
+int
+get_char2(LINE *lp, int offset)
+{
+    int result = -1;
+
+    if (offset >= 0 && offset < llength(lp)) {
+	result = CharOf(lgetc(lp, offset));
+#if OPT_MULTIBYTE
+	if (b_is_utfXX(curbp)) {
+	    UINT target;
+	    int used = vl_conv_to_utf32(&target,
+					lvalue(lp) + offset,
+					(B_COUNT) (llength(lp) - offset));
+	    if (used > 0) {
+		result = (int) target;
+	    }
+	}
+#endif
+    }
+    return result;
+}
+
 /*
  * Return the character at the given MARK, in the current buffer.
  */
@@ -437,18 +522,7 @@ char_at_mark(MARK mark)
     if (mark.o == llength(mark.l)) {
 	curchar = '\n';
     } else {
-	curchar = CharOf(char_at(mark));
-#if OPT_MULTIBYTE
-	if (b_is_utfXX(curbp)) {
-	    UINT target;
-	    int used = vl_conv_to_utf32(&target,
-					lvalue(mark.l) + mark.o,
-					llength(mark.l) - mark.o);
-	    if (used > 0) {
-		curchar = target;
-	    }
-	}
-#endif
+	curchar = get_char2(mark.l, mark.o);
     }
     return curchar;
 }
@@ -481,7 +555,7 @@ getcol(MARK mark, int actual)
 	     * If we are positioning the cursor on the right-side of a multi-
 	     * column character, update the goal to reflect that.
 	     */
-	    if (!global_g_val(GMDALTTABPOS)) {
+	    if (!global_g_val(GMDALTTABPOS) && i < llength(mark.l)) {
 		col = (next_column(mark.l, i, col) - 1);
 	    }
 	}
@@ -593,16 +667,17 @@ twiddle(int f GCC_UNUSED, int n GCC_UNUSED)
     if (llength(dot.l) <= 1)
 	return FALSE;
     if (is_at_end_of_line(dot))
-	--dot.o;
+	dot.o -= BytesBefore(dot.l, dot.o);
     if (dot.o == 0)
-	dot.o = 1;
-    cr = char_at(dot);
-    --dot.o;
-    cl = char_at(dot);
-    copy_for_undo(dot.l);
-    put_char_at(dot, cr);
-    ++dot.o;
-    put_char_at(dot, cl);
+	dot.o = BytesAt(dot.l, 0);
+    cr = get_char2(dot.l, dot.o);
+    dot.o -= BytesBefore(dot.l, dot.o);
+    cl = get_char2(dot.l, dot.o);
+    CopyForUndo(dot.l);
+    set_char2(dot.l, dot.o, cr);
+    dot.o += BytesAt(dot.l, dot.o);
+    set_char2(dot.l, dot.o, cl);
+    DOT.o = dot.o;
     chg_buff(curbp, WFEDIT);
     return (TRUE);
 }
@@ -622,9 +697,9 @@ forceblank(int f, int n)
     B_COUNT n_arg;
     C_NUM nchar;
     int s = TRUE;
-    B_COUNT len_rs = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
-    n_arg = need_at_least(f, n, 0);
+    n_arg = (B_COUNT) need_at_least(f, n, 0);
 
     lp1 = DOT.l;
     /* scan backward */
@@ -641,7 +716,7 @@ forceblank(int f, int n)
 	   firstchar(lp2) < 0) {
 	++nld;
 	if (nld > n_arg && lisreal(lp2))
-	    nchar += line_length(lp2);
+	    nchar += (C_NUM) line_length(lp2);
     }
 
     DOT.l = lforw(lp1);
@@ -998,7 +1073,7 @@ current_directory(int force)
     cwd = realpath(s ? s : ".", current_dirname);
     if (cwd == 0)
 #endif
-	cwd = getcwd(current_dirname, NFILEN);
+	cwd = getcwd(current_dirname, (size_t) NFILEN);
 #else
 # ifdef HAVE_GETWD
     cwd = getwd(current_dirname);
@@ -1175,7 +1250,7 @@ vl_chdir(int f GCC_UNUSED, int n GCC_UNUSED)
     status = mlreply_dir("Change to directory: ", &last, cdirname);
 #if SYS_UNIX || SYS_VMS || SYS_WINNT
     if (status == FALSE) {	/* empty reply, go HOME */
-	(void) strcpy(cdirname, "~");
+	(void) vl_strncpy(cdirname, "~", sizeof(cdirname));
 	status = TRUE;
     }
 #endif
@@ -1288,7 +1363,7 @@ set_directory(const char *dir)
     upmode();
 
     TRACE(("set_directory(%s)\n", dir));
-    exdp = strcpy(exdir, dir);
+    exdp = vl_strncpy(exdir, dir, sizeof(exdir));
 
     if (doglob(exdp)) {
 #if SYS_MSDOS || SYS_OS2
@@ -1319,7 +1394,7 @@ set_directory(const char *dir)
 	if (!dirs_add_active) {
 	    /* Save current directory for subsequent "cd -". */
 
-	    (void) strcpy(prevdir, current_directory(FALSE));
+	    (void) vl_strncpy(prevdir, current_directory(FALSE), sizeof(prevdir));
 	}
 #if OPT_VMS_PATH
 	if (!*exdp)
@@ -1344,7 +1419,7 @@ set_directory(const char *dir)
 	/*
 	 * chdir() failed.  Attempt to match dir using $CDPATH on a VMS
 	 * host.  Note that we can't test using is_pathname() because the
-	 * indirectly referenced is_absolute_pathname() code sez that any
+	 * indirectly referenced is_abs_pathname() code sez that any
 	 * filespec with an embedded '[' or ':' is absolute--not what we
 	 * want here.
 	 */
@@ -1361,7 +1436,7 @@ set_directory(const char *dir)
 		    char **globvec = glob_string(cdpathdir);
 
 		    if (glob_length(globvec) == 1) {
-			strcpy(cdpathdir, globvec[0]);
+			vl_strncpy(cdpathdir, globvec[0], sizeof(cdpathdir));
 			glob_free(globvec);
 			tmp = &cdpathdir[len - 1];
 			if (*tmp == R_BLOCK &&
@@ -1407,7 +1482,7 @@ set_directory(const char *dir)
 		char **globvec = glob_string(cdpathdir);
 
 		if (glob_length(globvec) == 1) {
-		    strcpy(cdpathdir, globvec[0]);
+		    vl_strncpy(cdpathdir, globvec[0], sizeof(cdpathdir));
 		    glob_free(globvec);
 		    if (cd_and_pwd(pathcat(cdpathdir, cdpathdir, exdp))) {
 			return TRUE;
@@ -1425,8 +1500,8 @@ set_directory(const char *dir)
     current_directory(TRUE);
 #endif
 
-    outlen = (term.cols - 1) - (sizeof(CHANGE_FAILED) - 3);
-    mlforce(CHANGE_FAILED, path_trunc(exdir, outlen, tmp, sizeof(tmp)));
+    outlen = (int) ((unsigned) (term.cols - 1) - (sizeof(CHANGE_FAILED) - 3));
+    mlforce(CHANGE_FAILED, path_trunc(exdir, outlen, tmp, (int) sizeof(tmp)));
     return FALSE;
 
 #undef CHANGE_FAILED
@@ -1438,7 +1513,7 @@ set_directory_from_file(BUFFER *bp)
     TRACE((T_CALLED "set_directory_from_file(%s)\n", bp->b_fname));
     if (!isInternalName(bp->b_fname)) {
 	char name[NFILEN];
-	char *leaf = pathleaf(vl_strncpy(name, bp->b_fname, NFILEN));
+	char *leaf = pathleaf(vl_strncpy(name, bp->b_fname, (size_t) NFILEN));
 	if (leaf != 0
 	    && leaf != name) {
 	    *leaf = EOS;
@@ -1515,10 +1590,10 @@ ch_fname(BUFFER *bp, const char *fname)
 }
 
 #if OPT_HOOKS
-static char *
+static const char *
 name_of_hook(HOOK * hook)
 {
-    char *result = "unknown";
+    const char *result = "unknown";
     if (hook == &cdhook)
 	result = "$cd-hook";
     else if (hook == &readhook)
@@ -1545,26 +1620,24 @@ run_a_hook(HOOK * hook)
 {
     int status = FALSE;
 
-    if (!hook->latch && hook->proc[0]) {
-	MARK save_pre_op_dot;	/* ugly hack */
-	int save_dotcmdactive;	/* another ugly hack */
+    if (hook->proc[0] && !doingopcmd) {
+	if (!DisableHook(hook)) {
+	    MARK save_pre_op_dot;	/* ugly hack */
+	    int save_dotcmdactive;	/* another ugly hack */
 
-	TRACE((T_CALLED "run_a_hook(%s)\n", name_of_hook(hook)));
+	    save_dotcmdactive = dotcmdactive;
+	    save_pre_op_dot = pre_op_dot;
 
-	save_dotcmdactive = dotcmdactive;
-	save_pre_op_dot = pre_op_dot;
+	    TPRINTF(("running %s HOOK with %s, current %s\n",
+		     name_of_hook(hook), hook->proc, curbp->b_bname));
+	    status = docmd(hook->proc, TRUE, FALSE, 1);
 
-	DisableHook(hook);
-	TPRINTF(("running %s HOOK with %s, current %s\n",
-		 name_of_hook(hook), hook->proc, curbp->b_bname));
-	status = docmd(hook->proc, TRUE, FALSE, 1);
+	    pre_op_dot = save_pre_op_dot;
+	    dotcmdactive = save_dotcmdactive;
+	}
 	EnableHook(hook);
-
-	pre_op_dot = save_pre_op_dot;
-	dotcmdactive = save_dotcmdactive;
-	returnCode(status);
     }
-    return FALSE;
+    return (status);
 }
 
 int
@@ -1581,37 +1654,178 @@ run_readhook(void)
 }
 #endif
 
+#if (OPT_AUTOCOLOR || OPT_ELAPSED) && defined(VL_ELAPSED)
+/*
+ * Returns elapsed time in milliseconds.
+ */
+double
+vl_elapsed(VL_ELAPSED * first, int begin)
+{
+    double result;
+
+#if defined(HAVE_GETTIMEOFDAY) && (SYS_UNIX && !SYS_MINGW)
+#define	SECS(tv)	((double) tv.tv_sec + ((double) tv.tv_usec / 1.0e6))
+    VL_ELAPSED tv1;
+    gettimeofday(&tv1, 0);
+    if (begin)
+	*first = tv1;
+    result = 1000.0 * (SECS(tv1) - SECS((*first)));
+#elif SYS_WINNT || SYS_MINGW
+    VL_ELAPSED tv1 = GetTickCount();
+    if (begin)
+	*first = tv1;
+    result = (tv1 - *first);
+#else
+    VL_ELAPSED tv1 = time((time_t *) 0);
+    if (begin)
+	*first = tv1;
+    result = 1000.0 * (tv1 - *first);
+#endif
+
+    return result;
+}
+#endif
+
+#if OPT_AUTOCOLOR
+static int
+can_autocolor(BUFFER *bp)
+{
+    int rc = FALSE;
+    if (b_val(bp, VAL_AUTOCOLOR) > 0
+#if OPT_MAJORMODE
+	&& (bp->majr != 0 || !b_is_temporary(bp))
+	&& b_val(bp, MDHILITE)
+#endif
+	) {
+	rc = TRUE;
+    }
+    return rc;
+}
+#endif
+
 void
 autocolor(void)
 {
-#if OPT_COLOR&&!SMALLER
+#if OPT_AUTOCOLOR
+    if (!(reading_msg_line || vile_is_busy || doingsweep)) {
+	BUFFER *bp;
+	WINDOW *wp;
+	VL_ELAPSED begin_time;
+	double elapsed_time;
+	int goal_time;
 
-    WINDOW *wp;
-    int do_update = FALSE;
-    if (reading_msg_line || vile_is_busy || doingsweep)
-	return;
-    for_each_visible_window(wp) {
-	BUFFER *bp = wp->w_bufp;
-	if (b_is_recentlychanged(bp)
-#if OPT_MAJORMODE
-	    && (bp->majr != 0 || !b_is_temporary(bp))
-	    && b_val(bp, MDHILITE)
-#endif
-	    && b_val(bp, VAL_AUTOCOLOR) > 0) {
-	    BUFFER *oldbp = curbp;
-	    WINDOW *oldwp = push_fake_win(bp);
-	    in_autocolor = TRUE;
-	    if (run_a_hook(&autocolorhook)) {
-		b_clr_recentlychanged(bp);
-		do_update = TRUE;
+	int do_update = 0;
+
+	(void) vl_elapsed(&begin_time, TRUE);
+
+	TRACE((T_CALLED "autocolor\n"));
+	for_each_visible_window(wp) {
+	    bp = wp->w_bufp;
+	    if (can_autocolor(bp)
+		&& b_is_recentlychanged(bp)) {
+		/*
+		 * Check if we're skipping autocolor (see below) because the
+		 * buffer is large enough to interfere with the autocolor
+		 * interval.
+		 */
+		if (bp->next_autocolor_time > 0) {
+		    bp->next_autocolor_time--;
+		}
+		if (bp->next_autocolor_time == 0) {
+		    BUFFER *oldbp = curbp;
+		    WINDOW *oldwp = push_fake_win(bp);
+
+		    in_autocolor = TRUE;
+		    if (run_a_hook(&autocolorhook)) {
+			b_clr_recentlychanged(bp);
+			++do_update;
+			bp->next_autocolor_time--;
+		    }
+		    in_autocolor = FALSE;
+		    pop_fake_win(oldwp, oldbp);
+		}
 	    }
-	    in_autocolor = FALSE;
-	    pop_fake_win(oldwp, oldbp);
 	}
-    }
-    if (do_update)
-	update(FALSE);
 
+	/*
+	 * If do_update is set, we colored at least one buffer.
+	 *
+	 * Now check to see if the process took longer than we would like.  If
+	 * the process lasts longer than the fraction of the autocolor interval
+	 * given by the percent-autocolor mode, we will set values to reduce
+	 * the number of times that we'll recolor the buffers that took a long
+	 * time.
+	 */
+	if (do_update) {
+	    elapsed_time = vl_elapsed(&begin_time, FALSE);
+
+	    goal_time = global_b_val(VAL_AUTOCOLOR);
+	    if (global_b_val(VAL_PERCENT_AUTOCOLOR) > 0) {
+		goal_time *= global_b_val(VAL_PERCENT_AUTOCOLOR);
+		if (goal_time <= 0)
+		    goal_time = 1;
+	    }
+	    if (elapsed_time > goal_time) {
+		for_each_visible_window(wp) {
+		    bp = wp->w_bufp;
+
+		    /*
+		     * Check if we have just autocolored the buffer.
+		     */
+		    if (can_autocolor(bp)
+			&& bp->next_autocolor_time < 0) {
+			/*
+			 * Compute the desired autocolor interval for the
+			 * buffer.  This might be different from the global
+			 * value, since each buffer can have its own
+			 * percent-autocolor value.
+			 *
+			 * Ignore the buffer's local autocolor value, since we
+			 * use that only for displaying results.
+			 */
+			goal_time = ((global_b_val(VAL_AUTOCOLOR)
+				      + do_update - 1)
+				     / do_update);
+			if (b_val(bp, VAL_PERCENT_AUTOCOLOR) > 0) {
+			    goal_time *= b_val(bp, VAL_PERCENT_AUTOCOLOR);
+			    goal_time /= 100;
+			    if (goal_time <= 0)
+				goal_time = 1;
+			}
+			if (bp->last_autocolor_time > goal_time) {
+			    int next_time = (int) bp->last_autocolor_time;
+			    int next_skip = ((next_time + goal_time - 1)
+					     / goal_time);
+
+			    /*
+			     * Make a local autocolor mode to show the user
+			     * what the actual interval is.
+			     */
+			    make_local_b_val(bp, VAL_AUTOCOLOR);
+			    set_b_val(bp, VAL_AUTOCOLOR, next_time);
+			    /*
+			     * Set counter to skip autocolor intervals.
+			     */
+			    bp->next_autocolor_time = next_skip;
+			}
+		    }
+		}
+	    }
+
+	    /*
+	     * Reset the state of buffers that were updated during this
+	     * interval.
+	     */
+	    for_each_visible_window(wp) {
+		bp = wp->w_bufp;
+		if (bp->next_autocolor_time < 0)
+		    bp->next_autocolor_time = 0;
+	    }
+
+	    update(FALSE);
+	}
+	TRACE((T_RETURN "\n"));
+    }
 #endif
 }
 
@@ -1643,12 +1857,13 @@ dirstack_extend(const char *dir, const char *fnname)
     if (dirs_idx >= dirs_len) {
 	if (dirs_len == 0) {
 	    dirs_len = 16;
-	    dirstack = castalloc(char *, dirs_len * sizeof(dirstack[0]));
+	    dirstack = castalloc(char *,
+				   ((unsigned) dirs_len * sizeof(dirstack[0])));
 	} else {
 	    dirs_len *= 2;
 	    dirstack = castrealloc(char *,
 				   dirstack,
-				   dirs_len * sizeof(dirstack[0]));
+				     ((unsigned) dirs_len * sizeof(dirstack[0])));
 	}
 	if (!dirstack)
 	    return (no_memory(fnname));
@@ -1709,7 +1924,7 @@ do_pushd(int uindx,		/* user-specified dirstack index */
     int rc;
     char oldcwd[NFILEN], *path;
 
-    strcpy(oldcwd, current_directory(TRUE));
+    vl_strncpy(oldcwd, current_directory(TRUE), sizeof(oldcwd));
 
     beginDisplay();
     if ((uindx == 0 && sign > 0) || (uindx == dirs_idx && sign < 0)) {
@@ -1756,6 +1971,7 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     long dirlocn;
     int failed, rc, sign;
+    int ch;
     char oldcwd[NFILEN], newcwd[NFILEN], *tmp, *cp;
     static TBUFF *last;
 
@@ -1780,7 +1996,7 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 
 	/* handle simple case first:  pushd without +|-n arg */
 	if (*cp != '+' && *cp != '-') {
-	    strcpy(oldcwd, current_directory(TRUE));
+	    vl_strncpy(oldcwd, current_directory(TRUE), sizeof(oldcwd));
 	    if ((rc = pushd_popd_set_dir(newcwd)) != TRUE)
 		return (rc);
 
@@ -1795,7 +2011,8 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 		sign = (*cp == '-') ? -1 : 1;
 		cp++;
 		rc = FALSE;
-		if (isDigit(*cp++)) {
+		ch = *cp++;
+		if (isDigit(ch)) {
 		    cp = skip_number(cp);
 		    if (*cp == '\0')
 			rc = TRUE;
@@ -1818,7 +2035,7 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 		    mlforce("[Pushd index out-of-range]");
 		    return (FALSE);
 		}
-		rc = do_pushd(dirlocn, sign);
+		rc = do_pushd((int) dirlocn, sign);
 	    }
 	}
     }
@@ -1837,6 +2054,7 @@ popd(int f GCC_UNUSED, int n GCC_UNUSED)
     long dirlocn;
     static TBUFF *last;
     int sign = 0, rc, failed;
+    int ch;
 
     rc = mlreply2("Directory stack entry: ", &last);
     if (rc == ABORT)
@@ -1870,7 +2088,8 @@ popd(int f GCC_UNUSED, int n GCC_UNUSED)
     if (*cp == '-' || *cp == '+') {
 	sign = (*cp == '-') ? -1 : 1;
 	cp++;
-	if (isDigit(*cp++)) {
+	ch = *cp++;
+	if (isDigit(ch)) {
 	    cp = skip_number(cp);
 	    if (*cp == '\0')
 		rc = TRUE;
@@ -1894,7 +2113,7 @@ popd(int f GCC_UNUSED, int n GCC_UNUSED)
 	mlforce("[Popd index out-of-range]");
 	return (FALSE);
     }
-    if ((rc = do_popd(dirlocn, sign)) == TRUE)
+    if ((rc = do_popd((int) dirlocn, sign)) == TRUE)
 	(void) vl_dirs(0, 0);
     return (rc);
 }
@@ -2033,7 +2252,7 @@ vl_dirs_add(int f GCC_UNUSED, int n GCC_UNUSED)
 	 * might be accessible via CDPATH
 	 */
 
-	strcpy(savedcwd, current_directory(TRUE));
+	vl_strncpy(savedcwd, current_directory(TRUE), sizeof(savedcwd));
 	dirs_add_active = TRUE;
 	if (set_directory(newdir)) {
 	    rc = dirstack_extend(current_directory(TRUE), "vl_dirs_add");
@@ -2081,7 +2300,7 @@ vl_atol(const char *str, int base, int *failed)
 	    if (errno == EDOM || errno == ERANGE) {
 		*failed = TRUE;
 	    } else {
-		if (*(prem = skip_blanks(prem)) != EOS) {
+		if (*skip_blanks(prem) != EOS) {
 		    *failed = TRUE;	/* trailing garbage */
 		}
 	    }
@@ -2121,7 +2340,7 @@ vl_atoul(const char *str, int base, int *failed)
 	    if (errno == EDOM || errno == ERANGE) {
 		*failed = TRUE;
 	    } else {
-		if (*(prem = skip_blanks(prem)) != EOS)
+		if (*skip_blanks(prem) != EOS)
 		    *failed = TRUE;	/* trailing garbage */
 	    }
 	}
@@ -2155,11 +2374,38 @@ vl_stricmp(const char *a, const char *b)
 	    ch2 = toUpper(ch2);
 	cmp = ch1 - ch2;
 	if (cmp != 0 || ch1 == EOS || ch2 == EOS)
-	    return cmp;
+	    break;
     }
+    return cmp;
 }
 #endif
 
+#if OPT_CASELESS
+#ifndef vl_strnicmp
+/*
+ * Compare two strings ignoring case
+ */
+int
+vl_strnicmp(const char *a, const char *b, size_t n)
+{
+    int ch1, ch2, cmp = 0;
+
+    while (n-- != 0) {
+	ch1 = CharOf(*a++);
+	ch2 = CharOf(*b++);
+	if (isLower(ch1))
+	    ch1 = toUpper(ch1);
+	if (isLower(ch2))
+	    ch2 = toUpper(ch2);
+	cmp = ch1 - ch2;
+	if (cmp != 0 || ch1 == EOS || ch2 == EOS)
+	    break;
+    }
+    return cmp;
+}
+#endif
+#endif /* OPT_CASELESS */
+
 /*
  * Format a visible representation of the given character, returns the buffer
  * address to simplify using this in formatting a whole buffer.
