@@ -3,7 +3,7 @@
  * and backward directions.
  *  heavily modified by Paul Fox, 1990
  *
- * $Header: /usr/build/vile/vile/RCS/search.c,v 1.145 2007/09/16 20:06:08 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/search.c,v 1.150 2010/05/03 10:56:36 tom Exp $
  *
  * original written Aug. 1986 by John M. Gamble, but I (pgf) have since
  * replaced his regex stuff with Henry Spencer's regexp package.
@@ -68,6 +68,8 @@ fsearch(int f, int n, int marking, int fromscreen)
     int didmark = FALSE;
     int didwrap;
 
+    assert(curwp != 0);
+
     if (f && n < 0)
 	return rsearch(f, -n, FALSE, FALSE);
 
@@ -92,6 +94,9 @@ fsearch(int f, int n, int marking, int fromscreen)
 
     ignorecase = window_b_val(curwp, MDIGNCASE);
 
+    if (curwp == 0)
+	return FALSE;
+
     curpos = DOT;
     scanboundry(wrapok, curpos, FORWARD);
     didwrap = FALSE;
@@ -158,6 +163,8 @@ forwhunt(int f, int n)
     MARK curpos;
     int didwrap;
 
+    assert(curwp != 0);
+
     wrapok = window_b_val(curwp, MDSWRAP);
 
     if (f && n < 0)		/* search backwards */
@@ -174,6 +181,9 @@ forwhunt(int f, int n)
 
     ignorecase = window_b_val(curwp, MDIGNCASE);
 
+    if (curwp == 0)
+	return FALSE;
+
     /* find n'th occurrence of pattern
      */
     curpos = DOT;
@@ -233,6 +243,8 @@ rsearch(int f, int n, int dummy GCC_UNUSED, int fromscreen)
     MARK curpos;
     int didwrap;
 
+    assert(curwp != 0);
+
     if (f && n < 0)		/* reverse direction */
 	return fsearch(f, -n, FALSE, fromscreen);
 
@@ -253,6 +265,9 @@ rsearch(int f, int n, int dummy GCC_UNUSED, int fromscreen)
 
     ignorecase = window_b_val(curwp, MDIGNCASE);
 
+    if (curwp == 0)
+	return FALSE;
+
     curpos = DOT;
     scanboundry(wrapok, DOT, REVERSE);
     didwrap = FALSE;
@@ -296,6 +311,8 @@ backhunt(int f, int n)
     MARK curpos;
     int didwrap;
 
+    assert(curwp != 0);
+
     wrapok = window_b_val(curwp, MDSWRAP);
 
     if (f && n < 0)		/* search forwards */
@@ -312,6 +329,9 @@ backhunt(int f, int n)
 
     ignorecase = window_b_val(curwp, MDIGNCASE);
 
+    if (curwp == 0)
+	return FALSE;
+
     /* find n'th occurrence of pattern
      */
     curpos = DOT;
@@ -493,8 +513,9 @@ scanner(
 		    returnCode(FALSE);
 		}
 	    } else if (llength(curpos.l) <= leftmargin
-		       || last < llength(curpos.l))
+		       || last < llength(curpos.l)) {
 		last--;
+	    }
 	    next = (C_NUM) (got - txt);
 	    if (next != last) {
 		DOT.l = curpos.l;
@@ -627,6 +648,8 @@ attrib_matches(void)
     REGIONSHAPE oregionshape = regionshape;
     VIDEO_ATTR vattr;
 
+    assert(curwp != 0);
+
     ignorecase = window_b_val(curwp, MDIGNCASE);
 
     if (!need_to_rehilite())
@@ -646,6 +669,9 @@ attrib_matches(void)
 
     (void) clear_match_attrs(TRUE, 1);
 
+    if (curwp == 0)
+	return;
+
     origdot = DOT;
     DOT.l = buf_head(curbp);
     DOT.o = 0;
@@ -684,7 +710,7 @@ attrib_matches(void)
 	    nextdot.o -= BytesBefore(nextdot.l, nextdot.o);
 
 	/* show highlighting from DOT to MK */
-	regionshape = EXACT;
+	regionshape = rgn_EXACT;
 	videoattribute |= VOWN_MATCHES;
 	status = attributeregion();
     } while (status == TRUE);
@@ -735,7 +761,7 @@ readpattern(const char *prompt,
     TRACE((T_CALLED "readpattern(%s, %s, %p, %d, %d)\n",
 	   prompt ? prompt : "",
 	   tb_visible(*apat),
-	   srchexpp,
+	   (void *) srchexpp,
 	   c,
 	   fromscreen));
 
@@ -827,14 +853,14 @@ movenext(MARK *pdot, int dir)
 	    curline = lforw(curline);	/* skip to next line */
 	    curoff = 0;
 	} else {
-	    curoff++;
+	    curoff += BytesAt(curline, curoff);
 	}
     } else {
 	if (curoff == 0) {
 	    curline = lback(curline);
 	    curoff = llength(curline);
 	} else {
-	    curoff--;
+	    curoff -= BytesBefore(curline, curoff);
 	}
     }
     pdot->l = curline;
