@@ -18,7 +18,7 @@
  * transferring the selection are not dealt with in this file.  Procedures
  * for dealing with the representation are maintained in this file.
  *
- * $Header: /usr/build/vile/vile/RCS/select.c,v 1.166 2007/09/16 20:05:13 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/select.c,v 1.177 2010/04/30 23:52:07 tom Exp $
  *
  */
 
@@ -30,6 +30,11 @@
 #include	<filters.h>
 #endif
 
+#ifdef WIN32
+#include	"w32vile.h"	/* for VL_ELAPSED */
+#undef WIN32_LEAN_AND_MEAN
+#endif
+
 #define BTN_BEGIN   1
 #define BTN_PASTE   2
 #define BTN_EXTEND  3
@@ -248,7 +253,7 @@ sel_begin(void)
 	startregion.ar_region.r_end = DOT;
     plus_region.o += BytesAt(plus_region.l, plus_region.o);
     startregion.ar_vattr = 0;
-    startregion.ar_shape = EXACT;
+    startregion.ar_shape = rgn_EXACT;
 #if OPT_HYPERTEXT
     startregion.ar_hypercmd = 0;
 #endif
@@ -271,6 +276,7 @@ dot_vs_mark(void)
 int
 sel_extend(int wiping, int include_dot)
 {
+    BUFFER *bp = curbp;
     REGIONSHAPE save_shape = regionshape;
     REGION a, b;
     WINDOW *wp;
@@ -307,7 +313,7 @@ sel_extend(int wiping, int include_dot)
      */
     working_dot = DOT;
     if (include_dot
-	&& (selregion.ar_shape == EXACT)
+	&& (selregion.ar_shape == rgn_EXACT)
 	&& dot_vs_mark() >= 0) {
 	if (samepoint(MK, orig_region)) {
 	    DOT.o += BytesAt(DOT.l, DOT.o);
@@ -316,7 +322,7 @@ sel_extend(int wiping, int include_dot)
 	    MK = orig_region;
 	}
     }
-    if (getregion(&a) == FALSE) {
+    if (getregion(bp, &a) == FALSE) {
 	return FALSE;
     }
     DOT = working_dot;
@@ -330,19 +336,19 @@ sel_extend(int wiping, int include_dot)
 	MK = selregion.ar_region.r_end;
 
     if (include_dot) {
-	if (selregion.ar_shape == EXACT) {
+	if (selregion.ar_shape == rgn_EXACT) {
 	    if (dot_vs_mark() <= 0) {
 		if (samepoint(MK, orig_region))
 		    MK.o += BytesAt(MK.l, MK.o);
 	    }
-	} else if (selregion.ar_shape == RECTANGLE) {
+	} else if (selregion.ar_shape == rgn_RECTANGLE) {
 	    if (samepoint(MK, DOT)) {	/* avoid making empty-region */
 		MK = orig_region;
 		DOT = plus_region;
 	    }
 	}
     }
-    if (getregion(&b) == FALSE) {
+    if (getregion(bp, &b) == FALSE) {
 	return FALSE;
     }
 
@@ -351,7 +357,7 @@ sel_extend(int wiping, int include_dot)
      * special case where we've extended one to the right to include the
      * right endpoint of the region.
      *
-     * For EXACT selections, setting 'whichend' to ORIG_FIXED means that
+     * For rgn_EXACT selections, setting 'whichend' to ORIG_FIXED means that
      * we're selecting from the anchor point right/down.  Conversely,
      * setting it to END_FIXED means that we selecting left/up.
      *
@@ -368,7 +374,7 @@ sel_extend(int wiping, int include_dot)
 	whichend = ORIG_FIXED;
 	selregion.ar_region = a;
     } else {
-	if (selregion.ar_shape == RECTANGLE) {
+	if (selregion.ar_shape == rgn_RECTANGLE) {
 	    if (dot_vs_mark() < 0)
 		whichend = END_FIXED;
 	    else
@@ -480,7 +486,7 @@ sel_all(int f GCC_UNUSED, int n GCC_UNUSED)
     sel_begin();
     gotoeob(0, 0);
     gotoeol(0, 0);
-    (void) sel_setshape(EXACT);
+    (void) sel_setshape(rgn_EXACT);
     rc = sel_extend(TRUE, TRUE);
     DOT = savedot;
     if (rc)
@@ -552,6 +558,7 @@ sel_setshape(REGIONSHAPE shape)
 static REGION *
 extended_region(void)
 {
+    BUFFER *bp = curbp;
     REGION *rp = NULL;
     static REGION a, b;
     MARK savemark;
@@ -560,13 +567,13 @@ extended_region(void)
     regionshape = selregion.ar_shape;
     MK = selregion.ar_region.r_orig;
     DOT.o += BytesAt(DOT.l, DOT.o);
-    if (getregion(&a) == TRUE) {
+    if (getregion(bp, &a) == TRUE) {
 	DOT.o -= BytesBefore(DOT.l, DOT.o);
 	MK = selregion.ar_region.r_end;
-	if (regionshape == FULLLINE)
+	if (regionshape == rgn_FULLLINE)
 	    MK.l = lback(MK.l);
 	/* region b is to the end of the selection */
-	if (getregion(&b) == TRUE) {
+	if (getregion(bp, &b) == TRUE) {
 	    /* if a is bigger, it's the one we want */
 	    if (a.r_size > b.r_size)
 		rp = &a;
@@ -630,6 +637,7 @@ sel_motion(int f GCC_UNUSED, int n GCC_UNUSED)
 static int
 selectregion(void)
 {
+    BUFFER *bp = curbp;
     int status;
     REGION region;
     MARK savedot;
@@ -649,7 +657,7 @@ selectregion(void)
     MK = savemark;
     if (status != TRUE)
 	return status;
-    if (hadregion || ((status = getregion(&region)) == TRUE)) {
+    if (hadregion || ((status = getregion(bp, &region)) == TRUE)) {
 	detach_attrib(startbufp, &startregion);
 	detach_attrib(selbufp, &selregion);
 	selbufp = curbp;
@@ -724,10 +732,10 @@ on_double_click(void)
     TRACE(("MOUSE double-click DOT %d.%d\n", line_no(curbp, DOT.l), DOT.o));
     sel_release();
     if (!is_at_end_of_line(DOT)
-	&& !isSpace(char_at(DOT))) {
+	&& !isSpace(CharAtDot())) {
 	while (DOT.o > 0) {
 	    DOT.o -= BytesBefore(DOT.l, DOT.o);
-	    if (isSpace(char_at(DOT))) {
+	    if (isSpace(CharAtDot())) {
 		DOT.o += BytesAt(DOT.l, DOT.o);
 		break;
 	    }
@@ -737,7 +745,7 @@ on_double_click(void)
 	while (!is_at_end_of_line(DOT)) {
 	    DOT.o += BytesAt(DOT.l, DOT.o);
 	    if (is_at_end_of_line(DOT)
-		|| isSpace(char_at(DOT))) {
+		|| isSpace(CharAtDot())) {
 		DOT.o -= BytesBefore(DOT.l, DOT.o);
 		break;
 	    }
@@ -881,7 +889,7 @@ on_mouse_click(int button, int y, int x)
 		}
 		do_sweep(SORTOFTRUE);
 		(void) sel_begin();
-		(void) sel_setshape(EXACT);
+		(void) sel_setshape(rgn_EXACT);
 		(void) setcursor(y, x);
 		status = multimotion(TRUE, 1);
 		TRACE(("MOUSE end multimotion after button%d-up\n", button));
@@ -936,11 +944,11 @@ multimotion(int f, int n)
     n = need_at_least(f, n, 1);
 
     if (n == 3)
-	regionshape = RECTANGLE;
+	regionshape = rgn_RECTANGLE;
     else if (n == 2)
-	regionshape = FULLLINE;
+	regionshape = rgn_FULLLINE;
     else
-	regionshape = EXACT;
+	regionshape = rgn_EXACT;
     shape = regionshape;
 
     sweephack = FALSE;
@@ -965,7 +973,7 @@ multimotion(int f, int n)
 	    do_sweep(TRUE);
 	}
 	sweepmsg("Begin cursor sweep...");
-	sel_extend(TRUE, (regionshape != RECTANGLE && sweephack));
+	sel_extend(TRUE, (regionshape != rgn_RECTANGLE && sweephack));
 	savedot = MK;
 	TRACE(("MOUSE BEGIN DOT: %d.%d MK %d.%d\n",
 	       line_no(curbp, DOT.l), DOT.o,
@@ -1067,7 +1075,7 @@ multimotion(int f, int n)
 		   selection (unless it's a rectangle, in
 		   which case it's taken care of elsewhere)
 		 */
-		sel_extend(TRUE, (regionshape != RECTANGLE &&
+		sel_extend(TRUE, (regionshape != rgn_RECTANGLE &&
 				  sweephack));
 		break;
 
@@ -1094,7 +1102,7 @@ multimotion(int f, int n)
     TRACE(("MOUSE SAVE DOT: %d.%d MK %d.%d\n",
 	   line_no(curbp, DOT.l), DOT.o,
 	   line_no(curbp, MK.l), MK.o));
-    if ((regionshape != RECTANGLE) && sweephack) {
+    if ((regionshape != rgn_RECTANGLE) && sweephack) {
 	if (dot_vs_mark() < 0)
 	    MK.o += BytesAt(MK.l, MK.o);
 	else
@@ -1150,20 +1158,22 @@ apply_attribute(void)
 int
 attributeregion(void)
 {
+    BUFFER *bp = curbp;
     int status;
     REGION region;
     AREGION *arp;
 
-    if ((status = getregion(&region)) == TRUE) {
+    if ((status = getregion(bp, &region)) == TRUE) {
 	if (apply_attribute()) {
-	    if (add_line_attrib(curbp, &region, regionshape, videoattribute,
+	    if (add_line_attrib(bp, &region, regionshape, videoattribute,
 #if OPT_HYPERTEXT
 				hypercmd
 #else
 				0
 #endif
-		))
+		)) {
 		return TRUE;
+	    }
 
 	    /* add new attribute-region */
 	    if ((arp = alloc_AREGION()) == NULL) {
@@ -1174,15 +1184,24 @@ attributeregion(void)
 	    arp->ar_shape = regionshape;
 #if OPT_HYPERTEXT
 	    arp->ar_hypercmd = 0;
-	    if (tb_length(hypercmd)) {
+	    if (tb_length(hypercmd) && *tb_values(hypercmd)) {
+#if OPT_EXTRA_COLOR
+		if (tb_length(hypercmd) > 4
+		    && !memcmp(tb_values(hypercmd), "view ", 4)) {
+		    int *newVideo = lookup_extra_color(XCOLOR_HYPERTEXT);
+		    if (!isEmpty(newVideo)) {
+			arp->ar_vattr = (VIDEO_ATTR) * newVideo;
+		    }
+		}
+#endif
 		arp->ar_hypercmd = strmalloc(tb_values(hypercmd));
 		tb_init(&hypercmd, 0);
 	    }
 #endif
-	    attach_attrib(curbp, arp);
+	    attach_attrib(bp, arp);
 	} else {		/* purge attributes in this region */
-	    L_NUM rls = line_no(curbp, region.r_orig.l);
-	    L_NUM rle = line_no(curbp, region.r_end.l);
+	    L_NUM rls = line_no(bp, region.r_orig.l);
+	    L_NUM rle = line_no(bp, region.r_end.l);
 	    C_NUM ros = region.r_orig.o;
 	    C_NUM roe = region.r_end.o;
 	    AREGION *p, *q, *n;
@@ -1190,9 +1209,9 @@ attributeregion(void)
 
 	    owner = VOWNER(videoattribute);
 
-	    purge_line_attribs(curbp, &region, regionshape, owner);
+	    purge_line_attribs(bp, &region, regionshape, owner);
 
-	    for (p = curbp->b_attribs; p != 0; p = q) {
+	    for (p = bp->b_attribs; p != 0; p = q) {
 		L_NUM pls, ple;
 		C_NUM pos, poe;
 
@@ -1204,14 +1223,14 @@ attributeregion(void)
 		if (owner != 0 && owner != VOWNER(p->ar_vattr))
 		    continue;
 
-		pls = line_no(curbp, p->ar_region.r_orig.l);
-		ple = line_no(curbp, p->ar_region.r_end.l);
+		pls = line_no(bp, p->ar_region.r_orig.l);
+		ple = line_no(bp, p->ar_region.r_end.l);
 		pos = p->ar_region.r_orig.o;
 		poe = p->ar_region.r_end.o;
 
 		/* Earlier the overlapping region check was made based only
 		 * on line numbers and so was right only for FULLINES shape
-		 * changed it to be correct for EXACT and RECTANGLE also
+		 * changed it to be correct for rgn_EXACT and rgn_RECTANGLE also
 		 * -kuntal 9/13/98
 		 */
 		/*
@@ -1222,18 +1241,18 @@ attributeregion(void)
 		if (ple < rls || pls > rle)
 		    continue;
 		/*
-		 * for EXACT 'p' region
+		 * for rgn_EXACT 'p' region
 		 */
-		if (p->ar_shape == EXACT) {
+		if (p->ar_shape == rgn_EXACT) {
 		    if (ple == rls && poe - 1 < ros)
 			continue;
 		    if (pls == rle && pos > roe)
 			continue;
 		}
 		/*
-		 * for RECTANGLE 'p' region
+		 * for rgn_RECTANGLE 'p' region
 		 */
-		if (p->ar_shape == RECTANGLE)
+		if (p->ar_shape == rgn_RECTANGLE)
 		    if (poe < ros || pos > roe)
 			continue;
 
@@ -1244,13 +1263,13 @@ attributeregion(void)
 
 		/*
 		 * we take care of this fix easily as long as neither of
-		 * 'p' or 'region' are RECTANGLE. we will need to create
+		 * 'p' or 'region' are rgn_RECTANGLE. we will need to create
 		 * at the most one new region in case 'region' is
 		 * completely contained within 'p'
 		 */
-		if (p->ar_shape != RECTANGLE && regionshape != RECTANGLE) {
+		if (p->ar_shape != rgn_RECTANGLE && regionshape != rgn_RECTANGLE) {
 		    if ((rls > pls) || (rls == pls && ros > pos)) {
-			p->ar_shape = EXACT;
+			p->ar_shape = rgn_EXACT;
 			if ((rle < ple) || (rle == ple && roe < poe)) {
 			    /* open a new region */
 			    if ((n = alloc_AREGION()) == NULL) {
@@ -1264,7 +1283,7 @@ attributeregion(void)
 #endif
 			    n->ar_region.r_orig.l = (region.r_end.l);
 			    n->ar_region.r_orig.o = (region.r_end.o);
-			    attach_attrib(curbp, n);
+			    attach_attrib(bp, n);
 			}
 			p->ar_region.r_end.l = (region.r_orig.l);
 			p->ar_region.r_end.o = (region.r_orig.o);
@@ -1278,7 +1297,7 @@ attributeregion(void)
 		    }
 		}
 
-		free_attrib(curbp, p);
+		free_attrib(bp, p);
 	    }
 	}
     }
@@ -1294,7 +1313,7 @@ attributeregion_in_region(REGION * rp,
     haveregion = rp;
     DOT = rp->r_orig;
     MK = rp->r_end;
-    if (shape == FULLLINE)
+    if (shape == rgn_FULLLINE)
 	MK.l = lback(MK.l);
     regionshape = shape;	/* Not that the following actually cares */
     videoattribute = vattr;
@@ -1472,7 +1491,7 @@ attribute_cntl_a_seqs_in_region(REGION * rp, REGIONSHAPE shape)
     haveregion = rp;
     DOT = rp->r_orig;
     MK = rp->r_end;
-    if (shape == FULLLINE)
+    if (shape == rgn_FULLLINE)
 	MK.l = lback(MK.l);
     regionshape = shape;	/* Not that the following actually cares */
     return attribute_cntl_a_sequences();
@@ -1484,11 +1503,12 @@ attribute_cntl_a_seqs_in_region(REGION * rp, REGIONSHAPE shape)
 LINE *
 setup_region(void)
 {
+    BUFFER *bp = curbp;
     LINE *pastline;		/* pointer to line just past EOP */
 
     if (!sameline(MK, DOT)) {
 	REGION region;
-	if (getregion(&region) != TRUE)
+	if (getregion(bp, &region) != TRUE)
 	    return 0;
 	if (sameline(region.r_orig, MK))
 	    swapmark();
@@ -1497,7 +1517,7 @@ setup_region(void)
     if (pastline != win_head(curwp))
 	pastline = lforw(pastline);
     DOT.o = 0;
-    regionshape = EXACT;
+    regionshape = rgn_EXACT;
 
     return pastline;
 }
@@ -1654,14 +1674,14 @@ set_mark_after(int count, int rslen)
 static int
 attribute_cntl_a_sequences(void)
 {
+    BUFFER *bp = curbp;
     LINE *pastline;
     C_NUM offset;		/* offset in cur line of place to attribute */
     int count;
 
 #if EFFICIENCY_HACK
-    AREGION *orig_attribs;
+    AREGION *orig_attribs = bp->b_attribs;
     AREGION *new_attribs;
-    orig_attribs = new_attribs = curbp->b_attribs;
 #endif
 
     if ((pastline = setup_region()) == 0)
@@ -1671,20 +1691,20 @@ attribute_cntl_a_sequences(void)
 	if (interrupted())
 	    return FALSE;
 	while (DOT.o < llength(DOT.l)) {
-	    if (char_at(DOT) == CONTROL_A) {
+	    if (CharAtDot() == CONTROL_A) {
 		offset = decode_attribute(lvalue(DOT.l), llength(DOT.l),
 					  DOT.o, &count);
 		if (offset > DOT.o) {
 #if EFFICIENCY_HACK
-		    new_attribs = curbp->b_attribs;
-		    curbp->b_attribs = orig_attribs;
+		    new_attribs = bp->b_attribs;
+		    bp->b_attribs = orig_attribs;
 		    ldel_bytes((B_COUNT) (offset - DOT.o), FALSE);
-		    curbp->b_attribs = new_attribs;
+		    bp->b_attribs = new_attribs;
 #else
 		    ldel_bytes((B_COUNT) (offset - DOT.o), FALSE);
 #endif
 		}
-		set_mark_after(count, len_record_sep(curbp));
+		set_mark_after(count, len_record_sep(bp));
 		if (apply_attribute())
 		    (void) attributeregion();
 	    } else {
@@ -1715,6 +1735,7 @@ discard_syntax_highlighting(void)
 static int
 attribute_from_filter(void)
 {
+    BUFFER *bp = curbp;
     LINE *pastline;
     int skip;
     size_t nbytes;
@@ -1723,11 +1744,12 @@ attribute_from_filter(void)
     int result = TRUE;
     int drained = FALSE;
 
+    TRACE((T_CALLED "attribute_from_filter\n"));
     if ((pastline = setup_region()) == 0) {
 	result = FALSE;
 
 #ifdef MDHILITE
-    } else if (!b_val(curbp, MDHILITE)) {
+    } else if (!b_val(bp, MDHILITE)) {
 	discard_syntax_highlighting();
 #endif
 
@@ -1774,13 +1796,13 @@ attribute_from_filter(void)
 	attach_attrib(selbufp, &selregion);
 	attach_attrib(startbufp, &startregion);
 #if OPT_HILITEMATCH
-	if (curbp->b_highlight & HILITE_ON) {
-	    curbp->b_highlight |= HILITE_DIRTY;
+	if (bp->b_highlight & HILITE_ON) {
+	    bp->b_highlight |= HILITE_DIRTY;
 	    attrib_matches();
 	}
 #endif
     }
-    return result;
+    returnCode(result);
 }
 
 int
@@ -1797,26 +1819,31 @@ operattrfilter(int f, int n)
 static int
 attribute_directly(void)
 {
+    BUFFER *bp = curbp;
     int code = FALSE;
 
 #if OPT_MAJORMODE
-    if (valid_buffer(curbp)) {
+    if (valid_buffer(bp)) {
+#if OPT_AUTOCOLOR
+	VL_ELAPSED begin_time;
+	(void) vl_elapsed(&begin_time, TRUE);
+#endif
 	discard_syntax_highlighting();
-	if (b_val(curbp, MDHILITE)) {
+	if (b_val(bp, MDHILITE)) {
 	    char *filtername = 0;
 	    TBUFF *token = 0;
 
 	    if (clexec || isnamedcmd)
-		filtername = mac_tokval(&token);
+		filtername = mac_unquotedarg(&token);
 
 	    if (filtername == 0
-		&& curbp->majr != 0)
-		filtername = curbp->majr->shortname;
+		&& bp->majr != 0)
+		filtername = bp->majr->shortname;
 
 	    if (filtername != 0
 		&& flt_start(filtername)) {
 		TRACE(("attribute_directly(%s) using %s\n",
-		       curbp->b_bname,
+		       bp->b_bname,
 		       filtername));
 		flt_finish();
 		code = TRUE;
@@ -1826,10 +1853,14 @@ attribute_directly(void)
 	attach_attrib(selbufp, &selregion);
 	attach_attrib(startbufp, &startregion);
 #if OPT_HILITEMATCH
-	if (curbp->b_highlight & HILITE_ON) {
-	    curbp->b_highlight |= HILITE_DIRTY;
+	if (bp->b_highlight & HILITE_ON) {
+	    bp->b_highlight |= HILITE_DIRTY;
 	    attrib_matches();
 	}
+#endif
+#if OPT_AUTOCOLOR
+	bp->last_autocolor_time = vl_elapsed(&begin_time, FALSE);
+	bp->next_autocolor_time = 0;
 #endif
     }
 #endif
@@ -2019,12 +2050,14 @@ add_line_attrib(BUFFER *bp, REGION * rp, REGIONSHAPE rs, VIDEO_ATTR vattr,
     int last;
 
     if (rp->r_orig.l != rp->r_end.l	/* must be confined to one line */
-	|| rs != EXACT		/* must be an exact region */
+	|| rs != rgn_EXACT	/* must be an exact region */
 	|| (hypercmdp && tb_length(hypercmdp) != 0)
     /* can't be a hypertext command */
 	|| vattr == 0		/* can't be normal */
-	|| (vattr & VASEL) != 0)	/* can't be a selection */
+	|| ((UCHAR) vattr) != vattr
+	|| (vattr & VASEL) != 0) {	/* can't be a selection */
 	return FALSE;
+    }
 
     beginDisplay();
 
@@ -2102,10 +2135,10 @@ purge_line_attribs(BUFFER *bp, REGION * rp, REGIONSHAPE rs, int owner)
 		    break;	/* at end of attrs */
 		if (lp->l_attrs[i] == 1)
 		    continue;	/* normal, so proceed to next one */
-		if (rs != FULLLINE) {
-		    if ((rs == RECTANGLE || lp == ls) && i < os)
+		if (rs != rgn_FULLLINE) {
+		    if ((rs == rgn_RECTANGLE || lp == ls) && i < os)
 			continue;
-		    if ((rs == RECTANGLE || lp == le) && i >= oe)
+		    if ((rs == rgn_RECTANGLE || lp == le) && i >= oe)
 			break;
 		}
 		if (owner == 0
