@@ -3,13 +3,14 @@
  *	for getting and setting the values of the vile state variables,
  *	plus helper utility functions.
  *
- * $Header: /usr/build/vile/vile/RCS/statevar.c,v 1.121 2008/04/15 22:49:06 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/statevar.c,v 1.151 2010/06/09 20:54:25 tom Exp $
  */
 
-#include	"estruct.h"
-#include	"edef.h"
-#include	"nevars.h"
-#include	"patchlev.h"
+#include	<estruct.h>
+#include	<edef.h>
+#include	<nevars.h>
+#include	<nefsms.h>
+#include	<patchlev.h>
 
 #if SYS_WINNT
 #include	<process.h>
@@ -39,12 +40,15 @@ static char *cdpath;		/* $CDPATH environment is "$cdpath" state variable. */
 static const char *
 DftEnv(const char *name, const char *dft)
 {
+    const char *result;
+
 #if OPT_EVAL && OPT_SHELL
     name = safe_getenv(name);
-    return isEmpty(name) ? dft : name;
+    result = isEmpty(name) ? dft : name;
 #else
-    return dft;
+    result = dft;
 #endif
+    return result;
 }
 
 static void
@@ -104,6 +108,19 @@ any_ro_INT(TBUFF **rp, const char *vp, int value)
     }
 }
 
+static int
+any_ro_ULONG(TBUFF **rp, const char *vp, ULONG value)
+{
+    if (rp) {
+	render_ulong(rp, value);
+	return TRUE;
+    } else if (vp) {
+	return ABORT;		/* read-only */
+    } else {
+	return FALSE;
+    }
+}
+
 static int
 any_rw_INT(TBUFF **rp, const char *vp, int *value)
 {
@@ -260,16 +277,18 @@ any_HOOK(TBUFF **rp, const char *vp, HOOK * hook)
 const char *
 safe_getenv(const char *s)
 {
+    const char *result;
 #if OPT_EVAL && OPT_SHELL
     char *v = getenv(s);
 #if defined(_WIN32)
     if (v == 0)
 	v = vile_getenv(s);
 #endif
-    return NONNULL(v);
+    result = NONNULL(v);
 #else
-    return "";
+    result = "";
 #endif
+    return result;
 }
 
 #if OPT_EVAL && OPT_SHELL
@@ -312,19 +331,25 @@ get_findpath(void)
 #endif
 
 #if OPT_EVAL && DISP_X11 && OPT_SHELL
+#ifndef DEFAULT_XSHELL
+#define DEFAULT_XSHELL "xterm"
+#endif
 char *
 get_xshell(void)
 {
     if (x_shell == 0)
-	SetEnv(&x_shell, DftEnv("XSHELL", "xterm"));
+	SetEnv(&x_shell, DftEnv("XSHELL", DEFAULT_XSHELL));
     return x_shell;
 }
 
+#ifndef DEFAULT_XSHELLFLAGS
+#define DEFAULT_XSHELLFLAGS "-e"
+#endif
 char *
 get_xshellflags(void)
 {
     if (x_shellflags == 0)
-	SetEnv(&x_shellflags, DftEnv("XSHELLFLAGS", "-e"));
+	SetEnv(&x_shellflags, DftEnv("XSHELLFLAGS", DEFAULT_XSHELLFLAGS));
     return x_shellflags;
 }
 
@@ -347,16 +372,15 @@ get_directory(void)
 }
 #endif
 
-/* access the default kill buffer */
-static void
-getkill(TBUFF **rp)
+static KILLREG *
+default_kill(void)
 {
-    tb_init(rp, EOS);
+    KILLREG *result = 0;
     if (kbs[0].kbufh != 0) {
 	int n = index2ukb(0);
-	tb_bappend(rp, (char *) (kbs[n].kbufh->d_chunk), kbs[n].kused);
+	result = kbs + n;
     }
-    tb_append(rp, EOS);
+    return result;
 }
 
 #if OPT_PATHLOOKUP
@@ -396,15 +420,36 @@ cfgopts(void)
 #if SYS_WINNT && defined(VILE_OLE)
 	"oleauto",
 #endif
+#if OPT_HYPERTEXT
+	"hypertext",
+#endif
 #if OPT_LOCALE
 	"locale",
 #endif
+#if OPT_ICONV_FUNCS
+	"iconv",
+#endif
+#if OPT_MULTIBYTE
+	"multibyte",
+#endif
 #if OPT_PERL
 	"perl",
 #endif
+#if DISP_ANSI
+	"ansi",
+#endif
+#if DISP_BORLAND
+	"borland",
+#endif
 #if DISP_CURSES
 	"curses",
 #endif
+#if DISP_NTCONS
+	"ntcons",
+#endif
+#if DISP_NTWIN
+	"ntwin",
+#endif
 #if DISP_TERMCAP
 # if USE_TERMINFO
 	"terminfo",
@@ -412,6 +457,12 @@ cfgopts(void)
 	"termcap",
 # endif
 #endif
+#if DISP_VIO
+	"os2vio",
+#endif
+#if DISP_VMSVT
+	"vmsvt",
+#endif
 #if DISP_X11
 # if OL_WIDGETS
 	"openlook",
@@ -472,10 +523,22 @@ var_BCHARS(TBUFF **rp, const char *vp)
 {
     bsizes(curbp);
     return (valid_buffer(curbp)
-	    ? any_ro_INT(rp, vp, curbp->b_bytecount)
+	    ? any_ro_ULONG(rp, vp, curbp->b_bytecount)
 	    : FALSE);
 }
 
+int
+var_BCHANGED(TBUFF **rp, const char *vp)
+{
+    return any_ro_BOOL(rp, vp, valid_buffer(curbp) && b_is_changed(curbp));
+}
+
+int
+var_FCHANGED(TBUFF **rp, const char *vp)
+{
+    return any_ro_BOOL(rp, vp, eval_fchanged(curbp));
+}
+
 int
 var_BFLAGS(TBUFF **rp, const char *vp)
 {
@@ -549,13 +612,17 @@ var_BUFHOOK(TBUFF **rp, const char *vp)
 int
 var_BUF_FNAME_EXPR(TBUFF **rp, const char *vp)
 {
-    return any_ro_STR(rp, vp, get_buf_fname_expr(curbp)->pat);
+    return any_ro_STR(rp, vp, (valid_buffer(curbp)
+			       ? get_buf_fname_expr(curbp)->pat
+			       : ""));
 }
 
 int
 var_BUFNAME(TBUFF **rp, const char *vp)
 {
-    return any_REGEX_MATCH(rp, vp, b_val_rexp(curbp, VAL_BUFNAME_EXPR));
+    return any_REGEX_MATCH(rp, vp, (valid_buffer(curbp)
+				    ? b_val_rexp(curbp, VAL_BUFNAME_EXPR)
+				    : (REGEXVAL *) 0));
 }
 #endif
 
@@ -563,7 +630,7 @@ int
 var_CBUFNAME(TBUFF **rp, const char *vp)
 {
     if (rp) {
-	tb_scopy(rp, curbp->b_bname);
+	tb_scopy(rp, valid_buffer(curbp) ? curbp->b_bname : "");
 	return TRUE;
     } else if (vp) {
 	if (valid_buffer(curbp)) {
@@ -580,7 +647,7 @@ int
 var_BWINDOWS(TBUFF **rp, const char *vp)
 {
     return (valid_buffer(curbp)
-	    ? any_ro_INT(rp, vp, curbp->b_nwnd)
+	    ? any_ro_INT(rp, vp, (int) curbp->b_nwnd)
 	    : FALSE);
 }
 
@@ -624,7 +691,7 @@ var_CHAR(TBUFF **rp, const char *vp)
 
     if (rp) {
 	if (valid_buffer(curbp) && !is_empty_buf(curbp)) {
-	    render_int(rp, char_at_mark(DOT));
+	    render_int(rp, CharAtDot());
 	} else {
 	    tb_scopy(rp, error_val);
 	}
@@ -669,7 +736,7 @@ int
 var_CURCHAR(TBUFF **rp, const char *vp)
 {
     if (rp) {
-	render_int(rp, vl_getcchar() + 1);
+	render_ulong(rp, vl_getcchar() + 1);
 	return TRUE;
     } else if (vp && valid_buffer(curbp)) {
 	return gotochr(TRUE, strtol(vp, 0, 0));
@@ -761,6 +828,14 @@ var_CMD_COUNT(TBUFF **rp, const char *vp)
     return any_ro_INT(rp, vp, cmd_count);
 }
 
+#if OPT_PROCEDURES
+int
+var_CMD_MOTION(TBUFF **rp, const char *vp)
+{
+    return any_ro_STR(rp, vp, fnc2engl(cmd_motion));
+}
+#endif
+
 int
 var_DISCMD(TBUFF **rp, const char *vp)
 {
@@ -917,32 +992,31 @@ var_FINDCMD(TBUFF **rp, const char *vp)
 }
 #endif
 
-#if DISP_X11||DISP_NTWIN
 int
 var_FONT(TBUFF **rp, const char *vp)
 {
+    int result = FALSE;
+
 #if SYS_WINNT && DISP_NTWIN
     if (rp) {
 	tb_scopy(rp, ntwinio_current_font());
-	return TRUE;
+	result = TRUE;
     } else if (vp) {
-	return ntwinio_font_frm_str(vp, FALSE);
-    } else {
-	return FALSE;
+	result = ntwinio_font_frm_str(vp, FALSE);
     }
-#endif
-#if DISP_X11
+#elif DISP_X11
     if (rp) {
 	tb_scopy(rp, x_current_fontname());
-	return TRUE;
+	result = TRUE;
     } else if (vp) {
-	return x_setfont(vp);
-    } else {
-	return FALSE;
+	result = x_setfont(vp);
     }
+#else
+    result = any_ro_STR(rp, vp, "fixed");
 #endif
+
+    return result;
 }
-#endif
 
 int
 var_FWD_SEARCH(TBUFF **rp, const char *vp)
@@ -976,7 +1050,9 @@ var_ICONNAME(TBUFF **rp, const char *vp)
 int
 var_IDENTIFIER(TBUFF **rp, const char *vp)
 {
-    return any_REGEX_MATCH(rp, vp, b_val_rexp(curbp, VAL_IDENTIFIER_EXPR));
+    return any_REGEX_MATCH(rp, vp, (valid_buffer(curbp)
+				    ? b_val_rexp(curbp, VAL_IDENTIFIER_EXPR)
+				    : (REGEXVAL *) 0));
 }
 #endif
 
@@ -997,7 +1073,28 @@ int
 var_KILL(TBUFF **rp, const char *vp)
 {
     if (rp) {
-	getkill(rp);
+	KILLREG *kr = default_kill();
+	KILL *kb;
+	int limit = kill_limit;
+	int used;
+
+	tb_init(rp, EOS);
+	if ((kr != 0) && (kb = kr->kbufh) != 0) {
+	    while (kb->d_next != 0) {
+		if ((used = KBLOCK) > limit)
+		    used = limit;
+		tb_bappend(rp, (char *) (kb->d_chunk), (size_t) used);
+		if ((limit -= used) <= 0)
+		    break;
+		kb = kb->d_next;
+	    }
+	    if (limit > 0) {
+		if ((used = (int) kr->kused) > limit)
+		    used = limit;
+		tb_bappend(rp, (char *) (kb->d_chunk), (size_t) used);
+	    }
+	}
+	tb_append(rp, EOS);
 	return TRUE;
     } else if (vp) {
 	return ABORT;		/* read-only */
@@ -1006,6 +1103,36 @@ var_KILL(TBUFF **rp, const char *vp)
     }
 }
 
+int
+var_KILL_SIZE(TBUFF **rp, const char *vp)
+{
+    if (rp) {
+	KILLREG *kr = default_kill();
+	KILL *kb;
+	int result = 0;
+	if ((kr != 0) && (kb = kr->kbufh) != 0) {
+	    while (kb->d_next != 0) {
+		result += KBLOCK;
+		kb = kb->d_next;
+	    }
+	    result += (int) kr->kused;
+	    return any_ro_INT(rp, vp, result);
+	} else {
+	    return FALSE;
+	}
+    } else if (vp) {
+	return ABORT;		/* read-only */
+    } else {
+	return FALSE;
+    }
+}
+
+int
+var_KILL_LIMIT(TBUFF **rp, const char *vp)
+{
+    return any_rw_INT(rp, vp, &kill_limit);
+}
+
 int
 var_LASTKEY(TBUFF **rp, const char *vp)
 {
@@ -1079,6 +1206,14 @@ var_MAJORMODEHOOK(TBUFF **rp, const char *vp)
 }
 #endif
 
+#if OPT_MODELINE
+int
+var_MODELINE_LIMIT(TBUFF **rp, const char *vp)
+{
+    return any_rw_INT(rp, vp, &modeline_limit);
+}
+#endif
+
 int
 var_MATCH(TBUFF **rp, const char *vp)
 {
@@ -1181,19 +1316,20 @@ var_OS(TBUFF **rp, const char *vp)
 int
 var_PAGELEN(TBUFF **rp, const char *vp)
 {
+    int result = FALSE;
+
     if (rp) {
 	render_int(rp, term.rows);
-	return TRUE;
+	result = TRUE;
     } else if (vp) {
 #if DISP_X11 || DISP_NTWIN
 	gui_resize(term.cols, strtol(vp, 0, 0));
-	return TRUE;
+	result = TRUE;
 #else
-	return newlength(TRUE, strtol(vp, 0, 0));
+	result = newlength(TRUE, strtol(vp, 0, 0));
 #endif
-    } else {
-	return FALSE;
     }
+    return result;
 }
 
 /* separator for lists of pathnames */
@@ -1252,19 +1388,20 @@ var_POSFORMAT(TBUFF **rp, const char *vp)
 int
 var_CURWIDTH(TBUFF **rp, const char *vp)
 {
+    int result = FALSE;
+
     if (rp) {
 	render_int(rp, term.cols);
-	return TRUE;
+	result = TRUE;
     } else if (vp) {
 #if DISP_X11 || DISP_NTWIN
 	gui_resize(strtol(vp, 0, 0), term.rows);
-	return TRUE;
+	result = TRUE;
 #else
-	return newwidth(TRUE, strtol(vp, 0, 0));
+	result = newwidth(TRUE, strtol(vp, 0, 0));
 #endif
-    } else {
-	return FALSE;
     }
+    return result;
 }
 
 int
@@ -1292,7 +1429,9 @@ var_PATCHLEVEL(TBUFF **rp, const char *vp)
 int
 var_PATHNAME(TBUFF **rp, const char *vp)
 {
-    return any_REGEX_MATCH(rp, vp, b_val_rexp(curbp, VAL_PATHNAME_EXPR));
+    return any_REGEX_MATCH(rp, vp, (valid_buffer(curbp)
+				    ? b_val_rexp(curbp, VAL_PATHNAME_EXPR)
+				    : (REGEXVAL *) 0));
 }
 #endif
 
@@ -1401,7 +1540,7 @@ var_SEED(TBUFF **rp, const char *vp)
 	return TRUE;
     } else if (vp) {
 	seed = strtol(vp, 0, 0);
-	srand(seed);
+	srand((UINT) seed);
 	return TRUE;
     } else {
 	return FALSE;
@@ -1482,8 +1621,10 @@ int
 var_BUF_ENCODING(TBUFF **rp, const char *vp GCC_UNUSED)
 {
     if (rp) {
-	int code = b_val(curbp, VAL_FILE_ENCODING);
-	char *value = 0;
+	int code = (valid_buffer(curbp)
+		    ? b_val(curbp, VAL_FILE_ENCODING)
+		    : enc_AUTO);
+	const char *value = 0;
 
 	switch ((ENC_CHOICES) (code)) {
 	case enc_POSIX:
@@ -1522,40 +1663,117 @@ var_BUF_ENCODING(TBUFF **rp, const char *vp GCC_UNUSED)
     }
 }
 
+int
+var_CMD_ENCODING(TBUFF **rp, const char *vp)
+{
+    if (rp) {
+	tb_scopy(rp, encoding2s(cmd_encoding));
+	return TRUE;
+    } else if (vp) {
+	int choice = choice_to_code(&fsm_cmd_encoding_blist, vp, strlen(vp));
+	cmd_encoding = (ENC_CHOICES) choice;
+	make_local_b_val(bminip, VAL_FILE_ENCODING);
+	set_b_val(bminip, VAL_FILE_ENCODING, vl_encoding);
+	return TRUE;
+    }
+    return FALSE;
+}
+
 int
 var_TERM_ENCODING(TBUFF **rp, const char *vp GCC_UNUSED)
 {
     if (rp) {
-	tb_scopy(rp, encoding2s(term.encoding));
+	tb_scopy(rp, encoding2s(term.get_enc()));
 	return TRUE;
     } else {
 	return FALSE;
     }
 }
+
+int
+var_LATIN1_EXPR(TBUFF **rp, const char *vp)
+{
+    int rc = any_rw_EXPR(rp, vp, &latin1_expr);
+    if (rc && !rp) {
+	char *wide = vl_wide_enc.locale;
+	char *narrow = vl_narrow_enc.locale;
+	if (isEmpty(narrow))
+	    narrow = vl_narrowed(wide);
+	if (isEmpty(wide))
+	    wide = narrow;
+	vl_init_8bit(wide, narrow);
+    }
+    return rc;
+}
 #endif
 
+int
+var_TERM_COLS(TBUFF **rp, const char *vp GCC_UNUSED)
+{
+    return any_ro_INT(rp, vp, term.cols);
+}
+
+int
+var_TERM_LINES(TBUFF **rp, const char *vp GCC_UNUSED)
+{
+    return any_ro_INT(rp, vp, term.rows);
+}
+
+int
+var_TERM_RESIZES(TBUFF **rp, const char *vp GCC_UNUSED)
+{
+    int result = FALSE;
+#if DISP_X11 || DISP_NTWIN || DISP_NTCONS
+    result = TRUE;
+#elif (DISP_TERMCAP || DISP_CURSES) && defined(SIGWINCH)
+    result = TRUE;
+#endif
+
+    return any_ro_BOOL(rp, vp, result);
+}
+
 #if OPT_TITLE
 int
 var_TITLE(TBUFF **rp, const char *vp)
 {
+    int result = FALSE;
     if (rp) {
 #if DISP_X11
 	tb_scopy(rp, x_get_window_name());
-	return TRUE;
+	result = TRUE;
 #elif SYS_WINNT
 	tb_scopy(rp, w32_wdw_title());
-	return TRUE;
+	result = TRUE;
 #endif
     } else if (vp) {
 #if DISP_X11 || SYS_WINNT
 	tb_scopy(&request_title, vp);
-	return TRUE;
+	result = TRUE;
 #else
-	return ABORT;		/* read-only */
+	result = ABORT;		/* read-only */
 #endif
     }
+    return result;
+}
+
+#if OPT_MULTIBYTE
+int
+var_TITLE_ENCODING(TBUFF **rp, const char *vp)
+{
+    if (rp) {
+	tb_scopy(rp, encoding2s(title_encoding));
+	return TRUE;
+    } else if (vp) {
+	int choice = choice_to_code(&fsm_title_encoding_blist, vp, strlen(vp));
+	title_encoding = (ENC_CHOICES) choice;
+	if (auto_set_title) {
+	    term.set_title(tb_values(current_title));
+	}
+	return TRUE;
+    }
     return FALSE;
 }
+#endif
 
 int
 var_TITLEFORMAT(TBUFF **rp, const char *vp)
@@ -1660,6 +1878,7 @@ ev_leaks(void)
     while ((p = temp_vars) != 0)
 	rmv_tempvar(p->u_name);
 
+#if OPT_EVAL && OPT_SHELL
     FreeAndNull(shell);
     FreeAndNull(directory);
 #if DISP_X11
@@ -1667,5 +1886,6 @@ ev_leaks(void)
     FreeAndNull(x_shell);
 #endif
 #endif
+#endif
 }
 #endif /* NO_LEAKS */
