@@ -3,7 +3,7 @@
  * characters, and write characters in a barely buffered fashion on the display.
  * All operating systems.
  *
- * $Header: /usr/build/vile/vile/RCS/termio.c,v 1.213 2007/12/24 01:52:17 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/termio.c,v 1.222 2010/09/07 00:35:30 tom Exp $
  *
  */
 
@@ -42,6 +42,7 @@ static void ttmiscinit(void);
 # define USE_POSIX_TERMIOS 1
 # define USE_FCNTL 1
 #else
+# define USE_POSIX_TERMIOS 0
 # ifdef HAVE_TERMIO_H
 #  define USE_TERMIO 1
 #  define USE_FCNTL 1
@@ -63,22 +64,19 @@ error "No termios or sgtty"
 #if !defined(FIONREAD)
 /* try harder to get it */
 # ifdef HAVE_SYS_FILIO_H
-#  include "sys/filio.h"
+#  include <sys/filio.h>
 # else				/* if you have trouble including ioctl.h, try "sys/ioctl.h" instead */
 #  ifdef HAVE_IOCTL_H
 #   include <ioctl.h>
 #  else
 #   ifdef HAVE_SYS_IOCTL_H
 #    include <sys/ioctl.h>
+#    define VL_sys_ioctl
 #   endif
 #  endif
 # endif
 #endif
 
-#if !defined(FIONREAD) && defined(HAVE_SYS_SOCKET_H)	/* SCO OpenServer v5 */
-# include <sys/socket.h>
-#endif
-
 #if DISP_X11			/* don't use either one */
 # undef USE_FCNTL
 # undef USE_FIONREAD
@@ -269,7 +267,7 @@ ttopen(void)
 # ifdef SETVBUF_REVERSED
     setvbuf(stdout, _IOFBF, tobuf, TBUFSIZ);
 # else
-    setvbuf(stdout, tobuf, _IOFBF, TBUFSIZ);
+    setvbuf(stdout, tobuf, _IOFBF, (size_t) TBUFSIZ);
 # endif
 #else /* !HAVE_SETVBUF */
     setbuffer(stdout, tobuf, TBUFSIZ);
@@ -326,7 +324,7 @@ ttopen(void)
 */
 /* ARGSUSED */
 int
-flow_control_enable(int f, int n GCC_UNUSED)
+flow_control_enable(int f GCC_UNUSED, int n GCC_UNUSED)
 {
 #if !DISP_X11
     if (!f) {
@@ -355,7 +353,7 @@ ttclose(void)
  */
 /*ARGSUSED*/
 void
-ttclean(int f)
+ttclean(int f GCC_UNUSED)
 {
 #if !DISP_X11
     if (f)
@@ -810,13 +808,13 @@ vl_getchar(void)
     char c;
     int n;
 
-    n = read(0, &c, 1);
+    n = (int) read(0, &c, (size_t) 1);
     if (n <= 0) {
 	if (n < 0 && errno == EINTR)
 	    return -1;
 	imdying(SIGINT);
     }
-    return char2int(c);
+    return (c & 0xff);
 }
 #endif
 
@@ -977,7 +975,7 @@ tttypahead(void)
 #  if	USE_FIONREAD
     {
 	long x;
-	return ((ioctl(0, FIONREAD, (caddr_t) & x) < 0) ? 0 : (int) x);
+	return ((ioctl(0, FIONREAD, (void *) &x) < 0) ? 0 : (int) x);
     }
 #  else
 #   if	USE_FCNTL
@@ -1149,6 +1147,35 @@ tttypahead(void)
 
 #endif /* not SYS_UNIX */
 
+#ifdef HAVE_SIZECHANGE
+
+#if !defined(VL_sys_ioctl)
+#if !defined(sun) || !defined(HAVE_TERMIOS_H)
+#include <sys/ioctl.h>
+#endif
+#endif
+
+/*
+ * SCO defines TIOCGSIZE and the corresponding struct.  Other systems (SunOS,
+ * Solaris, IRIX) define TIOCGWINSZ and struct winsize.
+ */
+#ifdef TIOCGSIZE
+# define IOCTL_GET_WINSIZE TIOCGSIZE
+# define IOCTL_SET_WINSIZE TIOCSSIZE
+# define STRUCT_WINSIZE struct ttysize
+# define WINSIZE_ROWS(n) n.ts_lines
+# define WINSIZE_COLS(n) n.ts_cols
+#else
+# ifdef TIOCGWINSZ
+#  define IOCTL_GET_WINSIZE TIOCGWINSZ
+#  define IOCTL_SET_WINSIZE TIOCSWINSZ
+#  define STRUCT_WINSIZE struct winsize
+#  define WINSIZE_ROWS(n) n.ws_row
+#  define WINSIZE_COLS(n) n.ws_col
+# endif
+#endif
+
+#endif /* HAVE_SIZECHANGE */
 /* Get terminal size from system, first trying the driver, and then
  * the environment.  Store number of lines into *heightp and width
  * into *widthp.  If zero or a negative number is stored, the value
@@ -1160,17 +1187,17 @@ void
 getscreensize(int *widthp, int *heightp)
 {
     char *e;
-#ifdef TIOCGWINSZ
-    struct winsize size;
+#ifdef HAVE_SIZECHANGE
+    STRUCT_WINSIZE size;
 #endif
     *widthp = 0;
     *heightp = 0;
-#ifdef TIOCGWINSZ
-    if (ioctl(0, TIOCGWINSZ, (caddr_t) & size) == 0) {
-	if ((int) (size.ws_row) > 0)
-	    *heightp = size.ws_row;
-	if ((int) (size.ws_col) > 0)
-	    *widthp = size.ws_col;
+#ifdef HAVE_SIZECHANGE
+    if (ioctl(0, (long) IOCTL_GET_WINSIZE, (void *) &size) == 0) {
+	if ((int) (WINSIZE_ROWS(size)) > 0)
+	    *heightp = WINSIZE_ROWS(size);
+	if ((int) (WINSIZE_COLS(size)) > 0)
+	    *widthp = WINSIZE_COLS(size);
     }
     if (*widthp <= 0) {
 	e = getenv("COLUMNS");
@@ -1206,20 +1233,20 @@ open_terminal(TERM * termp)
     static TERM save_term;
     static int initialized;
 
-    /*
-     * Help separate dumb_term from termio.c
-     */
-    if (termp != 0 && termp != &null_term) {
-	if (termp->clean == nullterm_clean)
-	    termp->clean = ttclean;
-	if (termp->unclean == nullterm_unclean)
-	    termp->unclean = ttunclean;
-	if (termp->openup == nullterm_openup)
-	    termp->openup = kbd_openup;
-    }
-
     if (!initialized++) {
 
+	/*
+	 * Help separate dumb_term from termio.c
+	 */
+	if (termp != &null_term) {
+	    if (termp->clean == nullterm_clean)
+		termp->clean = ttclean;
+	    if (termp->unclean == nullterm_unclean)
+		termp->unclean = ttunclean;
+	    if (termp->openup == nullterm_openup)
+		termp->openup = kbd_openup;
+	}
+
 	/*
 	 * If the open and/or close slots are empty, fill them in with
 	 * the screen driver's functions.
