@@ -1,7 +1,7 @@
 /*
  * debugging support -- tom dickey.
  *
- * $Header: /usr/build/vile/vile/RCS/trace.c,v 1.75 2007/11/25 16:06:56 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/trace.c,v 1.102 2010/05/01 15:34:13 tom Exp $
  *
  */
 
@@ -12,26 +12,13 @@
 
 #if OPT_ELAPSED
 
-#if SYS_UNIX
-typedef struct timeval ElapsedType;
-#define BI_DATA0 {0, 0}
-#elif SYS_WINNT
-typedef DWORD ElapsedType;
-#define BI_DATA0 0
-#else
-typedef time_t ElapsedType;
-#define BI_DATA0 0
-#endif
-
 typedef struct {
     long total_calls;
     double total_time;
-    ElapsedType this_time;	/* origin of begin_elapsed()    */
+    VL_ELAPSED this_time;	/* origin of begin_elapsed()    */
     int nesting_time;		/* recursion-guard */
 } ELAPSED_DATA;
 
-#define BI_DATA MY_BI_DATA
-
 typedef struct {
     char *bi_key;		/* the name of the command      */
     ELAPSED_DATA *data;		/* actual data                  */
@@ -93,7 +80,10 @@ static unsigned used_indent;
 
 static int trace_depth;
 
-#if OPT_ELAPSED
+static char null_string[] = "<null>";
+static char empty_string[] = "";
+
+#if OPT_ELAPSED && !defined(show_elapsed)
 
 #if 0
 static void ETrace(const char *fmt,...) GCC_PRINTFLIKE(1,2);
@@ -171,33 +161,6 @@ static BI_TREE elapsed_tree =
 static BI_DATA *elapsed_stack[100];
 static int elapsed_sp = 0;
 
-static double
-Elapsed(ElapsedType * first, int begin)
-{
-    double result;
-
-#if SYS_UNIX
-#define	SECS(tv)	(tv.tv_sec + (tv.tv_usec / 1.0e6))
-    ElapsedType tv1;
-    gettimeofday(&tv1, 0);
-    if (begin)
-	*first = tv1;
-    result = (SECS(tv1) - SECS((*first)));
-#elif SYS_WINNT
-    ElapsedType tv1 = GetTickCount();
-    if (begin)
-	*first = tv1;
-    result = ((tv1 - *first) / 1000.0);
-#else
-    ElapsedType tv1 = time((time_t *) 0);
-    if (begin)
-	*first = tv1;
-    result = (tv1 - *first);
-#endif
-
-    return result;
-}
-
 static BI_DATA *
 find_elapsed(const char *txt)
 {
@@ -227,7 +190,7 @@ begin_elapsed(const char *txt)
 	    ELAPSED_DATA *q = p->data;
 	    elapsed_stack[elapsed_sp] = p;
 	    if (!(q->nesting_time++)) {
-		(void) Elapsed(&(q->this_time), TRUE);
+		(void) vl_elapsed(&(q->this_time), TRUE);
 	    }
 	}
     }
@@ -244,7 +207,7 @@ end_elapsed(void)
 	    if (p != 0) {
 		ELAPSED_DATA *q = p->data;
 		if (!(--(q->nesting_time))) {
-		    q->total_time += Elapsed(&(q->this_time), FALSE);
+		    q->total_time += vl_elapsed(&(q->this_time), FALSE);
 		    q->total_calls += 1;
 		}
 	    }
@@ -305,13 +268,12 @@ Trace(const char *fmt,...)
 	    vfprintf(fp, fmt, ap);
 	    (void) fflush(fp);
 	} else if (fp != 0) {
-	    while (trace_depth > 0) {
-		fprintf(fp, "%s", trace_indent(trace_depth--, '|'));
-		fprintf(fp, T_RETURN "(close)\n");
-	    }
+	    fprintf(fp, "%s", trace_indent(trace_depth, '|'));
+	    fprintf(fp, T_RETURN "(close)\n");
 	    (void) fclose(fp);
 	    (void) fflush(stdout);
 	    (void) fflush(stderr);
+	    trace_depth = 0;
 	}
 
 	va_end(ap);
@@ -344,7 +306,15 @@ char *
 retrace_string(char *code)
 {
     Trace(T_RETURN "%s%s\n", TRACE_NULL(code),
-	  (code == error_val) ? " (ERROR)" : "");
+	  isErrorVal(code) ? " (ERROR)" : "");
+    return code;
+}
+
+const char *
+retrace_cstring(const char *code)
+{
+    Trace(T_RETURN "%s%s\n", TRACE_NULL(code),
+	  isErrorVal(code) ? " (ERROR)" : "");
     return code;
 }
 
@@ -371,10 +341,32 @@ alloc_visible(size_t need)
     return visible_result;
 }
 
+static char *
+my_vischr(char *buffer, int ch)
+{
+    if (ch == 127) {
+	strcpy(buffer, "^?");
+    } else if (ch < 32) {
+	sprintf(buffer, "^%c", ch | '@');
+#if OPT_MULTIBYTE
+    } else if (ch >= 256) {
+	sprintf(buffer, "\\u%04x", ch);
+#endif
+    } else if (ch >= 128) {
+	sprintf(buffer, "\\x%02x", ch);
+    } else if (ch == '\\') {
+	sprintf(buffer, "\\\\");
+    } else {
+	buffer[0] = (char) ch;
+	buffer[1] = '\0';
+    }
+    return buffer;
+}
+
 char *
 trace_indent(int level, int marker)
 {
-    unsigned need = 1 + (level * 3);
+    unsigned need = (unsigned) (1 + (level * 3));
     int n;
 
     if (need > used_indent) {
@@ -385,8 +377,9 @@ trace_indent(int level, int marker)
 	else
 	    visible_indent = typereallocn(char, visible_indent, need);
 
-	if (visible_indent != 0)
-	    memset(visible_indent, 0, need);
+	assert(visible_indent != 0);
+
+	memset(visible_indent, 0, need);
     }
 
     *visible_indent = EOS;
@@ -400,7 +393,7 @@ visible_buff(const char *buffer, int length, int eos)
 {
     int j;
     unsigned k = 0;
-    unsigned need = ((length > 0) ? (length * 6) : 0) + 1;
+    unsigned need = (unsigned) (((length > 0) ? (length * 6) : 0) + 1);
     char *result;
 
     beginDisplay();
@@ -412,12 +405,12 @@ visible_buff(const char *buffer, int length, int eos)
     result = alloc_visible(need);
 
     for (j = 0; j < length; j++) {
-	int c = buffer[j] & 0xff;
+	int c = CharOf(buffer[j]);
 	if (eos && !c) {
 	    break;
 	} else {
 	    char *dst = result + k;
-	    vl_vischr(dst, c);
+	    my_vischr(dst, c);
 	    k += (unsigned) strlen(dst);
 	}
     }
@@ -431,7 +424,7 @@ visible_video_text(const VIDEO_TEXT * buffer, int length)
 {
     int j;
     unsigned k = 0;
-    unsigned need = ((length > 0) ? (length * 6) : 0) + 1;
+    unsigned need = (unsigned) (((length > 0) ? (length * 6) : 0) + 1);
     char *result;
 
     beginDisplay();
@@ -447,7 +440,7 @@ visible_video_text(const VIDEO_TEXT * buffer, int length)
     for (j = 0; j < length; j++) {
 	int c = buffer[j] & ((1 << (8 * sizeof(VIDEO_TEXT))) - 1);
 	char *dst = result + k;
-	vl_vischr(dst, c);
+	my_vischr(dst, c);
 	k += (unsigned) strlen(dst);
     }
     result[k] = 0;
@@ -459,10 +452,10 @@ visible_video_text(const VIDEO_TEXT * buffer, int length)
  * Convert a string to visible form.
  */
 char *
-str_visible(char *s)
+str_visible(const char *s)
 {
     if (s == 0)
-	return "<null>";
+	return null_string;
     return visible_buff(s, (int) strlen(s), FALSE);
 }
 
@@ -470,10 +463,10 @@ str_visible(char *s)
  * Convert a string to visible form, but include the trailing null.
  */
 char *
-str_visible0(char *s)
+str_visible0(const char *s)
 {
     if (s == 0)
-	return "<null>";
+	return null_string;
     return visible_buff(s, (int) strlen(s), TRUE);
 }
 
@@ -524,7 +517,7 @@ char *
 lp_visible(LINE *p)
 {
     if (p == 0)
-	return "";
+	return empty_string;
     return visible_buff(lvalue(p), llength(p), FALSE);
 }
 
@@ -539,7 +532,7 @@ fail_alloc(char *msg, char *ptr)
 #if NO_LEAKS
     show_alloc();
 #endif
-    Trace(bad_form);
+    Trace("%s", bad_form);
     abort();
 }
 
@@ -829,17 +822,24 @@ trace_line(LINE *lp, BUFFER *bp)
 	const char *a = check_forw(lp) ? "" : "?";
 	const char *b = check_back(lp) ? "" : "?";
 	Trace("%4d%s:{%p, f %p%s, b %p%s}:%s\n",
-	      line_no(bp, lp), check_line(lp, bp) ? "" : "?", lp,
-	      lforw(lp), a,
-	      lback(lp), b,
+	      line_no(bp, lp),
+	      check_line(lp, bp) ? "" : "?",
+	      (void *) lp,
+	      (void *) lforw(lp), a,
+	      (void *) lback(lp), b,
 	      lp_visible(lp));
+#if OPT_LINE_ATTRS
+	if (lp->l_attrs != NULL) {
+	    Trace("l_attrs{%s}\n", lp->l_attrs);
+	}
+#endif
 	if (*a || *b)
 	    oops++;
     }
 }
 
 void
-trace_mark(char *name, MARK *mk, BUFFER *bp)
+trace_mark(const char *name, MARK *mk, BUFFER *bp)
 {
     Trace("%s %d.%d\n", name, line_no(bp, mk->l), mk->o);
 }
@@ -848,11 +848,11 @@ static const char *
 visible_shape(REGIONSHAPE shape)
 {
     return
-	(shape == EXACT
+	(shape == rgn_EXACT
 	 ? "exact"
-	 : (shape == FULLLINE
+	 : (shape == rgn_FULLLINE
 	    ? "full-line"
-	    : (shape == RECTANGLE
+	    : (shape == rgn_RECTANGLE
 	       ? "rectangle"
 	       : "?")));
 }
@@ -861,7 +861,7 @@ void
 trace_region(REGION * rp, BUFFER *bp)
 {
     B_COUNT total = 0;
-    B_COUNT len_rs = len_record_sep(bp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(bp);
 
     L_NUM no_1st = line_no(bp, rp->r_orig.l);
     L_NUM no_2nd = line_no(bp, rp->r_end.l);
@@ -917,7 +917,7 @@ trace_region(REGION * rp, BUFFER *bp)
 		  "->",
 		  visible_buff(text, size, 0));
 	}
-	total += (size + len_rs);
+	total += ((B_COUNT) size + len_rs);
 	if (total > rp->r_size) {
 	    if (lp == lp2 && (total - len_rs) == rp->r_size) {
 		total -= len_rs;
@@ -928,22 +928,47 @@ trace_region(REGION * rp, BUFFER *bp)
 	    break;
 	lp = lforw(lp);
     }
-    Trace("total %ld vs %ld\n", (long) total, (long) rp->r_size);
+    Trace("total %ld vs %ld%s\n",
+	  (long) total, (long) rp->r_size,
+	  total != rp->r_size ? " - OOPS" : "");
+}
+
+void
+trace_attribs(BUFFER *bp, char *fn, int ln)
+{
+    AREGION *arp;
+    for (arp = bp->b_attribs; arp != 0; arp = arp->ar_next) {
+	Trace("%s@%d r_attr_id %d\n", fn, ln, arp->ar_region.r_attr_id);
+	trace_region(&(arp->ar_region), bp);
+    }
 }
 
 void
 trace_buffer(BUFFER *bp)
 {
     LINE *lp;
-    Trace("trace_buffer(%s) dot=%p%s\n",
+    Trace("trace_buffer(%s)%s %s dot=%p%s fn=%s\n",
 	  bp->b_bname,
-	  bp->b_dot.l,
-	  bp == curbp ? " (curbp)" : "");
+	  (bp == bminip ? " MINI" : ""),
+	  encoding2s(b_val(bp, VAL_FILE_ENCODING)),
+	  (void *) bp->b_dot.l,
+	  bp == curbp ? " (curbp)" : "",
+	  NonNull(bp->b_fname));
     for_each_line(lp, bp) {
 	trace_line(lp, bp);
     }
 }
 
+void
+trace_all_buffers(const char *fn, int ln)
+{
+    BUFFER *bp;
+    Trace(T_CALLED "trace_all_buffers %s@%d\n", fn, ln);
+    for_each_buffer(bp)
+	trace_buffer(bp);
+    returnVoid();
+}
+
 void
 trace_window(WINDOW *wp)
 {
@@ -951,17 +976,23 @@ trace_window(WINDOW *wp)
     int got_dot = FALSE;
     int had_line = FALSE;
 
-    if (wp->w_bufp == 0
+    if (wp == 0
+	|| wp->w_bufp == 0
 	|| wp->w_bufp->b_bname == 0)
 	return;
 
-    Trace("trace_window(%s) top=%d, rows=%d, head=%p, line=%p dot=%p%s\n",
+    Trace("trace_window(%s)%s top=%d, rows=%d, head=%p, line=%p dot=%p%s\n",
 	  wp->w_bufp->b_bname,
+	  ((wp == wminip)
+	   ? " MINI"
+	   : ((wp == wnullp)
+	      ? " NULL"
+	      : "")),
 	  wp->w_toprow,
 	  wp->w_ntrows,
-	  win_head(wp),
-	  wp->w_line.l,
-	  wp->w_dot.l,
+	  (void *) win_head(wp),
+	  (void *) wp->w_line.l,
+	  (void *) wp->w_dot.l,
 	  wp == curwp ? " (curwp)" : "");
 
     for (lp = wp->w_line.l; lp != win_head(wp); lp = lforw(lp)) {
@@ -978,6 +1009,46 @@ trace_window(WINDOW *wp)
     }
 }
 
+void
+trace_all_windows(const char *fn, int ln)
+{
+    WINDOW *wp;
+    Trace(T_CALLED "trace_all_windows %s@%d\n", fn, ln);
+    for_each_window(wp)
+	trace_window(wp);
+    returnVoid();
+}
+
+#define TRACE_CTYPE(mask) if (table[n] & chrBIT(mask)) strcat(buffer, "|" #mask)
+void
+trace_ctype(CHARTYPE *table, int first, int last)
+{
+    int n;
+    char buffer[1024];
+    char temp[80];
+
+    if (first < 0)
+	first = 0;
+    if (last < 0)
+	last = N_chars;
+    for (n = first; n < last; ++n) {
+	*buffer = EOS;
+	TRACE_CTYPE(vl_UPPER);
+	TRACE_CTYPE(vl_LOWER);
+	TRACE_CTYPE(vl_DIGIT);
+	TRACE_CTYPE(vl_SPACE);
+	TRACE_CTYPE(vl_CNTRL);
+	TRACE_CTYPE(vl_PRINT);
+	TRACE_CTYPE(vl_PUNCT);
+#if OPT_WIDE_CTYPES
+	TRACE_CTYPE(vl_XDIGIT);
+#endif
+	Trace("(%s) /* %d:%s */\n",
+	      *buffer ? buffer + 1 : "0",
+	      n, my_vischr(temp, n));
+    }
+}
+
 #if OPT_WORKING && (OPT_TRACE > 2)
 #undef beginDisplay
 void
@@ -1003,14 +1074,15 @@ close_me(void)
 #if DOALLOC
     FreeAndNull(area);
 #endif
+#if OPT_ELAPSED
+    btree_freeup(&elapsed_tree);
+#endif
+    Trace("%s", bad_form);
+
     FreeAndNull(visible_result);
     FreeAndNull(visible_indent);
     used_visible = 0;
     used_indent = 0;
-#if OPT_ELAPSED
-    btree_freeup(&elapsed_tree);
-#endif
-    Trace(bad_form);
 }
 
 void
