@@ -2,7 +2,7 @@
  * Window management. Some of the functions are internal, and some are
  * attached to keys that the user actually types.
  *
- * $Header: /usr/build/vile/vile/RCS/window.c,v 1.112 2007/10/27 16:00:08 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/window.c,v 1.124 2010/09/07 00:36:17 tom Exp $
  *
  */
 
@@ -11,7 +11,7 @@
 
 #if OPT_PERL || OPT_TCL
 /* Window id to assign to w_id field for next visible window allocated */
-static ULONG w_id_next = 1;
+static WIN_ID w_id_next = 1;
 
 /* Fake windows are given a window id of 0 */
 #define FAKE_WINDOW_ID 0
@@ -51,9 +51,13 @@ alloc_WINDOW(void)
 static void
 free_WINDOW(WINDOW *wp)
 {
-    beginDisplay();
-    free(wp);
-    endofDisplay();
+    if (wp != 0) {
+	beginDisplay();
+	if (wp == wminip)
+	    wminip = 0;
+	free(wp);
+	endofDisplay();
+    }
 }
 
 /*
@@ -125,11 +129,11 @@ int
 reposition(int f, int n)
 {
     if (f) {
-	int an;
+	long an;
 	/* clamp the value at the size of the window */
-	an = absol(n);
+	an = absol((long) n);
 	if (an > curwp->w_ntrows)
-	    curwp->w_force = curwp->w_ntrows * (n / an);
+	    curwp->w_force = (int) (curwp->w_ntrows * (n / an));
 	else
 	    curwp->w_force = n;
 	curwp->w_flag |= WFFORCE;
@@ -558,7 +562,7 @@ save_window_modes(BUFFER *bp)
 
     if (bp->b_nwnd != 0) {
 	beginDisplay();
-	result = typecallocn(W_VALUES, bp->b_nwnd);
+	result = typecallocn(W_VALUES, (size_t) bp->b_nwnd);
 	if (result != 0) {
 	    n = 0;
 	    for_each_window(wp) {
@@ -1109,6 +1113,23 @@ clone_window(WINDOW *dst, WINDOW *src)
     dst->w_flag |= WFMODE | WFHARD;	/* Quite nasty.         */
 }
 
+static WINDOW *
+new_WINDOW(int screen)
+{
+    WINDOW *wp = typecalloc(WINDOW);
+
+    if (wp == NULL) {
+	fprintf(stderr, "Cannot allocate windows\n");
+	ExitProgram(BADEXIT);
+    } else {
+	init_window(wp, (BUFFER *) 0);
+	wp->w_ntrows = (screen
+			? (term.rows - 2)	/* "-1" for mode line.  */
+			: 1);	/* command-line         */
+    }
+    return wp;
+}
+
 /*
  * Initialize all of the windows.
  */
@@ -1119,21 +1140,10 @@ winit(int screen)
 
     TRACE((T_CALLED "winit(%d)\n", screen));
 
-    wp = typecalloc(WINDOW);	/* First window         */
-    if (wp == NULL)
-	ExitProgram(BADEXIT);
+    wp = new_WINDOW(screen);	/* First window         */
     wheadp = wp;
     curwp = wp;
 
-    init_window(wp, (BUFFER *) 0);
-    wp->w_wndp = NULL;		/* Initialize window    */
-    wp->w_toprow = 0;
-    wp->w_ntrows = (screen
-		    ? (term.rows - 2)	/* "-1" for mode line.  */
-		    : 1);	/* command-line         */
-    wp->w_force = 0;
-    wp->w_bufp = NULL;
-
 #if OPT_PERL || OPT_TCL
     wp->w_id = w_id_next;
     ++w_id_next;
@@ -1145,8 +1155,13 @@ winit(int screen)
 	       bminip->b_linecount,
 	       bminip->b_bytecount));
 	(void) delink_bp(bminip);
+#if OPT_MULTIBYTE
 	(void) delink_bp(btempp);
+#endif
     } else {
+	/* create a dummy window, to hold buffers before display is active */
+	wnullp = new_WINDOW(1);
+
 	/* create the command-buffer */
 	TRACE(("winit creating bminip & wminip\n"));
 	wminip = wp;
@@ -1164,7 +1179,9 @@ winit(int screen)
 	make_local_w_val(wminip, WMDLINEWRAP);
 	set_w_val(wminip, WMDLINEWRAP, FALSE);
 #endif
+#if OPT_MULTIBYTE
 	btempp = bfind("\tUTF-8", BFINVS);
+#endif
     }
     returnVoid();
 }
@@ -1248,7 +1265,7 @@ pop_fake_win(WINDOW *oldwp, BUFFER *oldbp)
 /* Find and return the window with the given window id.  Return NULL
    if not found */
 WINDOW *
-id2win(ULONG id)
+id2win(WIN_ID id)
 {
     WINDOW *wp;
     for_each_visible_window(wp) {
@@ -1259,7 +1276,7 @@ id2win(ULONG id)
 }
 
 /* Return the window id associated with the given window */
-ULONG
+WIN_ID
 win2id(WINDOW *wp)
 {
     return wp->w_id;
@@ -1308,11 +1325,16 @@ wp_leaks(void)
 {
     WINDOW *wp;
 
+    beginDisplay();
     while ((wp = wheadp) != 0) {
 	wp = wp->w_wndp;
 	free_WINDOW(wheadp);
-	wheadp = wp;
+	if (wp != wheadp)
+	    wheadp = wp;
+	else
+	    break;
     }
     free_WINDOW(wminip);
+    endofDisplay();
 }
 #endif
