@@ -3,7 +3,7 @@
  * paragraph at a time.  There are all sorts of word mode commands.  If I
  * do any sentence mode commands, they are likely to be put in this file.
  *
- * $Header: /usr/build/vile/vile/RCS/word.c,v 1.85 2007/09/16 20:09:35 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/word.c,v 1.101 2010/02/12 09:57:34 tom Exp $
  *
  */
 
@@ -11,6 +11,11 @@
 #include	"edef.h"
 #include	"nefunc.h"
 
+#define each_character(c) \
+	    c = DOT.o, once = TRUE; \
+		(c > 0) || once; \
+		    c -= BytesBefore(DOT.l, c), once = FALSE
+
 /* Word wrap on n-spaces. Back-over whatever precedes the point on the current
  * line and stop on the first word-break or the beginning of the line. If we
  * reach the beginning of the line, jump back to the end of the word and start
@@ -21,80 +26,102 @@
 int
 wrapword(int f, int n)
 {
+    int once;
+    int rc = FALSE;
+    int done = FALSE;
     int cnt = 0;		/* size of word wrapped to next line */
     int c;			/* character temporary */
-    B_COUNT to_delete;
+    B_COUNT to_delete = 0;
     C_NUM to_append = 0;
 
-    /* Back up from the <NL> 1 char */
-    if (!backchar(FALSE, 1))
-	return (FALSE);
+    TRACE((T_CALLED "wrapword(%d, %d)\n", f, n));
 
-    /* If parameter given, delete the trailing white space.  This is done
-     * to support a vi-like wrapmargin mode (see insert.c).  Unlike vi,
-     * we're deleting all of the trailing whitespace; vi deletes only the
-     * portion that was added during the current insertion.
-     *
-     * If there's no trailing whitespace, and we're not inserting blanks
-     * (n!=0), try to split the line at the last embedded whitespace.
-     */
-    if (f) {
-	LINE *lp = DOT.l;
-	to_delete = 0L;
-	if (DOT.o >= 0 && !n && !isSpace(lgetc(lp, DOT.o))) {
-	    for (c = DOT.o; c >= 0; c -= BytesBefore(DOT.l, c)) {
+    /* Back up from the <NL> 1 char */
+    if (backchar(FALSE, 1)) {
+
+	/* If parameter given, delete the trailing white space.  This is done
+	 * to support a vi-like wrapmargin mode (see insert.c).  Unlike vi,
+	 * we're deleting all of the trailing whitespace; vi deletes only the
+	 * portion that was added during the current insertion.
+	 *
+	 * If there's no trailing whitespace, and we're not inserting blanks
+	 * (n!=0), try to split the line at the last embedded whitespace.
+	 */
+	if (f) {
+	    LINE *lp = DOT.l;
+	    to_delete = 0L;
+	    if (DOT.o >= 0 && !n && !isSpace(lgetc(lp, DOT.o))) {
+		int tmp = 0;
+		for (each_character(c)) {
+		    if (isSpace(lgetc(lp, c))) {
+			to_append = n;
+			cnt = tmp;
+			DOT.o = c;
+			break;
+		    }
+		    ++tmp;
+		}
+	    }
+	    for (each_character(c)) {
 		if (isSpace(lgetc(lp, c))) {
-		    to_append = n;
-		    cnt = (DOT.o - c);
+		    to_delete++;
 		    DOT.o = c;
+		} else {
 		    break;
 		}
 	    }
-	}
-	for (c = DOT.o; c >= 0; c -= BytesBefore(DOT.l, c)) {
-	    if (isSpace(lgetc(lp, c))) {
-		to_delete++;
-		DOT.o = c;
-	    } else {
-		break;
+	    if (to_delete == 0) {
+		DOT.o += BytesAt(DOT.l, DOT.o);
+	    }
+	} else {
+	    /* Back up until we aren't in a word, make sure there is a
+	     * break in the line
+	     */
+	    while (c = char_at(DOT), !isSpace(c)) {
+		cnt++;
+		if (!backchar(FALSE, 1)) {
+		    rc = FALSE;
+		    done = TRUE;
+		    break;
+		}
+	    }
+	    if (!done) {
+		/* if we make it to the beginning, start a new line */
+		if (DOT.o == 0) {
+		    (void) gotoeol(FALSE, 0);
+		    rc = lnewline();
+		    done = TRUE;
+		} else {
+		    to_delete = 1L;
+		}
 	    }
 	}
-	if (to_delete == 0) {
-	    DOT.o += BytesAt(DOT.l, DOT.o);
-	}
-    } else {
-	/* Back up until we aren't in a word, make sure there is a
-	 * break in the line
-	 */
-	while (c = char_at(DOT), !isSpace(c)) {
-	    cnt++;
-	    if (!backchar(FALSE, 1))
-		return (FALSE);
-	    /* if we make it to the beginning, start a new line */
-	    if (DOT.o == 0) {
-		(void) gotoeol(FALSE, 0);
-		return (lnewline());
+
+	if (!done) {
+	    /* delete the forward white space */
+	    if (ldel_bytes(to_delete, FALSE)) {
+
+		/* put in newline */
+		if (newline(FALSE, 1)) {
+
+		    rc = TRUE;
+		    /* and past the first word */
+		    while (cnt-- > 0) {
+			if (forwchar(FALSE, 1) == FALSE) {
+			    rc = FALSE;
+			    break;
+			}
+		    }
+		    if (rc) {
+			if (to_append > 0) {
+			    lins_bytes(to_append, ' ');
+			}
+		    }
+		}
 	    }
 	}
-	to_delete = 1L;
     }
-
-    /* delete the forward white space */
-    if (!ldel_bytes(to_delete, FALSE))
-	return (FALSE);
-
-    /* put in newline */
-    if (!newline(FALSE, 1))
-	return FALSE;
-
-    /* and past the first word */
-    while (cnt-- > 0) {
-	if (forwchar(FALSE, 1) == FALSE)
-	    return (FALSE);
-    }
-    if (to_append > 0)
-	lins_bytes(to_append, ' ');
-    return (TRUE);
+    returnCode(rc);
 }
 
 /*
@@ -282,14 +309,15 @@ backword(int f, int n)
 static int
 join_region(int exact)
 {
+    BUFFER *bp = curbp;
     int status;
     int doto, c;
     LINE *end;
     REGION region;
     int done = FALSE;
 
-    if ((status = getregion(&region)) == TRUE
-	&& (status = !is_last_line(region.r_orig, curbp)) == TRUE) {
+    if ((status = getregion(bp, &region)) == TRUE
+	&& (status = !is_last_line(region.r_orig, bp)) == TRUE) {
 
 	DOT = region.r_orig;
 	end = region.r_end.l;
@@ -298,7 +326,7 @@ join_region(int exact)
 
 	    return (TRUE);
 	}
-	regionshape = EXACT;
+	regionshape = rgn_EXACT;
 
 	while (!done && status == TRUE) {
 	    c = EOS;
@@ -381,6 +409,59 @@ comment_prefix(void)
     return result;
 }
 
+#if OPT_MULTIBYTE
+static int
+byte_to_columns(char text, int size)
+{
+    int result = 0;
+
+    if (b_is_utfXX(curbp)) {
+	result = (mb_cellwidth(curwp, &text, 1) * size);
+    } else {
+	result = size;
+    }
+    return result;
+}
+
+static int
+any_wcs_width(const char *value, long bytes)
+{
+    int result = 0;
+
+    if (b_is_utfXX(curbp)) {
+
+	while (bytes > 0) {
+	    int step = vl_conv_to_utf32((UINT *) 0, value, (B_COUNT) bytes);
+	    if (step <= 0)
+		break;
+	    result += mb_cellwidth(curwp, value, step);
+	    value += step;
+	    bytes -= step;
+	}
+    } else {
+	result = bytes;
+    }
+
+    return result;
+}
+
+int
+str_wcs_width(const char *value)
+{
+    return any_wcs_width(value, (long) strlen(value));
+}
+
+int
+tb_wcs_width(TBUFF *text)
+{
+    return any_wcs_width(tb_values(text), (long) tb_length(text));
+}
+
+#define Byte2Column(s,n) byte_to_columns(s, n)
+#else
+#define Byte2Column(s,n) n
+#endif
+
 #define cplus_comment_start(c) \
 		((c == '/') \
 		&& DOT.o+1 < llength(DOT.l) \
@@ -391,11 +472,18 @@ comment_prefix(void)
 		&& DOT.o+1 < llength(DOT.l) \
 		&& lgetc(DOT.l,DOT.o+1) == '*')
 
-#define fmt_insert(count,chr) \
+#define fmt_insert_chr(count,chr) \
 		if ((s = lins_bytes(count,chr)) != TRUE) \
 			return s; \
 		else \
-			clength += count
+			clength += Byte2Column(chr, count)
+
+#define fmt_insert_str(str) \
+		for (i = 0; i < tb_length(str); ++i) { \
+		    if ((s = lins_bytes(1, tb_values(str)[i])) != TRUE) \
+			return s; \
+		} \
+		clength += tb_columns(str)
 
 #define fmt_c_preprocessor(cp) \
 		(tb_length(*cp) == 1 \
@@ -409,6 +497,7 @@ comment_prefix(void)
 static int
 do_formatting(TBUFF **wp, TBUFF **cp)
 {
+    BUFFER *bp = curbp;
     int c;			/* current char during scan     */
     int clength;		/* position on line during fill */
     int plength;		/* prefix to ignore during fill */
@@ -422,18 +511,18 @@ do_formatting(TBUFF **wp, TBUFF **cp)
     int sentence;		/* was the last char a period?  */
     int secondindent;
     int s;
-    int fillcolumn = getfillcol(curbp);
+    int fillcolumn = getfillcol(bp);
 
     if (!sameline(MK, DOT)) {
 	REGION region;
 
-	if (getregion(&region) != TRUE)
+	if (getregion(bp, &region) != TRUE)
 	    return FALSE;
 	if (sameline(region.r_orig, MK))
 	    swapmark();
     }
     pastline = MK.l;
-    if (pastline != buf_head(curbp))
+    if (pastline != buf_head(bp))
 	pastline = lforw(pastline);
 
     finished = FALSE;
@@ -509,7 +598,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 			|| (s > 0
 			    && memcmp(tb_values(*cp),
 				      lvalue(DOT.l) + DOT.o,
-				      s))) {
+				      (size_t) s))) {
 			finished = SORTOFTRUE;
 		    }
 
@@ -523,7 +612,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 			}
 			if (DOT.l != pastline
 			    && !dot_at_section_break()) {
-			    B_COUNT spcs = DOT.o;
+			    B_COUNT spcs = (B_COUNT) DOT.o;
 			    DOT.o = 0;
 			    s = ldel_bytes(spcs, FALSE);
 			    if (s != TRUE)
@@ -553,7 +642,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 	    if (!isBlank(c)) {
 		tb_append(wp, c);
 		sentence = FALSE;
-		if (global_g_val(GMDSPACESENT)) {
+		if (b_val(bp, MDSPACESENT)) {
 		    /* was it the end of a "sentence"? */
 		    if (c == '?' || c == ':' || c == '!') {
 			sentence = TRUE;
@@ -579,42 +668,38 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 		/* at a word break with a word waiting */
 		/* calculate tentative new length
 		   with word added */
-		newlen = (int) (clength + 1 + tb_length(*wp));
+		newlen = (int) (clength + 1 + tb_columns(*wp));
 		if (newlen <= fillcolumn) {
 		    /* add word to current line */
 		    if (!firstflag) {
-			fmt_insert(1, ' ');
+			fmt_insert_chr(1, ' ');
 		    }
 		} else {
 		    /* fix the leading indent now, if
 		       some spaces should be tabs */
-		    if (b_val(curbp, MDTABINSERT))
+		    if (b_val(bp, MDTABINSERT))
 			entabline((void *) TRUE, 0, 0);
 		    if (lnewline() == FALSE)
 			return FALSE;
 		    clength = 0;
-		    fmt_insert(secondindent, ' ');
+		    fmt_insert_chr(secondindent, ' ');
 		    firstflag = TRUE;
 		}
 
 		if (firstflag
 		    && is_comment
 		    && !word_finishes_c_comment(wp)) {
-		    for (i = 0; i < tb_length(*cp); i++) {
-			fmt_insert(1, tb_values(*cp)[i]);
-		    }
+		    fmt_insert_str(*cp);
 		    if (!fmt_c_preprocessor(cp)) {
-			fmt_insert(1, ' ');
+			fmt_insert_chr(1, ' ');
 		    }
 		}
 		firstflag = FALSE;
 
 		/* and add the word in in either case */
-		for (i = 0; i < tb_length(*wp); i++) {
-		    fmt_insert(1, tb_values(*wp)[i]);
-		}
+		fmt_insert_str(*wp);
 		if (finished == FALSE && sentence) {
-		    fmt_insert(1, ' ');
+		    fmt_insert_chr(1, ' ');
 		}
 		tb_init(wp, EOS);
 	    }
@@ -622,7 +707,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 	/* catch the case where we're done with a line not because
 	   there's no more room, but because we're done processing a
 	   section or the region */
-	if (b_val(curbp, MDTABINSERT))
+	if (b_val(bp, MDTABINSERT))
 	    entabline((void *) TRUE, 0, 0);
 	DOT.l = lforw(DOT.l);
     }
@@ -654,24 +739,25 @@ formatregion(void)
 int
 wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    BUFFER *bp = curbp;
     LINE *lp;			/* current line to scan */
     int offset;			/* current char to scan */
-    long size;			/* size of region left to count */
+    ULONG size;			/* size of region left to count */
     int ch;			/* current character to scan */
     int wordflag;		/* are we in a word now? */
     int lastflag;		/* were we just in a word? */
-    long nwhite;		/* total # of blanks */
+    B_COUNT nwhite;		/* total # of blanks */
     long nwords;		/* total # of words */
-    long nchars;		/* total number of chars */
+    B_COUNT nchars;		/* total number of chars */
     int nlines;			/* total number of lines in region */
     double avgch;		/* average number of chars/word */
     int status;			/* status return code */
     REGION region;		/* region to look at */
-    const char *ending = get_record_sep(curbp);
-    int len_rs = len_record_sep(curbp);
+    const char *ending = get_record_sep(bp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(bp);
 
     /* make sure we have a region to count */
-    if ((status = getregion(&region)) != TRUE) {
+    if ((status = getregion(bp, &region)) != TRUE) {
 	return (status);
     }
     lp = region.r_orig.l;
@@ -694,7 +780,7 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 	    ++nlines;
 	    if (len_rs > 1) {
 		nwhite += (len_rs - 1);
-		size -= (len_rs - 1);
+		size -= (ULONG) (len_rs - 1);
 	    }
 	} else {
 	    ch = lgetc(lp, offset);
@@ -717,7 +803,7 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
     else
 	avgch = 0.0;
 
-    mlforce("lines %d, words %ld, chars %ld, avg chars/word %f",
+    mlforce("lines %d, words %ld, chars %lu, avg chars/word %f",
 	    nlines, nwords, nchars + nwhite, avgch);
 
     return (TRUE);
